From C4.Util Require Import
  LibTactics.

From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair.

Record AnnMethod {A: Type}{M: Type -> Type} {T} := {
  annMethodAnn: A;
  annMethodMet: M T
}.

Arguments AnnMethod : clear implicits.

Definition annMethod {A M T} := @Build_AnnMethod A M T.


Definition Annotated (A: Type){I}(o: Spec I): Spec (AnnMethod A I) :=
  mapSpec o (fun _ => annMethodMet).


Definition erase {A}{M}{R} (p: Program (AnnMethod A M) R): Program M R :=
  mapProgram
    (fun _ m => annMethodMet m)
    p.

Definition eraseImpl {A}{M I} (impl: forall R, M R -> Program (AnnMethod A I) R):
  forall R, M R -> Program I R :=
  fun R m => erase (impl R m).


Definition annProgram {A M R} (a: A) (p : Program M R):
  Program (AnnMethod A M) R :=
  
  mapProgram (fun R' i => annMethod a i) p.


Definition annImplProg 
           {M I} (impl: forall R, M R -> Program I R) A:
  forall R, AnnMethod A M R -> Program (AnnMethod A I) R :=

  fun R m => annProgram (annMethodAnn m) (impl R (annMethodMet m)).


Definition annImpl 
           A {M} (impl: Object M):
  Object (AnnMethod A M) :=

  object
    (Annotated A impl.(objBase))
    (annImplProg impl.(objProgram) A).


(* ------------------------------- *)
(* Annotation on left *)

Definition eraseLeft
           {A}{M1 M2}{R}
           (p: Program (PairMethod (AnnMethod A M1) M2) R):
  Program (PairMethod M1 M2) R :=

  mapProgram
    (fun _ m =>
       match m with
       | inl m => inl (annMethodMet m)
       | inr m => inr m
       end)
    p.

Definition eraseLeftImplProg
           {A}{M I1 I2}
           (impl: forall R, M R -> Program (PairMethod (AnnMethod A I1) I2) R):
  forall R, M R -> Program (PairMethod I1 I2) R :=
  fun R m => eraseLeft (impl R m).


(* ------------------------------- *)
(* flatten ann *)

Definition flattenAnnProgram
           {M I A} (impl: forall R, M R -> Program I R)
           {R} (p: Program (AnnMethod A M) R):
  Program (AnnMethod A I) R :=

  flattenProgram (annImplProg impl A) p.


Definition flattenAnnObject
           {M} (base: Object M)
           {M' A} (impl: forall R, M' R -> Program (AnnMethod A M) R) :
  Object M' :=
  flattenObject (annImpl A base) _ impl.



Definition flattenAnnProgramLeft
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {A R} (p: Program (PairMethod (AnnMethod A M1) M2) R):
  Program (PairMethod (AnnMethod A I1) M2) R :=

  flattenProgramLeft (annImplProg impl A) p.

(*
Definition flattenAnnProgramLeft'
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {A R} (p: Program (AnnMethod A (PairMethod M1 M2)) R):
  Program (AnnMethod A (PairMethod I1 M2)) R :=

  flattenProgramLeft (annImplProg impl) p.
*)

Definition flattenAnnImplProgLeft
           {M1 M2 I1 M A} 
           (impl: forall R, M1 R -> Program I1 R)
           (impl': forall R, M R -> Program (PairMethod (AnnMethod A M1) M2) R):
  forall R, M R -> Program (PairMethod (AnnMethod A I1) M2) R :=

  fun R m => flattenAnnProgramLeft impl (impl' R m).


Definition flattenAnnProgramRight
           {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
           {A R} (p: Program (PairMethod M1 (AnnMethod A M2)) R):
  Program (PairMethod M1 (AnnMethod A I2)) R :=

  flattenProgramRight (annImplProg impl A) p.

Definition flattenAnnImplProgRight
           {M1 M2 I2 M A} 
           (impl: forall R, M2 R -> Program I2 R)
           (impl': forall R, M R -> Program (PairMethod M1 (AnnMethod A M2)) R):
  forall R, M R -> Program (PairMethod M1 (AnnMethod A I2)) R :=

  fun R m => flattenAnnProgramRight impl (impl' R m).


Definition bindFlattenAnnProgramLeft
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {R R' A} (p: Program (PairMethod (AnnMethod A I1) (AnnMethod A M2)) R)
           (f: R -> Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R'):
  Program (PairMethod (AnnMethod A I1) (AnnMethod A M2)) R' :=

  bindFlattenProgramLeft (annImplProg impl A) p f.


Definition bindFlattenAnnProgramRight
           {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
           {R R' A} (p: Program (PairMethod (AnnMethod A M1) (AnnMethod A I2)) R)
           (f: R -> Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R'):
  Program (PairMethod (AnnMethod A M1) (AnnMethod A I2)) R' :=

  bindFlattenProgramRight (annImplProg impl A) p f.


Definition distPairAnnProg
           {A M1 M2 R} (p: Program (AnnMethod A (PairMethod M1 M2)) R):
  Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R :=

  mapProgram
    (fun _ m =>
       let a := annMethodAnn m in
       match annMethodMet m with
       | inl m' => inl (annMethod a m')
       | inr m' => inr (annMethod a m')
       end)
    p.


Definition distPairAnnImplProg 
           {M A M1 M2} (impl: forall R, M R -> Program (AnnMethod A (PairMethod M1 M2)) R):
  forall R, M R -> Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R :=

  fun R m => distPairAnnProg (impl R m).


Definition factPairAnn {A I1 I2} Ret (m : PairInstruction (AnnMethod A I1) (AnnMethod A I2) Ret) :
  AnnMethod A (PairInstruction I1 I2) Ret :=
  match m with
  | inl m' => annMethod (annMethodAnn m') (inl (annMethodMet m'))
  | inr m' => annMethod (annMethodAnn m') (inr (annMethodMet m'))
  end.

Definition factPairAnnProg
           {A M1 M2 R} (p: Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R):
  Program (AnnMethod A (PairMethod M1 M2)) R :=

  mapProgram factPairAnn p.


Definition factPairAnnImplProg 
           {M A M1 M2} (impl: forall R, M R -> Program (PairMethod (AnnMethod A M1) (AnnMethod A M2)) R):
  forall R, M R -> Program (AnnMethod A (PairMethod M1 M2)) R :=

  fun R m => factPairAnnProg (impl R m).


