Require Import
        Coq.Logic.FunctionalExtensionality.

From C4.Util Require Import
  LibTactics.

From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Serializability
  ProgramEqSimulates
  FlattenSimulates
  MethodSimulates
  Pair
  Object.

From C4.App Require Import
  Annotated
  Dum.

Import SimulatesCoInd.


Lemma annImplConcRefImp:
  forall A I (o: Spec I) M (impl: forall R, M R -> Program (AnnMethod A I) R),
    ConcSimulates
      (object (Annotated A o) impl)
      (object o (eraseImpl impl)).
Proof.
  intros.
  eapply (simulatesCoInd
            (object (Annotated A o) impl)
            (object o (fun R m => erase (impl R m)))
            (fun _ p1 s1 p2 s2 => p2 = erase p1)
            (fun s1 s2 => s1 = s2)
            (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    right. left. eexists. splits.
    + progFrobSimplInAt H2 2. eassumption.
    + reflexivity.
    + auto.
  - introv H1. auto.
  - introv H1 H2. progFrobSimplInAt H1 2. assumption.
  - introv H1 H2 H3. subst.
    eexists. eexists. splits.
    Focus 2. reflexivity.
    Focus 2. reflexivity.
    + progFrobSimplAt 1. unfold erase. eapply ProgramStepImplProgramSteps. applys_eq (@EBind I) 2. eassumption. reflexivity.
    + auto.
Qed.

(* Todo: Refactor to inversion lemma. *)
Lemma annProgramEqRet:
  forall I R (p : Program I R) A (a: A) r,
    annProgram a p = Return r ->
    p = Return r.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  reflexivity.
Qed.

Lemma annProgramEqNoop:
  forall I R (p: Program I R) A (a: A) p',
    annProgram a p = Noop p' ->
    exists p'', p = Noop p'' /\ p' = annProgram a p''.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  eauto.
Qed.


Lemma annImplProgSerialImplProg:
  forall M A,
    annImplProg (serialImplProg M) A =
    serialImplProg (AnnMethod A M).
Proof.
  intros.
  unfold annImplProg. unfold serialImplProg.
  eapply functional_extensionality_dep. intros R.
  eapply functional_extensionality. intros m. destruct m.
  progFrobSimpl. f_equal. eapply functional_extensionality. intros R'.
  progFrobSimpl. reflexivity.
Qed.

Lemma flattenImplProgConcRefFlattenAnnImplProg:
  forall {M I} (iBase: forall R, M R -> Program I R)
    {M' A} (impl: forall R, M' R -> Program (AnnMethod A M) R) impl' o,

    impl' = eraseImpl impl ->
    ConcSimulates
      (object o (flattenImplProg iBase impl'))
      (object (Annotated A o) (flattenImplProg (annImplProg iBase A) impl)).

Proof.
  intros.
  subst.
  eapply simulatesTrans_, (mapMethodSimulates o _ _ nil).
  apply (programEqSimulates (I := I)).
  intros.
  unfold flattenImplProg, eraseImpl, erase, annImplProg, annProgram.
  rewrite mapFlattenProgram, flattenMapProgram.
  apply flattenProgramProgramEq.
  - intros.
    rewrite mapMapProgram.
    apply programEqSym, mapProgramRefl.
  - reflexivity.
Qed.


(* --------------------- *)

Lemma evalFactPairAnn I1 I2 A (o1 : Spec I1) (o2 : Spec I2) :
  EvalPairInstruction (Annotated A o1) (Annotated A o2) =
  fun R m => (pairSpec o1 o2).(EvalInstruction) (annMethodMet (factPairAnn R m)).
Proof.
  extensionality R.
  extensionality m.
  destruct m; reflexivity.
Qed.

Lemma flattenImplProgConRefFlattenImplProgDist:
  forall M M1 M2 A I1 I2
    (impl: forall R, M R -> Program (AnnMethod A (PairMethod M1 M2)) R)  
    (iBase1: forall R, M1 R -> Program I1 R)
    (iBase2: forall R, M2 R -> Program I2 R)
    (o1: Spec I1) (o2: Spec I2),
    ConcSimulates
      (object
        (Annotated A (pairSpec o1 o2))
        (flattenImplProg (annImplProg (pairImplProg iBase1 iBase2) A) impl))
      (object
        (pairSpec (Annotated A o1) (Annotated A o2))
        (flattenImplProg (pairImplProg (annImplProg iBase1 A) (annImplProg iBase2 A)) (distPairAnnImplProg impl))).

Proof.
  intros.
  eapply simulatesTrans with
  (spec2:=
     ThreadsBehavior
       (flattenObject
          (mapProgramImpl
             (Annotated A (pairSpec o1 o2))
             (pairImplProg (annImplProg iBase1 A) (annImplProg iBase2 A))
             factPairAnn)
          M
          (distPairAnnImplProg impl))).
  - apply programEqSimulates.
    intros.
    unfold flattenImplProg, distPairAnnImplProg, distPairAnnProg, annImplProg, annProgram.
    rewrite flattenMapProgram.
    apply flattenProgramProgramEq.
    + destruct m0 as [? []]; simpl; repeat rewrite mapMapProgram; reflexivity.
    + reflexivity.
  - apply (flattenSimulates {| objBase := _ |} {| objBase := _ |}).
    rewrite evalFactPairAnn.
    exact (mapMethodSimulates _ _ factPairAnn nil _).
Qed.

