From Coq Require Import
  FunctionalExtensionality.
From Paco Require Import paco.

From C4 Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Annotated
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Abortable
  Serializability
  ProgramEqSimulates
  FlattenSimulates
  MethodSimulates
  LibTactics
  Dum2
  LinearizabilityFacts2
  Util
  RSimulates Eutt.

Import SimulatesCoInd.


Lemma annImplConcRefImp:
  forall A I (o: Spec I) M (impl: forall R, M R -> Program (AnnMethod A I) R),
    ConcSimulates
      (object (Annotated A o) impl)
      (object o (eraseImpl impl)).

Proof.
  intros.
  eapply (simulatesCoInd
            (object (Annotated A o) impl)
            (object o (fun R m => erase (impl R m)))
            (fun _ p1 s1 p2 s2 => p2 = erase p1)
            (fun s1 s2 => s1 = s2)
            (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    right. left. eexists. splits.
    + progFrobSimplInAt H2 2. eassumption.
    + reflexivity.
    + auto.
  - introv H1. auto.
  - introv H1 H2. progFrobSimplInAt H1 2. assumption.
  - introv H1 H2 H3. subst.
    eexists. eexists. splits.
    Focus 2. reflexivity.
    Focus 2. reflexivity.
    + progFrobSimplAt 1. unfold erase. eapply ProgramStepImplProgramSteps. applys_eq (@EBind I) 2. eassumption. reflexivity.
    + auto.
Qed.

(*
Lemma annImplSeqRefImp:
  forall A I (o: Spec I) M (impl: forall R, M R -> Program (AnnMethod A I) R) init,
    SeqSimulates
      (object (Annotated A o) impl)
      (object o (eraseImpl impl))
      init init.

Todo: First, in ThreadsFacts.v, prove that ConcRef impl SeqRef.
*)

(*
Lemma implConcRefAnnImp:
  forall A I (o: Spec I) M (impl: forall R, M R -> Program (AnnMethod A I) R) init,
    ConcSimulates
      (object o (eraseImpl impl))
      (object (Annotated A o) impl)
      init init.

Proof.
Qed.
*)

(*
Lemma implConcRefAnnImp':
  forall A I (o: Spec I) M (impl: forall R, M R -> Program (AnnMethod A I) R) impl' init ,
    impl' = eraseImpl impl ->
    ConcSimulates
      (object o impl')
      (object (Annotated A o) impl)
      init init.

Proof.
Qed.
*)


(* Todo: Refactor to inversion lemma. *)
Lemma annProgramEqRet:
  forall I R (p : Program I R) A (a: A) r,
    annProgram a p = Return r ->
    p = Return r.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  reflexivity.
Qed.

Lemma annProgramEqNoop:
  forall I R (p: Program I R) A (a: A) p',
    annProgram a p = Noop p' ->
    exists p'', p = Noop p'' /\ p' = annProgram a p''.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  eauto.
Qed.


Lemma annImplProgSerialImplProg:
  forall M A,
    annImplProg (serialImplProg M) A =
    serialImplProg (AnnMethod A M).
Proof.
  intros.
  unfold annImplProg. unfold serialImplProg.
  eapply functional_extensionality_dep. intros R.
  eapply functional_extensionality. intros m. destruct m.
  progFrobSimpl. f_equal. eapply functional_extensionality. intros R'.
  progFrobSimpl. reflexivity.
Qed.

(* ---------------------------------- *)

(*
Lemma leftImplConcRefLeftAnnImp:
  forall A I1 I2 (o1: Spec I1) (o2: Spec I2)
    M (annImpl: forall R, M R -> Program (PairMethod (AnnMethod A I1) I2) R) s1 s2,
    
    ConcSimulates
      (object
         (pairSpec o1 o2)
         (eraseLeftImplProg annImpl))
      (object
         (pairSpec (Annotated A o1) o2)
         annImpl)
      (s1, s2) (s1, s2).
Proof.
Qed.
(* Similar to implConcRefAnnImp? *)


Lemma leftImplConcRefLeftAnnImp':
  forall A I1 I2 (o1: Spec I1) (o2: Spec I2)
    M (annImpl: forall R, M R -> Program (PairMethod (AnnMethod A I1) I2) R) s1 s2 impl',

    impl' = eraseLeftImplProg annImpl ->
    ConcSimulates
      (object
         (pairSpec o1 o2)
         impl')
      (object
         (pairSpec (Annotated A o1) o2)
         annImpl)
      (s1, s2) (s1, s2).      
Proof.
  intros.
  subst.
  eapply leftImplConcRefLeftAnnImp.
Qed.


Lemma eraseFlattenImplProgLeft:
  forall {M I1 I2 I1' A}
    (implProg: forall R, M R -> Program (PairMethod I1 I2) R)
    (baseAnnImplProg: forall R, I1 R -> Program (AnnMethod A I1') R)
    R m,

    programEq
      (eraseLeftImplProg
         (flattenImplProgLeft' baseAnnImplProg implProg) R m)
      (flattenImplProgLeft' (eraseImpl baseAnnImplProg) implProg R m).

Proof.
  (*
  intros.
  gen m. cofix CoInd. intros.
  destruct (implProg R m) as [ R' m' f' | ret | np ] eqn: E.
  - progFrobSimpl. rewrite E. destruct m'. constructor.

    Lemma L:
      forall (M I1 I2 I1' : Type -> Type) (A : Type)
             (implProg : forall R : Type, M R -> Program (PairMethod I1 I2) R)
             (baseAnnImplProg : forall R : Type, I1 R -> Program (AnnMethod A I1') R)
             (R : Type),
        
        (forall m : M R,
            programEq
              (eraseLeftImplProg (flattenImplProgLeft' baseAnnImplProg implProg) R m)
              (flattenImplProgLeft' (eraseImpl baseAnnImplProg) implProg R m)) ->
        
        forall (R' : Type) (i : I1 R') (f' : R' -> Program (PairMethod I1 I2) R),
          programEq
            (flattenProgramLeft' (bindFlattenProgramLeft' baseAnnImplProg (baseAnnImplProg R' i) f'))
            (bindFlattenProgramLeft' (eraseImpl baseAnnImplProg)
                                     (eraseImpl baseAnnImplProg R' i) f').
    
    Proof.
      intros M I1 I2 I1' A implProg baseAnnImplProg R CoInd R' i f'.
      

    progFrobSimpl. apply CoInd.
  -  
  
Qed.
*)
  
  
Lemma eraseFlattenImplProgLeft':
  forall {M I1 I2 I1' A}
    (implProg: forall R, M R -> Program (PairMethod I1 I2) R)
    (baseImplProg: forall R, I1 R -> Program I1' R)
    (baseAnnImplProg: forall R, I1 R -> Program (AnnMethod A I1') R),

    (forall R m,
        programEq
          (baseImplProg R m)
          ((eraseImpl baseAnnImplProg) R m)) ->
    (forall R m,
       programEq
         ((eraseLeftImplProg
            (flattenImplProgLeft' baseAnnImplProg implProg)) R m)
         ((flattenImplProgLeft' baseImplProg implProg) R m)).
(*
Proof.
  intros. subst.
  apply eraseFlattenImplProgLeft.
Qed.
*)

Lemma dumInstLeftAnnFlatImplConRefAnnImpl:
  forall A M M' T o (s: o.(State))
    (annImplProg: forall R, @TransMethod M T R -> Program (AnnMethod A M') R),

    ConcSimulates
      (dumInstLeftAnnFlatImpl (object o (eraseImpl annImplProg)))
      (dumInstLeftAnnFlatImpl (object (Annotated A o) annImplProg))
      (s, tt)
      (s, tt).

Proof.
  intros.
  unfold dumInstLeftAnnFlatImpl.
  eapply leftImplConcRefLeftAnnImp'.
  symmetry.
  eapply eraseFlattenImplProgLeft'.
  simpl.
  reflexivity.
Qed.      
*)

Section DILAFICRAI.

Context
  A M M' T (o : Spec M')
  (implProg : forall R, TransMethod M T R -> Program M' R)
  (annImplProg: forall R, @TransMethod M T R -> Program (AnnMethod A M') R)
  (H_implProg : forall R m,
    programEq
      (implProg R m)
      (eraseImpl annImplProg R m)).

Let RR R
    (p1 : Program (PairInstruction M' (AnnMethod T DumMethod)) R)
    (p2 : Program (PairInstruction (AnnMethod A M') (AnnMethod T DumMethod)) R)
  : Prop :=
  eutt ieq p1 (eraseLeft p2).

Inductive PreRR {R}
  : forall
    (p1 : Program (PairInstruction M' (AnnMethod T DumMethod)) R)
    (p2 : Program (PairInstruction (AnnMethod A M') (AnnMethod T DumMethod)) R), Prop :=
| PreRR_Base p
  : PreRR (flattenProgramLeft' implProg p)
          (flattenProgramLeft' annImplProg p)
| PreRR_Bind A0 p q (k : A0 -> _)
  : programEq p (erase q) ->
    PreRR
      (bindFlattenProgramLeft' implProg p k)
      (bindFlattenProgramLeft' annImplProg q k)
.
Hint Constructors PreRR : core.

Lemma RR_Spawn_ {R} : forall p q, PreRR p q -> RR R p q.
Proof.
  unfold RR; pcofix SELF; intros p q H.
  rewrite (frobProgramId p), (frobProgramId (eraseLeft q)).
  destruct H.
  - destruct p as [ ? [] | | ]; cbn; pfold; constructor; auto.
    + right; apply SELF; constructor. apply H_implProg.
    + right; eauto.
    + right; apply SELF; constructor.
  - remember (erase q).
    rewrite (frobProgramId (erase q)) in Heqp0.
    destruct H; destruct q; cbn in Heqp0; inversion Heqp0; subst; injpair; clear Heqp0; cbn.
    all: pfold; constructor; auto.
    all: right; eauto.
Qed.

Lemma RR_Spawn : forall (R : Type) (m : Abortable (Program M) R),
    RR R (flattenImplProgLeft' implProg dumInstLeftAnnImplProg R m)
         (flattenImplProgLeft' annImplProg dumInstLeftAnnImplProg R m).
Proof.
  intros; apply RR_Spawn_; constructor.
Qed.

Lemma eq_unit (x y : unit) : x = y.
Proof.
  destruct x, y; reflexivity.
Qed.

Lemma RR_Step
  : forall s1 (R : Type)
      (p1 : Program (ObjInstruction (dumInstLeftAnnFlatImpl (object o implProg)))
              R)
      (p2 : Program
              (ObjInstruction
                 (dumInstLeftAnnFlatImpl (object (Annotated A o) annImplProg))) R),
    RR R p1 p2 ->
    forall (s1' : State (objBase (dumInstLeftAnnFlatImpl (object o implProg))))
      (p1' : Program
               (ObjInstruction (dumInstLeftAnnFlatImpl (object o implProg))) R),
    OneProgramStep (objBase (dumInstLeftAnnFlatImpl (object o implProg))) s1 p1
      s1' p1' ->
    exists p1'' s2' p2',
    eutt ieq p1' p1'' /\
    ProgramSteps
      (objBase (dumInstLeftAnnFlatImpl (object (Annotated A o) annImplProg))) R
      p2 s1 p2' s2' /\ s1' = s2' /\ RR R p1'' p2'.
Proof.
  unfold RR; intros * H_RR * H_OPS.
  induction H_OPS.
  { apply IHH_OPS, eutt_Noop_L_inv, H_RR. }
  subst.
  match type of H_RR with
  | (_ ?t1 ?t2) => remember t1 eqn:EE1 in H_RR; remember t2 eqn:EE2 in H_RR
  end.
  revert p2 EE1 EE2; punfold H_RR; induction H_RR; try discriminate; intros.
  - inversion EE1; subst; injpair; clear EE1.
    destruct H0.
    rewrite (frobProgramId (eraseLeft p2)) in EE2; destruct p2 as [ ? [] | | ] ;
      cbn in EE2; inversion EE2; subst; injpair; clear EE2.
    all: change (mapProgram _) with (@eraseLeft A M' (AnnMethod T DumMethod) R) in *.
    all: pclearbot; cbn.
    + eexists _, _, _. split; [ eapply H1 | ].
      split; [ | split; [ reflexivity | apply Reflexive_eutt_ieq ]].
      eapply ProgramStepImplProgramSteps. constructor. constructor; [ apply H | apply eq_unit ].
    + eexists _, _, _. split; [ eapply H1 | ].
      split; [ | split; [ reflexivity | apply Reflexive_eutt_ieq ]].
      eapply ProgramStepImplProgramSteps. constructor. assumption.
  - subst. rewrite (frobProgramId (eraseLeft p0)) in EE2. destruct p0 as [ ? [] | | ];
      cbn in EE2; inversion EE2; subst; clear EE2.
    change (mapProgram _) with (@eraseLeft A M' (AnnMethod T DumMethod) R) in *.
    specialize (IHH_RR _ eq_refl eq_refl).
    destruct IHH_RR as (p1'' & s2' & p2' & ? & ? & -> & ?).
    exists p1'' s2' p2'.
    split; eauto.
    split; eauto.
    eapply ProgramStepStepsImplSteps; [ constructor | eassumption ].
Qed.

Lemma RR_Return
  : forall (R : Type)
      (p1 : Program (PairInstruction M' (AnnMethod T DumMethod)) R)
      (p2 : Program (PairInstruction (AnnMethod A M') (AnnMethod T DumMethod)) R)
      (r : R),
    RR R p1 p2 -> eutt ieq p1 (Return r) ->
    forall s0 : State o * DumState,
    ProgramSteps (pairSpec (Annotated A o) (Annotated T dumObj)) R
      p2 s0 (Return r) s0.
Proof.
  intros * H_RR EE.
  apply Symmetric_eutt_ieq in EE.
  eapply (Transitive_eutt_ieq _ _ _ EE) in H_RR; clear EE.
  remember (Return r) eqn:EEr; remember (eraseLeft p2) eqn:EEe.
  revert p2 EEr EEe; punfold H_RR; induction H_RR; intros; try discriminate.
  - inversion EEr; subst.
    rewrite (frobProgramId (eraseLeft p2)) in EEe.
    destruct p2; inversion EEe; subst.
    apply ProgramStepsRefl.
  - subst p0; rewrite (frobProgramId (eraseLeft p3)) in EEe.
    destruct p3 as [ ? [] | | ]; cbn in EEe; inversion EEe; subst; clear EEe.
    eapply ProgramStepStepsImplSteps; [ constructor | eauto ].
Qed.

Lemma dumInstLeftAnnFlatImplConRefAnnImpl'
  : ConcSimulates
      (dumInstLeftAnnFlatImpl (object o implProg))
      (dumInstLeftAnnFlatImpl (object (Annotated A o) annImplProg)).
Proof.
  intros. apply SMS.SMS_ConcS.
  constructor 1 with (RS := @eq (prod _ _)) (RR := RR).
  - reflexivity.
  - cbn. apply RR_Spawn.
  - intros ? _ []. apply RR_Step.
  - cbn. apply RR_Return.
Qed.

End DILAFICRAI.

Lemma flattenImplProgConcRefFlattenAnnImplProg {M I} (iBase: forall R, M R -> Program I R)
    {M' A} (impl: forall R, M' R -> Program (AnnMethod A M) R) impl' o :
    impl' = eraseImpl impl ->
    ConcSimulates
      (object o (flattenImplProg iBase impl'))
      (object (Annotated A o) (flattenImplProg (annImplProg iBase A) impl)).
Proof.
  intros ->.
  eapply simulatesTrans_, (mapMethodSimulates o _ _ nil).
  apply (programEqSimulates (I := I)).
  intros.
  unfold flattenImplProg, eraseImpl, erase, annImplProg, annProgram.
  rewrite mapFlattenProgram, flattenMapProgram.
  apply flattenProgramProgramEq.
  - intros.
    rewrite mapMapProgram.
    apply programEqSym, mapProgramRefl.
  - reflexivity.
Qed.

(* --------------------- *)

Lemma evalFactPairAnn I1 I2 A (o1 : Spec I1) (o2 : Spec I2) :
  EvalPairInstruction (Annotated A o1) (Annotated A o2) =
  fun R m => (pairSpec o1 o2).(EvalInstruction) (annMethodMet (factPairAnn R m)).
Proof.
  extensionality R.
  extensionality m.
  destruct m; reflexivity.
Qed.

Lemma flattenImplProgConRefFlattenImplProgDist M M1 M2 A I1 I2
    (impl: forall R, M R -> Program (AnnMethod A (PairMethod M1 M2)) R)  
    (iBase1: forall R, M1 R -> Program I1 R)
    (iBase2: forall R, M2 R -> Program I2 R)
    (o1: Spec I1) (o2: Spec I2) :

    ConcSimulates
      (object
        (Annotated A (pairSpec o1 o2))
        (flattenImplProg (annImplProg (pairImplProg iBase1 iBase2) A) impl))
      (object
        (pairSpec (Annotated A o1) (Annotated A o2))
        (flattenImplProg (pairImplProg (annImplProg iBase1 A) (annImplProg iBase2 A)) (distPairAnnImplProg impl))).

Proof.
  eapply simulatesTrans with
  (spec2:=
     ThreadsBehavior
       (flattenObject
          (mapProgramImpl
             (Annotated A (pairSpec o1 o2))
             (pairImplProg (annImplProg iBase1 A) (annImplProg iBase2 A))
             factPairAnn)
          M
          (distPairAnnImplProg impl))).
  - apply programEqSimulates.
    intros.
    unfold flattenImplProg, distPairAnnImplProg, distPairAnnProg, annImplProg, annProgram.
    rewrite flattenMapProgram.
    apply flattenProgramProgramEq.
    + destruct m0 as [? []]; simpl; repeat rewrite mapMapProgram; reflexivity.
    + reflexivity.
  - apply (flattenSimulates {| objBase := _ |} {| objBase := _ |}).
    rewrite evalFactPairAnn.
    exact (mapMethodSimulates _ _ factPairAnn nil _).
Qed.

