Require Import
        Coq.Logic.FunctionalExtensionality.

From C4.Util Require Import
  LibTactics.

From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Abortable
  ProgramEqSimulates
  FlattenSimulates
  MethodSimulates
  Serializability.

From C4.App Require Import
  Annotated
  SerializabilityAnn.

(* ---------------------------------------------- *)
(* Dum Noop *)

Definition DumState := unit.

(*
Definition DumMethod :=
  fun R: Type => unit.
*)

Inductive DumMethod: Type -> Type :=
| dumCall: DumMethod unit.


Definition dumObj: Spec DumMethod :=
  Build_Spec DumMethod DumState (fun _ _ _ _ _ => True) tt.

Definition dumState: dumObj.(State) := tt.

(*
        CoFixpoint denoop {I R} (p: Program I R): Program I R :=
          match p with
          | Bind i f => Bind i (fun r => denoop (f r))
          | Return a => Return a
          | Noop p' => p'
          end.
*)

CoFixpoint dumNoopProgram
           {I R} (p: Program I R):
  Program (PairMethod I DumMethod) R :=
  match p with
  | Bind i f => Bind (inl i) (fun r => dumNoopProgram (f r))
  | Return a => Return a
  | Noop p' => Bind (inr dumCall) (fun _: unit => dumNoopProgram p')
  end.

Theorem DumNoopOpStart:
  forall I R (p: Program I R),
    OpStartProgram (dumNoopProgram p).

Proof.
  intros. destruct p; progFrobSimpl; constructor.
Qed.

Definition dumNoopImplProg
           {M I} (impl: forall R, M R -> Program I R):
  forall R, M R -> Program (PairMethod I DumMethod) R :=
  fun R m => dumNoopProgram (impl R m).


Definition dumNoopImpl
           {M} (impl: Object M):
  Object M :=
  object
    (pairSpec impl.(objBase) dumObj)
    (dumNoopImplProg impl.(objProgram)).


Lemma dumNoopProgramEqRet:
  forall I R (p : Program I R) r,
    dumNoopProgram p = Return r ->
    p = Return r.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  reflexivity.
Qed.

Lemma dumNoopProgramEqNoop:
  forall I R (p: Program I R) p',
    dumNoopProgram p = Noop p' ->
    False.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
Qed.


CoInductive DumOrNoopEq {I R}: Program I R -> Program (PairMethod I DumMethod) R -> Prop :=
| bindDumOrNoopEq R' (i: I R') f f':
    (forall r, DumOrNoopEq (f r) (f' r)) ->
    DumOrNoopEq (Bind i f) (Bind (inl i) f')
| retDumOrNoopEq r:
    DumOrNoopEq (Return r) (Return r)
| noopDumOrNoopEq1 p p':
    DumOrNoopEq p p' ->
    DumOrNoopEq (Noop p) (Bind (inr dumCall) (fun r: unit => p'))
| noopDumOrNoopEq2 p p':
    DumOrNoopEq p p' ->
    DumOrNoopEq (Noop p) (Noop p').

Lemma dumNoopProgramDumOrNoopEq:
  forall {I R} (p: Program I R),
    DumOrNoopEq
      p
      (dumNoopProgram p).
Proof.
  intros.
  gen p. cofix CoInd.
  intros. destruct p.
  - rewrite frobProgramId. simpl. constructor. auto. 
  - rewrite frobProgramId. simpl. constructor.
  - rewrite frobProgramId. simpl. constructor. auto.   
Defined.


Lemma bindFlattenProgramDumOrNoopEqAux:
  (forall M R (p: Program M R)
          I (iBase: forall R, M R -> Program I R),
    DumOrNoopEq
      (flattenProgram iBase p)
      (flattenProgram (dumNoopImplProg iBase) p)) ->

  forall M I (iBase: forall R, M R -> Program I R)
         R R' (f: R -> Program M R') (p: Program I R) p',
    DumOrNoopEq p p' ->
    DumOrNoopEq 
      (bindFlattenProgram iBase f p)
      (bindFlattenProgram (dumNoopImplProg iBase) f p').

Proof.
  introv H1 H2.
  gen p p'. cofix CoInd.
  introv H2.
  destruct p.
  
  - inverts H2. 
    match goal with
    | [|- DumOrNoopEq ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
    
  - inverts H2. 
    match goal with
    | [|- DumOrNoopEq ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
    
  - inverts H2 as.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
Defined.

Lemma flattenProgramDumOrNoopEq:
  forall M R (p: Program M R)
         I (iBase: forall R, M R -> Program I R),
    DumOrNoopEq
      (flattenProgram iBase p)
      (flattenProgram (dumNoopImplProg iBase) p).
      
Proof.
  cofix CoInd.
  intros. destruct p.
  
  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. 
    eapply bindFlattenProgramDumOrNoopEqAux.
    eassumption. apply dumNoopProgramDumOrNoopEq.
    
  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.

  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. auto.
    
Qed.

Lemma bindFlattenProgramDumOrNoopEq:
  forall M I (iBase: forall R, M R -> Program I R)
         R R' (f: R -> Program M R') (p: Program I R) p',
    DumOrNoopEq p p' ->
    DumOrNoopEq 
      (bindFlattenProgram iBase f p)
      (bindFlattenProgram (dumNoopImplProg iBase) f p').

Proof.
  apply bindFlattenProgramDumOrNoopEqAux.
  apply flattenProgramDumOrNoopEq.
Qed.


Lemma flattenDumNoopConcRef:
  forall M M' (impl: forall R, M R -> Program M' R)
         I (iBase: forall R, M' R -> Program I R) o,
  ConcSimulates
    (object
      o
      (flattenImplProg iBase impl))
    (object
      (pairSpec o dumObj)
      (flattenImplProg (dumNoopImplProg iBase) impl)).
Proof.
  intros M M' impl I iBase o.
  eapply (SimulatesCoInd.simulatesCoInd
    (object
      o
      (flattenImplProg iBase impl))
    (object
      (pairSpec o dumObj)
      (flattenImplProg (dumNoopImplProg iBase) impl))
    (fun _ p1 s1 p2 s2 => 
      DumOrNoopEq p1 p2)
    (fun s1 s2 => s1 = fst s2)
    (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    inverts H2 as.
    + introv H2.
    right. right. eexists. splits.
      * apply ProgramStepImplProgramSteps. simpl.
        eapply (@EBind _ _ _ _ _ tt).
        simpl. auto.
      * eassumption.
      * auto.
    + right. left. eexists. splits.
      * reflexivity.
      * assumption. 
      * auto. 
  - introv H1. split.
    + apply flattenProgramDumOrNoopEq.
    + auto.
  - introv H1 H2. inverts H1. reflexivity.
  - introv H1 H2 H3. inverts H2. subst.
    eexists. evar (s2s : State o). exists (s2s, tt). subst s2s. splits.
    + eapply ProgramStepImplProgramSteps. constructor. simpl. split. eassumption. destruct s2. destruct d. reflexivity.
    + reflexivity. 
    + auto.
    + auto.
Qed.


CoInductive DumOrNoopEq2 {I1 I2 R}: Program (PairMethod I1 I2) R -> Program (PairMethod I1 (PairMethod I2 DumMethod)) R -> Prop :=
| bindDumOrNoopEq21 R' (i: I1 R') f f':
    (forall r, DumOrNoopEq2 (f r) (f' r)) ->
    DumOrNoopEq2 
      (Bind (inl i) f) 
      (Bind (inl i) f')
| bindDumOrNoopEq22 R' (i: I2 R') f f':
    (forall r, DumOrNoopEq2 (f r) (f' r)) ->
    DumOrNoopEq2 
      (Bind (inr i) f) 
      (Bind (inr (inl i)) f')
| retDumOrNoopEq2 r:
    DumOrNoopEq2 (Return r) (Return r)
| noopDumOrNoopEq21 p p':
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2
      (Noop p)
      (Bind (inr (inr dumCall)) (fun r: unit => p'))
| noopDumOrNoopEq22 p p':
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2 (Noop p) (Noop p').


Lemma DumOrNoopEq2Left:
  forall I1 I2 R (p: Program I1 R),
    @DumOrNoopEq2 I1 I2 R
      (mapProgram (fun _ => inl) p)
      (mapProgram (fun _ => inl) p).

Proof.
  intros. gen p.
  cofix CoInd.
  intros. destruct p.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
Qed.

Lemma DumOrNoopEq2Right:
  forall I1 I2 R (p: Program I2 R),
    @DumOrNoopEq2 I1 I2 R
      (mapProgram (fun _ => inr) p)
      (mapProgram (fun _ => inr) (dumNoopProgram p)).

Proof.
  intros.
  gen p. cofix CoInd.
  destruct p.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end. 
    constructor. auto.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end. 
    constructor.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end. 
    constructor. auto.
  Qed.

Lemma bindFlattenProgramDumOrNoopEq2Aux:
  (forall M1 M2 R (p: Program (PairMethod M1 M2) R)
          I1 (iBase1: forall R, M1 R -> Program I1 R)
          I2 (iBase2: forall R, M2 R -> Program I2 R),
    DumOrNoopEq2
      (flattenProgram (pairImplProg iBase1 iBase2) p)
      (flattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) p)) ->

  forall M1 M2 I1 I2
         (iBase1: forall R, M1 R -> Program I1 R)
         (iBase2: forall R, M2 R -> Program I2 R)
         R R' (f: R -> Program (PairMethod M1 M2) R') (p: Program (PairMethod I1 I2) R) p',
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2 
      (bindFlattenProgram (pairImplProg iBase1 iBase2) f p)
      (bindFlattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) f p').

Proof.
  introv H1 H2.
  gen p p'. cofix CoInd.
  introv H2.
  destruct p.
  
  - inverts H2. 
    + match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
 
  - inverts H2. 
    match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
    
  - inverts H2 as.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
Defined.

Lemma flattenProgramDumOrNoopEq2:
  forall M1 M2 R (p: Program (PairMethod M1 M2) R)
         I1 (iBase1: forall R, M1 R -> Program I1 R)
         I2 (iBase2: forall R, M2 R -> Program I2 R),
    DumOrNoopEq2
      (flattenProgram (pairImplProg iBase1 iBase2) p)
      (flattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) p).
      
Proof.
  cofix CoInd.
  intros. destruct p.
  
  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. 
    eapply bindFlattenProgramDumOrNoopEq2Aux.
    eassumption. 
    destruct p. 
    eapply DumOrNoopEq2Left.
    eapply DumOrNoopEq2Right.
        
  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.

  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. auto.
    
Qed.

Lemma bindFlattenProgramDumOrNoopEq2:
  forall M1 M2 I1 I2
         (iBase1: forall R, M1 R -> Program I1 R)
         (iBase2: forall R, M2 R -> Program I2 R)
         R R' (f: R -> Program (PairMethod M1 M2) R') (p: Program (PairMethod I1 I2) R) p',
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2 
      (bindFlattenProgram (pairImplProg iBase1 iBase2) f p)
      (bindFlattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) f p').
      
Proof.
  apply bindFlattenProgramDumOrNoopEq2Aux.
  apply flattenProgramDumOrNoopEq2.
Qed.


Lemma flattenDumNoopConcRefPair:
  forall M M1' M2' I1' I2' 
         (impl: forall R, M R -> Program (PairMethod M1' M2') R)
         (iBase1: forall R, M1' R -> Program I1' R)
         (iBase2: forall R, M2' R -> Program I2' R) o1 o2,
  ConcSimulates
    (object
      (pairSpec o1 o2)
      (flattenImplProg 
        (pairImplProg 
          iBase1
          iBase2)
        impl))
    (object
      (pairSpec
        o1 
        (pairSpec o2 dumObj))
      (flattenImplProg
        (pairImplProg
          iBase1
          (dumNoopImplProg iBase2))
        impl)).

Proof.
  intros.
  eapply (SimulatesCoInd.simulatesCoInd
    (object
      (pairSpec o1 o2)
      (flattenImplProg 
        (pairImplProg 
          iBase1
          iBase2)
        impl))
    (object
      (pairSpec
        o1 
        (pairSpec o2 dumObj))
      (flattenImplProg
        (pairImplProg
          iBase1
          (dumNoopImplProg iBase2))
        impl))
    (fun _ p1 s1 p2 s2 => 
      DumOrNoopEq2 p1 p2)
    (fun s1 s2 => s1 = (fst s2, fst (snd s2)))
    (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    inverts H2 as.
    + introv H2.
    right. right. eexists. splits.
      * apply ProgramStepImplProgramSteps.
        eapply (@EBind _ _ _ _ _ tt).
        simpl. auto.
      * eassumption.
      * auto.
    + right. left. eexists. splits.
      * reflexivity.
      * assumption. 
      * auto. 
  - introv H1. split.
    + apply flattenProgramDumOrNoopEq2.
    + auto.
  - introv H1 H2. inverts H1. reflexivity.
  - introv H1 H2 H3. subst.
    destruct m.
    + inverts H2. simpl in *. destruct H3 as (H31 & H32).
      exists (f' r). eexists (_, _). splits.
      * eapply ProgramStepImplProgramSteps. constructor. simpl. split. eassumption. reflexivity.      
      * simpl. destruct s1'. simpl in *. congruence.
      * auto.
      * auto.
    + inverts H2. simpl in *. destruct H3 as (H31 & H32). destruct s2 as [s21 [s221 s222]]. simpl in *.
      eexists. exists (s21, (snd s1', s222)). splits; simpl in *.
      * eapply ProgramStepImplProgramSteps. constructor. simpl. split. split. eassumption. reflexivity. reflexivity.
      * destruct s1'. simpl in *. congruence.
      * auto.
      * auto.
Qed.

Definition dumInstLeftImplProg {M TId}
           (instrument :
              forall R,
                Program (PairMethod M DumMethod) R ->
                Program (TransMethod (PairMethod M DumMethod) TId) (option R)) :
  forall R, (Abortable (Program M)) R ->
       Program (PairMethod (@TransMethod M TId) DumMethod) R :=

  fun R m =>
    match m with
    | abortable p =>
      mapProgramDep focusLeft (instrument _ (dumNoopProgram p))
    end.

Definition dumInstLeftFlatImpl
           {M} {TId}
           (alg: Object (@TransMethod M TId))
           (instrument :
              forall R,
                Program (PairMethod M DumMethod) R ->
                Program (TransMethod (PairMethod M DumMethod) TId) (option R)) :
  Object (Abortable (Program M)) :=

  flattenImplLeft'
    alg dumObj
    (dumInstLeftImplProg instrument).

CoInductive DumOrNoopNEq {I R}: Program I R -> Program (PairMethod I DumMethod) R -> Prop :=
| bindDumOrNoopNEq R' (i: I R') f f':
    (forall r, DumOrNoopNEq (f r) (f' r)) ->
    DumOrNoopNEq (Bind i f) (Bind (inl i) f')
| retDumOrNoopNEq r:
    DumOrNoopNEq (Return r) (Return r)
| noopDumOrNoopNEq1 f p:
    (* | noopDumOrNoopNEq1 p p':
    DumOrNoopNEq p p' ->
    DumOrNoopNEq (Noop p) (Bind (inr tt) (fun r: unit => p')) *)
    (forall r: unit, DumOrNoopNEq p (f r)) ->
    DumOrNoopNEq (Noop p) (Bind (inr dumCall) f)
| noopDumOrNoopNEq2 p p':
    DumOrNoopNEq p p' ->
    DumOrNoopNEq (Noop p) (Noop p').
(*| noopDumOrNoopNEq3 p p':
    (forall r, p <> Return r) ->
    DumOrNoopNEq p p' ->
    DumOrNoopNEq p (Noop p'). *)
(* | noopDumOrNoopNEq3 p f:
    (forall r: unit, DumOrNoopNEq (Noop p) (f r)) ->
    DumOrNoopNEq (Noop p) (Noop (Bind (inr tt) f)). *)


Lemma bindFlattenProgramSer1:
  forall (M M' : Type -> Type) (TId: Type)
    (implProg : forall R : Type, @TransMethod M TId R -> Program M' R)
    R (r: R) R' (p': Program M' R'),

    DumOrNoopNEq
      (bindFlattenProgram
         implProg
         (fun ret => Return r)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret => mapProgramDep focusLeft (Return r))).

Proof.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.
  - progFrobSimpl. constructor.
  - apply CoInd.
Qed.

Lemma bindFlattenProgramSer2:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R)
    p' R (r: R),

    DumOrNoopNEq
      (bindFlattenProgram
         implProg
         (fun ret : unit + TId =>
            match ret with
            | inl _ => Return (Some r)
            | inr t' => _ <- abort t'; Return None
            end)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun b : unit + TId =>
            mapProgramDep
              focusLeft
              match b with
              | inl _ => Return (Some r)
              | inr t' => _ <- abort t'; Return None
              end)).

Proof.
  intros.
  gen p' r.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.      
  - destruct s.
    + progFrobSimpl. constructor.
    + progFrobSimpl. constructor. apply bindFlattenProgramSer1.
  - apply CoInd.
Qed.

Lemma bindFlattenProgramSer3:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R),
    
    (forall R (p : Program M R) (t: TId),
        DumOrNoopNEq
          (flattenProgram implProg (instrument' R p t))
          (flattenProgramLeft' implProg (instrumentLeft' R (dumNoopProgram p) t))) ->

    forall R A p p',
      DumOrNoopNEq
        (bindFlattenProgram
           implProg
           (fun a : option A * TId =>
              instrumentWrap'
                R
                (foldProgram
                   (stateful TId (Abortable M))
                   match fst a with
                   | Some a1 => optionProgram (@abortable M) (p a1)
                   | None => Return None
                   end
                   (snd a)))
           p')
        (bindFlattenProgramLeft'
           implProg
           p'
           (fun b : option A * TId =>
              instrumentLeftWrap'
                R
                (foldProgram
                   (stateful TId (Abortable (PairMethod M DumMethod)))
                   match fst b with
                   | Some a0 =>
                     optionProgram
                       (@abortable (PairMethod M DumMethod))
                       (dumNoopProgram (p a0))
                   | None => Return None
                   end
                   (snd b)))).

Proof.
  introv H1.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.      
  - destruct p0.
    destruct o.
    + eapply H1.
    + progFrobSimpl. constructor. apply bindFlattenProgramSer1.
  - apply CoInd.
Defined.

Lemma flattenProgramSer1:
  forall (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, @TransMethod M TId R -> Program M' R)
    (p : Program M R) t,
    DumOrNoopNEq
      (flattenProgram implProg (instrument' R p t))
      (flattenProgramLeft'
         implProg
         (instrumentLeft' R (dumNoopProgram p) t)).

Proof.
  intros.
  gen R p t.    
  cofix CoInd.
  intros.
  destruct p.
  
  - match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.
    eapply bindFlattenProgramSer3.    
    apply CoInd.    
    
  - match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.    
    eapply bindFlattenProgramSer2.
  - match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    
    constructor.
    intros.

    apply CoInd.
Qed.

Lemma flattenProgramSer2:
  forall (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, TransMethod M TId R -> Program M' R)
    (p : Program M R) (p': Program M' TId),

    DumOrNoopNEq
      (bindFlattenProgram
         implProg
         (instrument' R p)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (instrumentLeft' R (dumNoopProgram p))).

Proof.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.
  - intros. apply flattenProgramSer1.
  - apply CoInd.
Qed.

Lemma flattenProgramSer3:
  forall M M' TId R
    (implProg: forall R, @TransMethod M TId R -> Program M' R)
    (p : Program M R),
    
    DumOrNoopNEq
      (flattenProgram implProg (instrument R p))
      (flattenImplProgLeft' implProg (dumInstLeftImplProg instrument) (option R) (abortable p)).

Proof.
  intros.
  match goal with
  | [ |- DumOrNoopNEq ?p ?p'] =>
    rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
  end. constructor.
  apply flattenProgramSer2.
Qed.


Lemma dumNoopSerializable:
  forall M TId
    (alg: Object (@TransMethod M TId)),
    ConcSimulates
      (instFlatImpl alg instrument)
      (dumInstLeftFlatImpl alg instrument).

Proof.
  intros.
  eapply
    (SimulatesCoInd.simulatesCoInd
       (instFlatImpl alg instrument)
       (dumInstLeftFlatImpl alg instrument)
       (fun _ p1 s1 p2 s2 => 
          DumOrNoopNEq p1 p2)
       (fun s1 s2 => s1 = fst s2)
       (fun _ _ _ _ _ _ => True)); simpl in *.

  - auto.
  - reflexivity.        

  - introv s1 H2.
    inverts H2 as.
    + introv H1.
      right. right.
      exists (f tt). splits.
      * apply ProgramStepImplProgramSteps. constructor. simpl. auto.
      * auto.
      * auto.
    + introv H1.
      right. left.
      eexists. splits.
      * reflexivity.
      * assumption.
      * auto.

  - intros. split.
    + destruct m.
      simpl.
      apply flattenProgramSer3.
    + auto.
      
  - introv H1 H2. inverts H1. reflexivity.
    
  - introv H1 H2 H3. subst.
    inverts H2.
    + eexists. exists (s1', tt). splits.
      * eapply ProgramStepImplProgramSteps.
        constructor. simpl.        
        split.
        eassumption.
        destruct (snd s2). reflexivity.
      * reflexivity.
      * auto.
      * auto.        
Qed.

Lemma dumNoopSerializable':
  forall M TId (alg: Object (@TransMethod M TId)),
    ConcSimulates
      (dumInstLeftFlatImpl alg instrument)
      (instFlatImpl alg instrument).

Proof.
  intros.
  eapply
    (SimulatesCoInd2.simulatesCoInd
       (dumInstLeftFlatImpl alg instrument)
       (instFlatImpl alg instrument)
       (fun _ p1 p2 =>
          DumOrNoopNEq p2 p1)
       (fun s1 s2 => s2 = fst s1)).
  
  - reflexivity.        

  - introv H1.
    inverts H1 as.
    + introv H1.
      right. eexists. split. reflexivity. assumption.
      
  - intros.
    destruct m.
    simpl.
    apply flattenProgramSer3.

  - introv H1. inverts H1. reflexivity.
    
  - introv H1 H2 H3. subst.
    inverts H1.
    + inverts H3 as H3 H4. 
      eexists. eexists. splits.
      * eapply ProgramStepImplProgramSteps.
        constructor. simpl. eassumption.
      * auto.
      * congruence.
    + eexists. eexists. splits.
      * eapply ProgramStepImplProgramSteps.
        constructor.
      * auto.
      * inverts H3 as H3 H4. assumption.
Qed.


(* ========================================================= *)
(* To pass t on the annotation of the right hand side. *)


Definition dumInstLeftAnnImplProg {M TId}
           (instrument :
              forall R,
                Program (PairMethod M DumMethod) R ->
                Program (TransMethod (PairMethod M DumMethod) TId) (option R)) :
  forall R, (Abortable (Program M)) R ->
       Program (PairMethod (@TransMethod M TId) (AnnMethod TId DumMethod)) R :=

  fun R m =>
    match m with
    | abortable p =>
      mapProgramDep focusLeftAnn (instrument _ (dumNoopProgram p))
    end.

Definition dumInstLeftAnnFlatImpl
           {M} {TId}
           (alg : Object (@TransMethod M TId))
           (instrument :
              forall R,
                Program (PairMethod M DumMethod) R ->
                Program (TransMethod (PairMethod M DumMethod) TId) (option R)) :
  Object (Abortable (Program M)) :=

  flattenImplLeft'
    alg (Annotated TId dumObj)
    (dumInstLeftAnnImplProg instrument).

Definition dumInstLeftAnnFlatProg
           {M} {TId}
           {M'} (conc: forall R, (@TransMethod M TId) R -> Program M' R)
           (instrument :
              forall R,
                Program (PairMethod M DumMethod) R ->
                Program (TransMethod (PairMethod M DumMethod) TId) (option R))
           {R} (p: Program M R):
  Program (PairMethod M' (AnnMethod TId DumMethod)) (option R) :=
  
  flattenProgramLeft'
    conc
    (mapProgramDep focusLeftAnn (instrument R (dumNoopProgram p))).


CoInductive DumOrNoopAnnNEq {I R TId}: Program I R -> Program (PairMethod I (AnnMethod TId DumMethod)) R -> Prop :=
| bindDumOrNoopAnnNEq R' (i: I R') f f':
    (forall r, DumOrNoopAnnNEq (f r) (f' r)) ->
    DumOrNoopAnnNEq (Bind i f) (Bind (inl i) f')
| retDumOrNoopAnnNEq r:
    DumOrNoopAnnNEq (Return r) (Return r)
| noopDumOrNoopAnnNEq1 t f p:
    (forall r: unit, DumOrNoopAnnNEq p (f r)) ->
    DumOrNoopAnnNEq (Noop p) (Bind (inr (annMethod t dumCall)) f)
| noopDumOrNoopAnnNEq2 p p':
    DumOrNoopAnnNEq p p' ->
    DumOrNoopAnnNEq (Noop p) (Noop p').

Lemma bindFlattenProgramAnnSer1:
  forall (M M' : Type -> Type) (TId: Type)
    (implProg : forall R : Type, @TransMethod M TId R -> Program M' R)
    R (r: R) R' (p': Program M' R'),

    @DumOrNoopAnnNEq M' R TId
      (bindFlattenProgram
         implProg
         (fun ret => Return r)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret => mapProgramDep focusLeftAnn (Return r))).

Proof.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.
  - progFrobSimpl. constructor.
  - apply CoInd.
Qed.

Lemma bindFlattenProgramAnnSer2:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R)
    p' R (r: R),

    @DumOrNoopAnnNEq _ _ TId
      (bindFlattenProgram
         implProg
         (fun ret : unit + TId =>
            match ret with
            | inl _ => Return (Some r)
            | inr t' => _ <- abort t'; Return None
            end)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun b : unit + TId =>
            mapProgramDep
              focusLeftAnn
              match b with
              | inl _ => Return (Some r)
              | inr t' => _ <- abort t'; Return None
              end)).

Proof.
  intros.
  gen p' r.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.      
  - destruct s.
    + progFrobSimpl. constructor.
    + progFrobSimpl. constructor. apply bindFlattenProgramAnnSer1.
  - apply CoInd.
Qed.

Lemma bindFlattenProgramAnnSer3:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R),
    
    (forall R (p : Program M R) (t: TId),
        DumOrNoopAnnNEq
          (flattenProgram implProg (instrument' R p t))
          (flattenProgramLeft'
             implProg
             (instrumentLeftAnn' R (dumNoopProgram p) t))) ->

    forall R A p p',
      DumOrNoopAnnNEq
        (bindFlattenProgram
           implProg
           (fun a : option A * TId =>
              instrumentWrap'
                R
                (foldProgram
                   (stateful TId (Abortable M))
                   match fst a with
                   | Some a1 => optionProgram (@abortable M) (p a1)
                   | None => Return None
                   end
                   (snd a)))
           p')
        (bindFlattenProgramLeft'
           implProg
           p'
           (fun b : option A * TId =>
              instrumentLeftAnnWrap'
                R
                (foldProgram
                   (stateful TId (Abortable (PairMethod M DumMethod)))
                   match fst b with
                   | Some a0 =>
                     optionProgram
                       (@abortable (PairMethod M DumMethod))
                       (dumNoopProgram (p a0))
                   | None => Return None
                   end
                   (snd b)))).
Proof.
  introv H1.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.      
  - destruct p0.
    destruct o.
    + eapply H1.
    + progFrobSimpl. constructor. apply bindFlattenProgramAnnSer1.
  - apply CoInd.
Defined.

Lemma flattenProgramAnnSer1:
  forall (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, @TransMethod M TId R -> Program M' R)
    (p : Program M R) t,
    DumOrNoopAnnNEq
      (flattenProgram implProg (instrument' R p t))
      (flattenProgramLeft' implProg (instrumentLeftAnn' R (dumNoopProgram p) t)).

Proof.
  intros.
  gen R p t.    
  cofix CoInd.
  intros.
  destruct p.
  
  - match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.
    eapply bindFlattenProgramAnnSer3.
    apply CoInd.    
    
  - match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.    
    eapply bindFlattenProgramAnnSer2.
  - match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    
    constructor.
    intros.

    apply CoInd.
Qed.

Lemma flattenProgramAnnSer2:
  forall (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, TransMethod M TId R -> Program M' R)
    (p : Program M R) (p': Program M' TId),

    DumOrNoopAnnNEq
      (bindFlattenProgram
         implProg
         (instrument' R p)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (instrumentLeftAnn' R (dumNoopProgram p))).

Proof.
  intros.
  gen p'.
  cofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- DumOrNoopAnnNEq ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end;
    constructor.
  - intros. apply CoInd.
  - intros. apply flattenProgramAnnSer1.
  - apply CoInd.
Qed.

Lemma flattenProgramAnnSer3:
  forall M M' TId R
    (implProg: forall R, @TransMethod M TId R -> Program M' R)
    (p : Program M R),
    
    DumOrNoopAnnNEq
      (flattenProgram implProg (instrument R p))
      (flattenImplProgLeft' implProg (dumInstLeftAnnImplProg instrument) (option R) (abortable p)).

Proof.
  intros.
  match goal with
  | [ |- DumOrNoopAnnNEq ?p ?p'] =>
    rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
  end. constructor.
  apply flattenProgramAnnSer2.
Qed.


Lemma dumNoopAnnSerializable:
  forall M TId
    (alg: Object (@TransMethod M TId)),

    ConcSimulates
      (instFlatImpl alg instrument)
      (dumInstLeftAnnFlatImpl alg instrument).

Proof.
  intros.
  eapply
    (SimulatesCoInd.simulatesCoInd
       (instFlatImpl alg instrument)
       (dumInstLeftAnnFlatImpl alg instrument)
       (fun _ p1 s1 p2 s2 => 
          DumOrNoopAnnNEq p1 p2)
       (fun s1 s2 => s1 = fst s2)
       (fun _ _ _ _ _ _ => True)); simpl in *.

  - auto.
  - reflexivity.        

  - introv s1 H2.
    inverts H2 as.
    + introv H1.
      right. right.
      exists (f tt). splits.
      * apply ProgramStepImplProgramSteps. constructor. simpl. auto.
      * auto.
      * auto.
    + introv H1.
      right. left.
      eexists. splits.
      * reflexivity.
      * assumption.
      * auto.

  - intros. split.
    + destruct m.
      simpl.
      apply flattenProgramAnnSer3.
    + auto.
      
  - introv H1 H2. inverts H1. reflexivity.
    
  - introv H1 H2 H3. subst.
    inverts H2.
    + eexists. exists (s1', tt). splits.
      * eapply ProgramStepImplProgramSteps.
        constructor. simpl.        
        split.
        eassumption.
        destruct (snd s2). reflexivity.
      * reflexivity.
      * auto.
      * auto.        
Qed.


Lemma dumNoopAnnSerializable':
  forall M TId
    (alg: Object (@TransMethod M TId)),

    ConcSimulates
      (dumInstLeftAnnFlatImpl alg instrument)
      (instFlatImpl alg instrument).

Proof.
  intros.
  eapply
    (SimulatesCoInd2.simulatesCoInd
       (dumInstLeftAnnFlatImpl alg instrument)
       (instFlatImpl alg instrument)
       (fun _ p1 p2 => 
          DumOrNoopAnnNEq p2 p1)
       (fun s1 s2 => s2 = fst s1)); simpl in *.
       (* (fun _ _ _ _ _ _ => True)); simpl in *. *)
  
  - reflexivity.

  - introv H1.
    inverts H1 as.
    + introv H1.
      right. eexists. split. reflexivity. assumption.

  - intros.
    destruct m.
    simpl.
    apply flattenProgramAnnSer3.

  - introv H1. inverts H1. reflexivity.
    
  - introv H1 H2 H3. subst.
    inverts H1.
    + inverts H3 as H3 H4. 
      eexists. eexists. splits.
      * eapply ProgramStepImplProgramSteps.
        constructor. simpl. eassumption.
      * auto.
      * reflexivity.
    + eexists. eexists. splits.
      * eapply ProgramStepImplProgramSteps.
        constructor.
      * auto.
      * inverts H3 as H3 H4. assumption.
Qed.
