From Coq Require Import FunctionalExtensionality.
From Paco Require Import paco.
From C4 Require Import
  Annotated
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Simulates
  SimulatesFacts
  Threads
  LinCoInd
  Abortable
  ProgramEqSimulates
  FlattenSimulates
  MethodSimulates
  Serializability
  SerializabilityAnn2
  LinearizabilityFacts2
  LibTactics
  Util
  Eutt RSimulates.

(* ---------------------------------------------- *)
(* Dum Noop *)

Definition DumState := unit.

Inductive DumMethod: Type -> Type :=
| dumCall: DumMethod unit.


Definition dumObj: Spec DumMethod :=
  Build_Spec DumMethod DumState (fun _ _ _ _ _ => True) tt.

Definition dumState: dumObj.(State) := tt.

CoFixpoint dumNoopProgram
           {I R} (p: Program I R):
  Program (PairMethod I DumMethod) R :=
  match p with
  | Bind i f => Bind (inl i) (fun r => dumNoopProgram (f r))
  | Return a => Return a
  | Noop p' => Bind (inr dumCall) (fun _: unit => dumNoopProgram p')
  end.

Theorem DumNoopOpStart:
  forall I R (p: Program I R),
    OpStartProgram (dumNoopProgram p).

Proof.
  intros. destruct p; progFrobSimpl; constructor.
Qed.

Definition dumNoopImplProg
           {M I} (impl: forall R, M R -> Program I R):
  forall R, M R -> Program (PairMethod I DumMethod) R :=
  fun R m => dumNoopProgram (impl R m).


Definition dumNoopImpl
           {M} (impl: Object M):
  Object M :=
  object
    (pairSpec impl.(objBase) dumObj)
    (dumNoopImplProg impl.(objProgram)).


Lemma dumNoopProgramEqRet:
  forall I R (p : Program I R) r,
    dumNoopProgram p = Return r ->
    p = Return r.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  reflexivity.
Qed.

Lemma dumNoopProgramEqNoop:
  forall I R (p: Program I R) p',
    dumNoopProgram p = Noop p' ->
    False.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
Qed.


CoInductive DumOrNoopEq {I R}: Program I R -> Program (PairMethod I DumMethod) R -> Prop :=
| bindDumOrNoopEq R' (i: I R') f f':
    (forall r, DumOrNoopEq (f r) (f' r)) ->
    DumOrNoopEq (Bind i f) (Bind (inl i) f')
| retDumOrNoopEq r:
    DumOrNoopEq (Return r) (Return r)
| noopDumOrNoopEq1 p p':
    DumOrNoopEq p p' ->
    DumOrNoopEq (Noop p) (Bind (inr dumCall) (fun r: unit => p'))
| noopDumOrNoopEq2 p p':
    DumOrNoopEq p p' ->
    DumOrNoopEq (Noop p) (Noop p').

Lemma dumNoopProgramDumOrNoopEq:
  forall {I R} (p: Program I R),
    DumOrNoopEq
      p
      (dumNoopProgram p).
Proof.
  intros.
  gen p. cofix CoInd.
  intros. destruct p.
  - rewrite frobProgramId. simpl. constructor. auto.
  - rewrite frobProgramId. simpl. constructor.
  - rewrite frobProgramId. simpl. constructor. auto.
Defined.


Lemma bindFlattenProgramDumOrNoopEqAux:
  (forall M R (p: Program M R)
          I (iBase: forall R, M R -> Program I R),
    DumOrNoopEq
      (flattenProgram iBase p)
      (flattenProgram (dumNoopImplProg iBase) p)) ->

  forall M I (iBase: forall R, M R -> Program I R)
         R R' (f: R -> Program M R') (p: Program I R) p',
    DumOrNoopEq p p' ->
    DumOrNoopEq
      (bindFlattenProgram iBase f p)
      (bindFlattenProgram (dumNoopImplProg iBase) f p').

Proof.
  introv H1 H2.
  gen p p'. cofix CoInd.
  introv H2.
  destruct p.

  - inverts H2.
    match goal with
    | [|- DumOrNoopEq ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.

  - inverts H2.
    match goal with
    | [|- DumOrNoopEq ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.

  - inverts H2 as.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
Defined.

Lemma flattenProgramDumOrNoopEq:
  forall M R (p: Program M R)
         I (iBase: forall R, M R -> Program I R),
    DumOrNoopEq
      (flattenProgram iBase p)
      (flattenProgram (dumNoopImplProg iBase) p).

Proof.
  cofix CoInd.
  intros. destruct p.

  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.
    eapply bindFlattenProgramDumOrNoopEqAux.
    eassumption. apply dumNoopProgramDumOrNoopEq.

  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.

  - match goal with
      | [|- DumOrNoopEq ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. auto.

Qed.

Lemma bindFlattenProgramDumOrNoopEq:
  forall M I (iBase: forall R, M R -> Program I R)
         R R' (f: R -> Program M R') (p: Program I R) p',
    DumOrNoopEq p p' ->
    DumOrNoopEq
      (bindFlattenProgram iBase f p)
      (bindFlattenProgram (dumNoopImplProg iBase) f p').

Proof.
  apply bindFlattenProgramDumOrNoopEqAux.
  apply flattenProgramDumOrNoopEq.
Qed.


Lemma flattenDumNoopConcRef M M' (impl: forall R, M R -> Program M' R)
         I (iBase: forall R, M' R -> Program I R) o :
  ConcSimulates
    (object
      o
      (flattenImplProg iBase impl))
    (object
      (pairSpec o dumObj)
      (flattenImplProg (dumNoopImplProg iBase) impl)).
Proof.
  eapply (SimulatesCoInd.simulatesCoInd
    (object
      o
      (flattenImplProg iBase impl))
    (object
      (pairSpec o dumObj)
      (flattenImplProg (dumNoopImplProg iBase) impl))
    (fun _ p1 s1 p2 s2 =>
      DumOrNoopEq p1 p2)
    (fun s1 s2 => s1 = fst s2)
    (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    inverts H2 as.
    + introv H2.
    right. right. eexists. splits.
      * apply ProgramStepImplProgramSteps. simpl.
        eapply (@EBind _ _ _ _ _ tt).
        simpl. auto.
      * eassumption.
      * auto.
    + right. left. eexists. splits.
      * reflexivity.
      * assumption.
      * auto.
  - introv H1. split.
    + apply flattenProgramDumOrNoopEq.
    + auto.
  - introv H1 H2. inverts H1. reflexivity.
  - introv H1 H2 H3. inverts H2. subst.
    eexists. evar (s2s : State o). exists (s2s, tt). subst s2s. splits.
    + eapply ProgramStepImplProgramSteps. constructor. simpl. split. eassumption. destruct s2. destruct d. reflexivity.
    + reflexivity.
    + auto.
    + auto.
Qed.


CoInductive DumOrNoopEq2 {I1 I2 R}: Program (PairMethod I1 I2) R -> Program (PairMethod I1 (PairMethod I2 DumMethod)) R -> Prop :=
| bindDumOrNoopEq21 R' (i: I1 R') f f':
    (forall r, DumOrNoopEq2 (f r) (f' r)) ->
    DumOrNoopEq2
      (Bind (inl i) f)
      (Bind (inl i) f')
| bindDumOrNoopEq22 R' (i: I2 R') f f':
    (forall r, DumOrNoopEq2 (f r) (f' r)) ->
    DumOrNoopEq2
      (Bind (inr i) f)
      (Bind (inr (inl i)) f')
| retDumOrNoopEq2 r:
    DumOrNoopEq2 (Return r) (Return r)
| noopDumOrNoopEq21 p p':
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2
      (Noop p)
      (Bind (inr (inr dumCall)) (fun r: unit => p'))
| noopDumOrNoopEq22 p p':
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2 (Noop p) (Noop p').


Lemma DumOrNoopEq2Left:
  forall I1 I2 R (p: Program I1 R),
    @DumOrNoopEq2 I1 I2 R
      (mapProgram (fun _ => inl) p)
      (mapProgram (fun _ => inl) p).

Proof.
  intros. gen p.
  cofix CoInd.
  intros. destruct p.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
Qed.

Lemma DumOrNoopEq2Right:
  forall I1 I2 R (p: Program I2 R),
    @DumOrNoopEq2 I1 I2 R
      (mapProgram (fun _ => inr) p)
      (mapProgram (fun _ => inr) (dumNoopProgram p)).

Proof.
  intros.
  gen p. cofix CoInd.
  destruct p.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor.
  - match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.
  Qed.

Lemma bindFlattenProgramDumOrNoopEq2Aux:
  (forall M1 M2 R (p: Program (PairMethod M1 M2) R)
          I1 (iBase1: forall R, M1 R -> Program I1 R)
          I2 (iBase2: forall R, M2 R -> Program I2 R),
    DumOrNoopEq2
      (flattenProgram (pairImplProg iBase1 iBase2) p)
      (flattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) p)) ->

  forall M1 M2 I1 I2
         (iBase1: forall R, M1 R -> Program I1 R)
         (iBase2: forall R, M2 R -> Program I2 R)
         R R' (f: R -> Program (PairMethod M1 M2) R') (p: Program (PairMethod I1 I2) R) p',
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2
      (bindFlattenProgram (pairImplProg iBase1 iBase2) f p)
      (bindFlattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) f p').

Proof.
  introv H1 H2.
  gen p p'. cofix CoInd.
  introv H2.
  destruct p.

  - inverts H2.
    + match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.

  - inverts H2.
    match goal with
    | [|- DumOrNoopEq2 ?p ?p'] =>
      rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
    end.
    constructor. auto.

  - inverts H2 as.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
    + introv H2.
      match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
      end.
      constructor. auto.
Defined.

Lemma flattenProgramDumOrNoopEq2:
  forall M1 M2 R (p: Program (PairMethod M1 M2) R)
         I1 (iBase1: forall R, M1 R -> Program I1 R)
         I2 (iBase2: forall R, M2 R -> Program I2 R),
    DumOrNoopEq2
      (flattenProgram (pairImplProg iBase1 iBase2) p)
      (flattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) p).

Proof.
  cofix CoInd.
  intros. destruct p.

  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.
    eapply bindFlattenProgramDumOrNoopEq2Aux.
    eassumption.
    destruct p.
    eapply DumOrNoopEq2Left.
    eapply DumOrNoopEq2Right.

  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor.

  - match goal with
      | [|- DumOrNoopEq2 ?p ?p'] =>
        rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        simpl
    end.
    constructor. auto.

Qed.

Lemma bindFlattenProgramDumOrNoopEq2:
  forall M1 M2 I1 I2
         (iBase1: forall R, M1 R -> Program I1 R)
         (iBase2: forall R, M2 R -> Program I2 R)
         R R' (f: R -> Program (PairMethod M1 M2) R') (p: Program (PairMethod I1 I2) R) p',
    DumOrNoopEq2 p p' ->
    DumOrNoopEq2
      (bindFlattenProgram (pairImplProg iBase1 iBase2) f p)
      (bindFlattenProgram (pairImplProg iBase1 (dumNoopImplProg iBase2)) f p').

Proof.
  apply bindFlattenProgramDumOrNoopEq2Aux.
  apply flattenProgramDumOrNoopEq2.
Qed.


Lemma flattenDumNoopConcRefPair M M1' M2' I1' I2'
         (impl: forall R, M R -> Program (PairMethod M1' M2') R)
         (iBase1: forall R, M1' R -> Program I1' R)
         (iBase2: forall R, M2' R -> Program I2' R) o1 o2 :
  ConcSimulates
    (object
      (pairSpec o1 o2)
      (flattenImplProg
        (pairImplProg
          iBase1
          iBase2)
        impl))
    (object
      (pairSpec
        o1
        (pairSpec o2 dumObj))
      (flattenImplProg
        (pairImplProg
          iBase1
          (dumNoopImplProg iBase2))
        impl)).
Proof.
  eapply (SimulatesCoInd.simulatesCoInd
    (object
      (pairSpec o1 o2)
      (flattenImplProg
        (pairImplProg
          iBase1
          iBase2)
        impl))
    (object
      (pairSpec
        o1
        (pairSpec o2 dumObj))
      (flattenImplProg
        (pairImplProg
          iBase1
          (dumNoopImplProg iBase2))
        impl))
    (fun _ p1 s1 p2 s2 =>
      DumOrNoopEq2 p1 p2)
    (fun s1 s2 => s1 = (fst s2, fst (snd s2)))
    (fun _ _ _ _ _ _ => True)); simpl in *.

  - intros. auto.
  - reflexivity.
  - introv H1 H2.
    inverts H2 as.
    + introv H2.
    right. right. eexists. splits.
      * apply ProgramStepImplProgramSteps.
        eapply (@EBind _ _ _ _ _ tt).
        simpl. auto.
      * eassumption.
      * auto.
    + right. left. eexists. splits.
      * reflexivity.
      * assumption.
      * auto.
  - introv H1. split.
    + apply flattenProgramDumOrNoopEq2.
    + auto.
  - introv H1 H2. inverts H1. reflexivity.
  - introv H1 H2 H3. subst.
    destruct m.
    + inverts H2. simpl in *. destruct H3 as (H31 & H32).
      eexists. evar (s21 : State o1). exists (s21, snd s2). subst s21. splits.
      * eapply ProgramStepImplProgramSteps. constructor. simpl. split. eassumption. reflexivity.
      * simpl. destruct s1'. simpl in *. congruence.
      * auto.
      * auto.

    + inverts H2. simpl in *. destruct H3 as (H31 & H32). destruct s2 as [s21 [s221 s222]]. simpl in *.
      eexists. exists (s21, (snd s1', s222)). splits; simpl in *.
      * eapply ProgramStepImplProgramSteps. constructor. simpl. split. split. eassumption. reflexivity. reflexivity.
      * destruct s1'. simpl in *. congruence.
      * auto.
      * auto.
Qed.

Definition dumInstLeftImplProg
           {M}{TId}:
  forall R, (Abortable (Program M)) R ->
       Program (PairMethod (@TransMethod M TId) DumMethod) R :=

  fun R m =>
    match m with
    | abortable p =>
      instLeft (dumNoopProgram p) TId
    end.

Definition dumInstLeftFlatImpl
           {M} {TId}
           (alg: Object (@TransMethod M TId)):
  Object (Abortable (Program M)) :=

  flattenImplLeft'
    alg dumObj
    dumInstLeftImplProg.

Definition dumret {R} (d : DumMethod R) : R :=
  match d with
  | dumCall => tt
  end.

CoFixpoint nodum {I R} (p : Program (PairMethod I DumMethod) R) : Program I R :=
  match p with
  | Noop p => Noop (nodum p)
  | Return r => Return r
  | Bind (inl i) k => Bind i (fun x => nodum (k x))
  | Bind (inr d) k => Noop (nodum (k (dumret d)))
  end.

Definition DumOrNoopNEq {I R}: Program I R -> Program (PairMethod I DumMethod) R -> Prop :=
  fun p p' => eutt ieq p (nodum p').

Lemma bindFlattenProgramSer1:
  forall (M M' : Type -> Type) (TId: Type)
    (implProg : forall R : Type, @TransMethod M TId R -> Program M' R)
    R (r: R) R' (p': Program M' R'),

    DumOrNoopNEq
      (bindFlattenProgram
         implProg
         (fun ret => Return r)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret => Return r)).
Proof.
  intros.
  gen p'.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  progFrobSimpl. left; pfold; constructor.
Qed.

Lemma bindFlattenProgramSer2:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R)
    p' R (r: R),

    DumOrNoopNEq
      (bindFlattenProgram
         implProg
         (fun ret : bool * TId =>
            if fst ret
            then Return (Some r)
            else _ <- abort (snd ret); Return None)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret : bool * TId =>
            if fst ret
            then Return (Some r)
            else _ <- inl (abort (snd ret)); Return None)).

Proof.
  intros.
  gen p' r.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  destruct (fst p).
  + progFrobSimpl. left; pfold; constructor.
  + progFrobSimpl. left; pfold; constructor. left. eapply paco2_mon_bot; [ apply bindFlattenProgramSer1 | ]; auto.
Qed.

Lemma bindFlattenProgramSer3:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R),

    (forall R (p : Program M R) (t: TId),
        DumOrNoopNEq
          (flattenProgram implProg (instrument' _ p t))
          (flattenProgramLeft' implProg (instLeft' t (dumNoopProgram p)))) ->

    forall R R' p' (f : R -> Program M R'),
      DumOrNoopNEq
        (bindFlattenProgram
           implProg
           (fun ret =>
              match fst ret with
              | Some r => instrument' _ (f r) (snd ret)
              | None => _ <- abort (snd ret); Return None
              end)
           p')
        (bindFlattenProgramLeft'
           implProg
           p'
           (fun ret =>
              match fst ret with
              | Some r => instLeft' (snd ret) (dumNoopProgram (f r))
              | None => _ <- inl (abort (snd ret)); Return None
              end)).
Proof.
  introv H1.
  intros.
  gen p' f.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  destruct (fst p).
  + left. eapply paco2_mon_bot; [ eapply H1 | auto].
  + progFrobSimpl. left; pfold; constructor. left. eapply paco2_mon_bot; [ apply bindFlattenProgramSer1 | auto].
Defined.

(* ========================================================= *)
(* To pass t on the annotation of the right hand side. *)


Definition dumInstLeftAnnImplProg
           {M}{TId}:
  forall R, (Abortable (Program M)) R ->
       Program (PairMethod (@TransMethod M TId) (AnnMethod TId DumMethod)) R :=

  fun R m =>
    match m with
    | abortable p =>
      instLeftAnn (dumNoopProgram p) TId
    end.

Definition dumInstLeftAnnFlatImpl
           {M} {TId}
           (alg: Object (@TransMethod M TId)):
  Object (Abortable (Program M)) :=

  flattenImplLeft'
    alg (Annotated TId dumObj)
    dumInstLeftAnnImplProg.

Definition dumInstLeftAnnFlatProg
           {M} {TId}
           {M'} (conc: forall R, (@TransMethod M TId) R -> Program M' R)
           {R} (p: Program M R):
  Program (PairMethod M' (AnnMethod TId DumMethod)) (option R) :=

  flattenProgramLeft'
    conc
    (instLeftAnn (dumNoopProgram p) TId).

Definition anndumret {TId R} (m : AnnMethod TId DumMethod R) : R :=
  match annMethodMet m with
  | dumCall => tt
  end.

CoFixpoint noanndum {TId I R} (p : Program (PairMethod I (AnnMethod TId DumMethod)) R) : Program I R :=
  match p with
  | Noop p => Noop (noanndum p)
  | Return r => Return r
  | Bind (inl i) k => Bind i (fun x => noanndum (k x))
  | Bind (inr d) k => Noop (noanndum (k (anndumret d)))
  end.

Definition DumOrNoopAnnNEq {TId I R}
  : Program I R -> Program (PairMethod I (AnnMethod TId DumMethod)) R -> Prop :=
  fun p p' => eutt ieq p (noanndum p').

Lemma DONAN_Noop_inv_l {TId I R} (p1 : Program I R) p2 : DumOrNoopAnnNEq (Noop p1) p2 -> DumOrNoopAnnNEq (TId := TId) p1 p2.
Proof.
  unfold DumOrNoopNEq.
  apply eutt_Noop_L_inv.
Qed.

Lemma bindFlattenProgramAnnSer1:
  forall (M M' : Type -> Type) (TId: Type)
    (implProg : forall R : Type, @TransMethod M TId R -> Program M' R)
    R (r: R) R' (p': Program M' R'),

    @DumOrNoopAnnNEq TId M' R
      (bindFlattenProgram
         implProg
         (fun ret => Return r)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret => Return r)).
Proof.
  intros.
  gen p'.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  progFrobSimpl. left; pfold; constructor.
Qed.

Lemma bindFlattenProgramAnnSer2:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R)
    p' R (r: R),

    @DumOrNoopAnnNEq TId _ _
      (bindFlattenProgram
         implProg
         (fun ret : bool * TId =>
            if fst ret
            then Return (Some r)
            else _ <- abort (snd ret); Return None)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun ret : bool * TId =>
            if fst ret
            then Return (Some r)
            else _ <- inl (abort (snd ret)); Return None)).
Proof.
  intros.
  gen p' r.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p);
        rewrite (frobProgramId p');
        cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  destruct (fst p).
  + progFrobSimpl. left; pfold; constructor.
  + progFrobSimpl. left; pfold; constructor. left. eapply paco2_mon_bot; [ apply bindFlattenProgramAnnSer1 | auto ].
Qed.

Lemma bindFlattenProgramAnnSer3:
  forall (M M' : Type -> Type) TId
    (implProg : forall R : Type, TransMethod M TId R -> Program M' R),

    (forall R (p : Program M R) (t: TId),
        DumOrNoopAnnNEq
          (flattenProgram implProg (instrument' _ p t))
          (flattenProgramLeft' implProg (instLeftAnn' t (dumNoopProgram p)))) ->

    forall R R' p' (f : R -> Program M R'),
      DumOrNoopAnnNEq
        (bindFlattenProgram
           implProg
           (fun ret =>
              match fst ret with
              | Some r => instrument' _ (f r) (snd ret)
              | None => _ <- abort (snd ret); Return None
              end)
           p')
        (bindFlattenProgramLeft'
           implProg
           p'
           (fun ret =>
              match fst ret with
              | Some r => instLeftAnn' (snd ret) (dumNoopProgram (f r))
              | None => _ <- inl (abort (snd ret)); Return None
              end)).
Proof.
  introv H1.
  intros.
  gen p' f.
  pcofix SELF.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p); rewrite (frobProgramId p'); cbn
    end;
    pfold; constructor; try solve [reflexivity | right; auto].
  destruct (fst p).
  + left; eapply paco2_mon_bot; [ eapply H1 | auto ].
  + progFrobSimpl. left; pfold; constructor.
    left; eapply paco2_mon_bot; [ apply bindFlattenProgramAnnSer1 | auto ].
Defined.

Section FPAS1Proof.

Context (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, @TransMethod M TId R -> Program M' R).

Inductive FPAS1 : Program M' (option R) -> Program (PairMethod M' (AnnMethod TId DumMethod)) (option R) -> Prop :=
| FPAS1_base p t
  : FPAS1
      (flattenProgram implProg (instrument' _ p t))
      (flattenProgramLeft' implProg (instLeftAnn' t (dumNoopProgram p)))
| FPAS1_bind A p k
  : FPAS1
      (bindFlattenProgram implProg
         (fun a : option A * TId =>
          x <--
          mapProgram instrumentMethod
            (foldProgram (stateful TId (Abortable M))
               match fst a with
               | Some a0 => optionProgram (@abortable M) (k a0)
               | None => Return None
               end (snd a)); instrumentEnd R x)
         p)
       (bindFlattenProgramLeft' implProg p
            (fun ret : option A * TId =>
             match fst ret with
             | Some ret0 => instLeftAnn' (snd ret) (dumNoopProgram (k ret0))
             | None => _ <- inl (abort (snd ret)); Return None
             end))
| FPAS1_commit r0 p
  : FPAS1
      (bindFlattenProgram implProg
         (fun ret : unit + TId =>
          match ret with
          | inl _ => Return (Some r0)
          | inr t' => _ <- abort t'; Return None
          end) p)
       (bindFlattenProgramLeft' implProg p
         (fun ret : unit + TId =>
           match ret with
           | inl _ => Return (Some r0)
           | inr t' => _ <- inl (abort t'); Return None
           end))
| FPAS1_abort p
  : FPAS1
      (bindFlattenProgram implProg (fun _ : unit => Return None) p)
      (bindFlattenProgramLeft' implProg p (fun _ : unit => Return None))
.
Hint Constructors FPAS1 : core.

Lemma flattenProgramAnnSer1_: forall p q,
    FPAS1 p q ->
    DumOrNoopAnnNEq p q.
Proof.
  pcofix SELF.
  intros _ _ [].
  all: match goal with
  | [ |- _ ?p ?p'] => rewrite (frobProgramId p); rewrite (frobProgramId p')
  end.
  - destruct p; cbn; pfold; constructor; right; auto.
    apply SELF; constructor.
  - destruct p; cbn; pfold; constructor; auto.
    match goal with
    | [ |- _ ?p _ ] => rewrite (frobProgramId p); cbn
    end. destruct p as [[] ?]; cbn.
    { rewrite <- frobProgramId.
      right; apply SELF; constructor. }
    { match goal with
      | [ |- _ ?p ] => rewrite (frobProgramId p); cbn
      end.
      left; pfold; constructor; right; auto.
    }
  - destruct p; cbn; pfold; constructor; auto.
    left.
    match goal with
    | [ |- _ ?p ?p'] => rewrite (frobProgramId p); rewrite (frobProgramId p')
    end.
    destruct s; cbn.
    { pfold; constructor. }
    { pfold; constructor; auto. }
  - destruct p; cbn; pfold; constructor; auto.
    match goal with
    | [ |- _ ?p ?p'] => rewrite (frobProgramId p); rewrite (frobProgramId p'); cbn
    end.
    left; pfold; constructor.
Qed.

Lemma flattenProgramAnnSer1: forall (p : Program M R) t,
    DumOrNoopAnnNEq
      (flattenProgram implProg (instrument' _ p t))
      (flattenProgramLeft' implProg (instLeftAnn' t (dumNoopProgram p))).
Proof.
  intros. apply flattenProgramAnnSer1_; eauto.
Qed.

End FPAS1Proof.

Lemma flattenProgramAnnSer2:
  forall (M M' : Type -> Type) (TId R : Type)
    (implProg : forall R, TransMethod M TId R -> Program M' R)
    (p : Program M R) (p': Program M' TId),

    DumOrNoopAnnNEq
      (bindFlattenProgram
         implProg
         (fun t => instrument' _ p t)
         p')
      (bindFlattenProgramLeft'
         implProg
         p'
         (fun t : TId => instLeftAnn' t (dumNoopProgram p))).

Proof.
  intros.
  gen p'.
  pcofix CoInd.
  intros.
  destruct p';
    match goal with
    | [ |- _ ?p ?p'] =>
      rewrite (frobProgramId p); rewrite (frobProgramId p'); cbn
    end;
    pfold; constructor; try solve [ reflexivity | right; auto].
  left; eapply paco2_mon_bot; [ apply flattenProgramAnnSer1 | auto ].
Qed.

Lemma flattenProgramAnnSer3:
  forall M M' TId R
    (implProg: forall R, @TransMethod M TId R -> Program M' R)
    (p : Program M R),

    DumOrNoopAnnNEq
      (flattenProgram implProg (instrument _ p))
      (flattenProgramLeft' implProg (instLeftAnn (dumNoopProgram p) TId)).

Proof.
  intros. red.
  match goal with
  | [ |- _ ?p ?p'] =>
    rewrite (frobProgramId p);
      rewrite (frobProgramId p');
      simpl
  end; pfold; constructor.
  left; eapply paco2_mon_bot; [ apply flattenProgramAnnSer2 | auto ].
Qed.

Inductive dumSteps_ {M TId S R}
    (e : M S) (k : S -> Program (PairMethod M (AnnMethod TId DumMethod)) R)
  : Program (PairMethod M (AnnMethod TId DumMethod)) R -> Prop :=
| dumSteps_Noop p
  : dumSteps_ e k p -> dumSteps_ e k (Noop p)
| dumSteps_Bind_r X (z : AnnMethod TId DumMethod X) k'
  : dumSteps_ e k (k' (anndumret z)) -> dumSteps_ e k (Bind (inr z) k')
| dumSteps_Bind_l
  : dumSteps_ e k (Bind (inl e) k)
.

Lemma DONANE_dumSteps {M TId S R} (e : M S) (k : S -> Program M R)
    (p2 : Program (PairMethod M (AnnMethod TId DumMethod)) R)
  : DumOrNoopAnnNEq (x <- e; k x) p2 ->
    exists k',
      dumSteps_ e k' p2 /\ (forall x, DumOrNoopAnnNEq (k x) (k' x)).
Proof.
  unfold DumOrNoopAnnNEq; remember (Bind e k); remember (noanndum p2); intros EE.
  revert S e k Heqp p2 Heqp0.
  punfold EE; induction EE; intros; try discriminate.
  - rewrite (frobProgramId (noanndum _)) in Heqp0; destruct p2 as [ ? [] k2' | | ]; cbn in Heqp0; try discriminate.
    inversion H; inversion Heqp0; inversion Heqp; subst; injpair; clear H Heqp Heqp0.
    exists k2'; split; [ constructor | pclearbot; assumption ].
  - rewrite (frobProgramId (noanndum _)) in Heqp0; destruct p0 as [ ? [] | | ]; cbn in Heqp0; try discriminate.
    + inversion Heqp0; subst.
      edestruct IHEE as (k' & ? & ?); try reflexivity.
      exists k'; split; auto.
      constructor; assumption.
    + inversion Heqp0; subst.
      edestruct IHEE as (k' & ? & ?); try reflexivity.
      exists k'; split; auto.
      constructor; assumption.
Qed.

Lemma dumNoopAnnSerializable M TId (alg: Object (@TransMethod M TId)) :
    ConcSimulates
      (instFlatImpl alg instrument)
      (dumInstLeftAnnFlatImpl alg).
Proof.
  apply SMS_ConcS.
  constructor 1 with (RS := fun s s' => s = fst s') (RR := fun R => DumOrNoopAnnNEq); cbn.
  - auto.
  - unfold flattenImplProgLeft'.
    unfold flattenImplProg.
    intros ? []; cbn.
    apply flattenProgramAnnSer3.
  - intros.
    induction H1 as [ * ? IH | ].
    + apply DONAN_Noop_inv_l in H0.
      specialize (IH H0). auto.
    + subst. apply DONANE_dumSteps in H0. destruct H0 as [k' [STEPS HH]].
      induction STEPS.
      * destruct IHSTEPS as (p1'' & s2' & p2' & EUTT & PSTEPS & -> & EE).
        exists p1'' s2' p2'; split; eauto.
        split; eauto. eapply ProgramStepStepsImplSteps; [ constructor | eauto ].
      * destruct IHSTEPS as (p1'' & s2' & p2' & EUTT & PSTEPS & -> & EE).
        exists p1'' s2' p2'; split; eauto.
        split; eauto. eapply ProgramStepStepsImplSteps; [ constructor | eauto ].
        constructor; constructor.
      * exists (k r0) (s1', tt) (k' r0).
        split; [ apply Reflexive_eutt_ieq | ].
        split; auto.
        eapply ProgramStepImplProgramSteps; constructor.
        constructor; auto. destruct s2 as [ ? []]; reflexivity.
  - intros. apply Symmetric_eutt_ieq in H0.
    apply eutt_Ret_inv in H0.
    induction H0 as [ * ? IH | ].
    + apply DONAN_Noop_inv_l in H.
      specialize (IH H); auto.
    + subst. apply eutt_Ret_inv in H. remember (noanndum p2) in H.
      revert p2 Heqp.
      induction H; cbn.
      * intros p2; rewrite (frobProgramId (noanndum p2)); destruct p2 as [ ? [] | | ]; cbn; try discriminate.
        all: injection 1; intros ->.
        all: eapply ProgramStepStepsImplSteps; [ constructor | eauto ].
        constructor; constructor.
      * intros p2 ->. rewrite (frobProgramId (noanndum p2)) in H; destruct p2 as [ ? [] | | ]; cbn in H; try discriminate.
        inversion H. apply ProgramStepsRefl.
Qed.
