From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program.

Section LockSec.
  
  Definition LockState := bool.
  
  Inductive LockMethod: Type -> Type :=
  | doLock: LockMethod unit
  | doUnlock: LockMethod unit.
  
  Inductive LockSem:
    forall {R}, LockMethod R -> LockState -> R -> LockState -> Prop :=
  | lockSem:
      LockSem doLock false tt true
  | unlockSem:
      LockSem doUnlock true tt false.

  Definition atomicLock : Spec LockMethod :=
    Build_Spec LockMethod LockState (@LockSem) false.

  Lemma invertLockSem R m s r s':
    @LockSem R m s r s' ->
    
    match m in LockMethod R' return R' -> Prop with
    | doLock =>
      fun r =>
        r = tt
        /\ s = false
        /\ s' = true
              
    | doUnlock =>
      fun r =>
        r = tt
        /\ s = true
        /\ s' = false

    end r.

  Proof.
    intros H1.
    destruct H1; auto.
  Qed.

  
End LockSec.


Section TLockSec.

  Context {TId: Type}.  
  Definition TLockState := option TId.


  Inductive TLockMethod: Type -> Type :=
  | doTLock: TId -> TLockMethod unit
  | doTUnlock: TId -> TLockMethod unit.
  
  Inductive TLockSem:
    forall {R}, TLockMethod R -> TLockState -> R -> TLockState -> Prop :=
  | lockTSem t:
      TLockSem (doTLock t) None tt (Some t)
  | lockTSem' t:
      TLockSem (doTLock t) (Some t) tt (Some t)               
  | unlockTSem t:
      TLockSem (doTUnlock t) (Some t) tt None.

  Definition atomicTLock: Spec TLockMethod :=
    Build_Spec TLockMethod TLockState (@TLockSem) None.
  
  Lemma invertTLockSem R m s r s':
    @TLockSem R m s r s' ->
    
    match m in TLockMethod R' return R' -> Prop with
    | doTLock t =>
      fun r =>
        r = tt
        /\ (s = None \/ s = Some t)
        /\ s' = Some t
              
    | doTUnlock t =>
      fun r =>
        r = tt
        /\ s = Some t
        /\ s' = None

    end r.

  Proof.
    intros H1.
    destruct H1; auto.
  Qed.

End TLockSec.
