From C4 Require Import
  Util.LibTactics
  Core.Program.

Class Default (Val : Type) : Type :=
  default : Val.

Section RegisterSec.
  
  Context {Val: Type}.

  Definition RegisterState := Val.
  
  Inductive RegisterMethod: Type -> Type :=
  | read: RegisterMethod Val
  | write: Val -> RegisterMethod unit
  | cas: Val -> Val -> RegisterMethod bool.
  
  Inductive RegSem:
    forall {A}, RegisterMethod A -> RegisterState -> A -> RegisterState -> Prop :=
  | readSem v:
      RegSem read v v v
  | writeSem v v':
      RegSem (write v') v tt v'
  | casTrueSem v v':
      RegSem (cas v v') v true v' 
  | casFalseSem v v' v'':
      v <> v' ->
      RegSem (cas v' v'') v false v.

  Definition atomicReg `{Default Val} : Spec RegisterMethod :=
    Build_Spec RegisterMethod RegisterState (@RegSem) default.

  Lemma invertRegSem R m s r s':
    @RegSem R m s r s' ->
    
    match m in RegisterMethod R' return R' -> Prop with
    | read =>
      fun r =>
        r = s
        /\ s = s'
              
    | write v' =>
      fun r =>
        r = tt
        /\ s' = v'
                 
    | cas v v' =>
      fun r =>
        (r = true /\ s = v /\ s' = v') \/
        (r = false /\ s <> v /\ s' = s)

    end r.

  Proof.
    intros H1.
    destruct H1; auto.
  Qed.

End RegisterSec.

Instance Default_nat : Default nat := 0.
