Set Asymmetric Patterns.
Require Import
        Coq.Sets.Ensembles
        Coq.Lists.List.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Pair
  Program
  ProgramOps
  Object
  Threads
  Simulates
  SimulatesFacts
  Abortable
  Linearizability
  Serializability.
From C4.App Require Import
  Annotated.

Section SerializabilityLeft.

  Definition focusLeft {M' M TLocal} A (m : TransMethod (PairMethod M M') TLocal A) :
    let B := match m with
             | liftMethod A' t (inl _) => A
             | liftMethod A' t (inr _) => A'
             | _ => A
             end in
    PairMethod (TransMethod M TLocal) M' B * (B -> A) :=
    match m with
    | init => (inl init, fun a => a)
    | abort t => (inl (abort t), fun a => a)
    | commit t => (inl (commit t), fun a => a)
    | liftMethod _ t (inl m') => (inl (liftMethod t m'), fun a => a)
    | liftMethod _ t (inr m') => (inr m', fun a' => (Some a', t))
    end.

  Definition instrumentLeftWrap' {M' M TLocal} R
             (p : Program (Stateful TLocal (Abortable (PairMethod M M'))) (option R * TLocal)) :
    Program (PairMethod (TransMethod M TLocal) M') (option R) :=
    mapProgramDep focusLeft (instrumentWrap' R p).

  Definition instrumentLeft' {M' M TLocal} R (p : Program (PairMethod M M') R) (t : TLocal) :
    Program (PairMethod (TransMethod M TLocal) M') (option R) :=
    mapProgramDep focusLeft (instrument' R p t).

  Definition instrumentLeft {M' M TLocal} R (p : Program (PairMethod M M') R) :
    Program (PairMethod (TransMethod M TLocal) M') (option R) :=
    mapProgramDep focusLeft (instrument R p).

  (* ========================================================= *)
  (* To pass t on the annotation of the right hand side. *)

  Definition focusLeftAnn {M' M TLocal} A (m : TransMethod (PairMethod M M') TLocal A) :
    let B := match m with
             | liftMethod A' t (inl _) => A
             | liftMethod A' t (inr _) => A'
             | _ => A
             end in
    PairMethod (TransMethod M TLocal) (AnnMethod TLocal M') B * (B -> A) :=
    match m with
    | init => (inl init, fun a => a)
    | abort t => (inl (abort t), fun a => a)
    | commit t => (inl (commit t), fun a => a)
    | liftMethod _ t (inl m') => (inl (liftMethod t m'), fun a => a)
    | liftMethod _ t (inr m') => (inr (annMethod t m'), fun a' => (Some a', t))
    end.

  Definition instrumentLeftAnnWrap' {M' M TLocal} R
             (p : Program (Stateful TLocal (Abortable (PairMethod M M'))) (option R * TLocal)) :
    Program (PairMethod (TransMethod M TLocal) (AnnMethod TLocal M')) (option R) :=
    mapProgramDep focusLeftAnn (instrumentWrap' R p).

  Definition instrumentLeftAnn' {M' M TLocal} R (p : Program (PairMethod M M') R) (t : TLocal) :
    Program (PairMethod (TransMethod M TLocal) (AnnMethod TLocal M')) (option R) :=
    mapProgramDep focusLeftAnn (instrument' R p t).

  Definition instrumentLeftAnn {M' M TLocal} R (p : Program (PairMethod M M') R) :
    Program (PairMethod (TransMethod M TLocal) (AnnMethod TLocal M')) (option R) :=
    mapProgramDep focusLeftAnn (instrument R p).

End SerializabilityLeft.


Section SerializabilityRight.

  Definition focusRight {M' M TLocal} A (m : TransMethod (PairMethod M' M) TLocal A) :
    let B := match m with
             | liftMethod A' t (inl _) => A'
             | liftMethod A' t (inr _) => A
             | _ => A
             end in
    PairMethod M' (TransMethod M TLocal) B * (B -> A) :=
    match m with
    | init => (inr init, fun a => a)
    | abort t => (inr (abort t), fun a => a)
    | commit t => (inr (commit t), fun a => a)
    | liftMethod _ t (inl m') => (inl m', fun a' => (Some a', t))
    | liftMethod _ t (inr m') => (inr (liftMethod t m'), fun a => a)
    end.

  Definition instrumentRight {M' M} R (p : Program (PairMethod M' M) R) TLocal :
    Program (PairMethod M' (TransMethod M TLocal)) (option R) :=
    mapProgramDep focusRight (instrument R p).

End SerializabilityRight.
