From Coq Require Import Ensembles List.
From C4 Require Import
  Annotated
  Pair
  Program
  ProgramOps
  Object
  Threads
  Linearizability
  Serializability
  Abortable.

Section SerializabilityLeft.

  CoFixpoint instLeft'
             {M' M R} {TId}
             (t: TId) (p: Program (PairMethod M M') R):
    Program (PairMethod (@TransMethod M TId) M') (option R) :=

    match p with

    | Bind (inl m) f =>

      ret <- inl (liftMethod t m);
      let retOpt := fst ret in
      let t' := snd ret in      
      match retOpt with
      | Some ret =>
        instLeft' t' (f ret)
      | None =>
        _ <- inl (abort t');
        Return None
      end

    | Bind (inr m) f =>
      r <- inr m;
      instLeft' t (f r)
        
    | Return r =>
      ret <- inl (commit t);
      match ret with
      | inl _ => Return (Some r)
      | inr t' =>
        _ <- inl (abort t');
        Return None
      end
               
    | Noop p' =>
      Noop (instLeft' t p')
    end.

  
  Definition instLeft
             {M' M R}
             (p: Program (PairMethod M M') R) TId:
    Program (PairMethod (@TransMethod M TId) M') (option R) :=

    t <- inl (@init M TId);
    instLeft' t p.

  
  Definition instLeftImplProg
             {M M'}{TId}:
    forall R, (Abortable (Program (PairMethod M M'))) R ->
         Program (PairMethod (@TransMethod M TId) M') R :=
    
    fun R m' =>
       match m' with
       | abortable p =>
         instLeft p TId
       end.
  
  
  Definition instLeftFlatImpl
             {M M'} {TId}
             (alg: Object (@TransMethod M TId))
             (o: Spec M'):
    Object (Abortable (Program (PairMethod M M'))) :=

    flattenImplLeft
      alg o
      instLeftImplProg.

  Definition StrictSerializableLeft
             {TId}
             {M Instruction}
             (o: Spec Instruction)
             (rObj: Spec M)             
             (alg: Object (@TransMethod Instruction TId)) : Prop :=
    Linearizable
      (abortableSpec (programSpec (pairSpec o rObj)))
      (instLeftFlatImpl alg rObj).

  (* To inst and flatten a program. *)  
  Definition instLeftFlatProg
             {M MR M' R} {TId}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R)
             (p: Program (PairMethod M MR) R):
    Program (PairMethod M' MR) (option R) :=
    
    flattenProgramLeft
      conc
      (instLeft p TId).


  (* To instrument' and flatten a program. *)
  Definition instLeftFlatProg'
             {M MR} {TId} {M'}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R)
             (t: TId) {R} (p: Program (PairMethod M MR) R):
    Program (PairMethod M' MR) (option R) :=
    
    flattenProgramLeft
      conc
      (instLeft' t p).


  (* To instrument and flatten the objProgram of an Object. *)
  Definition instLeftFlatImplProg
             {TId}{M MR}{M'}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R):
    forall R, Abortable (Program (PairMethod M MR)) R -> Program (PairMethod M' MR) R :=

    flattenImplProgLeft conc instLeftImplProg.

  (* ========================================================= *)
  (* To pass t on the annotation of the right hand side. *)

  CoFixpoint instLeftAnn' {M' M R} {TId}
             (t: TId) (p: Program (PairMethod M M') R):
    Program (PairMethod (@TransMethod M TId) (AnnMethod TId M')) (option R) :=
    match p with
    | Bind (inl m) f =>
      ret <- inl (liftMethod t m);
      let retOpt := fst ret in
      let t' := snd ret in
      match retOpt with
      | Some ret =>
        instLeftAnn' t' (f ret)
      | None =>
        _ <- inl (abort t');
        Return None
      end

    | Bind (inr m) f =>
      r <- inr (annMethod t m);
      instLeftAnn' t (f r)

    | Return r =>
      ret <- inl (commit t);
      match ret with
      | inl _tt => Return (Some r)
      | inr t' =>
        _ <- inl (abort t');
        Return None
      end

    | Noop p' =>
      Noop (instLeftAnn' t p')
    end.

  Definition instLeftAnn {M' M R} (p: Program (PairMethod M M') R) TId:
    Program (PairMethod (@TransMethod M TId) (AnnMethod TId M')) (option R) :=
    t <- inl (@init M TId);
    instLeftAnn' t p.
  
  Definition instLeftAnnImplProg
             {M M'}{TId}:
    forall R, (Abortable (Program (PairMethod M M'))) R ->
         Program (PairMethod (@TransMethod M TId) (AnnMethod TId M')) R :=
    
    fun R m' =>
       match m' with
       | abortable p =>
         instLeftAnn p TId
       end.
  
  
  Definition instLeftAnnFlatImpl
             {M M'} {TId}
             (alg: Object (@TransMethod M TId))
             (o: Spec M'):
    Object (Abortable (Program (PairMethod M M'))) :=

    flattenImplLeft
      alg (Annotated TId o)
      instLeftAnnImplProg.


  Definition StrictSerializableLeftAnn
             {TId}
             {M Instruction}
             (o: Spec Instruction)
             (rObj: Spec M)             
             (alg: Object (@TransMethod Instruction TId)) :=

    Linearizable
      (abortableSpec (programSpec (pairSpec o rObj)))
      (instLeftAnnFlatImpl alg rObj).

  (* To inst and flatten a program. *)  
  Definition instLeftAnnFlatProg
             {M MR M' R} {TId}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R)
             (p: Program (PairMethod M MR) R):
    Program (PairMethod M' (AnnMethod TId MR)) (option R) :=
    
    flattenProgramLeft
      conc
      (instLeftAnn p TId).


  (* To instrument' and flatten a program. *)
  Definition instLeftAnnFlatProg'
             {M MR} {TId} {M'}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R)
             (t: TId) {R} (p: Program (PairMethod M MR) R):
    Program (PairMethod M' (AnnMethod TId MR)) (option R) :=
    
    flattenProgramLeft
      conc
      (instLeftAnn' t p).


  (* To instrument and flatten the objProgram of an Object. *)
  Definition instLeftAnnFlatImplProg
             {TId}{M MR}{M'}
             (conc: forall R, (@TransMethod M TId) R -> Program M' R):
    forall R, Abortable (Program (PairMethod M MR)) R -> Program (PairMethod M' (AnnMethod TId MR)) R :=

    flattenImplProgLeft conc instLeftAnnImplProg.

End SerializabilityLeft.
