From C4 Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  Abortable
  Linearizability
  Serializability
  LibTactics
  Util.

Inductive wrapMethod
          (M: Type -> Type)
          (pred: forall (R: Type), M R -> Prop): Type -> Type :=
| wrapMethod1 R (m: M R):
    pred R m ->
    wrapMethod M pred R.


Definition wrapImpl
           {M}
           (pred: forall (R: Type), M R -> Prop)
           (obj: Object M):
  Object (wrapMethod M pred) :=
  
  object
    (objBase obj)
    (fun R (m: wrapMethod M pred R) =>
       match m with
       | wrapMethod1 _ _ R m _ => 
         objProgram obj _ m
       end).

Definition ConcSimulatesWrap
           {M}
           (pred: forall (R: Type), M R -> Prop)
           (obj obj': Object M) :=
  ConcSimulates
    (wrapImpl pred obj)
    (wrapImpl pred obj').

(*
    Lemma ConcRefWrapImplConRef:
      forall M
        (pred: forall (R: Type), M R -> Prop)
        (obj: Object M)
        (s: obj.(objBase).(State)),

        ConcSimulates
          (wrapImpl pred obj)
          obj
          s s.
 *)

Fixpoint mapInteraction {M1 M2} (f : forall X, M1 X -> M2 X) {R} (m : Interaction M1 R)
  : Interaction M2 R :=
  match m with
  | Spawn m => Spawn (f _ m)
  | Join => Join
  | Up m => Up (mapInteraction f m)
  end.

Definition unwrap {M} {pred : forall R, M R -> Prop} R (m : wrapMethod M pred R)
  : M R :=
  match m with
  | wrapMethod1 _ _ _ m _ => m
  end.

Lemma ThreadAdvance_Wrap {M} (obj1 : Object M) pred ts1 ts2
  : ThreadAdvance (wrapImpl pred obj1) ts1 ts2 ->
    ThreadAdvance obj1 ts1 ts2.
Proof.
  induction 1; cbn; try constructor; auto.
Qed.

Lemma ThreadsAdvance_Wrap {M} (obj1 : Object M) pred ts1 ts2
  : ThreadsAdvance (wrapImpl pred obj1) ts1 ts2 ->
    ThreadsAdvance obj1 ts1 ts2.
Proof.
  induction 1; econstructor.
  2: eapply IHclos_refl_trans_1n.
  apply ThreadAdvance_Wrap; auto.
Qed.

Lemma ThreadsInteraction_Wrap {M} (obj1 : Object M) pred ts1 ts2 R (r : R) int
  : ThreadsInteraction (wrapImpl pred obj1) R int ts1 r ts2 ->
    ThreadsInteraction obj1 R (mapInteraction unwrap int) ts1 r ts2.
Proof.
  intros [* H]; cbn; constructor.
  induction H; cbn.
  1: destruct m.
  all: constructor; auto.
Qed.

Lemma ThreadAdvance_Wrap_unwrap {M} (obj1 : Object M) pred ts1 ts2
  : ThreadAdvance obj1 ts1 ts2 ->
    ThreadAdvance (wrapImpl pred obj1) ts1 ts2.
Proof.
  induction 1; cbn; try constructor; auto.
  - apply (TACons (Method := wrapMethod M pred)). auto.
  - apply (TABind (Method := wrapMethod M pred)). auto.
  - apply (TANoop (Method := wrapMethod M pred)).
Qed.

Lemma ThreadsAdvance_Wrap_unwrap {M} (obj1 : Object M) pred ts1 ts2
  : ThreadsAdvance obj1 ts1 ts2 ->
    ThreadsAdvance (wrapImpl pred obj1) ts1 ts2.
Proof.
  induction 1; econstructor.
  2: eapply IHclos_refl_trans_1n.
  apply ThreadAdvance_Wrap_unwrap; auto.
Qed.

Lemma ThreadsInteraction_Wrap_unwrap {M} (obj1 : Object M) pred ts1 ts2 R (r : R) int
  : ThreadsInteraction obj1 R (mapInteraction unwrap int) ts1 r ts2 ->
    ThreadsInteraction (wrapImpl pred obj1) R int ts1 r ts2.
Proof.
  intros [* H]; constructor 1 with (Method := wrapMethod _ _).
  revert ps ps' H. induction int; intros * H; inversion H; injpair; subst.
  - destruct m; cbn. eapply TISpawn with (impl := wrapImpl pred obj1).
  - eapply TIJoin with (impl := wrapImpl pred obj1).
  - eapply TICons with (impl := wrapImpl pred obj1); auto.
Qed.

(*
Lemma AdvInt {M} (obj : Object M) R (int : _ R) (r : R) ts1 ts2 ts3
    (H : ThreadsAdvance obj ts1 ts2)
    (H0 : ThreadsInteraction obj R int ts2 r ts3)
  : ThreadsInteraction obj R int ts1 r ts3.
Proof.
  inversion H0; subst.
  destruct ts1; econstructor.

Qed.
*)

Lemma EvalInstruction_Wrap {M} (obj1 : Object M) pred ts1 ts2 R (r : R) int
  : EvalInstruction (ThreadsBehavior (wrapImpl pred obj1)) int ts1 r ts2 ->
    EvalInstruction (ThreadsBehavior obj1) (mapInteraction unwrap int) ts1 r ts2.
Proof.
  intros [].
  econstructor.
  - apply ThreadsAdvance_Wrap; eassumption.
  - apply ThreadsInteraction_Wrap; eassumption.
Qed.

Lemma EvalInstruction_Wrap_unwrap {M} (obj1 : Object M) pred ts1 ts2 R (r : R) int
  : EvalInstruction (ThreadsBehavior obj1) (mapInteraction unwrap int) ts1 r ts2 ->
    EvalInstruction (ThreadsBehavior (wrapImpl pred obj1)) int ts1 r ts2.
Proof.
  intros [].
  econstructor.
  - apply ThreadsAdvance_Wrap_unwrap; eassumption.
  - apply ThreadsInteraction_Wrap_unwrap; eassumption.
Qed.

Lemma Simulates_Wrap_ (M : Type -> Type) (pred : forall R : Type, M R -> Prop)
    (obj1 obj2 : Object M)
  : forall ts1 ts2,
    Simulates_ (ThreadsBehavior obj1) (ThreadsBehavior obj2)
      ts1 ts2 ->
    Simulates_ (ThreadsBehavior (wrapImpl pred obj1)) (ThreadsBehavior (wrapImpl pred obj2))
      ts1 ts2.
Proof.
  cofix SELF; intros ? ? [STEP]. constructor; intros * EVAL.
  apply EvalInstruction_Wrap in EVAL.
  apply STEP in EVAL.
  destruct EVAL; econstructor; eauto.
  apply EvalInstruction_Wrap_unwrap.
  auto.
Qed.

Lemma ConcRefWrapImplConRef:
  forall M
    (pred: forall (R: Type), M R -> Prop)
    (obj1 obj2: Object M),
    ConcSimulates obj1 obj2 ->
    ConcSimulatesWrap pred obj1 obj2.
Proof.
  unfold ConcSimulatesWrap, ConcSimulates.
  intros *; apply Simulates_Wrap_.
Qed.

Definition LinearizableWrap
           {M}
           (pred: forall (R: Type), M R -> Prop)
           (spec: Spec M)
           (obj: Object M) :=
  ConcSimulatesWrap pred obj (serialObject spec).

Definition StrictSerializableWrap
           {TId} {M}
           (pred: forall (R: Type), Abortable (Program M) R -> Prop)
           (o: Spec M)
           (conc: Object (@TransMethod M TId)) : Prop :=
  LinearizableWrap
    pred
    (abortableSpec (programSpec o))
    (instFlatImpl conc instrument).

Definition SelfStrictSerializableWrap
           {TId M}
           (pred: forall (R: Type), Abortable (Program M) R -> Prop)
           (conc: Object (@TransMethod M TId)) : Prop :=
  LinearizableWrap
    pred
    (coarsen (instFlatImpl conc instrument))
    (instFlatImpl conc instrument).
