From Coq Require Import List.

From Paco Require Import paco.

From C4 Require Import
  Util
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  Linearizability
  Abortable
  Serializability
  LibTactics
  Wrap
  LinearizabilityFacts2
  Eutt RSimulates.

Set Implicit Arguments.

Lemma Proper_bindProgram {M R S} (p1 p2 : Program M R) (k1 k2 : R -> Program M S)
  : eutt ieq p1 p2 -> (forall x, eutt ieq (k1 x) (k2 x)) ->
    eutt ieq (bindProgram p1 k1) (bindProgram p2 k2).
Proof.
  intros Hp Hk; revert p1 p2 Hp; pcofix SELF; intros.
  punfold Hp; induction Hp;
    match goal with
    | [ |- _ ?t1 ?t2 ] =>
      rewrite (frobProgramId t1), (frobProgramId t2); cbn
    end.
  1:{ unfold frobProgram; cbn; fold (frobProgram (k1 r0)); fold (frobProgram (k2 r0)).
    rewrite <- 2 frobProgramId; specialize (Hk r0).
    eapply (paco2_mon _ Hk); contradiction. }
  all: rewrite <- ?frobProgramId; pfold; constructor; eauto.
  1,2: right; apply SELF; pclearbot; auto; apply H0.
  1,2: pfold_reverse; auto.
Qed.

Lemma Proper_mapProgram {M N R} (f : forall X, M X -> N X)
  : forall (p1 p2 : Program M R), eutt ieq p1 p2 -> eutt ieq (mapProgram f p1) (mapProgram f p2).
Proof.
  pcofix SELF; intros p1 p2 Hp.
  punfold Hp; induction Hp;
    match goal with
    | [ |- _ ?t1 ?t2 ] =>
      rewrite (frobProgramId t1), (frobProgramId t2); cbn; rewrite <- ?frobProgramId
    end; pfold; constructor; auto.
  - destruct H; auto.
  - right; apply SELF; pclearbot; apply H0.
  - right; apply SELF; pclearbot; apply H.
  - pfold_reverse; auto.
  - pfold_reverse; auto.
Qed.

Lemma Proper_foldProgram (M M' : Type -> Type) (S : Type)
    (alg : forall A : Type, M' A -> S -> M (A * S)%type)
    (Ret : Type)
  : forall (s : S) (p1 p2 : Program M' Ret),
    eutt ieq p1 p2 -> eutt ieq (foldProgram alg p1 s) (foldProgram alg p2 s).
Proof.
  pcofix SELF; intros s p1 p2 Hp.
  punfold Hp; induction Hp;
    match goal with
    | [ |- _ ?t1 ?t2 ] =>
      rewrite (frobProgramId t1), (frobProgramId t2); cbn; rewrite <- ?frobProgramId
    end; pfold; constructor; auto.
  - destruct H; auto.
  - right; apply SELF; pclearbot; apply H0.
  - right; apply SELF; pclearbot; apply H.
  - pfold_reverse; auto.
  - pfold_reverse; auto.
Qed.

Lemma Proper_optionProgram (M M' : Type -> Type) (f : forall A, M A -> M' (option A)) (R : Type)
  : forall (p1 p2 : Program M R), eutt ieq p1 p2 ->
    eutt ieq (optionProgram f p1) (optionProgram f p2).
Proof.
  pcofix SELF; intros p1 p2 Hp.
  punfold Hp; induction Hp;
    match goal with
    | [ |- _ ?t1 ?t2 ] =>
      rewrite (frobProgramId t1), (frobProgramId t2); cbn; rewrite <- ?frobProgramId
    end; pfold; constructor; auto.
  - destruct H; auto.
  - intros [ | ]; [ right; apply SELF; pclearbot; apply H0 | left; pfold; constructor ].
  - right; apply SELF; pclearbot; apply H.
  - pfold_reverse; auto.
  - pfold_reverse; auto.
Qed.

Lemma Proper_instrument {M T R} p1 p2 :
  eutt ieq p1 p2 -> eutt ieq (@instrument M T R p1) (instrument _ p2).
Proof.
  intros H.
  unfold instrument, instrumentWrap, instrumentBody, instrumentWrap'.
  pfold; constructor; [ constructor | intros ?; left ].
  eapply Proper_bindProgram; [ | intros; apply Reflexive_eutt_ieq ].
  apply Proper_mapProgram, Proper_foldProgram, Proper_optionProgram, H.
Qed.

Section SSU.

Context
  {TId} {M}
  (pred: forall (R: Type), Abortable (Program M) R -> Prop)
  (o: Spec M)
  (conc: Object (@TransMethod M TId))
  (PRED : forall R p, exists p', pred R p' /\ eq_Abortable (eutt ieq) _ p' p).

(* Spec object *)
Notation so := (serialObject (abortableSpec (programSpec o))).
(* Conc object *)
Notation co := (instFlatImpl conc instrument).

Inductive RR {M} : forall (p1 p2 : {X : Type & Program M X}), Prop :=
| MkRR X (p1 p2 : Program M X) : eutt ieq p1 p2 -> RR (existT _ X p1) (existT _ X p2)
.

Lemma ForallRR_Sym {M0} p1 p2 : Forall2 (@RR M0) p1 p2 -> Forall2 RR p2 p1.
Proof.
  induction 1; auto. constructor; auto. destruct H; constructor; auto using Symmetric_eutt_ieq.
Qed.

Inductive RR' {M} : forall (p1 p2 : {X : Type & Program M X}), Prop :=
| MkRR' X (p1 p2 : Program M X) : (forall r, p1 = Return r -> p2 = Return r) -> eutt ieq p1 p2 -> RR' (existT _ X p1) (existT _ X p2)
.

Lemma SS_Advance1 {M0} ww (o0 : Object M0) t t' s x
  : Forall2 RR t' t ->
    ThreadAdvance o0 (t', s) x ->
    exists t_, ThreadsAdvance (wrapImpl ww o0) (t, s) (t_, snd x) /\ Forall2 RR (fst x) t_.
Proof.
  intros H_RR ADV; revert t H_RR. remember (t', s); revert t' s Heqp.
  induction ADV; intros; inversion Heqp; clear Heqp; subst; cbn.
  - inversion H_RR; subst.
    specialize (IHADV _ _ eq_refl _ H3).
    destruct IHADV as [t_ [? ?]].
    exists (y :: t_).
    split; auto. apply (advanceCons (wrapImpl ww o0)); auto.
  - inversion H_RR; subst; clear H_RR.
    inversion H2; injpair; clear H2.
    apply eutt_Bind_inv in H5.
    destruct H5 as (? & ? & [] & ? & UT).
    induction UT.
    + destruct IHUT as (? & ? & ?).
      eexists; split; eauto.
      eapply ThreadsAdvanceTrans; [ | eassumption ].
      apply ThreadAdvanceImplThreadsAdvance. apply (TANoop (wrapImpl ww o0)).
    + subst; exists (existT _ _ (x0 a) :: l').
      split; [ | constructor; [ constructor; auto | auto]].
      apply ThreadAdvanceImplThreadsAdvance. apply (TABind (wrapImpl ww o0)); eauto.
  - exists t; split.
    + apply Reflexive_ThreadsAdvance.
    + inversion H_RR; subst; constructor; auto.
      inversion H1; injpair.
      constructor. apply eutt_Noop_L_inv; auto.
Qed.

Lemma SS_Advance {M0} ww (o0 : Object M0) t t' s x
  : Forall2 RR t' t ->
    ThreadsAdvance o0 (t', s) x ->
    exists t_, ThreadsAdvance (wrapImpl ww o0) (t, s) (t_, snd x) /\ Forall2 RR (fst x) t_.
Proof.
  intros H_RR ADV; revert t H_RR. remember (t', s); revert t' s Heqp. induction ADV; intros; subst; cbn.
  - eexists; split; [constructor | auto ].
  - eapply SS_Advance1 in H; eauto.
    destruct H as (? & ? & ?).
    destruct y.
    eapply IHADV in H0; eauto.
    destruct H0 as (? & ? & ?).
    eexists; split; [ | eauto ].
    eapply ThreadsAdvanceTrans; eauto.
Qed.

Inductive RR_2 : forall (p1 p2 : {X : Type & Program (Abortable (Program M)) X}), Prop :=
| RR_2_Spawned X (p1 p2 : Program M X)
  : eutt ieq p1 p2 -> RR_2 (existT _ _ (Bind (abortable p1) Return)) (existT _ _ (Bind (abortable p2) Return))
| RR_2_Returned X (x : option X)
  : RR_2 (existT _ _ (Return x)) (existT _ _ (Return x))
.

Lemma AbortableEvalInst_ {I S} (EvalInst : forall R, I R -> S -> R -> S -> Prop)
    T (y : Abortable I T) (s : S) (r : T) (s' : S)
  : AbortableEvalInst EvalInst T y s r s' ->
    match y in Abortable _ T return T -> _ with
    | abortable w => fun r => (r = None /\ s = s') \/ (exists r', r = Some r' /\ EvalInst _ w s r' s')
    end r.
Proof.
  intros []; cbn; eauto.
Qed.

Lemma EvalProgram_eutt {X} {p1 p2 : Program _ X} {x s s'}
  : eutt ieq p1 p2 ->
    EvalProgram o X p2 s x s' ->
    EvalProgram o X p1 s x s'.
Proof.
  unfold EvalProgram; eauto using ProgramSteps_eutt_Return, Symmetric_eutt_ieq.
Qed.

Lemma SS_Advance1_2 t t' s x
  : Forall2 RR_2 t t' ->
    ThreadAdvance (wrapImpl pred so) (t', s) x ->
    exists t_, ThreadsAdvance so (t, s) (t_, snd x) /\ Forall2 RR_2 t_ (fst x).
Proof.
  intros H_RR ADV; revert t H_RR. remember (t', s); revert t' s Heqp.
  induction ADV; intros; inversion Heqp; clear Heqp; subst; cbn.
  - inversion H_RR; subst.
    specialize (IHADV _ _ eq_refl _ H3).
    destruct IHADV as [t_ [? ?]].
    eexists (_ :: t_).
    split; [ apply advanceCons; auto | constructor; auto ].
  - inversion H_RR; subst; clear H_RR.
    inversion H3; injpair; clear H3.
    exists (existT _ (option X) (Return a) :: l).
    split; [ | constructor; [ constructor | auto ]].
    apply ThreadAdvanceImplThreadsAdvance. constructor.
    cbn in H. apply AbortableEvalInst_ in H. destruct H as [[] | [? []]]; subst;
      constructor.
    eapply EvalProgram_eutt; eauto.
  - exists t; split.
    + apply Reflexive_ThreadsAdvance.
    + inversion H_RR; subst; constructor; auto.
      inversion H2; injpair.
Qed.

Lemma SS_Advance_2 t t' s x
  : Forall2 RR_2 t t' ->
    ThreadsAdvance (wrapImpl pred so) (t', s) x ->
    exists t_, ThreadsAdvance so (t, s) (t_, snd x) /\ Forall2 RR_2 t_ (fst x).
Proof.
  intros H_RR ADV; revert t H_RR. remember (t', s); revert t' s Heqp. induction ADV; intros; subst; cbn.
  - eexists; split; [constructor | auto ].
  - eapply SS_Advance1_2 in H; eauto.
    destruct H as (? & ? & ?).
    destruct y.
    eapply IHADV in H0; eauto.
    destruct H0 as (? & ? & ?).
    eexists; split; [ | eauto ].
    eapply ThreadsAdvanceTrans; eauto.
Qed.

Lemma SS_AdvanceRR {M0} (o0 : Object M0) t t'
  : Forall2 RR t t' ->
    exists t_, (forall s, ThreadsAdvance o0 (t', s) (t_, s)) /\ Forall2 RR' t t_.
Proof.
  induction 1.
  - eexists nil; split; auto.
  - destruct IHForall2 as [? []].
    destruct H.
    assert (HHH : (exists r, p1 = Return r) \/ (forall r, p1 <> Return r)).
    { rewrite (frobProgramId p1); destruct p1; cbn; eauto; right; discriminate. }
    destruct HHH as [[r ->] | ].
    + exists (existT _ _ (Return r) :: x0).
      split; [ | constructor; auto ].
      2: constructor; auto; apply Reflexive_eutt_ieq.
      intros s.
      eapply ThreadsAdvanceTrans; [ | eapply advanceCons; eauto ].
      apply ProgramStepsImplThreadsAdvance.
      apply Symmetric_eutt_ieq in H; revert H; apply eutt_ProgramSteps.
    + eexists; split; [ | constructor; eauto ].
      { intros; apply advanceCons; eauto. }
      constructor; eauto. intros; contradiction (H3 r).
Qed.

Lemma RRRR1 {M0} (t t' : {X : Type & Program M0 X})
  : RR' t t' -> RR t t'.
Proof.
  intros []; constructor; auto.
Qed.

Lemma RRRR {M0} (t t' : list {X : Type & Program M0 X})
  : Forall2 RR' t t' -> Forall2 RR t t'.
Proof.
  apply Util.Forall2_impl_; intros * _; apply RRRR1.
Qed.

Lemma SS_Int t t' t_ s R int (r : R)
  : Forall2 RR' t t_ ->
    ThreadsInteraction' co R int t r t' ->
    exists int' t_',
      ThreadsInteraction (wrapImpl pred co) R int' (t_, s) r (t_', s) /\ Forall2 RR t' t_' /\
      eq_Interaction (fun R y1 y2 => eq_Abortable (eutt ieq) _ y1 (unwrap _ y2)) R int int'.
Proof.
  intros W U. revert t_ W; induction U; intros.
  - specialize (PRED m). destruct PRED as [z [PRED_z PRED_e]].
    exists (Spawn (wrapMethod1 _ _ _ z PRED_z)).
    destruct PRED_e.
    eexists; split; [ | split ].
    + constructor 1 with (impl := wrapImpl pred co).
      constructor.
    + constructor; auto using RRRR.
      constructor. cbn. unfold flattenImplProg.
      eapply Proper_flattenProgram; eauto using Proper_instrument, Symmetric_eutt_ieq.
      intros ? ? ? []. apply Reflexive_eutt_ieq.
    + constructor; constructor; auto using Symmetric_eutt_ieq.
  - inversion W; subst.
    inversion H1; subst; injpair.
    rewrite (H2 _ eq_refl).
    exists (Join (Method := wrapMethod _ pred) (Ret := Ret)) l'; split; [ | split ].
    + constructor 1 with (impl := wrapImpl pred co).
      constructor 2 with (impl := wrapImpl pred co).
    + auto using RRRR.
    + constructor.
  - inversion W; subst.
    apply IHU in H3.
    destruct H3 as (? & ? & ? & ? & ?).
    eexists _, _; split; [ | split ].
    + inversion H; subst.
      constructor 1 with (impl := wrapImpl pred co).
      constructor 3 with (impl := wrapImpl pred co).
      eauto.
    + constructor; eauto using RRRR1.
    + constructor; auto.
Qed.

Lemma SS_Int_2 t t' t_ s R (int : Interaction (Abortable _) R) (int' : Interaction (wrapMethod (Abortable _) pred) R) (r : R)
  : Forall2 RR_2 t_ t ->
    eq_Interaction (fun R y1 y2 => eq_Abortable (eutt ieq) _ y1 (unwrap _ y2)) R int int' ->
    ThreadsInteraction' (wrapImpl pred so) R int' t r t' ->
    exists t_', ThreadsInteraction so R int (t_, s) r (t_', s) /\ Forall2 RR_2 t_' t'.
Proof.
  intros W EE U. revert t_ int W EE; induction U; intros; cbn.
  - generalize tt.
    refine (match EE in eq_Interaction _ X m n return
        match n return Prop with
        | Spawn _ => forall (u : X), exists _, ThreadsInteraction _ X m _ u _ /\ _
        | _ => True
        end with
      | eq_Interaction_Spawn _ _ _ _ _ => _
      | _ => I
      end); clear int EE.
    intros [].
    destruct w; cbn in e |- *; destruct e; cbn.
    eexists; split.
    + constructor. constructor.
    + constructor; cbn; auto. constructor; auto using Symmetric_eutt_ieq.
  - revert ret W.
    refine (match EE in eq_Interaction _ X m n return
        match n return Prop with
        | Join => forall ret, _ -> exists _, ThreadsInteraction _ X m _ ret _ /\ _
        | _ => True
        end with
      | eq_Interaction_Join _ _ => _
      | _ => I
      end); clear int EE; cbn.
    intros ? W.
    inversion W; subst.
    inversion H2; injpair.
    exists l; split; [ | auto using RRRR ].
    constructor. constructor.
  - inversion EE; subst; injpair; clear EE.
    inversion W; subst.
    eapply IHU in H2; eauto.
    destruct H2 as (? & ? & ?).
    eexists _; split; [ | constructor; eauto using RRRR1 ].
    inversion H; subst.
    cbn; do 2 constructor; auto.
Qed.

Lemma SSU_ : forall ta ta' sa tc tc' sc,
    Forall2 RR_2 ta ta' -> Forall2 RR tc tc' ->
    Simulates_ (ThreadsBehavior (wrapImpl pred (instFlatImpl conc instrument)))
      (ThreadsBehavior
         (wrapImpl pred (serialObject (abortableSpec (programSpec o)))))
      (tc', sc)
      (ta', sa) ->
    Simulates_ (ThreadsBehavior (instFlatImpl conc instrument))
      (ThreadsBehavior (serialObject (abortableSpec (programSpec o))))
      (tc, sc)
      (ta, sa).
Proof.
  cbn; cofix SELF; intros.
  constructor; intros * [? ADV INT].
  eapply (SS_Advance _ (o0 := instFlatImpl conc instrument)) in ADV; eauto.
  destruct ADV as [? [ADV H_RR1]].
  eapply (SS_AdvanceRR (wrapImpl pred (instFlatImpl conc instrument))) in H_RR1.
  destruct H_RR1 as [? [ADV' H_RR1]].
  inversion INT; cbn in *; subst; cbn in *.
  eapply SS_Int with (s := st) in H2; eauto.
  destruct H2 as (? & ? & INT' & H_RR2 & EI).
  destruct H1 as [[? [? ADV2 INT2] SIM]].
  { cbn. eexists.
    + eapply ThreadsAdvanceTrans; eauto.
    + eassumption. }
  cbn in ADV2.
  eapply SS_Advance_2 in ADV2; eauto. destruct ADV2 as (? & ADV2 & H_RR3).
  inversion INT2; subst; clear INT2.
  eapply SS_Int_2 in H1; eauto. destruct H1 as (? & INT2 & ?).
  eauto.
Qed.

Theorem StrictSerializable_unwrap
  : StrictSerializableWrap pred o conc ->
    StrictSerializable o conc instrument.
Proof.
  unfold StrictSerializableWrap, LinearizableWrap, ConcSimulatesWrap, StrictSerializable, Linearizable, ConcSimulates.
  apply SSU_.
  - constructor.
  - constructor.
Qed.

End SSU.
