From C4.Core Require Import
  Program
  ProgramOps
  Threads
  LinCoInd
  Object
  Linearizability
  LinearizabilityFacts
  FlattenSimulates
  LinearizabilityTactics.

Require Import C4.App.Register. (* "Register" is a keyword. *)
From C4.AppExtra Require Import
  Counter
  CounterImpl.

Import SimulatesCoInd2.

Set Implicit Arguments.

Module Type CounterImplLinParams.

  Parameter reg: Object (@RegisterMethod nat).
  Parameter regInit: State (objBase reg).
  Parameter regLin: Linearizable atomicReg reg.

End CounterImplLinParams.

Module CounterImplLinFun (CounterImplLinArgs: CounterImplLinParams).

  Import CounterImplLinArgs.

  Module CounterImpl := CounterImplFun CounterImplLinArgs.
  Import CounterImpl.

  
  Definition concImpl := {| objBase := @atomicReg nat _; objProgram := CounterImpl.impl |}.
  Definition abstImpl := (serialObject atomicCounter).
  
  Inductive PRel : forall R, Program concImpl.(ObjInstruction) R
                             -> Program abstImpl.(ObjInstruction) R -> Prop :=
  | PRelReturn : forall r : nat, PRel (Return r) (Return r)
  | PRelReadFetch : PRel (r <- read; Return r) (r' <- fetch; Return r')
  | PRelLoop1 : PRel readAndCasLoop (r' <- incAndFetch; Return r')
  | PRelLoop2 : forall v, PRel (b <- cas v (S v);
                                if b: bool then
                                  Return (S v)
                                else
                                  readAndCasLoop) (r' <- incAndFetch; Return r').

  Hint Constructors PRel CounterSem.

  Definition DRel (s1: concImpl.(objBase).(State)) (s2: abstImpl.(objBase).(State)): Prop :=
    s1 = s2.

  Definition init1: DataState concImpl := 0.
  Definition init2: DataState abstImpl := 0.

  Ltac t0 :=
    match goal with
    | [ H : PRel _ _ |- _ ] => inv H
    | [ H : RegSem _ _ _ _ |- _ ] => apply invertRegSem in H; inv'
    | [ H : EvalInstruction concImpl.(objBase) _ _ _ _ |- _ ] => apply invertRegSem in H; inv'
    | [ H : readAndCasLoop = _ |- _ ] => rewrite (frobProgramId readAndCasLoop) in H; simpl in H; inv'
    | [ m : CounterMethod _ |- _ ] => destruct m
    | [ m : RegisterMethod _ |- _ ] => destruct m; inv'
    | [ H : _ /\ _ |- _ ] => destruct H
    end.

  Ltac t := intros; repeat t0; eauto 10.

  Lemma CounterImplLin:
    Linearizable atomicCounter counterImpl.
  Proof.
    flatten regLin.
    apply (simulatesCoInd concImpl abstImpl PRel DRel);
      unfold DRel; t.
  Qed.

  (* Print Assumptions CounterImplLin. *)

End CounterImplLinFun.
