Require Import
        Coq.Lists.List
        Coq.Sorting.Permutation.

From C4.CoreExtra Require Import 
  Predefs.

Inductive Prog {Inst} {R : Type}: Type :=
| Bind {A} : Inst A -> (A -> Prog) -> Prog
| Return : R -> Prog.

Arguments Prog : clear implicits.
Notation "x <- a ; f" := (Bind a (fun x => f)) (at level 80, right associativity).

Record SpecType := {
  State: Type;
  Method: Type -> Type;
  Exec {A}: State -> Method A -> A -> State -> Prop
}.
Definition specType := Build_SpecType.

Section Exec.

  Variable obj: SpecType.

  (* Execution *)
  Inductive progExecStep R:
    Prog (Method obj) R -> State obj -> Prog (Method obj) R -> State obj -> Prop :=
  | progExecBind {s R'} {m} {r: R'} {s'} {f}:
      (Exec obj) s m r s' ->
      progExecStep R (Bind m f) s (f r) s'.
  
  Inductive progExec R:
    Prog (Method obj) R -> State obj -> Prog (Method obj) R -> State obj -> Prop :=
  | progExecRefl p s: 
      progExec R p s  p s
  | progExecSteps s p p' s' p'' s'': 
      progExec R p s p' s'
      -> progExecStep R p' s' p'' s''
      -> progExec R p s p'' s''.

  Hint Constructors progExec.

  (*
  Definition CProg R := TId -> Prog (Method obj) R. 
  *)

  Definition Trans := Prog (Method obj) (option nat).
  Definition CTrans := TId -> Trans.
  
  (* Concurrent execution *)
  (*
  Inductive cTransExecStep:
    forall R, CTrans R -> State obj -> CLabel -> CTrans R -> State obj -> Prop :=
  | cTransExecBind {R t cp s m l} {r: R} {s' f}:
      In t tids ->
      (Exec obj) s m l r s' ->
      cTransExecStep R (override cp t (Bind m f)) s (cLabel l t) (override cp t (f r)) s'.

  Inductive cTransExec: 
    forall R, CTrans R -> State obj -> list CLabel -> CTrans R -> State obj -> Prop :=
  | cTransExecRefl R p s: 
      cTransExec R p s nil p s
  | cTransExecSteps R s p l p' s' ls p'' s'': 
      cTransExecStep R p s l p' s'
      -> cTransExec R p' s' ls p'' s''
      -> cTransExec R p s (l :: ls) p'' s''.
  *)

  Inductive cCTransExecStep:
    CTrans -> State obj -> CTrans -> State obj -> Prop :=
  | cCTransExecBind {R t cp s m} {r: R} {s' f}:
      In t tids ->
      (Exec obj) s m r s' ->
      cCTransExecStep (override cp t (Bind m f)) s (override cp t (f r)) s'.

  Inductive cCTransExec: 
    CTrans -> State obj -> CTrans -> State obj -> Prop :=
  | cCTransExecRefl cp s: 
      cCTransExec cp s cp s
  | cCTransExecSteps cp s cp' s' cp'' s'': 
      cCTransExec cp s cp' s'
      -> cCTransExecStep cp' s' cp'' s''
      -> cCTransExec cp s cp'' s''.

  (*
  | cTransExecSteps cp s l cp' s' ls cp'' s'': 
      cTransExecStep cp s l cp' s'
      -> cTransExec cp' s' ls cp'' s''
      -> cTransExec cp s (l :: ls) cp'' s''.
  *)

  (* Sequential execution *)
  (*
  Inductive sTransExecStep:
    forall R, CTrans (option R) -> State obj -> list CLabel -> CTrans (option R) -> State obj -> Prop :=
  | sTransExecStepExec {R t cp p s ls r s'}:
      In t tids ->
      progExec (option R) p s ls (Return r) s' ->
      sTransExecStep R 
                    (override cp t p) s 
                    (map (fun l => (cLabel l t)) ls)
                    (override cp t (Return r)) s'

  | sTransExecStepAbort {R t cp p s p' s' ls}:
      In t tids ->
      progExec (option R) p s ls p' s' ->
      sTransExecStep R 
                    (override cp t p) s 
                    (map (fun l => (cLabel l t)) ls)
                    (override cp t (Return None)) s.

  Inductive sTransExec: 
    forall R, CTrans (option R) -> State obj -> list CLabel -> CTrans (option R) -> State obj -> Prop :=
  | sTransExecRefl R p s: 
      sTransExec R p s nil p s
  | sTransExecSteps R s p ls p' s' ls' p'' s'': 
      sTransExecStep R p s ls p' s'
      -> sTransExec R p' s' ls' p'' s''
      -> sTransExec R p s (ls ++ ls') p'' s''.
  *)

  (*
  Inductive sTransExecStep:
    CTrans -> State obj -> list CLabel -> CTrans -> State obj -> Prop :=
  | sTransExecStepExec {t cp p s ls r s'}:
      In t tids ->
      progExec (option nat) p s ls (Return r) s' ->
      sTransExecStep (override cp t p) s 
                    (map (fun l => (cLabel l t)) ls)
                    (override cp t (Return r)) s'

  | sTransExecStepAbort {t cp p s p' s' ls}:
      In t tids ->
      progExec (option nat) p s ls p' s' ->
      sTransExecStep (override cp t p) s 
                    (map (fun l => (cLabel l t)) ls)
                    (override cp t (Return None)) s.
  *)

  Inductive sTransExecStep:
    Trans -> State obj -> Trans -> State obj -> Prop :=
  | sTransExecStepExec {p s r s'}:
      progExec (option nat) p s (Return (Some r)) s' ->
      sTransExecStep p s 
                    (Return (Some r)) s'

  | sTransExecStepAbort {p s}:
      sTransExecStep p s
                     (Return None) s.


  Inductive sCTransExecStep:
    CTrans -> State obj -> CTrans -> State obj -> Prop :=
  | sCTransExecStepExec {t cp p s p' s'}:
      In t tids ->
      sTransExecStep p s p' s' ->
      sCTransExecStep (override cp t p) s 
                      (override cp t p') s'.

  Inductive sCTransExec: 
    CTrans -> State obj -> CTrans -> State obj -> Prop :=
  | sTransExecRefl cp s: 
      sCTransExec cp s cp s
  | sCTransExecSteps s cp cp' s' cp'' s'': 
      sCTransExec cp s cp' s'
      -> sCTransExecStep cp' s' cp'' s''
      -> sCTransExec cp s cp'' s''.
  
End Exec.



Module Type SerializabilityPars.

  Parameter absObj: SpecType.
  Parameter absInit: State absObj.

  Parameter concObj: SpecType.
  Parameter concInit: State concObj.

  (*
  Parameter translateTrans:
    forall R, Trans (Method absObj) R -> Trans (Method concObj) (option R).
  *)

  Parameter translateTrans: TId -> Trans absObj -> Trans concObj.

End SerializabilityPars.


Module Serializability (SerializabilityArgs: SerializabilityPars).

  Import SerializabilityArgs.

  (*
  Fixpoint liftTrans {R} (p: Trans (Method absObj) R): Trans (Method absObj) (option R) :=
    match p with
    | Bind _ m f => Bind m (fun x => liftTrans (f x))
    | Return r => Return (Some r)
    end.

  Definition liftCTrans {R} (cp: @CTrans absObj R): @CTrans absObj (option R) :=
    fun t => liftTrans (cp t).
  *)  

  (* mapsto *)
  (* fold_right *)

  (*  
  Definition translateCTrans {R} (cp: @CTrans absObj R) :=
    fun t => translateTrans R (cp t).    
  *)

  Definition translateCTrans (cp: CTrans absObj) :=
    fun t => translateTrans t (cp t).

  (*
  Definition finished {R} (cp: @CTrans concObj R) :=
    forall t, In t tids -> exists r, cp t = Return r.
  *)

  (*
  Definition serializable :=
    forall R cp s lc cp', 
      cTransExec concObj (option R) 
                (translateCTrans cp) concInit
                lc
                cp' s
      -> finished cp'
      -> exists la cp'' s', 
          Permutation lc la /\
          sTransExec absObj R 
                    (liftCTrans cp) absInit
                    (labelsMap la)
                    cp'' s'.
  *)
  (* Permutation (labelsMap lc) la /\ *) 


  
  Definition finished (cp: CTrans concObj) :=
    forall t, In t tids -> exists r, cp t = Return r.
  
  Definition resEq (ccp: CTrans concObj) (acp: CTrans absObj) :=
    forall (t: TId), 
      In t tids -> 
      exists r, ccp t = Return r
           /\ acp t = Return r.

  Definition serializable :=
    forall acp cs ccp', 
      cCTransExec concObj
                  (translateCTrans acp) concInit
                  ccp' cs
      -> finished ccp'
      -> exists acp' as', 
          sCTransExec absObj
                      acp absInit
                      acp' as'
          /\ resEq ccp' acp'.

  
End Serializability.
