Set Asymmetric Patterns.
Require Import
        Coq.Structures.Equalities.
From C4.Core Require Import
  Program.


Definition override
           {Key}{Val}(eq_key_dec: forall k k': Key, {k = k'} + {k <> k'})
           (m: Key -> Val) (k: Key) (val: Val): Key -> Val :=
  fun (k': Key) =>
    if (eq_key_dec k' k) then val else (m k').

Lemma override_new_val:
  forall K V eq_key_dec m (k: K) (v: V),
    (override eq_key_dec m k v) k = v.

Proof.
  intros.
  unfold override.
  destruct (eq_key_dec k k).  
  reflexivity.
  exfalso. apply n. reflexivity.
Qed.

Lemma override_new_val':
  forall K V eq_key_dec m (k1 k2: K) (v: V),
    k1 = k2 ->
    (override eq_key_dec m k1 v) k2 = v.

Proof.
  intros.
  subst.
  apply override_new_val.
Qed.

Lemma override_new_val'':
  forall K V eq_key_dec m (k1 k2: K) (v: V),
    k2 = k1 ->
    (override eq_key_dec m k1 v) k2 = v.

Proof.
  intros.
  subst.
  apply override_new_val.
Qed.


Lemma override_old_val:
  forall K V eq_key_dec m (k k': K) (v: V),
    (k <> k')
    -> (override eq_key_dec m k v) k' = m k'.

Proof.
  intros.
  unfold override.
  destruct (eq_key_dec k' k).  
  subst. exfalso. apply H. reflexivity.
  reflexivity.
Qed.

Lemma override_old_val':
  forall K V eq_key_dec m (k k': K) (v: V),
    (k' <> k)
    -> (override eq_key_dec m k v) k' = m k'.

Proof.
  intros.
  unfold override.
  destruct (eq_key_dec k' k).
  subst. exfalso. apply H. reflexivity.
  reflexivity.
Qed.


Inductive HistMethod {Key}: Type -> Type :=
| get: Key -> HistMethod (option nat)
| inc: Key -> HistMethod nat.


Module HistFun (E : MiniDecidableType).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.

  Definition HistState :=
    Key -> option nat.  

  Inductive HistSem:
    forall {R}, (@HistMethod Key R) -> HistState -> R -> HistState -> Prop :=
  | getSem map k:
      HistSem (get k) map (map k) map
  | incSem1 map k v:
      map k = Some v ->
      HistSem (inc k)
              map
              (v + 1)
              (override eq_key_dec map k (Some (v + 1)))
  | incSem2 map k:
      map k = None ->
      HistSem (inc k)
              map
              1
              (override eq_key_dec map k (Some 1)).
              

  Definition atomicHist: Spec HistMethod :=
    Build_Spec HistMethod HistState (@HistSem) (fun _ => None).

End HistFun.
