From Coq Require Import
        Peano_dec
        Equalities.

From C4.Core Require Import
  Program
  ProgramOps
  Threads
  LinCoInd
  Object
  Linearizability
  LinearizabilityFacts
  FlattenSimulates
  LinearizabilityTactics.
From C4.App Require Import
  Maps.
From C4.AppExtra Require Import
  Hist
  HistImpl.

Import EMap.
Import SimulatesCoInd2.


Set Implicit Arguments.

Module Type HistImplLinParams (E : MiniDecidableType).

  Parameter map: Object (EMapMethod E.t nat).
  Parameter mapInit: State (objBase map).

  Module EMap := EMapFun E.
  Export EMap.

  Parameter mapLin : Linearizable (atomicEMap nat) map.

End HistImplLinParams.


Module HistImplLinFun (E : MiniDecidableType) (HistImplLinArgs : HistImplLinParams E).

  Import HistImplLinArgs.

  Module Hist := HistFun E.
  Import Hist.

  Module HistImpl := HistImplFun E HistImplLinArgs.
  Import HistImpl.

  Definition Val := nat.


  Definition concImpl := {| objBase := atomicEMap Val; objProgram := HistImpl.impl |}.
  Definition abstImpl := (serialObject atomicHist).
  
  Inductive PRel : forall R, Program concImpl.(ObjInstruction) R ->
                        Program abstImpl.(ObjInstruction) R -> Prop :=
  | PRelReturn : forall R (r: R), PRel (Return r)
                                  (Return r)
  | PRelGetGet : forall k: Key, PRel (r' <- EMap.get k; Return r')
                                (r <- Hist.get k; Return r)
  | PRelLoop1 : forall k : Key, PRel (incLoop k)
                                (r' <- inc k ; Return r')
  | PRelLoop2 : forall k v, PRel (s <- EMap.replaceIt k v (v + 1);
                             if s:bool then
                               Return (v + 1)
                             else
                               incLoop k)
                            (r' <- inc k; Return r')
  | PRelLoop3 : forall k, PRel (jOpt <- EMap.putIfAbsent k 1;
                           match jOpt with      
                           | Some j =>
                             incLoop k
                           | None =>
                             Return 1
                           end)
                          (r' <- inc k; Return r').

  (*
  Hint Resolve ProgramStepsRefl.
  Hint Resolve ProgramStepImplProgramSteps.
  Hint Constructors ProgramStep.
  *)
  
  Hint Constructors HistSem.
  Hint Constructors PRel.
  
  Definition DRel (s1: concImpl.(objBase).(State)) (s2: abstImpl.(objBase).(State)): Prop :=
    s1 = s2.

  Definition init1: DataState concImpl := fun k => None.
  Definition init2: DataState abstImpl := fun k => None.

  Ltac t0 :=
    match goal with
    | [ H : PRel _ _ |- _ ] => inv H
    | [ H : EMapSem _ _ _ _ _ _ |- _ ] => apply invertEMapSem in H; inv'
    | [ H : EvalInstruction concImpl.(objBase) _ _ _ _ |- _ ] => apply invertEMapSem in H; inv'
    | [ H : incLoop ?k = _, k : Key |- _ ] => rewrite (frobProgramId (incLoop k)) in H; simpl in H; inv'
    | [ m : HistMethod _ |- _ ] => destruct m; simpl
    | [ m : EMapMethod _ _ _ |- _ ] => destruct m; inv'
    | [ s : EMapState _, k : Key |- _ ] =>
      match goal with
      | [ _ : s k = _ |- _ ] => fail 1
      | _ => let N := fresh in destruct (s k) eqn:N
      end
    | [ s : HistState, k : Key |- _ ] =>
      match goal with
      | [ _ : s k = _ |- _ ] => fail 1
      | _ => let N := fresh in destruct (s k) eqn:N
      end
    | [ s : State abstImpl.(objBase), k : Key |- _ ] =>
      match goal with
      | [ _ : s k = _ |- _ ] => fail 1
      | _ => let N := fresh in destruct (s k) eqn:N
      end
    | H : get _ = get _ |- _ => injection H; clear H; inv'
    | [ H : _ = _ |- _ ] => rewrite H
    | [ H : _ /\ _ |- _ ] => destruct H
    end.

  Ltac t := intros; repeat t0; try solve [ exfalso; eauto ]; eauto 10.

  Lemma PRelReturn' : forall R (r1 r2: R), r2 = r1 -> PRel (Return r1) (Return r2).
  Proof.
    intros; subst; constructor.
  Qed.

  Hint Immediate PRelReturn'.

  Hint Extern 1 (HistSem (inc _) _ _ _) => eapply incSem1.

  Lemma HistImplLin : Linearizable atomicHist histImpl.
  Proof.
    flatten mapLin.
    apply (simulatesCoInd concImpl abstImpl PRel DRel);
      unfold DRel; t.
  Qed.

  (* Print Assumptions HistImplLin. *)
  
End HistImplLinFun.
