Require Import
        Coq.Lists.List.

From C4.AppExtra Require Import
  Exec.

Module Type MapPars.

  Parameter Key Val: Set.
  Parameter eq_key_dec: forall k k': Key, {k = k'} + {not (k = k')}.
  Parameter eq_val_dec: forall v v': Val, {v = v'} + {not (v = v')}.

End MapPars.


Module Map (MapArgs: MapPars).

  Import MapArgs.

  Definition koverride {Val}(m: Key -> Val) (k: Key) (val: Val): Key -> Val :=
    fun (k': Key) =>
      if (eq_key_dec k' k) then val else (m k').

  Inductive MapMethod: Type -> Type :=
  | mapGet: Key -> MapMethod (option Val)
  | mapPut: Key -> option Val -> MapMethod unit.

  Definition MapState := (Key -> option Val).

  Inductive MapExec:
    forall A, MapState -> MapMethod A -> A -> MapState -> Prop :=
  | EGet k m:
      MapExec _ m (mapGet k) (m k) m
  | EPut k v m:
      MapExec _ m (mapPut k v) tt (koverride m k v).

  Definition MapType := specType MapState MapMethod MapExec.
  
  Definition mapInitState :=
    fun k: Key => @None Val.

  (* override *)  
  Lemma eq_key_dec_eq:
    forall (A: Type) n (s1 s2: A),
      (if eq_key_dec n n then s1 else s2) =
      s1.

  Proof.
    intros.
    destruct (eq_key_dec n n).
    reflexivity.

    exfalso. apply n0. reflexivity.
  Qed.

  Lemma eq_key_dec_neq:
    forall (A: Type) n n' (s1 s2: A),
      (not (n = n'))
      -> (if eq_key_dec n n' then s1 else s2) =
        s2.

  Proof.
    intros.

    destruct (eq_key_dec n n').
    exfalso. apply H in e. assumption.

    reflexivity.    
  Qed.

  Lemma koverride_new_val:
    forall (V: Type) m k (v: V),
      (koverride m k v) k = v.

    Proof.
      intros.
      unfold koverride.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.

  Lemma koverride_new_val':
    forall (V: Type) m k1 k2 (v: V),
      k1 = k2 ->
      (koverride m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold koverride.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.

  Lemma koverride_new_val'':
    forall (V: Type) m k1 k2 (v: V),
      k2 = k1 ->
      (koverride m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold koverride.
      rewrite eq_key_dec_eq.
      reflexivity.
    Qed.


  Lemma koverride_old_val:
    forall (V: Type) m k k' (v: V),
      (not (k = k'))
      -> (koverride m k v) k' = m k'.

    Proof.
      intros.
      unfold koverride.
      rewrite eq_key_dec_neq.
      reflexivity.
      unfold not. intro. symmetry in H0. apply H in H0. assumption.
    Qed.

  Lemma koverride_old_val':
    forall (V: Type) m k k' (v: V),
      (not (k' = k))
      -> (koverride m k v) k' = m k'.

  Proof.
      intros.
      unfold koverride.
      rewrite eq_key_dec_neq.
      reflexivity.
      assumption.
    Qed.


  Ltac simpl_koverride :=
    repeat (
       rewrite koverride_new_val ||
       (rewrite koverride_new_val'; [ | eassumption]) ||
       (rewrite koverride_new_val''; [ | eassumption]) ||
       (rewrite koverride_old_val; [ | eassumption]) ||
       (rewrite koverride_old_val'; [ | eassumption]));
    simpl.

  Ltac simpl_koverride_in H :=
    repeat (
       rewrite koverride_new_val in H ||
       (rewrite koverride_new_val' in H; [ | eassumption]) ||
       (rewrite koverride_new_val'' in H; [ | eassumption]) ||
       (rewrite koverride_old_val in H; [ | eassumption]) ||
       (rewrite koverride_old_val' in H; [ | eassumption]));
    simpl in H.

  Ltac simpl_koverride_in_star :=
    simpl_koverride;
    repeat
      match goal with
      | [ H: context[koverride ?m ?k ?v ?k'] |- _ ] => simpl_koverride_in H
      end.

End Map.
