Set Asymmetric Patterns.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts.

Inductive Abortable {M: Type -> Type}: Type -> Type :=
| abortable {R}: M R -> Abortable (option R).

Arguments Abortable : clear implicits.

Definition mapAbortable {M M'} (f : forall R, M R -> M' R)
           R (am : Abortable M R) :
  Abortable M' R :=
  let 'abortable R' m := am in abortable (f R' m).


Inductive AbortableEvalInst
          {I}{State}
          (EvalInst: forall R, I R -> State -> R -> State -> Prop):
  forall R, Abortable I R -> State -> R -> State -> Prop :=
  
| abortableEvalInstS R i s r s':
    EvalInst R i s r s' ->
    AbortableEvalInst EvalInst (option R) (abortable i) s (Some r) s'
| abortableEvalInstA R i s :
    AbortableEvalInst EvalInst (option R) (abortable i) s None s.


Definition abortableSpec {Instruction} (spec: Spec Instruction): Spec (Abortable Instruction) := {|
  State := spec.(State);  
  EvalInstruction := AbortableEvalInst (fun R => @EvalInstruction Instruction spec R);
  initialState := spec.(initialState)
|}.

Lemma abortableSimulates_ Method (spec1 spec2 : Spec Method) st1 st2 :
  Simulates_ spec1 spec2 st1 st2 ->
  Simulates_ (abortableSpec spec1) (abortableSpec spec2) st1 st2.
Proof.
  intro S.
  eapplyRelationSimulates (Simulates_ spec1 spec2).
  - intros * H1 H2.
    destruct int.
    inverts H2.
    + destruct H1 as [H1'].
      edestruct H1'.
      * eassumption.
      * eexists.
        -- constructor 1.
           eassumption.
        -- assumption.
    + eexists.
      * constructor 2.
      * assumption.
  - assumption.
Qed.

Lemma abortableSimulates Method (spec1 spec2 : Spec Method) :
  Simulates spec1 spec2 ->
  Simulates (abortableSpec spec1) (abortableSpec spec2).
Proof.
  apply abortableSimulates_.
Qed.

Lemma mapAbortableSimulates_ Method Method'
      (spec : Spec Method)
      (f : forall Ret, Method' Ret -> Method Ret)
      st :
  Simulates_
    (mapSpec (abortableSpec spec) (mapAbortable f))
    (abortableSpec (mapSpec spec f))
    st st.
Proof.
  eapplyRelationSimulates eq.
  - intros * H1 H2.
    subst.
    destruct int.
    inverts H2.
    + eexists.
      * constructor 1.
        simpl.
        eassumption.
      * reflexivity.
    + eexists.
      * constructor 2.
      * reflexivity.
  - reflexivity.
Qed.

Lemma mapAbortableSimulates Method Method'
      (spec : Spec Method)
      (f : forall Ret, Method' Ret -> Method Ret) :
  Simulates
    (mapSpec (abortableSpec spec) (mapAbortable f))
    (abortableSpec (mapSpec spec f)).
Proof.
  apply mapAbortableSimulates_.
Qed.
