From Coq Require Import List Relations.

From C4.Util Require Import LibTactics.

From C4.Core Require Import
  Program
  ProgramOps
  Object
  Simulates
  Threads
  ThreadsFacts
  MapState.

Section BaseSimulates.

  Context
    {Instruction Method : Type -> Type}
    (obj1 obj2 : Spec Instruction)
    (impl : forall Ret, Method Ret -> Program Instruction Ret).

  Notation o1 := (Build_Object _ Instruction obj1 impl).
  Notation o2 := (Build_Object _ Instruction obj2 impl).

  Definition BaseRel
             (ts1 : ThreadsState {| objBase := obj1; objProgram := impl |})
             (ts2 : ThreadsState {| objBase := obj2; objProgram := impl |}) :=
    fst ts1 = fst ts2 /\ Simulates_ obj1 obj2 (snd ts1) (snd ts2).

  Lemma baseThreadAdvance ps st1 ps' st1' st2 :
    ThreadAdvance {| objBase := obj1; objProgram := impl |} (ps, st1) (ps', st1') ->
    Simulates_ obj1 obj2 st1 st2 ->
    exists2 st2',
      ThreadAdvance {| objBase := obj2; objProgram := impl |} (ps, st2) (ps', st2') &
      Simulates_ obj1 obj2 st1' st2'.
  Proof.
    remember (ps, st1) as ts1.
    remember (ps', st1') as ts1'.
    intro ta1.
    revert ps ps' st2 Heqts1 Heqts1'.
    induction ta1; intros; inversions Heqts1; inversions Heqts1'.
    - edestruct IHta1.
      + reflexivity.
      + reflexivity.
      + eassumption.
      + eexists.
        * eapply (TACons {| objBase := obj2 |}).
          eassumption.
        * assumption.
    - destruct H0.
      edestruct H0.
      * eassumption.
      * eexists.
        -- eapply (TABind o2).
           eassumption.
        -- assumption.
    - eexists.
      -- eapply (TANoop {| objBase := obj2 |}).
      -- assumption.
  Qed.

  Lemma baseThreadsInteraction' Ret int ps (ret : Ret) ps' :
    ThreadsInteraction' {| objBase := obj1; objProgram := impl |} Ret int ps ret ps' ->
    ThreadsInteraction' {| objBase := obj2; objProgram := impl |} Ret int ps ret ps'.
  Proof.
    induction 1; simpl; constructor; assumption.
  Qed.

  Lemma baseThreadsInteraction Ret int ps st1 (ret : Ret) ps' st1' st2 :
    ThreadsInteraction {| objBase := obj1; objProgram := impl |} Ret int (ps, st1) ret (ps', st1') ->
    Simulates_ obj1 obj2 st1 st2 ->
    exists2 st2',
      ThreadsInteraction {| objBase := obj2; objProgram := impl |} Ret int (ps, st2) ret (ps', st2') &
      Simulates_ obj1 obj2 st1' st2'.
  Proof.
    intros ti' S.
    inversions ti'.
    eexists st2.
    - apply (TI {| objBase := obj2 |}).
      apply baseThreadsInteraction'.
      eassumption.
    - assumption.
  Qed.

  Theorem baseSimulates_
          (ps : list (sigT (Program Instruction)))
          (st1 : State obj1)
          (st2 : State obj2) :
    Simulates_ obj1 obj2 st1 st2 ->
    Simulates_
      (ThreadsBehavior {| objBase := obj1; objProgram := impl |})
      (ThreadsBehavior {| objBase := obj2; objProgram := impl |})
      (ps, st1)
      (ps, st2).
  Proof.
    intro.
    apply mapStateSimulates with BaseRel.
    - clear ps st1 st2 H.
      intros [ps st1] [ps' st1'] [ps2 st2] [E H] ta1.
      simpl in *.
      subst.
      edestruct baseThreadAdvance.
      + eassumption.
      + eassumption.
      + eexists.
        * apply simulateAdvance, clos_rt1n_step.
          eassumption.
        * split.
          -- reflexivity.
          -- assumption.
    - clear ps st1 st2 H.
      intros Ret [ps st1] ret [ps' st1'] [ps2 st2] [E H] int ti1.
      simpl in *.
      subst.
      edestruct baseThreadsInteraction.
      + eassumption.
      + eassumption.
      + eexists.
        * eexists.
          -- apply rt1n_refl.
          -- eassumption.
        * split.
          -- reflexivity.
          -- assumption.
    - split.
      + reflexivity.
      + assumption.
  Qed.

  Theorem baseSimulates :
    Simulates obj1 obj2 ->
    Simulates
      (ThreadsBehavior {| objBase := obj1; objProgram := impl |})
      (ThreadsBehavior {| objBase := obj2; objProgram := impl |}).
  Proof.
    apply baseSimulates_.
  Qed.

End BaseSimulates.
