From Coq Require Import List.

From C4.Util Require Import LibTactics.

From C4.Core Require Import
  Tactics
  Program
  Object
  Simulates
  SimulatesFacts
  Threads.

Lemma programStepAdvance
      Method (impl : Object Method) Ret p st p' st' :
  ProgramStep impl.(objBase) Ret p st p' st' ->
  ThreadAdvance
    impl
    (existT (Program impl.(ObjInstruction)) Ret p :: nil, st)
    (existT (Program impl.(ObjInstruction)) Ret p' :: nil, st').
Proof.
  intro H.
  destruct H.
  + constructor 2.
    assumption.
  + constructor 3.
Qed.

Lemma programStepsAdvances
      Method (impl : Object Method) Ret pst pst' :
  ProgramSteps' (objBase impl) Ret pst pst' ->
  ThreadsAdvance
    impl
    (existT _ Ret (fst pst) :: nil, snd pst)
    (existT _ Ret (fst pst') :: nil, snd pst').
Proof.
  induction 1.
  + constructor.
  + econstructor.
    * apply programStepAdvance.
      eassumption.
    * assumption.
Qed.

Lemma advanceProgramStep
      Method (impl : Object Method) Ret p st ts' :
  ThreadAdvance
    impl
    (existT (Program (ObjInstruction impl)) Ret p :: nil, st)
    ts' ->
  exists p' st',
    ts' = (existT (Program (ObjInstruction impl)) Ret p' :: nil, st') /\
    ProgramStep (objBase impl) Ret p st p' st'.
Proof.
  intro ta. inversions ta; injpair.
  + inversions H3.
  + repeat econstructor.
    assumption.
  + repeat econstructor.
Qed.

Lemma advancesProgramSteps
      Method (impl : Object Method) Ret pst ps' st' :
  ThreadsAdvance impl (existT _ Ret (fst pst) :: nil, snd pst) (ps', st') ->
  exists p',
    ps' = existT _ Ret p' :: nil /\
    ProgramSteps' (objBase impl) Ret pst (p', st').
Proof.
  intro tas.
  remember (_, _) as ts in tas.
  remember (_, _) as ts' in tas.
  revert pst Heqts ps' st' Heqts'.
  induction tas; intros; subst.
  - inversions Heqts'. destruct pst.
    repeat econstructor.
  - edestruct advanceProgramStep as [? [? [? ?]]].
    + eassumption.
    + subst.
      edestruct IHtas as [? [? ?]].
      * instantiate (1:=(_, _)).
        reflexivity.
      * reflexivity.
      * eexists.
        split.
        -- eassumption.
        -- econstructor 2 with (_, _); eassumption.
Qed.

Theorem coarsenSimulates
        Method
        (impl1 impl2 : Object Method) :
  ConcSimulates impl1 impl2 ->
  Simulates (coarsen impl1) (coarsen impl2).
Proof.
  unfold ConcSimulates, Simulates. cbn.
  generalize (impl1.(objBase).(initialState)).
  generalize (impl2.(objBase).(initialState)).
  intros ? ?.
  apply relationSimulates with (Rel := fun s0 s => Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2) (nil, s0) (nil, s)).
  clear.
  intros Ret m st1 ret st1' st2 [S] E1.
  edestruct S with (int:=Spawn m) as [st2'' [st2' ta2 ti2] [S']].
  - cbn. eexists (nil, st1).
    + constructor.
    + split.
      constructor.
  - inversions ta2; [| inversions H].
    inversions ti2.
    inversions H3. injpair.
    edestruct S' with (int:=@Join Method Ret) as [st2'' [st2' ta2 ti2] S''].
    + simpl.
      eexists (existT _ Ret (Return ret) :: nil, st1').
      * apply programStepsAdvances with (pst:=(_, _)) (pst':=(_, _)).
        assumption.
      * split.
        constructor.
    + inversions ti2. inversions H; injpair.
      edestruct (advancesProgramSteps Method) as [? [? ?]].
      * instantiate (6:=(_, _)).
        eassumption.
      * inversions H; injpair. eexists.
        -- simpl.
           eassumption.
        -- assumption.
Qed.
