From Coq Require Import
     List
     Program
     Relations
     RelationClasses.

From Paco Require Import paco.

From C4.Util Require Import
  Util
  LibTactics.

From C4.Core Require Import
  Program
  ProgramOps
  ProgramFacts
  Object.

Ltac inv H := inversion H; clear H; subst.

Inductive euttF {M1 M2} (RR : IRel M1 M2) {R}
  (_eutt : Program M1 R -> Program M2 R -> Prop)
  : Program M1 R -> Program M2 R -> Prop :=
| euttF_Return r : euttF RR _eutt (Return r) (Return r)
| euttF_Bind X (m1 : M1 X) (m2 : M2 X) k1 k2
  : RR X m1 m2 ->
    (forall x, _eutt (k1 x) (k2 x)) ->
    euttF RR _eutt (Bind m1 k1) (Bind m2 k2)
| euttF_Noop p1 p2 : _eutt p1 p2 -> euttF RR _eutt (Noop p1) (Noop p2)
| euttF_L p1 p2 : euttF RR _eutt p1 p2 -> euttF RR _eutt (Noop p1) p2
| euttF_R p1 p2 : euttF RR _eutt p1 p2 -> euttF RR _eutt p1 (Noop p2)
.

Definition eutt {M1 M2} (RR : IRel M1 M2) R
  : Program M1 R -> Program M2 R -> Prop := paco2 (euttF RR) bot2.

Arguments eutt {M1 M2} RR [R].

Lemma monotone_euttF {M1 M2} (RR : IRel M1 M2) R
  : monotone2 (euttF RR (R := R)).
Proof.
  induction 1; constructor; auto.
Qed.
Hint Resolve monotone_euttF : paco.

Lemma eutt_Noop {M1 M2} (RR : IRel M1 M2) {R}
  : forall (p1 : Program M1 R) (p2 : Program M2 R),
    eutt RR p1 p2 ->
    eutt RR (Noop p1) p2.
Proof.
  intros. punfold H. pfold.
  constructor. assumption.
Qed.

Lemma eutt_Noop_r (M1 M2 : Type -> Type) (RR : forall X : Type, M1 X -> M2 X -> Prop)
    (R : Type) (p1 : Program M1 R) (p2 : Program M2 R)
  : eutt RR p1 p2 -> eutt RR p1 (Noop p2).
Proof.
  intros. punfold H. pfold.
  constructor. assumption.
Qed.

Lemma Reflexive_eutt_ieq {M R}
  : forall (p : Program M R), eutt ieq p p.
Proof.
  pcofix self.
  pfold.
  intros []; repeat (constructor; auto).
Qed.

Global Hint Resolve Reflexive_eutt_ieq : core.
(***************)
(* BOILERPLATE *)

Inductive untaus1 {M R} (P : Program M R -> Prop) : Program M R -> Prop :=
| untaus1_Noop p : untaus1 P p -> untaus1 P (Noop p)
| untaus1_Id p : P p -> untaus1 P p
.

Lemma eutt_Ret_inv {M1 M2} (RR : IRel M1 M2) {R} (r : R) (p2 : Program M2 R)
  : eutt RR (Return r) p2 ->
    untaus1 (fun p => Return r = p) p2.
Proof.
  intros H. punfold H.
  remember (Return r) as p1; revert dependent r.
  induction H.
  - constructor. inv Heqp1. reflexivity.
  - discriminate.
  - discriminate.
  - discriminate.
  - constructor. eauto.
Qed.

Lemma eq_Bind_inv {M R} {X1} (m1 : M X1) k1 {X2} (m2 : M X2) k2
  : (x <- m1; k1 x) = (x <- m2; k2 x) ->
    forall (P : forall X (m : M X) (k : X -> Program M R), Prop),
      P _ m1 k1 -> P _ m2 k2.
Proof.
  intros H P.
  change (P _ m2 k2) with
      (match Bind m2 k2 with
       | Bind m k => P _ m k
       | _ => False
       end).
  remember (Bind m2 k2).
  destruct H.
  auto.
Qed.

Lemma eutt_Bind_inv {M1 M2} (RR : IRel M1 M2)
      {X} (m1 : M1 X) {R} (k1 : X -> Program M1 R) (p2 : Program M2 R)
  : eutt RR (Bind m1 k1) p2 ->
    exists m2 k2,
      RR _ m1 m2 /\
      (forall x, eutt RR (k1 x) (k2 x)) /\
      untaus1 (fun p => Bind m2 k2 = p) p2.
Proof.
  intros H. punfold H.
  remember (Bind m1 k1) as p1.
  revert m1 k1 Heqp1.
  induction H; intros.
  - discriminate.
  - pclearbot.
    inv Heqp1. injpair_; subst.
    exists m2 k2.
    split; [ assumption | ].
    split; [ assumption | ].
    constructor. reflexivity.
  - discriminate.
  - discriminate.
  - apply IHeuttF in Heqp1.
    destruct Heqp1 as (m2 & k2 & ? & ? & ?).
    exists m2 k2.
    split; [ assumption | ].
    split; [ assumption | ].
    constructor. auto.
Qed.

Lemma eutt_Noop_L_inv {M1 M2} (RR : IRel M1 M2) R
  : forall (p1 : Program M1 R) (p2 : Program M2 R),
    eutt RR (Noop p1) p2 ->
    eutt RR p1 p2.
Proof.
  pcofix self.
  intros p1 p2 H. remember (Noop p1) as Np1. revert dependent p1.
  punfold H. induction H.
  - discriminate.
  - discriminate.
  - pclearbot. injection 1. intros [].
    punfold H. pfold. constructor.
    eapply monotone_euttF; eauto.
    intros; eapply upaco2_mon_bot; eauto.
  - injection 1. intros []. pfold.
    eapply monotone_euttF; eauto.
    intros; eapply upaco2_mon_bot; eauto.
  - intros ? ?; subst p1. pfold. constructor.
    specialize (IHeuttF _ eq_refl). punfold IHeuttF.
Qed.

Lemma eutt_Ret_R_RR {M1 M2 M2'} (RR : IRel M1 M2) (RR' : IRel M1 M2') {R}
      (r : R) (p : Program M1 R)
  : eutt RR p (Return r) -> eutt RR' p (Return r).
Proof.
  intros H; punfold H; pfold.
  remember (Return r). revert Heqp0.
  induction H; try (discriminate + constructor; auto).
  injection 1; intros; subst; constructor.
Qed.

Lemma eutt_Ret_L_RR {M1 M1' M2} (RR : IRel M1 M2) (RR' : IRel M1' M2) {R}
      (r : R) (p : Program M2 R)
  : eutt RR (Return r) p -> eutt RR' (Return r) p.
Proof.
  intros H; punfold H; pfold.
  remember (Return r). revert Heqp0.
  induction H; try (discriminate + constructor; auto).
  injection 1; intros; subst; constructor.
Qed.

Lemma Symmetric_eutt_ {M1 M2} (RR : IRel M1 M2) (RR' : IRel M2 M1) {R}
      (SYM : forall X m1 m2, RR X m1 m2 -> RR' X m2 m1)
  : forall (p1 : Program M1 R) (p2 : Program M2 R),
    eutt RR p1 p2 -> eutt RR' p2 p1.
Proof.
  pcofix self.
  intros p1 p2 H; punfold H; pfold.
  induction H; constructor; pclearbot; auto.
  right; apply self, H0.
Qed.

Lemma Symmetric_eutt {M} (RR : IRel M M) {R}
      (SYM : forall X m1 m2, RR X m1 m2 -> RR X m2 m1)
  : forall (p1 : Program M R) (p2 : Program M R),
    eutt RR p1 p2 -> eutt RR p2 p1.
Proof.
  revert SYM; apply Symmetric_eutt_.
Qed.

Lemma Symmetric_eutt_ieq {M} {R}
  : forall (p1 : Program M R) (p2 : Program M R),
    eutt ieq p1 p2 -> eutt ieq p2 p1.
Proof.
  apply Symmetric_eutt. intros ? ? ? []; constructor.
Qed.

Section Trans.

Context {M1 M2 M3 : Type -> Type}
  (R12 : IRel M1 M2)
  (R23 : IRel M2 M3)
  (R13 : IRel M1 M3).

Context
  (RRR : forall A m1 m2 m3, R12 A m1 m2 -> R23 A m2 m3 -> R13 A m1 m3).

Lemma inv_eutt_Noop_left {R}
  : forall (p1 : Program M2 R) (p2 : Program M3 R),
    euttF R23 (upaco2 (euttF R23 (R := R)) bot2) (Noop p1) p2 ->
    euttF R23 (upaco2 (euttF R23 (R := R)) bot2) p1 p2.
Proof.
  intros p1 p2 H; remember (Noop p1) as Noop_p1 eqn:EQp1.
  induction H; try discriminate.
  - constructor.
    destruct H; [| contradiction ].
    punfold H.
    inversion EQp1.
    subst; auto.
  - inversion EQp1. subst. auto.
  - constructor; auto.
Qed.

Lemma inv_eutt_Noop_right {R}
  : forall (p1 : Program M1 R) (p2 : Program M2 R),
    euttF R12 (upaco2 (euttF R12 (R := R)) bot2) p1 (Noop p2) ->
    euttF R12 (upaco2 (euttF R12 (R := R)) bot2) p1 p2.
Proof.
  intros p1 p2 H; remember (Noop p2) as Noop_p1 eqn:EQp1.
  induction H; try discriminate.
  - constructor.
    destruct H; [| contradiction ].
    punfold H.
    inversion EQp1.
    subst; auto.
  - constructor; auto.
  - inversion EQp1. subst. auto.
Qed.

Lemma Transitive_eutt {R}
 : forall (p1 : Program M1 R) (p2 : Program M2 R) (p3 : Program M3 R),
   eutt R12 p1 p2 ->
   eutt R23 p2 p3 ->
   eutt R13 p1 p3.
Proof.
  pcofix SELF.
  intros p1 p2 p3 I1 I2.
  punfold I1; punfold I2.
  remember p2 as p2' in I2.
  pfold.
  revert p2' Heqp2' I2.
  induction I1 as [ | | p1 p2 | | ]; intros p2' Heqp2' I2.
  - (* Return *)
    induction I2; try discriminate;
    try lazymatch goal with
    | [ H : ?con _ = ?con _ |- _ ] => inversion H; clear H; subst
    end;
    solve [ constructor; auto ].

  - (* Bind *)
    induction I2; try (discriminate + solve [ constructor; eauto ]).

    (* Bind = Bind *)
    inversion Heqp2'; clear Heqp2'. subst.
    injpair_; subst.
    constructor.
    { apply (RRR _ m1 m2 m3); auto. }
    intros x.
    specialize (H0 x).
    destruct H0; [| contradiction ].
    specialize (H2 x).
    destruct H2; [| contradiction ].
    right; eauto.

  - (* Noop p1 | Noop p2 | p3 *)
    destruct H; [ | contradiction ].
    revert p2 Heqp2' H.
    induction I2 as [ | | | p2' p3 | ];
      intros p2'' EQp2 REL12; try rename p2'' into p2;
      try discriminate.
    + (* (p2, p3) -> (Noop p2, Noop p3) *)
      destruct H; [ | contradiction ].
      inversion EQp2; subst.
      constructor; eauto.
    + (* (p2, p3) -> (Noop p2, p3) *)
      inversion EQp2; subst.
      clear IHI2 EQp2.
      punfold REL12.
      { remember p2 as p2' in REL12. revert p2' Heqp2' REL12.
        induction I2; intros.
        - constructor.
          induction REL12; try discriminate;
          try lazymatch goal with
          | [ H : ?con _ = ?con _ |- _ ] => inversion H; clear H; subst
          end;
          solve [ constructor; auto ].

        - constructor.
          induction REL12; try (discriminate + solve [ constructor; eauto ]).

          (* Bind = Bind *)
          inversion Heqp2'; clear Heqp2'. subst.
          injpair_; subst.
          constructor.
          { eapply RRR; eauto. }
          intros x.
          specialize (H0 x).
          destruct H0; [| contradiction ].
          specialize (H2 x).
          destruct H2; [| contradiction ].
          right; eauto.

        - constructor.
          destruct H; [| contradiction ].
          subst p2'.
          apply inv_eutt_Noop_right in REL12.
          right; eapply SELF; eauto; pfold; eauto.

        - subst p2'; apply inv_eutt_Noop_right in REL12. eauto.

        - constructor.
          subst p2'.
          right; eapply SELF; pfold; eauto.
      }

    + (* (p2, p3) -> (p2, Noop p3) *)
      apply euttF_R. eauto.

  - (* Noop p1 | p2 | p3 *)
    constructor; eauto.

  - subst. apply inv_eutt_Noop_left in I2. eauto.
Qed.

End Trans.

Lemma Transitive_eutt_ieq {M R}
 : forall (p1 : Program M R) (p2 : Program M R) (p3 : Program M R),
   eutt ieq p1 p2 ->
   eutt ieq p2 p3 ->
   eutt ieq p1 p3.
Proof.
  apply Transitive_eutt.
  intros ? ? ? ? [] []. constructor.
Qed.

Local Notation fI := flattenObject.

Inductive flattenProgram_bisim {M1 M2 N1 N2}
          (RM : IRel M1 M2) (RN : IRel N1 N2)
          (f1 : forall R, M1 R -> Program N1 R)
          (f2 : forall R, M2 R -> Program N2 R)
          {R}
  : Program N1 R -> Program N2 R -> Prop :=
| flattenProgram_bisim_base p1 p2
  : eutt RM p1 p2 ->
    flattenProgram_bisim RM RN f1 f2 (flattenProgram f1 p1) (flattenProgram f2 p2)
| flattenProgram_bisim_bind {S} k1 k2 (p1 : Program N1 S) p2
  : (forall x, eutt RM (k1 x) (k2 x)) ->
    eutt RN p1 p2 ->
    flattenProgram_bisim RM RN f1 f2
      (bindFlattenProgram f1 k1 p1)
      (bindFlattenProgram f2 k2 p2)
.

Lemma euttF_Noop_L {M1 M2} (RR : IRel M1 M2) {R}
      (_eutt : Program M1 R -> Program M2 R -> Prop)
  : forall (p1 : Program M1 R) (p2 : Program M2 R),
    paco2 (euttF RR) _eutt p1 p2 -> paco2 (euttF RR) _eutt (Noop p1) p2.
Proof.
  pcofix self.
  intros p1 p2 H. pfold. punfold H. induction H.
  - constructor. constructor.
  - constructor. constructor; auto. intros.
    eapply upaco2_mon; eauto.
  - apply euttF_L. constructor.
    eapply upaco2_mon; eauto.
  - constructor; auto.
  - constructor. left. pfold.
    eapply monotone_euttF; eauto.
    intros; eapply upaco2_mon; eauto.
Qed.

Lemma Proper_flattenProgram {M1 M2 N1 N2}
      (RM : IRel M1 M2) (RN : IRel N1 N2)
      (f1 : forall R, M1 R -> Program N1 R)
      (f2 : forall R, M2 R -> Program N2 R)
      R
  : (RM ==> eutt RN)%isig f1 f2 ->
    forall (p1 : Program M1 R) (p2 : Program M2 R),
    eutt RM p1 p2 ->
    eutt RN (flattenProgram f1 p1) (flattenProgram f2 p2).
Proof.
  intros Hf.
  enough (forall (q1 : Program N1 R) (q2 : Program N2 R),
    flattenProgram_bisim RM RN f1 f2 q1 q2 -> eutt RN q1 q2).
  { intros p1 p2 I; apply H; constructor; apply I. }
  pcofix self.
  intros q1 q2 [].
  - punfold H. induction H.
    + rewrite (frobProgramId (_ f1 _ _)), (frobProgramId (_ f2 _ _)).
      pfold; constructor.
    + rewrite (frobProgramId (_ f1 _ _)), (frobProgramId (_ f2 _ _)).
      cbn. pfold; constructor. right. apply self.
      pclearbot; constructor; auto.
    + rewrite (frobProgramId (_ f1 _ _)), (frobProgramId (_ f2 _ _)).
      cbn. pclearbot; pfold; constructor; right; apply self; constructor.
      auto.
    + rewrite (frobProgramId (_ f1 _ _)).
      pfold. punfold IHeuttF.
      cbn. constructor. auto.
    + rewrite (frobProgramId (_ f2 _ _)).
      pfold. punfold IHeuttF.
      cbn. constructor; auto.
  - punfold H0. induction H0.
    + rewrite (frobProgramId (_ f1 _ _ _ _)), (frobProgramId (_ f2 _ _ _ _)).
      cbn. pfold. constructor.
      right; apply self; constructor. auto.
    + rewrite (frobProgramId (_ f1 _ _ _ _)), (frobProgramId (_ f2 _ _ _ _)).
      cbn. pfold. constructor; auto.
      pclearbot. right. apply self; constructor; auto.
      apply H1.
    + rewrite (frobProgramId (_ f1 _ _ _ _)), (frobProgramId (_ f2 _ _ _ _)).
      cbn. pfold. constructor; auto.
      pclearbot. right. apply self; constructor; auto.
    + rewrite (frobProgramId (_ f1 _ _ _ _)).
      cbn. punfold IHeuttF; pfold; constructor; auto.
    + rewrite (frobProgramId (_ f2 _ _ _ _)).
      cbn. punfold IHeuttF; pfold; constructor; auto.
Qed.

Inductive flatten_flatten_bisim {M N P}
          (f : forall R, M R -> Program N R) (g : forall R, N R -> Program P R)
          {R}
  : Program P R -> Program P R -> Prop :=
| flatten_flatten_bisim_base p
  : flatten_flatten_bisim f g
      (flattenProgram g (flattenProgram f p))
      (flattenProgram (fun _ m => flattenProgram g (f _ m)) p)
| flatten_flatten_bisim_bind {S} (p : _ S) k
  : flatten_flatten_bisim f g
      (flattenProgram g (bindFlattenProgram f k p))
      (bindFlattenProgram
         (fun (A0 : Type) (m0 : M A0) => flattenProgram g (f A0 m0)) k
         (flattenProgram g p))
| flatten_flatten_bisim_bind_2 {S T} (p : _ S) (h : S -> _ T) k
  : flatten_flatten_bisim f g
      (bindFlattenProgram g (fun r0 : _ => bindFlattenProgram f k (h r0)) p)
      (bindFlattenProgram
         (fun (A0 : Type) (m0 : M A0) => flattenProgram g (f A0 m0)) k
         (bindFlattenProgram g h p))
.

Inductive all_eutt {M1 M2} (RR : IRel M1 M2)
  : list (sigT (Program M1)) -> list (sigT (Program M2)) -> Prop :=
| all_eutt_nil : all_eutt RR [] []
| all_eutt_cons T p1 p2 t1 t2 :
  eutt RR p1 p2 ->
  all_eutt RR t1 t2 ->
  all_eutt RR (existT _ T p1 :: t1) (existT _ T p2 :: t2)
.

(* interp f . interp g = interp (interp f . g) *)
Lemma flatten_flatten {M N P}
      (f : forall R, M R -> Program N R) (g : forall R, N R -> Program P R)
  : (ieq ==> eutt ieq)%isig
      (fun _ (m : Program M _) => flattenProgram g (flattenProgram f m))
      (@flattenProgram _ _ (fun _ (m : M _) =>
                         flattenProgram g (f _ m))).
Proof.
  intros ? e _ [].
  enough (forall (p1 p2 : Program P T),
    flatten_flatten_bisim f g p1 p2 ->
    eutt ieq p1 p2).
  { apply H; constructor. }
  clear. pcofix self; intros _ _ []; pfold.
  all:
    match goal with
    | [ |- _ ?p1 ?p2 ] => rewrite (frobProgramId p1), (frobProgramId p2)
    end;
    match goal with
    | [ p : Program _ _ |- _ ] => destruct p; cbn
    end; constructor.
  all: try reflexivity.
  all: right; apply self; constructor.
Qed.

Lemma eutt_ProgramSteps {M} (o : Spec M) (Ret : Type) M2 {RR : IRel M M2}
      (s : State o) (x : Ret)
      (x0 : Program M Ret)
  : eutt RR x0 (Return x) -> ProgramSteps o Ret x0 s (Return x) s.
Proof.
  intros H; punfold H.
  remember (Return x) as p eqn:Hp.
  revert Hp; induction H; try discriminate.
  - injection 1; intros []. apply ProgramStepsRefl.
  - intros I; apply IHeuttF in I.
    eapply ProgramStepStepsImplSteps; [ constructor | eassumption ].
Qed.

Lemma flattenId {M R} (p : Program M R)
  : eutt ieq (flattenProgram (fun _ m => Bind m Return) p) p.
Proof.
  enough (J : forall q p : Program M R,
             (q = flattenProgram (fun _ m => Bind m Return) p \/
              q = Noop (flattenProgram (fun _ m => Bind m Return) p)) ->
             eutt ieq q p).
  { apply J. left; reflexivity. }
  clear. pcofix SELF. intros q p.
  enough (J : paco2 (euttF ieq) r (flattenProgram (fun _ m => Bind m Return) p) p).
  { intros []; subst; [ apply J | apply euttF_Noop_L, J ]. }
  pfold.
  destruct p.
  - rewrite (frobProgramId (flattenProgram _ _)). cbn. constructor.
    rewrite (frobProgramId (bindFlattenProgram _ _ _)). cbn.
    constructor; [ constructor | ].
    intros x. right.
    rewrite (frobProgramId (bindFlattenProgram _ _ _)). cbn.
    apply SELF. right. reflexivity.
  - rewrite (frobProgramId (flattenProgram _ _)). cbn. constructor.
  - rewrite (frobProgramId (flattenProgram _ _)). cbn. constructor.
    right. apply SELF. left. reflexivity.
Defined.

Lemma flattenTrigger {M N} (h : forall X, M X -> Program N X)
  {R} (m : M R)
  : eutt ieq (flattenProgram h (Bind m Return)) (h _ m).
Proof.
  rewrite (frobProgramId (flattenProgram _ _)). cbn.
  apply eutt_Noop.
  remember (h R m) as p eqn:E0. clear E0 m.
  revert p. pcofix SELF. intros p.
  pfold.
  destruct p;
    rewrite (frobProgramId (bindFlattenProgram _ _ _)); cbn; constructor.
  - constructor.
  - right; auto.
  - rewrite (frobProgramId (flattenProgram _ _)); cbn; constructor.
  - right; auto.
Qed.

Section FLATTENMAP.

Context {M M' I : Type -> Type}
  (impl : forall R : Type, M' R -> Program I R)
  (f : forall R : Type, M R -> M' R) (R : Type).

Inductive _flattenMap_sim : Program I R -> Program I R -> Prop :=
| _flattenMap_Noop p q : _flattenMap_sim p q -> _flattenMap_sim (Noop p) (Noop q)
| _flattenMap_flatten p :
    _flattenMap_sim
      (flattenProgram impl (mapProgram f p))
      (flattenProgram (fun (R0 : Type) (m : M R0) => impl R0 (f R0 m)) p)
| _flattenMap_bind S (p : Program I S) (k : S -> _) :
    _flattenMap_sim
      (bindFlattenProgram impl (fun x => mapProgram f (k x)) p)
      (bindFlattenProgram (fun _ m => impl _ (f _ m)) k p).

Lemma flattenMap (p : Program M R)
  : eutt ieq
      (flattenProgram impl (mapProgram f p))
      (flattenProgram (fun (R0 : Type) (m : M R0) => impl R0 (f R0 m)) p).
Proof.
  enough (J : forall p q, _flattenMap_sim p q -> eutt ieq p q).
  { apply J. constructor. }
  clear. pcofix SELF. intros p q; induction 1.
  - pfold; constructor; auto.
  - pfold.
    match goal with
    | [ |- _ ?t ?u ] => rewrite (frobProgramId t), (frobProgramId u)
    end.
    destruct p; cbn; constructor;
      right; apply SELF; constructor.
  - pfold.
    match goal with
    | [ |- _ ?t ?u ] => rewrite (frobProgramId t), (frobProgramId u)
    end.
    destruct p; cbn; constructor;
      try (right; apply SELF; constructor);
      try constructor.
Qed.

End FLATTENMAP.

