From Coq Require Import List Relations.

From C4.Util Require Import LibTactics.

From C4.Core Require Import
  Tactics
  Program
  ProgramOps
  Object
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  MapState
  FlattenBehavior
  ProgramFacts
  ProgramOpsFacts.

Section FlattenObjectBehavior.

  Context
    {BaseMethod Method : Type -> Type}
    (base : Object BaseMethod)
    (impl : forall Ret, Method Ret -> Program BaseMethod Ret).

  Inductive CombineFlattenThreads :
    list (FlattenThread BaseMethod) ->
    list (sigT (Program base.(ObjInstruction))) ->
    list (sigT (Program base.(ObjInstruction))) ->
    Prop :=
  | ftsNil :
      CombineFlattenThreads nil nil nil
  | ftsProgram Ret p fts' ps1 ps2 :
      CombineFlattenThreads fts' ps1 ps2 ->
      CombineFlattenThreads
        (ftProgram Ret p :: fts')
        ps1
        (existT _ Ret (flattenProgram base.(objProgram) p) :: ps2)
  | ftsWait A Ret f fts' p ps1 ps2 :
      CombineFlattenThreads fts' ps1 ps2 ->
      CombineFlattenThreads
        (ftWait A Ret f :: fts')
        (existT _ A p :: ps1)
        (existT _ Ret (bindFlattenProgram base.(objProgram) f p) :: ps2).

  Definition FBIRel
    (fs : list (FlattenThread BaseMethod) * ThreadsState base)
    (ts : ThreadsState (flattenObject base Method impl)) :
    Prop :=
    let '(fts, (ps1, st1)) := fs in
    let '(ps2, st2) := ts in
    CombineFlattenThreads fts ps1 ps2 /\ st1 = st2.

  Lemma flattenThreadAdvance ps1 st ps1' st' fts ps2 :
    CombineFlattenThreads fts ps2 ps1 ->
    ThreadAdvance (flattenObject base Method impl) (ps1, st) (ps1', st') ->
    exists fts' ps2',
      CombineFlattenThreads fts' ps2' ps1' /\
      (FlattenAdvance (ThreadsBehavior base) (fts, (ps2, st)) (fts', (ps2', st')) \/
       (fts = fts' /\ ThreadAdvance base (ps2, st) (ps2', st'))).
  Proof.
    remember (ps1, st) as ts1.
    remember (ps1', st') as ts1'.
    intros F ta.
    revert ps1 ps1' fts ps2 Heqts1 Heqts1' F.
    induction ta; intros; inversions Heqts1; inversions Heqts1'.
    - inversions F.
      + edestruct IHta as [fts'' [ps2' [F' H]]].
        * reflexivity.
        * reflexivity.
        * eassumption.
        * eexists _, _.
          split.
          -- constructor.
             eassumption.
          -- destruct H.
             ++ left.
                inversions H.
                esplit.
                ** constructor.
                   eassumption.
                ** assumption.
             ++ right.
                destruct H.
                subst.
                split.
                ** reflexivity.
                ** assumption.
      + edestruct IHta as [fts'' [ps2' [F' H]]].
        * reflexivity.
        * reflexivity.
        * eassumption.
        * eexists _, _.
          esplit.
          -- constructor.
             eassumption.
          -- destruct H.
             ++ left.
                inversions H.
                destruct int.
                ** inversion H6.
                   destruct x.
                   inversions H0.
                   esplit.
                   --- constructor.
                       eassumption.
                   --- esplit.
                       +++ apply (advanceCons (Method:=BaseMethod)).
                           eassumption.
                       +++ esplit with (Method:=BaseMethod).
                           constructor.
                           assumption.
                ** inversions H6.
                   esplit.
                   --- constructor.
                       eassumption.
                   --- reflexivity.
             ++ right.
                destruct H.
                subst.
                split.
                ** reflexivity.
                ** constructor 1 with (Method:=BaseMethod).
                   assumption.
    - inversions F; injpair.
      + frobinvl p0 H4.
      + frobinvl p H4.
        simpl in *.
        eexists _, _.
        esplit.
        * constructor.
          eassumption.
        * right.
          split.
          -- reflexivity.
          -- apply (TABind _ _ _ _ _ _ (fun _ => _)).
             assumption.
    - inversions F; injpair.
      + frobinvl p1 H3.
        * eexists _, _.
          split.
          -- constructor.
             eassumption.
          -- left.
             esplit.
             ++ constructor.
             ++ econstructor.
                ** constructor.
                ** esplit with (Method:=BaseMethod).
                   constructor.
        * eexists _, _.
          split.
          -- constructor.
             eassumption.
          -- left.
             esplit.
             ++ constructor.
             ++ reflexivity.
      + frobinvl p0 H3.
        * eexists _, _.
          split.
          -- constructor.
             eassumption.
          -- left.
             esplit.
             ++ constructor.
             ++ simpl.
                eexists.
                ** constructor.
                ** split.
                   constructor.
        * eexists _, _.
          split.
          -- constructor.
             eassumption.
          -- right.
             split.
             ++ reflexivity.
             ++ econstructor 3 with (Method:=BaseMethod).
  Qed.

  Lemma flattenThreadInteraction ps1 ps1' fts ps2 Ret int a :
    CombineFlattenThreads fts ps2 ps1 ->
    ThreadsInteraction' (flattenObject base Method impl) Ret int ps1 a ps1' ->
    exists fts',
      CombineFlattenThreads fts' ps2 ps1' /\
      FlattenInteraction impl Ret int fts a fts'.
  Proof.
    intros F ti.
    revert fts ps2 F.
    induction ti; intros.
    - eexists.
      split.
      + constructor.
        eassumption.
      + constructor.
    - inversions F; injpair.
      + frobinvl p0 H3.
        eexists.
        split.
        * eassumption.
        * constructor.
      + frobinvl p H3.
    - inversions F.
      + edestruct IHti as [? [? ?]].
        * eassumption.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- constructor.
             assumption.
      + edestruct IHti as [? [? ?]].
        * eassumption.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- constructor.
             assumption.
  Qed.

  Theorem flattenObjectBehavior_ st :
    Simulates_
      (ThreadsBehavior (flattenObject base Method impl))
      (FlattenBehavior (ThreadsBehavior base) impl)
      (nil, st)
      (nil, (nil, st)).
  Proof.
    apply mapStateSimulates with (fun ts fs => FBIRel fs ts).
    - clear.
      intros [ps1 st1] [ps1' st1'] [fts [ps2 st2]] [F E] ta.
      subst.
      edestruct flattenThreadAdvance as [fts' [ps2' [S [| [? ?]]]]].
      + eassumption.
      + eassumption.
      + eexists.
        * apply simulateFlattenAdvance.
          eassumption.
        * repeat esplit; assumption.
      + subst.
        eexists.
        * apply flattenBehaviorSimulates'_, simulateAdvance, clos_rt1n_step.
          eassumption.
        * repeat split; eassumption.
    - clear.
      intros Ret [ps1 st1] a [ps1' st1'] [fts [ps2 st2]] [F E] int ti.
      subst.
      inversions ti.
      edestruct flattenThreadInteraction as [? [? ?]].
      + eassumption.
      + eassumption.
      + eexists.
        * esplit.
          -- apply rt1n_refl.
          -- eassumption.
        * split.
          -- assumption.
          -- reflexivity.
    - split; constructor.
  Qed.

  Theorem flattenObjectBehavior :
    Simulates
      (ThreadsBehavior (flattenObject base Method impl))
      (FlattenBehavior (ThreadsBehavior base) impl).
  Proof.
    apply flattenObjectBehavior_.
  Qed.

  Lemma combineFlattenThreadAdvance fts ps1 st ps1' st' ps2 :
    CombineFlattenThreads fts ps1 ps2 ->
    ThreadAdvance base (ps1, st) (ps1', st') ->
    exists ps2',
      CombineFlattenThreads fts ps1' ps2' /\
      ThreadAdvance (flattenObject base Method impl) (ps2, st) (ps2', st').
  Proof.
    intro F.
    revert ps1'.
    induction F; intros ps1' ta; inversions ta; injpair.
    - edestruct IHF as [? [? ?]].
      + constructor.
        eassumption.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          assumption.
    - edestruct IHF as [? [? ?]].
      + econstructor 2.
        eassumption.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          assumption.
    - edestruct IHF as [? [? ?]].
      + econstructor 3.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          assumption.
    - edestruct IHF as [? [? ?]].
      + eassumption.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          assumption.
    - eexists.
      split.
      * constructor.
        eassumption.
      * rewrite (frobProgramId (bindFlattenProgram _ _ _)).
        simpl.
        apply (TABind (flattenObject base Method impl) _ _ _ _ _ (fun _ => _)).
        assumption.
    - eexists.
      split.
      * constructor.
        eassumption.
      * rewrite (frobProgramId (bindFlattenProgram _ _ _)).
        constructor 3 with (Method:=Method).
  Qed.

  Lemma combineFlattenThreadsAdvance fts ps1 st ps1' st' ps2 :
    CombineFlattenThreads fts ps1 ps2 ->
    ThreadsAdvance base (ps1, st) (ps1', st') ->
    exists ps2',
      CombineFlattenThreads fts ps1' ps2' /\
      ThreadsAdvance (flattenObject base Method impl) (ps2, st) (ps2', st').
  Proof.
    intros F tas.
    remember (ps1, st) as ts.
    remember (ps1', st') as ts'.
    revert fts ps1 st ps1' st' ps2 F Heqts Heqts'.
    induction tas; intros; subst.
    - inversions Heqts'. eexists.
      split.
      + eassumption.
      + constructor.
    - destruct y as [ps1'' st''].
      edestruct combineFlattenThreadAdvance as [? [? ?]].
      + eassumption.
      + eassumption.
      + edestruct IHtas as [? [? ?]].
        * eassumption.
        * reflexivity.
        * reflexivity.
        * eexists; split; [ eassumption | ].
          econstructor; eassumption.
  Qed.

  Lemma combineFlattenThreadsInteraction fts st fts' ps1' ps2 Ret int a ps1 :
    CombineFlattenThreads fts ps1 ps2 ->
    FlattenAdvance' Ret (Some int) fts a fts' ->
    ThreadsInteraction' base Ret int ps1 a ps1' ->
   exists ps2',
     CombineFlattenThreads fts' ps1' ps2' /\
     ThreadAdvance (flattenObject base Method impl) (ps2, st) (ps2', st).
  Proof.
    intros F fa.
    remember (Some int) as ints in fa.
    revert st ps1' ps2 ps1 F Heqints.
    induction fa; intros; subst; try inversions Heqints; inversions F; injpair.
    - inversions H; injpair. eexists.
      split.
      + constructor. eassumption.
      + rewrite (frobProgramId (flattenProgram _ _)).
        constructor 3 with (Method:=Method).
    - inversions H; injpair. eexists.
      split.
      + constructor.
        eassumption.
      + rewrite (frobProgramId (bindFlattenProgram _ _ _)).
        constructor 3 with (Method:=Method).
    - edestruct IHfa as [? [? ?]].
      + eassumption.
      + reflexivity.
      + eassumption.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          eassumption.
    - destruct int0; [ | discriminate]. inversions H1. inversions H. injpair.
      edestruct IHfa as [? [? ?]].
      + eassumption.
      + reflexivity.
      + eassumption.
      + eexists.
        split.
        * constructor.
          eassumption.
        * constructor 1 with (Method:=Method).
          eassumption.
  Qed.

  Lemma flattenThreadAdvance_r fts ps1 st fts' ps1' st' ps2 :
    CombineFlattenThreads fts ps1 ps2 ->
    FlattenAdvance (ThreadsBehavior base) (fts, (ps1, st)) (fts', (ps1', st')) ->
    exists ps2',
      CombineFlattenThreads fts' ps1' ps2' /\
      ThreadsAdvance (flattenObject base Method impl) (ps2, st) (ps2', st').
  Proof.
    intros F fa.
    inversions fa.
    destruct int.
    - destruct H4.
      inversions H0.
      edestruct combineFlattenThreadsAdvance as [? [? ?]].
      + eassumption.
      + eassumption.
      + edestruct combineFlattenThreadsInteraction as [? [? ?]].
        * eassumption.
        * eassumption.
        * eassumption.
        * eexists.
          split.
          -- eassumption.
          -- eapply clos_rt1n_trans, clos_rt1n_step; eassumption.
    - inversions H4.
      remember None as ints in H2.
      revert ps1' st' ps2 F Heqints.
      induction H2; intros; try inversions Heqints; inversions F; injpair.
      + eexists.
        split.
        * constructor.
          eassumption.
        * apply clos_rt1n_step.
          rewrite (frobProgramId (flattenProgram _ _)).
          constructor 3 with (Method:=Method).
      + edestruct IHFlattenAdvance' as [? [? ?]].
        * eassumption.
        * reflexivity.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- eapply (advanceCons (Method:=Method)).
             eassumption.
      + destruct int; [discriminate |].
        edestruct IHFlattenAdvance' as [? [? ?]].
        * eassumption.
        * eassumption.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- eapply (advanceCons (Method:=Method)).
             eassumption.
  Qed.

  Lemma flattenThreadsAdvance_r fts ps1 st fts' ps1' st' ps2 :
    CombineFlattenThreads fts ps1 ps2 ->
    FlattenAdvances (ThreadsBehavior base) (fts, (ps1, st)) (fts', (ps1', st')) ->
    exists ps2',
      CombineFlattenThreads fts' ps1' ps2' /\
      ThreadsAdvance (flattenObject base Method impl) (ps2, st) (ps2', st').
  Proof.
    intros F fas.
    remember (fts, (ps1, st)) as ts.
    remember (fts', (ps1', st')) as ts'.
    revert fts ps1 st Heqts fts' ps1' st' Heqts' ps2 F.
    induction fas; intros; subst; try inversions Heqts'.
    - eexists.
      split.
      + eassumption.
      + constructor.
    - destruct y as [fts'' [ps1'' st'']].
      edestruct flattenThreadAdvance_r as [? [? ?]].
      + eassumption.
      + eassumption.
      + edestruct IHfas as [? [? ?]].
        * reflexivity.
        * reflexivity.
        * eassumption.
        * eexists.
          split.
          -- eassumption.
          -- eapply clos_rt1n_trans; eassumption.
  Qed.

  Lemma flattenThreadsInteraction_r A int fts ps1 ps2 a fts' :
    CombineFlattenThreads fts ps1 ps2 ->
    FlattenInteraction impl A int fts a fts' ->
    exists ps2',
      CombineFlattenThreads fts' ps1 ps2' /\
      ThreadsInteraction' (flattenObject base Method impl) A int ps2 a ps2'.
  Proof.
    intros F fi.
    revert ps1 ps2 F.
    induction fi; intros.
    - eexists.
      split.
      + econstructor.
        eassumption.
      + constructor.
    - inversions F; injpair.
      eexists.
      split.
      + eassumption.
      + rewrite (frobProgramId (flattenProgram _ _)).
        constructor 2 with (Method:=Method).
    - inversions F.
      + edestruct IHfi as [? [? ?]].
        * eassumption.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- constructor 3 with (Method:=Method).
             assumption.
      + edestruct IHfi as [? [? ?]].
        * eassumption.
        * eexists.
          split.
          -- constructor.
             eassumption.
          -- constructor 3 with (Method:=Method).
             assumption.
  Qed.

  Theorem flattenBehaviorObject st :
    Simulates_
      (FlattenBehavior (ThreadsBehavior base) impl)
      (ThreadsBehavior (flattenObject base Method impl))
      (nil, (nil, st))
      (nil, st).
  Proof.
    apply relationSimulates with FBIRel.
    - clear.
      simpl in *.
      intros A int [fts [ps1 st1]] a [fts' [ps1' st1']] [ps2 st2] [F E] H.
      subst.
      inversions H.
      edestruct flattenThreadsAdvance_r as [? [? ?]].
      + eassumption.
      + eassumption.
      + edestruct flattenThreadsInteraction_r as [? [? ?]].
        * eassumption.
        * eassumption.
        * repeat esplit; eassumption.
    - split; constructor.
  Qed.

End FlattenObjectBehavior.


Section FlattenSimulates.

  Context
    {BaseMethod Method : Type -> Type}
    (base1 base2 : Object BaseMethod)
    (program : forall Ret, Method Ret -> Program BaseMethod Ret).

  Theorem flattenSimulates :
    ConcSimulates base1 base2 ->
    ConcSimulates (flattenObject base1 Method program) (flattenObject base2 Method program).
  Proof.
    intro.
    eapply simulatesTrans, simulatesTrans.
    - eapply flattenObjectBehavior.
    - eapply flattenBehaviorSimulates'.
      eassumption.
    - apply @flattenBehaviorObject.
  Qed.


  Theorem flattenSimulates2 impl :
    ConcSimulates base1 base2 ->
    ConcSimulates (flattenObject base2 Method program) impl ->
    ConcSimulates (flattenObject base1 Method program) impl.

  Proof.
    intros H1 H2.
    eapply simulatesTrans.
    Focus 2. exact H2.
    eapply flattenSimulates.
    assumption.    
  Qed.
  
End FlattenSimulates.

Lemma flattenSeqRef__
    (I : Type -> Type)
    (impl1 impl2 : Object I)
    (A : Type)
    (m m' : Program I A)
    (s1 : State (coarsen impl1))
    (s1' : State (coarsen impl1))
    (s2 : State (coarsen impl2))
    (H0 : Simulates_ (coarsen impl1) (coarsen impl2) s1 s2)
    (H1 : ProgramBigSteps (objBase impl1) _ (objProgram impl1) A m s1 m' s1')
  : exists s2',
      ProgramBigSteps (objBase impl2) _ (objProgram impl2) A m s2 m' s2'
      /\ Simulates_ (coarsen impl1) (coarsen impl2) s1' s2'.
Proof.
  unfold ProgramBigSteps, ProgramBigSteps' in H1.
  revert H1 s2 H0.
  enough (forall ms1 ms1',
    clos_refl_trans_1n (Program I A * State (objBase impl1))
      (ProgramBigStep' (objBase impl1) I (objProgram impl1) A)
      ms1 ms1' ->
    forall s2,
    Simulates_ (coarsen impl1) (coarsen impl2) (snd ms1) s2 ->
    exists (s2' : State (objBase impl2)),
      ProgramBigSteps (objBase impl2) _ (objProgram impl2) A (fst ms1) s2 (fst ms1') s2'
        /\ Simulates_ (coarsen impl1) (coarsen impl2) (snd ms1') s2').
  { apply (H (_, _) (_, _)). }
  clear; induction 1.
  - destruct x; cbn. intros s2; exists s2.
    split; [ apply rt1n_refl | auto ].
  - destruct H.
    + intros s2 [SIM]. apply SIM in H.
      destruct H as [s2' Hm SIM'].
      apply IHclos_refl_trans_1n in SIM'.
      destruct SIM' as [s2'' [Hp SIM'']].
      exists s2''; split; auto.
      eapply ProgramBigStepBigStepsImplBigSteps; eassumption + constructor; auto.
    + intros s2 SIM; specialize (IHclos_refl_trans_1n s2 SIM).
      destruct IHclos_refl_trans_1n as [s2' [Hp SIM']].
      exists s2'; split; auto.
      eapply ProgramBigStepBigStepsImplBigSteps; eassumption + constructor; auto.
Qed.

Lemma ProgramStepsBindProgram I (obj : Spec I)
  : forall R S (m : Program I R) (f : R -> Program I S) s r s' p s'',
    EvalProgram obj R m s r s' ->
    ProgramSteps obj S (f r) s' p s'' ->
    ProgramSteps obj S (bindProgram m f) s p s''.
Proof.
  intros until 1.
  revert R m s r s' H f. apply (EvalProgramInd I obj (fun R m s r s' =>
forall f : R -> Program I S,
ProgramSteps obj S (f r) s' p s'' ->
ProgramSteps obj S (x <-- m; f x) s p s''
)); intros.
  all: rewrite (frobProgramId (bindProgram _ _)); cbn.
  - change (frobProgram (bindProgram _ _)) with (frobProgram (f r)).
    rewrite <- frobProgramId.
    auto.
  - eapply ProgramStepStepsImplSteps.
    { constructor. eassumption. }
    auto.
  - eapply ProgramStepStepsImplSteps. { constructor. } auto.
Qed.

Lemma BigEvalProgramEvalProgramFlatten I (obj : Spec I) M (impl : _)
  : forall R p s r s',
    BigEvalProgram obj M impl R p s r s' ->
    EvalProgram obj R (flattenProgram impl p) s r s'.
Proof.
  apply BigEvalProgramInd; intros;
    unfold EvalProgram;
    rewrite (frobProgramId (flattenProgram _ _)); cbn.
  - apply ProgramStepsRefl.
  - eapply ProgramStepStepsImplSteps.
    { constructor. }
    eapply EvalProgramProgramEq.
    { symmetry. apply bindFlattenProgramBindProgram. }
    red. eapply ProgramStepsBindProgram; try eassumption.
    eapply ProgramStepStepsImplSteps.
    { constructor. }
    assumption.
  - eapply ProgramStepStepsImplSteps.
    { constructor. }
    assumption.
Qed.

Lemma flattenSeqRef___
    (I : Type -> Type)
    (impl1 impl2 : Object I)
    (M : Type -> Type)
    (implProg : forall R : Type, M R -> Program I R)
    (A : Type)
    (a : A)
    (int : M A)
    (s1 : State (coarsen (flattenObject impl1 M implProg)))
    (s1' : State (coarsen (flattenObject impl1 M implProg)))
    (s2 : State (coarsen (flattenObject impl2 M implProg)))
    (H0 : Simulates_ (coarsen impl1) (coarsen impl2) s1 s2)
    (H1 : EvalProgram (objBase impl1) A
       (flattenImplProg (objProgram impl1) implProg A int) s1 a s1')
  : exists2 s2',
      EvalProgram (objBase impl2) A (flattenImplProg (objProgram impl2) implProg A int) s2 a s2'
      & Simulates_ (coarsen impl1) (coarsen impl2) s1' s2'.
Proof.
  unfold flattenImplProg in *.
  apply EvalProgramFlattenBigEvalProgram in H1.
  eapply flattenSeqRef__ in H1; try eassumption.
  destruct H1 as [s2' [Hp SIM]].
  exists s2'; auto.
  apply BigEvalProgramEvalProgramFlatten.
  auto.
Qed.

Lemma flattenSeqRef_:
  forall I (impl1 impl2: Object I)
    M (implProg: forall R, M R -> Program I R)
    st1 st2,
    Simulates_
      (coarsen impl1) (coarsen impl2)
      st1 st2 ->
    Simulates_
      (coarsen (flattenObject impl1 M implProg))
      (coarsen (flattenObject impl2 M implProg))
      st1 st2.
Proof.
  intros.
  apply (relationSimulates (coarsen (flattenObject impl1 M implProg)) (coarsen (flattenObject impl2 M implProg)) (Simulates_ (coarsen impl1) (coarsen impl2))).
  { intros.  cbn in *.  eapply flattenSeqRef___; eauto. }
  auto.
Qed.

Lemma flattenSeqRef :
  forall I (impl1 impl2: Object I)
    M (implProg: forall R, M R -> Program I R),
    Simulates (coarsen impl1) (coarsen impl2) ->
    Simulates (coarsen (flattenObject impl1 M implProg)) (coarsen (flattenObject impl2 M implProg)).
Proof.
  intros *; apply flattenSeqRef_.
Qed.
