Set Asymmetric Patterns.
From Coq Require Import
  Arith.Peano_dec
  Sets.Ensembles
  Sets.Image
  Logic.FunctionalExtensionality
  Structures.Equalities
  Setoid
  Morphisms.
From C4.Util Require Import
  Util
  LibTactics
  MTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  Superposition
  Pair
  PairFacts
  MethodSimulates
  ProgramOps
  ProgramOpsFacts
  Object
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  Linearizability
  LinearizabilityTactics
  MapSimulates
  ImplSimulates
  RSimulates
  RSimulatesFacts.

Notation HConcSimulates i1 i2 := (HSimulates (ThreadsBehavior i1) (ThreadsBehavior i2)).

Notation fI := flattenObject.

Lemma simulatesHSimulates M (o1 o2 : Spec M) s1 s2
  : Simulates_ o1 o2 s1 s2 -> HSimulates_ o1 o2 s1 (eq s2).
Proof.
  intro H.
  eapply simulatesTrans_.
  - apply H.
  - apply superpositionSingleSimulates; eauto.
Qed.

Lemma HConcSimulatesEq {M} (i1 i2 : Object M)
  : HImplSimulates i1 i2 -> HConcSimulates i1 i2.
Proof.
  apply Proper_ThreadsBehavior.
Qed.

Lemma mapSerial {M N} (f : forall R, M R -> N R) (o : Spec N)
  : HConcSimulates (mapObject f (serialObject o)) (serialObject (mapSpec o f)).
Proof.
  apply HConcSimulatesEq, mapSerialEq.
Qed.

Lemma HConcSimulatesConcSimulates {M} (i1 i2 : Object M)
  : ConcSimulates i1 i2 -> HConcSimulates i1 i2.
Proof.
  apply simulatesHSimulates.
Qed.

