Require Import
        Coq.Logic.Eqdep
        Coq.Lists.List
        Coq.omega.Omega
        Coq.Relations.Relations
        Coq.Program.Equality.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  Object
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts.

Module SimulatesCoIndN.
Section SimulatesCoIndN.

  Context
    {Method : Type -> Type}
    (impl1 impl2 : Object Method).

  Definition ProgsState {M} (impl : Object M) :=
    list (sigT (Program impl.(ObjInstruction))).

  Definition DataState {M} (impl : Object M) :=
    impl.(objBase).(State).

  Variable (PDRel: forall R,
               Program impl1.(ObjInstruction) R -> impl1.(objBase).(State) ->
               Program impl2.(ObjInstruction) R -> impl2.(objBase).(State) ->
               Prop).
  Arguments PDRel {R} _ _ _ _.

  (* The concrete program can have extra Noops at the beginning. *)
  Inductive NPDRel: forall R,
      Program impl1.(ObjInstruction) R -> impl1.(objBase).(State) ->
      Program impl2.(ObjInstruction) R -> impl2.(objBase).(State) ->
      Prop :=
  | NPDRel1 R p1 s1 p2 s2:
      @PDRel R p1 s1 p2 s2 ->
      NPDRel R p1 s1 p2 s2
  | NPDRel2 R p1 s1 p2 s2:
      NPDRel R p1 s1 p2 s2 ->
      NPDRel R (Noop p1) s1 p2 s2.

  Arguments NPDRel {R} _ _ _ _.

  Variable (DRel: impl1.(objBase).(State) -> impl2.(objBase).(State) -> Prop).

  (*
   * DRel cannot be defined as part of PDRel because then
   * invRel requires that it be proved for every s1 and s2.
   *)

  Variable (PPRel: forall R R',
               Program impl1.(ObjInstruction) R  -> Program impl2.(ObjInstruction) R ->
               Program impl1.(ObjInstruction) R' -> Program impl2.(ObjInstruction) R' ->
               DataState impl1 -> DataState impl2 ->
               Prop).
  (* Arguments PPRel {R R'} _ _ _ _ _ _. *)
  Arguments PPRel {R R'} _ _ _ _ _ _.


  (*
  Hypothesis PDRelNoop:
    forall R (p1: Program impl1.(ObjInstruction) R) p2 s1 s2,
      PDRel (Noop p1) s1 p2 s2 ->
      PDRel p1 s1 p2 s2.

  Lemma NPDRelNoop:
    forall R (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      s1 s2,

      NPDRel (Noop p1) s1 p2 s2 ->
      NPDRel p1 s1 p2 s2.

  Proof.
    introv H1.
    clear -H1 PDRelNoop.
    remember (Noop p1) as np1.
    inverts H1; subst.
    - constructor. apply PDRelNoop. assumption.
    - inverts H. assumption.
  Qed.
  *)

  Hypothesis PDRelNoop:
    forall R (p1: Program impl1.(ObjInstruction) R) p2 s1 s2,
      PDRel (Noop p1) s1 p2 s2 ->
      False.

  Lemma NPDRelNoop:
    forall R (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      s1 s2,

      NPDRel (Noop p1) s1 p2 s2 ->
      NPDRel p1 s1 p2 s2.

  Proof.
    introv H1.
    clear -H1 PDRelNoop.
    remember (Noop p1) as np1.
    inverts H1; subst.
    - exfalso. eapply PDRelNoop. eassumption.
    - inverts H. assumption.
  Qed.

  Hypothesis PPRelSym:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2' s1 s2,

      PPRel p1 p2 p1' p2' s1 s2 ->
      PPRel p1' p2' p1 p2 s1 s2.

  (*
  Hypothesis PPRelNoop:
    forall R R'
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R') s1 s2,

      PPRel (Noop p1) p2 p1' p2' s1 s2 ->
      PPRel p1 p2 p1' p2' s1 s2.
  *)


  Hypothesis PPRelNoop:
    forall R R'
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R') s1 s2,

      PPRel (Noop p1) p2 p1' p2' s1 s2 ->
      False.

  Inductive NPPRel: forall R R',
      Program impl1.(ObjInstruction) R ->
      Program impl2.(ObjInstruction) R ->
      Program impl1.(ObjInstruction) R' ->
      Program impl2.(ObjInstruction) R' ->
      DataState impl1 ->
      DataState impl2 ->
      Prop :=
  | NPPRel1 R R' p1 p2 p1' p2' s1 s2:
      @PPRel R R' p1 p2 p1' p2' s1 s2 ->
      NPPRel R R' p1 p2 p1' p2' s1 s2
  | NPPRel2 R R' p1 p2 p1' p2' s1 s2:
      NPPRel R R' p1 p2 p1' p2' s1 s2 ->
      NPPRel R R' (Noop p1) p2 p1' p2' s1 s2
  | NPPRel3 R R' p1 p2 p1' p2' s1 s2:
      NPPRel R R' p1 p2 p1' p2' s1 s2 ->
      NPPRel R R' p1 p2 (Noop p1') p2' s1 s2.
  (*
  | NPPRel4 R R' p1 p2 p1' p2':
      NPPRel R R' p1 p2 p1' p2' ->
      NPPRel R R' (Noop p1) p2 (Noop p1') p2'.
  *)

  Arguments NPPRel {R R'} _ _ _ _ _ _.

  Hint Constructors NPPRel.

  Lemma NPPRelSym:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2'
      s1 s2,

      NPPRel p1 p2 p1' p2' s1 s2 ->
      NPPRel p1' p2' p1 p2 s1 s2.

  Proof.
    introv H1.
    induction H1; eauto.
  Qed.

  Lemma NPPRelNoop1:
    forall R R'
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R') s1 s2,

      NPPRel (Noop p1) p2 p1' p2' s1 s2->
      NPPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1.
    clear -H1 PPRelNoop.
    remember (Noop p1) as np1.
    induction H1; subst.
    - exfalso. eapply PPRelNoop. eassumption.
    - congruence.
    - apply NPPRel3. auto.
  Qed.

  Lemma NPPRelNoop2:
    forall R R'
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R') s1 s2,

      NPPRel p1 p2 (Noop p1') p2' s1 s2 ->
      NPPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1.
    clear -H1 PPRelNoop PPRelSym.
    remember (Noop p1') as np1'.
    induction H1; subst.
    - exfalso. eapply PPRelNoop. eapply PPRelSym. eassumption.
    - apply NPPRel2. auto.
    - congruence.
  Qed.

  Lemma NPPRelExistPPRel:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2' s1 s2,
      PDRel p1 s1 p2 s2 ->
      PDRel p1' s1 p2' s2 ->
      NPPRel p1 p2 p1' p2' s1 s2 ->
      PPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1 H2 H3.
    clear -H1 H2 H3 PDRelNoop.
    induction H3.
    - eassumption.
    - exfalso. eapply PDRelNoop. eassumption.
    - exfalso. eapply PDRelNoop. eassumption.
  Qed.

  Lemma NPPRelExistPPRel':
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R'' (p1'': Program impl1.(ObjInstruction) R'') p2'' s1 s2,
      PDRel p1'' s1 p2'' s2 ->
      NPPRel p1 p2 p1'' p2'' s1 s2 ->
      exists (p1': Program impl1.(ObjInstruction) R) p2',
        PPRel p1' p2' p1'' p2'' s1 s2.

  Proof.
    introv H1 H2.
    clear -H1 H2 PDRelNoop.
    induction H2.
    - eexists. eexists. eassumption.
    - eauto.
    - exfalso. eapply PDRelNoop. eassumption.
  Qed.

  Lemma NPPRelExistPPRel'':
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2' s1 s2,
      (forall np1, p1 <> Noop np1) ->
      PDRel p1' s1 p2' s2 ->
      NPPRel p1 p2 p1' p2' s1 s2 ->
      PPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1 H2 H3.
    clear -H1 H2 H3 PDRelNoop.
    induction H3.
    - eassumption.
    - exfalso. eapply H1. reflexivity.
    - exfalso. eapply PDRelNoop. eassumption.
  Qed.

  Lemma NPPRelExistPPRel''':
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2' s1 s2,

      PDRel p1 s1 p2 s2 ->
      (forall np1', p1' <> Noop np1') ->
      NPPRel p1 p2 p1' p2' s1 s2 ->
      PPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1 H2 H3.
    apply PPRelSym.
    apply NPPRelSym in H3.
    eapply NPPRelExistPPRel''; eassumption.
  Qed.

  Lemma NPDRelExistPDRel:
    forall R (p1: Program impl1.(ObjInstruction) R) p1'
      R' (p2: Program impl1.(ObjInstruction) R') p2' s1 s2,
      NPDRel p1 s1 p1' s2 ->
      PPRel p1 p1' p2 p2' s1 s2 ->
      PDRel p1 s1 p1' s2.

  Proof.
    introv H1 H2.
    clear -H1 H2 PPRelNoop.
    induction H1.
    - assumption.
    - exfalso. eapply PPRelNoop. eassumption.
  Qed.

  Hypothesis initRel: DRel impl1.(objBase).(initialState) impl2.(objBase).(initialState).

  (*
   * If a method m is invoked in both the concrete and abstract objects, then
   * (1) PDRel holds between them and
   * (2) For any pair of concrete and abstract programs in the pool that PDRel holds, PPRel holds between the that pair and the new pair.
   *)
  Hypothesis invRel:
    forall R m s1 s2,
      DRel s1 s2 ->
      NPDRel (objProgram impl1 R m) s1
             (objProgram impl2 R m) s2
      /\ forall R' (p1: Program impl1.(ObjInstruction) R') p2,
          PDRel p1 s1 p2 s2 ->
          NPPRel (objProgram impl1 R m) (objProgram impl2 R m)
                 p1 p2
                 s1 s2.

  Lemma NPDRelNPPRel:
    forall s1 s2 R' (p1': Program impl1.(ObjInstruction) R') p2',
    (forall R (p1: Program impl1.(ObjInstruction) R) p2,
        PDRel p1 s1 p2 s2 ->
        NPPRel p1' p2' p1 p2 s1 s2) ->
    (forall R (p1: Program impl1.(ObjInstruction) R) p2,
        NPDRel p1 s1 p2 s2 ->
        NPPRel p1' p2' p1 p2 s1 s2).

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H2.
    - eauto.
    - apply NPPRel3. eauto.
  Qed.

  Lemma invRel':
    forall R m s1 s2,
      DRel s1 s2 ->
      NPDRel (objProgram impl1 R m) s1
             (objProgram impl2 R m) s2
      /\ forall R' (p1: Program impl1.(ObjInstruction) R') p2,
          NPDRel p1 s1 p2 s2 ->
          NPPRel (objProgram impl1 R m) (objProgram impl2 R m)
                 p1 p2
                 s1 s2.

  Proof.
    introv H1.
    clear -H1 invRel.
    splits.
    - apply invRel. assumption.
    - eapply invRel in H1; destruct H1 as (_ & H1); clear invRel.
      eapply NPDRelNPPRel. eassumption.
  Qed.

  Hypothesis retRel:
    forall R (r: R) s1 p2 s2,
      PDRel (Return r) s1 p2 s2 ->
      DRel s1 s2 ->
      p2 = Return r.

  Hypothesis stepRel:
    forall R' (m: impl1.(ObjInstruction) R') R (f: R' -> Program impl1.(ObjInstruction) R) s1 p2 s2 s1' r,
      DRel s1 s2 ->
      PDRel (Bind m f) s1 p2 s2 ->
      impl1.(objBase).(EvalInstruction) m s1 r s1' ->
      exists p2' s2',
        ProgramSteps impl2.(objBase) _ p2 s2 p2' s2'
        /\ DRel s1' s2'
        /\ NPDRel (f r) s1' p2' s2'

        /\ (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
              PDRel p1'' s1 p2'' s2 ->
              PPRel (Bind m f) p2 p1'' p2'' s1 s2 ->
              PDRel p1'' s1' p2'' s2')

        /\ (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
              PDRel p1'' s1 p2'' s2 ->
              PPRel (Bind m f) p2 p1'' p2'' s1 s2 ->
              NPPRel (f r) p2' p1'' p2'' s1' s2')

(*
        /\ (forall R (p11: Program impl1.(ObjInstruction) R) p21
             R' (p12: Program impl1.(ObjInstruction) R') p22,
              PDRel p11 s1 p21 s2 ->
              PDRel p12 s1 p22 s2 ->
              PPRel p11 p21 p12 p22 s1 s2 ->
              PPRel p11 p21 p12 p22 s1' s2').
*)

        /\ (forall R (p11: Program impl1.(ObjInstruction) R) p21
             R' (p12: Program impl1.(ObjInstruction) R') p22,
              PDRel p11 s1 p21 s2 ->
              PDRel p12 s1 p22 s2 ->
              (*
              PPRel p11 p21 (Bind m f) p2 s1 s2 ->
              PPRel p12 p22 (Bind m f) p2 s1 s2 ->
              *)
              PPRel p11 p21 p12 p22 s1 s2 ->
              PPRel p11 p21 p12 p22 s1' s2').




  Inductive NPPsRel: forall R,
      Program impl1.(ObjInstruction) R ->
      Program impl2.(ObjInstruction) R ->
      ProgsState impl1 ->
      ProgsState impl2 ->
      DataState impl1 ->
      DataState impl2 ->
      Prop :=

  | NilNPPsRel:
      forall R p1 p2 s1 s2,
        NPPsRel R p1 p2 nil nil s1 s2
  | ConsNPPsRel:
      forall R p1 ps1 p2 ps2 R' p1' p2' s1 s2,
        NPPRel p1 p2 p1' p2' s1 s2 ->
        NPPsRel R p1 p2 ps1 ps2 s1 s2 ->
        NPPsRel R p1 p2 (existT _ R' p1' :: ps1) (existT _ R' p2' :: ps2) s1 s2.
  Arguments NPPsRel {R} _ _ _ _ _ _.

  Lemma invertNPPsRel
        (R: Type)
        (p1: Program impl1.(ObjInstruction) R)
        (p2: Program impl2.(ObjInstruction) R)
        (ps1: ProgsState impl1)
        (ps2: ProgsState impl2)
        (ds1: DataState impl1)
        (ds2: DataState impl2):
    @NPPsRel R p1 p2 ps1 ps2 ds1 ds2 ->
    (ps1 = nil
     /\ ps2 = nil) \/
    (exists R p1' p2' ps1' ps2',
        existT _ R p1' :: ps1' = ps1
        /\ existT _ R p2' :: ps2' = ps2
        /\ NPPRel p1 p2 p1' p2' ds1 ds2
        /\ NPPsRel p1 p2 ps1' ps2' ds1 ds2).

  Proof.
    clear.
    introv H1.
    destruct H1;
      [left | right];
      jauto.
  Qed.

  Lemma NPPsRelInd:
    forall (P: forall R,
          Program impl1.(ObjInstruction) R ->
          Program impl2.(ObjInstruction) R ->
          ProgsState impl1 ->
          ProgsState impl2 ->
          DataState impl1 ->
          DataState impl2 ->
          Prop),

      (forall R p1 p2 s1 s2,
          P R p1 p2 nil nil s1 s2) ->
      (forall R p1 ps1 p2 ps2 R' p1' p2' s1 s2,
          NPPRel p1 p2 p1' p2' s1 s2 ->
          NPPsRel p1 p2 ps1 ps2 s1 s2 ->
          P R p1 p2 ps1 ps2 s1 s2 ->
          P R p1 p2 (existT _ R' p1' :: ps1) (existT _ R' p2' :: ps2) s1 s2) ->
      forall R p1 p2 ps1 ps2 s1 s2,
        NPPsRel p1 p2 ps1 ps2 s1 s2 ->
        P R p1 p2 ps1 ps2 s1 s2.

  Proof.
    clear invRel retRel stepRel PDRelNoop initRel.
    introv H1 H2 H3.
    induction H3; eauto.
  Qed.

  Lemma NPPsRelNoop:
    forall (Ret : Type)
      (p1 : Program (ObjInstruction impl1) Ret)
      (ps1 : ProgsState impl1)
      (p2 : Program (ObjInstruction impl2) Ret)
      (ps2 : ProgsState impl2)
      s1 s2,

      NPPsRel (Noop p1) p2 ps1 ps2 s1 s2 ->
      NPPsRel p1 p2 ps1 ps2 s1 s2.

  Proof.
    introv H1.
    clear -H1 PPRelNoop.
    remember (Noop p1) as np1.
    induction H1; subst.
    - constructor.
    - constructor.
      + eapply NPPRelNoop1. assumption.
      + auto.
  Qed.

  Inductive PIn:
    forall R,
      Program impl1.(ObjInstruction) R ->
      Program impl2.(ObjInstruction) R ->
      ProgsState impl1 ->
      ProgsState impl2 ->
      Prop :=
  | PIn1 R p1 p2 ps1 ps2:
      PIn R p1 p2 (existT _ R p1 :: ps1) (existT _ R p2 :: ps2)
  | PIn2 R R' p1 p2 p1' p2' ps1 ps2:
      PIn R p1 p2 ps1 ps2 ->
      PIn R p1 p2 (existT _ R' p1' :: ps1) (existT _ R' p2' :: ps2).

  Arguments PIn {R} _ _ _ _.

  Lemma invertPIn
        (R: Type)
        (p1: Program impl1.(ObjInstruction) R)
        (p2: Program impl2.(ObjInstruction) R)
        (ps1: ProgsState impl1)
        (ps2: ProgsState impl2):
    @PIn R p1 p2 ps1 ps2 ->
    (exists ps1' ps2',
        existT _ R p1 :: ps1' = ps1 /\ existT _ R p2 :: ps2' = ps2) \/
    (exists p1' p2' ps1' ps2',
        p1' :: ps1' = ps1 /\ p2' :: ps2' = ps2 /\
        PIn p1 p2 ps1' ps2').

  Proof.
    clear.
    introv H1.
    destruct H1.
    - left.
      repeat esplit.
    - right.
      repeat esplit.
      assumption.
  Qed.


  Lemma NPPsRelImplNPPRel:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2'
      (ps1: ProgsState impl1) ps2 s1 s2,

      NPPsRel p1 p2 ps1 ps2 s1 s2 ->
      PIn p1' p2' ps1 ps2 ->
      NPPRel p1 p2 p1' p2' s1 s2.

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H2.
    - apply invertNPPsRel in H1.
      destruct H1 as [(? & ?) | (R' & p1' & p2' & ps1' & ps2' & Hps1' & Hps2' & H11 & H12)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      assumption.
    - apply invertNPPsRel in H1.
      destruct H1 as [(? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H11 & H12)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      eauto.
  Qed.


  Inductive NPsDRel :
    ProgsState impl1 -> DataState impl1 ->
    ProgsState impl2 -> DataState impl2 -> Prop :=
  | NilNPsDRel:
      forall s1 s2,
        DRel s1 s2 ->
        NPsDRel nil s1 nil s2
  | ConsNPsDRel:
      forall R p1 s1 ps1 p2 s2 ps2,
        (* DRel s1 s2 -> *)
        (* The concrete can have extra Noops at the beginning. *)
        NPDRel p1 s1 p2 s2 ->
        NPPsRel p1 p2 ps1 ps2 s1 s2 ->
        NPsDRel ps1 s1 ps2 s2 ->
        NPsDRel (existT _ R p1 :: ps1) s1 (existT _ R p2 :: ps2) s2.
  Arguments NPsDRel _ _ _ _.

  Lemma NPsDRelImplDRel:
    forall (ps1 : ProgsState impl1) s1
      ps2 s2,

      NPsDRel ps1 s1 ps2 s2 ->
      DRel s1 s2.

  Proof.
    introv H1.
    clear -H1.
    induction H1; assumption.
  Qed.

  Lemma NPsDRelInd:
    forall (P : ProgsState impl1 -> DataState impl1 ->
                ProgsState impl2 -> DataState impl2 -> Prop),

      (forall s1 s2,
          DRel s1 s2 ->
          P nil s1 nil s2) ->
      (forall R p1 s1 ps1 p2 s2 ps2,
          DRel s1 s2 ->
          NPDRel p1 s1 p2 s2 ->
          NPPsRel p1 p2 ps1 ps2 s1 s2 ->
          NPsDRel ps1 s1 ps2 s2 ->
          P ps1 s1 ps2 s2 ->
          P (existT _ R p1 :: ps1) s1 (existT _ R p2 :: ps2) s2) ->
      forall p s p' s',
        NPsDRel p s p' s' -> P p s p' s'.

  Proof.
    introv H1 H2 H3.
    induction H3.
    - auto.
    - eapply H2; try eassumption.
      eapply NPsDRelImplDRel. eassumption.
  Qed.

  Lemma invertNPsDRel (ps1 : ProgsState impl1) ds1 (ps2 : ProgsState impl2) ds2 :
    NPsDRel ps1 ds1 ps2 ds2 ->
    (ps1 = nil
     /\ ps2 = nil
     /\ DRel ds1 ds2) \/
    (exists R p1 p2 ps1' ps2',
        existT _ R p1 :: ps1' = ps1
        /\ existT _ R p2 :: ps2' = ps2
        /\ NPDRel p1 ds1 p2 ds2
        /\ DRel ds1 ds2
        /\ NPPsRel p1 p2 ps1' ps2' ds1 ds2
        /\ NPsDRel ps1' ds1 ps2' ds2).

  Proof.
    clear.
    introv H1.
    destruct H1; [left | right; eexists _, _, _, _, _]; splits; jauto.
    induction H1 using NPsDRelInd; assumption.
  Qed.

  Lemma NPsDRelImplNPDRel:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      (ps1: ProgsState impl1) ps2 s1 s2,

      NPsDRel ps1 s1 ps2 s2 ->
      PIn p1 p2 ps1 ps2 ->
      NPDRel p1 s1 p2 s2.

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H2.
    - apply invertNPsDRel in H1.
      destruct H1 as [(? & ?) | (R' & p3 & p4 & ps1' & ps2' & Hps1' & Hps2' & H11 & H12 & H13 & H14)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      assumption.
    - apply invertNPsDRel in H1.
      destruct H1 as [(? & ?) | (R'' & p3 & p4 & ps1' & ps2' & Hps1' & Hps2' & H11 & H12 & H13 & H14)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      eauto.
  Qed.

  Lemma NPDRelPres:
    forall s1 s2 s1' s2',

      (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
          PDRel p1'' s1 p2'' s2 ->
          PDRel p1'' s1' p2'' s2') ->

      (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
          NPDRel p1'' s1 p2'' s2 ->
          NPDRel p1'' s1' p2'' s2').

  Proof.
    introv H1 H2.
    clear -H1 H2 PDRelNoop PPRelSym PPRelNoop.
    - induction H2.
      + constructor. auto.
      + apply NPDRel2. apply IHNPDRel; try assumption.
  Qed.


  Lemma NPDRelPres2Co:
    forall s1 s2 s1' s2',

      (forall R (p1: Program impl1.(ObjInstruction) R) p2
         R' (p1': Program impl1.(ObjInstruction) R') p2',
          PDRel p1' s1 p2' s2 ->
          NPPRel p1 p2 p1' p2' s1 s2 ->
          PDRel p1' s1' p2' s2') ->

      (forall R (p1: Program impl1.(ObjInstruction) R) p2
         R' (p1': Program impl1.(ObjInstruction) R') p2',
          NPDRel p1' s1 p2' s2 ->
          NPPRel p1 p2 p1' p2' s1 s2 ->
          NPDRel p1' s1' p2' s2').

  Proof.
    introv H1 H2 H3.
    clear -H1 H2 H3 PDRelNoop PPRelSym PPRelNoop.
    - induction H2.
      + constructor. eapply H1.
        * assumption.
        * eassumption.
      + apply NPDRel2. apply IHNPDRel; try assumption. apply NPPRelNoop2. assumption.
  Qed.


  Lemma NPDRelPres2Co':
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      s1 s2 s1' s2',

      (forall p1', p1 <> Noop p1') ->
      (forall R' (p1': Program impl1.(ObjInstruction) R') p2',
          PDRel p1' s1 p2' s2 ->
          PPRel p1 p2 p1' p2' s1 s2 ->
          PDRel p1' s1' p2' s2') ->

      (forall R' (p1': Program impl1.(ObjInstruction) R') p2',
          NPDRel p1' s1 p2' s2 ->
          NPPRel p1 p2 p1' p2' s1 s2 ->
          NPDRel p1' s1' p2' s2').

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    - induction H3.
      + constructor. eapply H2.
        * assumption.
        * eapply NPPRelExistPPRel''; assumption.
      + apply NPDRel2. apply IHNPDRel; try assumption. apply NPPRelNoop2. assumption.
  Qed.


  Lemma NPPRelPres:
    forall s1 s2 s1' s2',

      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p2
         (p1' : Program impl1.(ObjInstruction) R2) p2',
          PPRel p1 p2 p1' p2' s1 s2 ->
          PPRel p1 p2 p1' p2' s1' s2') ->

      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p2
         (p1' : Program impl1.(ObjInstruction) R2) p2',
          NPPRel p1 p2 p1' p2' s1 s2 ->
          NPPRel p1 p2 p1' p2' s1' s2').

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H2.
    * constructor. eauto.
    * apply NPPRel2. auto.
    * apply NPPRel3. auto.
  Qed.

  Lemma NPPRelPresCo:
    forall s1 s2 s1' s2',

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 s1 p1' s2 ->
          PDRel p2 s1 p2' s2 ->
          PPRel p1 p1' p2 p2' s1 s2 ->
          NPPRel p1 p1' p2 p2' s1' s2') ->

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 s1 p1' s2 ->
          NPDRel p2 s1 p2' s2 ->
          NPPRel p1 p1' p2 p2' s1 s2 ->
          NPPRel p1 p1' p2 p2' s1' s2').

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    induction H4.
    - apply H1.
      + eapply NPDRelExistPDRel; eassumption.
      + eapply NPDRelExistPDRel. eassumption. apply PPRelSym. eassumption.
      + assumption.
    - apply NPPRel2. eapply IHNPPRel. assumption. apply NPDRelNoop. assumption. assumption.
    - apply NPPRel3. eapply IHNPPRel. assumption. assumption. apply NPDRelNoop. assumption.
  Qed.

  (*
  Lemma NPPRelPresCoT:
    forall s1 s2 s1' s2' Rs (ps1: ProgsState impl1 Rs) ps2,

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 s1 p1' s2 ->
          PDRel p2 s1 p2' s2 ->
          PIn p1 p1' ps1 ps2 ->
          PIn p2 p2' ps1 ps2 ->
          PPRel p1 p1' p2 p2' s1 s2 ->
          PPRel p1 p1' p2 p2' s1' s2') ->

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 s1 p1' s2 ->
          NPDRel p2 s1 p2' s2 ->
          PIn p1 p1' ps1 ps2 ->
          PIn p2 p2' ps1 ps2 ->
          NPPRel p1 p1' p2 p2' s1 s2 ->
          NPPRel p1 p1' p2 p2' s1' s2').

  Proof.
    introv H1 H2 H3 H4 H5 H6.
    clear -H1 H2 H3 H4 H5 H6 PDRelNoop PPRelSym PPRelNoop.
    induction H6.

    - constructor. apply H1.
      + eapply NPDRelExistPDRel; eassumption.
      + eapply NPDRelExistPDRel. eassumption. apply PPRelSym. eassumption.
      + assumption.
      + assumption.
      + assumption.

    - apply NPPRel2. eapply IHNPPRel. assumption. apply NPDRelNoop. assumption. assumption. assumption.

    - apply NPPRel3. eapply IHNPPRel. assumption. assumption. apply NPDRelNoop. assumption.
  Qed.
*)

  Lemma NPPRelPresCoRR:
    forall s1 s2 s1' s2' (ps1: ProgsState impl1) ps2,

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 s1 p1' s2 ->
          PDRel p2 s1 p2' s2 ->
          NPPsRel p1 p1' ps1 ps2 s1 s2 ->
          NPPsRel p2 p2' ps1 ps2 s1 s2 ->
          PPRel p1 p1' p2 p2' s1 s2 ->
          PPRel p1 p1' p2 p2' s1' s2') ->

      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 s1 p1' s2 ->
          NPDRel p2 s1 p2' s2 ->
          NPPsRel p1 p1' ps1 ps2 s1 s2 ->
          NPPsRel p2 p2' ps1 ps2 s1 s2 ->
          NPPRel p1 p1' p2 p2' s1 s2 ->
          NPPRel p1 p1' p2 p2' s1' s2').

  Proof.
    introv H1 H2 H3 H4 H5 H6.
    clear -H1 H2 H3 H4 H5 H6 PDRelNoop PPRelSym PPRelNoop.
    induction H6.
    - constructor. apply H1.
      + eapply NPDRelExistPDRel; eassumption.
      + eapply NPDRelExistPDRel. eassumption. apply PPRelSym. eassumption.
      + assumption.
      + assumption.
      + assumption.
    - apply NPPRel2. eapply IHNPPRel. assumption. apply NPDRelNoop; assumption. assumption.
      apply NPPsRelNoop; assumption. assumption.
    - apply NPPRel3. eapply IHNPPRel. assumption. assumption. apply NPDRelNoop. assumption.
      assumption. apply NPPsRelNoop; assumption.
  Qed.

  (*
  Lemma NPPRelPresCoR:
    forall s1 s2 s1' s2'
      R (m: impl1.(ObjInstruction) R)
      R1t (p1t : Program impl1.(ObjInstruction) R1t) p2t,

      (forall R1 (p1 : Program impl1.(ObjInstruction) R1) p1'
         R2 (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 s1 p1' s2 ->
          PDRel p2 s1 p2' s2 ->
          PPRel p1 p1' p1t p2t s1 s2 ->
          PPRel p2 p2' p1t p2t s1 s2 ->
          PPRel p1 p1' p2 p2' s1 s2 ->
          PPRel p1 p1' p2 p2' s1' s2') ->

      (forall R1 (p1 : Program impl1.(ObjInstruction) R1) p1'
         R2 (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 s1 p1' s2 ->
          NPDRel p2 s1 p2' s2 ->
          NPPRel p1 p1' p1t p2t s1 s2 ->
          NPPRel p2 p2' p1t p2t s1 s2 ->
          PPRel p1 p1' p2 p2' s1 s2 ->
          PPRel p1 p1' p2 p2' s1' s2').

  Proof.
    introv H1 H2 H3 H4 H5 H6.
    clear -H1 H2 H3 H4 H5 H6 PDRelNoop PPRelSym PPRelNoop.
    forwards H6': PPRelSym H6.
    forwards H2': NPDRelExistPDRel H2 H6.
    forwards H3': NPDRelExistPDRel H3 H6'.
    forwards H4': NPPRelExistPPRel''' H2' H4. intros. intro. inverts H.
    forwards H5': NPPRelExistPPRel''' H3' H5. intros. intro. inverts H.
    eauto.
  Qed.
  *)

  Lemma NPPRelPres2Co:
    forall R (p1: Program impl1.(ObjInstruction) R) p2 s1 s2
      R' (p1': Program impl1.(ObjInstruction) R') p2' s1' s2',

      (forall p1', p1 <> Noop p1') ->

      (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
          PDRel p1'' s1 p2'' s2 ->
          PPRel p1 p2 p1'' p2'' s1 s2 ->
          NPPRel p1' p2' p1'' p2'' s1' s2') ->

      (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
          NPDRel p1'' s1 p2'' s2 ->
          NPPRel p1 p2 p1'' p2'' s1 s2 ->
          NPPRel p1' p2' p1'' p2'' s1' s2').

  Proof.
    introv H0 H1 H2 H3.
    clear -H0 H1 H2 H3 PDRelNoop PPRelSym PPRelNoop.
    - induction H3.
      + apply H1. eapply NPDRelExistPDRel. eassumption. apply PPRelSym. eassumption. assumption.
      + exfalso. eapply H0. reflexivity.
      + apply NPPRel3. apply IHNPPRel.
        * assumption.
        * assumption.
        * apply NPDRelNoop. assumption.
  Qed.


  Lemma NPPsRelPres:
    forall (R: Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PPRel p1 p1' p2 p2' ds1 ds2 ->
          PPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1 p2 ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H1 using NPPsRelInd.
    - constructor.
    - constructor.
      + eapply NPPRelPres; eassumption.
      + eauto.
  Qed.

  Lemma NPPsRelPresCo:
    forall (R: Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPDRel p1 ds1 p2 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 ds1 p1' ds2 ->
          NPDRel p2 ds1 p2' ds2 ->
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1 p2 ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    induction H1 using NPPsRelInd.
    - constructor.
    - apply invertNPsDRel in H3.
      destruct H3 as [(? & ? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H31 & H32 & H33 & H34)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      constructor; eauto.
  Qed.

  (*
  Lemma NPPsRelPresCoR:
    forall (R: Type) (Rs: list Type)
      (p1t : Program impl1.(ObjInstruction) R)
      (p2t : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1 Rs) (ps2 : ProgsState impl2 Rs)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1t p2t ps1 ps2 ds1 ds2 ->
      NPDRel p1t ds1 p2t ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 ds1 p1' ds2 ->
          NPDRel p2 ds1 p2' ds2 ->
          NPPRel p1t p2t p1 p1' ds1 ds2 ->
          NPPRel p1t p2t p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1t p2t ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    induction H1 using NPPsRelInd.
    - constructor.
    - apply invertNPsDRel in H3. simpl in H3. destruct H3 as (H31 & H32 & H33 & H34).
      constructor; eauto.
      eapply H4; try eassumption.
  Qed.
  *)

  Lemma NPPsRelPresCoT:
    forall (R: Type)
      (p1t : Program impl1.(ObjInstruction) R)
      (p2t : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1t p2t ps1 ps2 ds1 ds2 ->
      NPDRel p1t ds1 p2t ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 ds1 p1' ds2 ->
          NPDRel p2 ds1 p2' ds2 ->
          PIn p1 p1' (existT _ R p1t :: ps1) (existT _ R p2t :: ps2) ->
          PIn p2 p2' (existT _ R p1t :: ps1) (existT _ R p2t :: ps2) ->
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1t p2t ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    induction H1 using NPPsRelInd.
    - constructor.
    - apply invertNPsDRel in H3.
      destruct H3 as [(? & ? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H31 & H32 & H33 & H34)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      constructor.
      + apply H4; try eassumption.
        * constructor.
        * constructor. constructor.
      + apply IHNPPsRel; try eassumption.
        introv N1 N2 N3 N4 N5.
        apply H4; try eassumption.
        * apply invertPIn in N3.
          destruct N3 as [(ps1' & ps2' & N31 & N32) | (p1'' & p2'' & ps1'' & ps2'' & N31 & N32 & N33)].
          -- inverts N31.
             inverts N32.
             constructor.
          -- inverts N31.
             inverts N32.
             constructor.
             constructor.
             assumption.
        * apply invertPIn in N4.
          destruct N4 as [(ps1' & ps2' & N41 & N42) | (p1'' & p2'' & ps1'' & ps2'' & N41 & N42 & N43)].
          -- inverts N41.
             inverts N42.
             constructor.
          -- inverts N41.
             inverts N42.
             constructor.
             constructor.
             assumption.
  Qed.


  Lemma NPPsRelPresCo':
    forall (R: Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPDRel p1 ds1 p2 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 ds1 p1' ds2 ->
          PDRel p2 ds1 p2' ds2 ->
          PPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1 p2 ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    eapply NPPsRelPresCo; try eassumption.
    eapply NPPRelPresCo. assumption.
  Qed.

  (*
  mymark
  Lemma NPPsRelPresCoR':
    forall (R: Type) (Rs: list Type)
      (p1t : Program impl1.(ObjInstruction) R)
      (p2t : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1 Rs) (ps2 : ProgsState impl2 Rs)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      NPPsRel p1t p2t ps1 ps2 ds1 ds2 ->
      NPDRel p1t ds1 p2t ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          PDRel p1 ds1 p1' ds2 ->
          PDRel p2 ds1 p2' ds2 ->
          PPRel p1t p2t p1 p1' ds1 ds2 ->
          PPRel p1t p2t p2 p2' ds1 ds2 ->
          PPRel p1 p1' p2 p2' ds1' ds2') ->
      NPPsRel p1t p2t ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    eapply NPPsRelPresCo; try eassumption.
    eapply NPPRelPresCo. assumption.
  Qed.
  *)

  Lemma NPPsRelPres2:
    forall (R R': Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R')
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2)
      s1 s2 s1' s2',

      NPPsRel p1 p2 ps1 ps2 s1 s2 ->
      (forall (Rt : Type)
         (p1t : Program impl1.(ObjInstruction) Rt)
         (p2t : Program impl2.(ObjInstruction) Rt),
          NPPRel p1 p2 p1t p2t s1 s2 ->
          NPPRel p1' p2' p1t p2t s1' s2') ->
      NPPsRel p1' p2' ps1 ps2 s1' s2'.

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H1.
    - constructor.
    - constructor; auto.
  Qed.

  Lemma NPPsRelPres2Co:
    forall R
      (p1 : Program impl1.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p2' : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1)) ds1'
      (ds2 : State (objBase impl2)) ds2',

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall (R : Type)
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          (* NPDRel p1 ds1 p2 ds2 -> *)
          NPDRel p1t ds1 p2t ds2 ->
          NPPRel p1 p2 p1t p2t ds1 ds2 ->
          NPPRel p1' p2' p1t p2t ds1' ds2') ->
      NPPsRel p1' p2' ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3.
    clear -H1 H2 H3.
    induction H1 using NPPsRelInd.
    - constructor.
    - apply invertNPsDRel in H2.
      simpl in *.
      destruct H2 as [(? & ? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H21 & H22 & H23 & H24)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      constructor.
      + apply H3; assumption.
      + eauto.
  Qed.

  Lemma NPPsRelPres2Co':
    forall R
      (p1 : Program impl1.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p2' : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1)) ds1'
      (ds2 : State (objBase impl2)) ds2',

      (forall p1', p1 <> Noop p1') ->
      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall R (p1'': Program impl1.(ObjInstruction) R) p2'',
          PDRel p1'' ds1 p2'' ds2 ->
          PPRel p1 p2 p1'' p2'' ds1 ds2 ->
          NPPRel p1' p2' p1'' p2'' ds1' ds2') ->
      NPPsRel p1' p2' ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2 H3 H4.
    eapply NPPsRelPres2Co; try eassumption.
    eapply NPPRelPres2Co. assumption. try eassumption.
  Qed.



  Lemma NPsDRelPres:
    forall
      (ps1: ProgsState impl1)
      (ps2: ProgsState impl2)
      (ds1 ds1': State (objBase impl1))
      (ds2 ds2': State (objBase impl2)),

      NPsDRel ps1 ds1 ps2 ds2 ->
      (forall R (p1 : Program impl1.(ObjInstruction) R) p2,
          NPDRel p1 ds1 p2 ds2 ->
          NPDRel p1 ds1' p2 ds2') ->
      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      DRel ds1' ds2' ->
      NPsDRel ps1 ds1' ps2 ds2'.

  Proof.
    clear PDRelNoop invRel retRel stepRel.
    introv H1 H2 H3 H4.
    induction H1.
    - constructor. assumption.
    - constructor; eauto.
      eapply NPPsRelPres2. eassumption. auto.
  Qed.

  Lemma NPsDRelPresCo:
    forall R
      (p1: Program impl1.(ObjInstruction) R)
      (p2: Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1))
      (ds2 : State (objBase impl2))
      ds1' ds2',

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      DRel ds1' ds2' ->
      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 ds1 p1' ds2 ->
          NPDRel p2 ds1 p2' ds2 ->
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      (forall R
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          NPDRel p1t ds1 p2t ds2 ->
          NPPRel p1 p2 p1t p2t ds1 ds2 ->
          NPDRel p1t ds1' p2t ds2') ->
      NPsDRel ps1 ds1' ps2 ds2'.

  Proof.
    introv H1 H2 H3 H4 H5.
    clear -H1 H2 H3 H4 H5 PDRelNoop PPRelSym PPRelNoop.
    induction H2 as [ s1 s2 H2 | R' Rs p1t s1 ps1 p2t s2 ps2 H21 H22 H23 H24] using NPsDRelInd.
    - constructor. assumption.
    - apply invertNPPsRel in H1.
      destruct H1 as [(? & ?) | (R'' & p1' & p2' & ps1' & ps2' & Hps1' & Hps2' & H11 & H12)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      constructor.
      + eapply H5; eassumption.
      + eapply NPPsRelPresCo; eauto.
      + eauto.
  Qed.

  Lemma NPsDRelPres2Co':
    forall R
      (p1: Program impl1.(ObjInstruction) R)
      (p2: Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1))
      (ds2 : State (objBase impl2))
      ds1' ds2',

      (forall p1', p1 <> Noop p1') ->
      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      DRel ds1' ds2' ->
      (forall R (p11: Program impl1.(ObjInstruction) R) p21
         R' (p12: Program impl1.(ObjInstruction) R') p22,
          PDRel p11 ds1 p21 ds2 ->
          PDRel p12 ds1 p22 ds2 ->

          PPRel p11 p21 p12 p22 ds1 ds2 ->
          PPRel p11 p21 p12 p22 ds1' ds2') ->
      (forall (R : Type)
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          PDRel p1t ds1 p2t ds2 ->
          PPRel p1 p2 p1t p2t ds1 ds2 ->
          PDRel p1t ds1' p2t ds2') ->
      NPsDRel ps1 ds1' ps2 ds2'.

  Proof.
    introv H1 H2 H3 H4 H5 H6.
    clear -H1 H2 H3 H4 H5 H6 PDRelNoop PPRelSym PPRelNoop.
    eapply NPsDRelPresCo; try eassumption.
    - eapply NPPRelPresCo; eauto.
    - eapply NPDRelPres2Co'; eauto.
  Qed.

  Lemma NPsDRelPresCoT:
    forall R
      (p1: Program impl1.(ObjInstruction) R)
      (p2: Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1))
      (ds2 : State (objBase impl2))
      ds1' ds2',

      NPPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      NPsDRel ps1 ds1 ps2 ds2 ->
      DRel ds1' ds2' ->
      (forall R1 R2
         (p1 : Program impl1.(ObjInstruction) R1) p1'
         (p2 : Program impl1.(ObjInstruction) R2) p2',
          NPDRel p1 ds1 p1' ds2 ->
          NPDRel p2 ds1 p2' ds2 ->
          PIn p1 p1' ps1 ps2 ->
          PIn p2 p2' ps1 ps2 ->
          NPPRel p1 p1' p2 p2' ds1 ds2 ->
          NPPRel p1 p1' p2 p2' ds1' ds2') ->
      (forall R
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          NPDRel p1t ds1 p2t ds2 ->
          NPPRel p1 p2 p1t p2t ds1 ds2 ->
          NPDRel p1t ds1' p2t ds2') ->
      NPsDRel ps1 ds1' ps2 ds2'.

  Proof.
    introv H0 H1 H2 H3 H4.
    clear -H0 H1 H2 H3 H4 PDRelNoop PPRelSym PPRelNoop.
    (* gen ps1 ps2. *)
    induction H1 as [ s1 s2 H1 | R' Rs p1t s1 ps1 p2t s2 ps2 H11 H12 H13 H14] using NPsDRelInd.
    - constructor. assumption.
    - apply invertNPPsRel in H0.
      destruct H0 as [(? & ?) | (R'' & p1' & p2' & ps1' & ps2' & Hps1' & Hps2' & H01 & H02)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      constructor.
      + eapply H4; eassumption.
      + eapply NPPsRelPresCoT; eauto.
      + apply H14; try eassumption.
        introv N1 N2 N3 N4 N5.
        eapply H3; try eassumption.
        constructor. assumption.
        constructor. assumption.
  Qed.

  Definition Rel (s1 : ThreadsState impl1) (s2 : ThreadsState impl2) : Prop :=
    NPsDRel (fst s1) (snd s1)
            (fst s2) (snd s2).

  Lemma ThreadAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'
        (* /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
            PDRel p1 (snd s1') p2 (snd s2')) *)
        /\ (forall R R'
             (p1: Program impl1.(ObjInstruction) R) p2
             (p1': Program impl1.(ObjInstruction) R') p2',
              PDRel p1 (snd s1) p2 (snd s2) ->
              PDRel p1' (snd s1) p2' (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPPsRel p1' p2' (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPRel p1 p2 p1' p2' (snd s1) (snd s2) ->
              PDRel p1 (snd s1') p2 (snd s2')
              /\ PDRel p1' (snd s1') p2' (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')
              /\ NPPsRel p1' p2' (fst s1') (fst s2') (snd s1') (snd s2')
              /\ PPRel p1 p2 p1' p2' (snd s1') (snd s2'))
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
              NPDRel p1 (snd s1) p2 (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPDRel p1 (snd s1') p2 (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')).

  (*
          /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2
                     R' (p3: Program impl1.(ObjInstruction) R') p4,
              PDRel p1 (snd s1) p2 (snd s2) ->
              PDRel p3 (snd s1) p4 (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPPsRel p3 p4 (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPRel p3 p4 p1 p2 (snd s1) (snd s2) ->
              PPRel p3 p4 p1 p2 (snd s1') (snd s2')).
  *)

  Proof.
    introv H1 H2.
    (* revert PPRelSym stepRel PDRelNoop PPRelNoop.
    clear -H1 H2.
    introv PPRelSym stepRel PDRelNoop PPRelNoop. *)
    unfold Rel in *.

    gen s2.
    induction H2 as [ p1 ps1 ds1 ps1' ds1'
                    | Ret ps1 ds1 R m f r ds1'
                    | Ret ps1 ds1 p1 ];
      introv H11;
      destruct s2 as (pps2 & ds2); destruct pps2 as [| p2 ps2]; simpl in *.

    - apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.
    - (* TACons *)
      apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (R1 & p1' & p2' & p1'' & p2'' & Hp1'' & Hp2'' & H111 & H112 & H113 & H114)]; [discriminate |].
      inverts Hp1''.
      inverts Hp2''.
      forwards (s2' & H3 & H4 & H4' & H5): IHThreadAdvance (ps2, ds2); simpl; clear IHThreadAdvance.
      + assumption.
      + destruct s2' as (ps2' & ds2'). simpl in *.
        eexists (existT _ _ p2' :: ps2', ds2'). simpl in *. splits.
        * apply advanceCons. eassumption.
        * { econstructor.
            - apply H5; assumption.
            - apply H5; assumption.
            - assumption.
          }

        (* * assumption. *)
        * { introv N1 N2 N3 N4 N5.
            apply invertNPPsRel in N3.
            destruct N3 as [(? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & N31 & N32)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            apply invertNPPsRel in N4.
            destruct N4 as [(? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & N41 & N42)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            forwards (M1 & M2 & M3 & M4 & M5): H4' N1 N2 N32 N42 N5.
            splits; try eassumption.
            - constructor.
              + eapply NPPRelPresCoRR. introv L1 L2 L3. eapply H4'. exact L1. assumption.
                assumption. constructor. assumption. assumption. assumption. assumption. assumption.
              + assumption.
            - constructor.
              + eapply NPPRelPresCoRR. introv L1 L2 L3. eapply H4'. exact L1. assumption.
                assumption. constructor. assumption. assumption. assumption. assumption. assumption.
              + assumption.
          }

        * { intros R p3 p4 H7 H8.
            apply invertNPPsRel in H8.
            destruct H8 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H81 & H82)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            simpl in *.
            split.
            - apply H5; assumption.
            - constructor.
              + eapply NPPRelPresCoRR. introv N1 N2 N3. eapply H4'. exact N1. assumption.
                assumption. assumption. assumption. assumption. assumption. assumption.
              + apply H5; assumption.
          }
        (* * intros R p3 p4 R' p5 p6 H7 H8 H9 H10 H11.
          apply invertNPPsRel in H9. destruct H9 as (H91 & H92). simpl in *.
          apply invertNPPsRel in H10. destruct H10 as (H101 & H102). simpl in *.
          eauto.
        *)

    - apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.
    - (* TABind *)
      apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (R' & p1'' & p2'' & ps1' & ps2' & Hps1' & Hps2' & H111 & H112 & H113 & H114)]; [discriminate |].

      inverts Hps1'.
      inverts Hps2'.
      inverts H111 as H111.
      forwards (p2' & ds2' & H2 & H3 & H4 & H5 & H6 & H7): stepRel. (*; clear stepRel. *)
      + eassumption.
      + eassumption.
      + eassumption.
      + exists (existT _ _ p2' :: ps2, ds2').
        simpl. splits.
        * eapply ProgramStepsImplThreadsAdvance in H2. exact H2.
        * { econstructor.
            - assumption.
            - eapply NPPsRelPres2Co'; try eassumption.
              intros. intro N. inverts N.
            - eapply NPsDRelPres2Co'; try eassumption.
              intros. intro N. inverts N.
          }

        (* * introv N1 N2 N3. *)
        * { introv N1 N2 N3 N4 N5.
            apply invertNPPsRel in N3.
            destruct N3 as [(? & ?) | (R'' & p1''' & p2''' & ps1''' & ps2''' & Hps1''' & Hps2''' & N31 & N32)]; [discriminate |].
            inverts Hps1'''.
            inverts Hps2'''.
            apply NPPRelSym in N31.
            forwards N31': NPPRelExistPPRel'' N31. intros. intro N. inverts N. assumption.
            apply invertNPPsRel in N4.
            destruct N4 as [(? & ?) | (R'' & p1''' & p2''' & ps1''' & ps2''' & Hps1''' & Hps2''' & N41 & N42)]; [discriminate |].
            inverts Hps1'''.
            inverts Hps2'''.
            apply NPPRelSym in N41.
            forwards N41': NPPRelExistPPRel'' N41. intros. intro N. inverts N. assumption.
            splits.
            - eauto.
            - eauto.
            - constructor.
              apply NPPRelSym. eauto.
              eapply NPPsRelPresCo'; eauto. constructor. assumption.
            - constructor.
              apply NPPRelSym. eauto.
              eapply NPPsRelPresCo'; eauto. constructor. assumption.
            - eauto.
          }

        * { intros R' p1t p2t H8 H9.
            apply invertNPPsRel in H9.
            destruct H9 as [(? & ?) | (R'' & p1''' & p2''' & ps1' & ps2' & Hps1' & Hps2' & H91 & H92)]; [discriminate |].
            inverts Hps1'.
            inverts Hps2'.
            split.
            - eapply NPDRelPres2Co'; try eassumption.
              intros. intro N. inverts N.
              apply NPPRelSym. assumption.
            - constructor.
              + apply NPPRelSym. apply NPPRelSym in H91.
                applys NPPRelPres2Co H6; try eassumption.
                intros. intro N. inverts N.
              + eapply NPPsRelPresCo'; eauto.
          }


    - apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.
    - (* TANoop *)
      apply invertNPsDRel in H11.
      destruct H11 as [(? & ? & ?) | (R & p1' & p2' & ps1' & ps2' & Hps1' & Hps2' & H111 & H112 & H113 & H114)]; [discriminate |].
      inverts Hps1'.
      inverts Hps2'.
      { inverts H111 as H111.
        - exfalso. eapply PDRelNoop. eassumption.

        - exists (existT _ _ p2' :: ps2, ds2). simpl in *.
          splits.
          * constructor.
          * { econstructor.
              - assumption.
              - apply NPPsRelNoop.
                eapply NPPsRelPres2Co; eauto.
              - assumption.
            }

          * { introv N1 N2 N3 N4 N5.
              apply invertNPPsRel in N3.
              destruct N3 as [(? & ?) | (R'' & p1'' & p2'' & ps1' & ps2' & Hps1' & Hps2' & N31 & N32)]; [discriminate |].
              inverts Hps1'.
              inverts Hps2'.
               apply NPPRelNoop2 in N31.
              apply invertNPPsRel in N4.
              destruct N4 as [(? & ?) | (R'' & p1'' & p2'' & ps1' & ps2' & Hps1' & Hps2' & N41 & N42)]; [discriminate |].
              inverts Hps1'.
              inverts Hps2'.
              apply NPPRelNoop2 in N41.
              splits; try assumption.
              - constructor; assumption.
              - constructor; assumption.
            }
          * { intros R' p1t p2t H2 H3.
              apply invertNPPsRel in H3.
              destruct H3 as [(? & ?) | (R & p1'' & p2'' & ps1' & ps2' & Hps1' & Hps2' & H31 & H32)]; [discriminate |].
              inverts Hps1'.
              inverts Hps2'.
              split.
              - assumption.
              - constructor.
                + apply NPPRelNoop2. assumption.
                (* eauto. *)
                + assumption.
            }
      }
  Qed.

  Lemma ThreadsAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadsAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'
        (* /\ forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            PDRel p1 (snd s1') p2 (snd s2'). *)
        /\ (forall R R'
             (p1: Program impl1.(ObjInstruction) R) p2
             (p1': Program impl1.(ObjInstruction) R') p2',
              PDRel p1 (snd s1) p2 (snd s2) ->
              PDRel p1' (snd s1) p2' (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPPsRel p1' p2' (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPRel p1 p2 p1' p2' (snd s1) (snd s2) ->
              PDRel p1 (snd s1') p2 (snd s2')
              /\ PDRel p1' (snd s1') p2' (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')
              /\ NPPsRel p1' p2' (fst s1') (fst s2') (snd s1') (snd s2')
              /\ PPRel p1 p2 p1' p2' (snd s1') (snd s2'))
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
              NPDRel p1 (snd s1) p2 (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPDRel p1 (snd s1') p2 (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')).
  Proof.
    introv H1 H2.
    (* clear -H1 H2 PPRelSym PDRelNoop stepRel. *)
    unfold ThreadsAdvance in H2.
    gen s2.
    induction H2 as [s1 | s1 s1m s1']; introv H1.
    - exists s2. splits. constructor. assumption.
      eauto. eauto.
      (* intros. splits. constructor. assumption. assumption. *)
    - forwards (s2m & H3 & H4 & H5 & H5'): ThreadAdvanceRel.
      + eassumption.
      + eassumption.
      + forwards (s2' & H6 & H7 & H8 & H8'): IHclos_refl_trans_1n.
        * eassumption.
        * { exists s2'. splits.
            - unfold ThreadsAdvance. eapply ThreadsAdvanceTrans; eassumption.
            - eassumption.

            - introv H9 H10 H11 H12 H13.
              forwards (N1 & N2 & N3 & N4 & N5): H5 H9 H10 H11 H12 H13; try eassumption.
              eapply H8; eassumption.

            - introv H9 H10.
              eapply H8'; eapply H5'; eassumption.
          }
  Qed.

  Lemma ThreadsInteractionRel:
    forall Ret s1 s2 s1' int a,
      Rel s1 s2 ->
      ThreadsInteraction impl1 Ret int s1 a s1' ->
      exists s2',
        ThreadsInteraction impl2 Ret int s2 a s2'
        /\ Rel s1' s2'
        (* /\ forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            PDRel p1 (snd s1') p2 (snd s2'). *)
        /\ (forall R R'
             (p1: Program impl1.(ObjInstruction) R) p2
             (p1': Program impl1.(ObjInstruction) R') p2',
              PDRel p1 (snd s1) p2 (snd s2) ->
              PDRel p1' (snd s1) p2' (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPPsRel p1' p2' (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPRel p1 p2 p1' p2' (snd s1) (snd s2) ->
              PDRel p1 (snd s1') p2 (snd s2')
              /\ PDRel p1' (snd s1') p2' (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')
              /\ NPPsRel p1' p2' (fst s1') (fst s2') (snd s1') (snd s2')
              /\ PPRel p1 p2 p1' p2' (snd s1') (snd s2'))
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
              NPDRel p1 (snd s1) p2 (snd s2) ->
              NPPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              NPDRel p1 (snd s1') p2 (snd s2')
              /\ NPPsRel p1 p2 (fst s1') (fst s2') (snd s1') (snd s2')).

  Proof.
    introv H1 H2.
    revert invRel retRel PPRelSym PDRelNoop PPRelNoop.
    clear -H1 H2.
    introv invRel retRel PPRelSym PDRelNoop PPRelNoop.
    destruct H2.

    revert s2 H1.
    induction H as [ Ret m ps1
                   | Ret r ps1
                   | Ret int p1 ps1 r ps1'];
      intros s2 H1;
      destruct s2 as (ps2 & ds2);
      unfold Rel in *;
      simpl in *.

    - (* TISpawn *)
      clear retRel.
      exists (existT _ _ (objProgram impl2 Ret m) :: ps2, ds2). simpl.
      splits.
      + constructor. constructor.
      + econstructor.
        * apply invRel.
          apply invertNPsDRel in H1.
          destruct H1; jauto.
        * { induction H1 using NPsDRelInd.
            - constructor.
            - constructor.
              + eapply invRel'. eassumption. eassumption.
              + assumption.
          }
        * assumption.

      + { introv N1 N2 N3 N4 N5.
          splits; try eassumption.
          - constructor.
            + apply NPPRelSym. eapply invRel.
              eapply NPsDRelImplDRel. eassumption. assumption.
            + assumption.
          - constructor.
            + apply NPPRelSym. eapply invRel.
              eapply NPsDRelImplDRel. eassumption. assumption.
            + assumption.
        }

      + intros R' p1t p2t H2 H3.
        splits.
        * assumption.
        * { constructor. apply NPPRelSym. eapply invRel'.
            - apply invertNPsDRel in H1.
              destruct H1; jauto.
            - assumption.
            - assumption.
          }

    - (* TIJoin *)
      clear invRel.
      apply invertNPsDRel in H1.
      destruct H1 as [(? & ? & ?) | (R & p1 & p2 & ps1' & ps2' & Hps1' & <- & H11 & H12 & H13 & H14)]; [discriminate |].
      inverts Hps1'.
      inverts H11 as H11.
      forwards A: retRel. eassumption. assumption. subst.
      exists (ps2', ds2).
      simpl. splits.
      + constructor. constructor.
      + assumption.

      + { introv N1 N2 N3 N4 N5.
          splits; try assumption.
          - apply invertNPPsRel in N3.
            destruct N3 as [(? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & N31 & N32)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            assumption.
          - apply invertNPPsRel in N4.
            destruct N4 as [(? & ?) | (R'' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & N41 & N42)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            assumption.
        }

      + intros R' p1t p2t H2 H3.
        splits.
        * assumption.
        * apply invertNPPsRel in H3.
          destruct H3 as [(? & ?) | (R & p1' & p2' & ps1'' & ps2'' & Hps1'' & Hps2'' & H31 & H32)]; [discriminate |].
          inverts Hps1''.
          inverts Hps2''.
          assumption.

    - (* TICons *)
      clear invRel retRel.
      apply invertNPsDRel in H1.
      destruct H1 as [(? & ? & ?) | (R1 & p1' & p2' & ps1'' & ps2'' & Hps1'' & <- & H11 & H12 & H13 & H14)]; [discriminate |].
      inverts Hps1''.
      forwards (s2' & H2 & H3 & H4' & H4): IHThreadsInteraction' (ps2'', ds2).
      (* forwards (s2' & H2 & H3): IHThreadsInteraction' (ps2, ds2). *)
      + simpl. assumption.
      + destruct s2' as [ps2' ds2']. simpl in *.
        exists (existT _ _ p2' :: ps2', ds2'). splits.
        * (* econstructor. *)
          eapply (invertThreadsInteraction impl2) in H2.
          destruct H2 as (H21 & H22).
          simpl in *. subst.
          econstructor. econstructor. assumption.
        * simpl in *.
          { econstructor.
            + apply H4; assumption.
            + apply H4; assumption.
            + assumption.
          }


        * { simpl in *.
            introv N1 N2 N3 N4 N5.
            apply invertNPPsRel in N3.
            destruct N3 as [(? & ?) | (R'' & p1'' & p2'' & ps1''' & ps2''' & Hps1''' & Hps2''' & N31 & N32)]; [discriminate |].
            inverts Hps1'''.
            inverts Hps2'''.
            apply invertNPPsRel in N4.
            destruct N4 as [(? & ?) | (R'' & p1'' & p2'' & ps1''' & ps2''' & Hps1''' & Hps2''' & N41 & N42)]; [discriminate |].
            inverts Hps1'''.
            inverts Hps2'''.
            forwards (M1 & M2 & M3 & M4 & M5): H4' N1 N2 N32 N42 N5.
            splits; try eassumption.
            - constructor.
              + eapply NPPRelPresCoRR. introv L1 L2 L3. eapply H4'. exact L1. assumption.
                assumption. constructor. assumption. assumption. assumption. assumption. assumption.
              + assumption.
            - constructor.
              + eapply NPPRelPresCoRR. introv L1 L2 L3. eapply H4'. exact L1. assumption.
                assumption. constructor. assumption. assumption. assumption. assumption. assumption.
              + assumption.
          }


        * { simpl in *.
            intros R' p1t p2t H5 H6.
            apply invertNPPsRel in H6.
            destruct H6 as [(? & ?) | (R & p1'' & p2'' & ps1''' & ps2''' & Hps1''' & Hps2''' & H61 & H62)]; [discriminate |].
            inverts Hps1'''.
            inverts Hps2'''.
            splits.
            + apply H4; assumption.
            + constructor.
              * eapply NPPRelPresCoRR. introv N1 N2 N3. eapply H4'. exact N1. assumption.
                assumption. assumption. assumption. assumption. assumption. assumption.
              * apply H4; assumption.
          }
  Qed.


  Theorem simulatesCoInd':
    forall (s1 : ThreadsState impl1) (s2 : ThreadsState impl2),
      Rel s1 s2 ->
      Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2) s1 s2.

  Proof.
    (* clear init1 init2 initRel. *)
    cofix self.
    introv H1.
    constructor.
    introv H2.
    unfold ThreadsBehavior in H2.
    destruct H2 as [ts1m H2 H3].
    forwards (ts2m & H4 & H5 & _): ThreadsAdvanceRel.
    - eassumption.
    - eassumption.
    - forwards (ts2' & H6 & H7 & _): ThreadsInteractionRel.
      + eassumption.
      + eassumption.
      + econstructor.
        * unfold ThreadsBehavior.
          econstructor. exact H4. eassumption.
        * eapply self.
          eassumption.
  Qed.

  Theorem simulatesCoInd:
    ConcSimulates
      impl1 impl2.
  Proof.
    eapply simulatesCoInd'.
    unfold Rel. simpl. constructor.
    assumption.
  Qed.

End SimulatesCoIndN.
End SimulatesCoIndN.


Module SimulatesCoInd.
Section SimulatesCoInd.

  Context
    {Method : Type -> Type}
    (impl1 impl2 : Object Method).

  Definition ProgsState {M} (impl : Object M) :=
    list (sigT (Program impl.(ObjInstruction))).

  Definition DataState {M} (impl : Object M) :=
    impl.(objBase).(State).

  Variable (PDRel: forall R,
               Program impl1.(ObjInstruction) R ->
               impl1.(objBase).(State) ->
               Program impl2.(ObjInstruction) R ->
               impl2.(objBase).(State) ->
               Prop).
  Arguments PDRel {R} _ _ _ _.

  Variable (DRel:
              impl1.(objBase).(State) ->
              impl2.(objBase).(State) ->
              Prop).

  (*
   * DRel cannot be defined as part of PDRel because then
   * invRel requires that it be proved for every s1 and s2.
   *)

  Variable (PPRel: forall R R',
               Program impl1.(ObjInstruction) R ->
               Program impl2.(ObjInstruction) R ->
               Program impl1.(ObjInstruction) R' ->
               Program impl2.(ObjInstruction) R' ->
               (* DataState impl1 ->
               DataState impl2 -> *)
               Prop).
  (* Arguments PPRel {R R'} _ _ _ _ _ _. *)
  Arguments PPRel {R R'} _ _ _ _.

  Hypothesis PPRelSym:
    forall R (p1: Program impl1.(ObjInstruction) R) p2
      R' (p1': Program impl1.(ObjInstruction) R') p2',

      PPRel p1 p2 p1' p2' ->
      PPRel p1' p2' p1 p2.

  Hypothesis initRel: DRel impl1.(objBase).(initialState) impl2.(objBase).(initialState).

  Hypothesis noopRel:
    forall R (p1: Program impl1.(ObjInstruction) R) p2 s1 s2,
      PDRel (Noop p1) s1 p2 s2 ->

      (PDRel p1 s1 p2 s2
       /\ (forall R (p1t: Program impl1.(ObjInstruction) R) p2t,
             PDRel p1t s1 p2t s2 ->
             PPRel (Noop p1) p2 p1t p2t ->
             PPRel p1 p2 p1t p2t))

      \/ (exists p2',
            p2 = Noop p2'
            /\ PDRel p1 s1 p2' s2
            /\ (forall R (p1t: Program impl1.(ObjInstruction) R) p2t,
                  PDRel p1t s1 p2t s2 ->
                  PPRel (Noop p1) p2 p1t p2t ->
                  PPRel p1 p2' p1t p2t))

      \/ (exists p2',
          ProgramSteps impl2.(objBase) _ p2 s2 p2' s2
          /\ PDRel p1 s1 p2' s2
          /\ (forall R (p1t: Program impl1.(ObjInstruction) R) p2t,
                PDRel p1t s1 p2t s2 ->
                PPRel (Noop p1) p2 p1t p2t ->
                PPRel p1 p2' p1t p2t)).

  (*
   * If a method m is invoked in both the concrete and abstract objects, then
   * (1) PDRel holds between them and
   * (2) For any pair of concrete and abstract programs in the pool that PDRel holds, PPRel holds between the that pair and the new pair.
   *)
  Hypothesis invRel:
    forall R m s1 s2,
      DRel s1 s2 ->
      PDRel (objProgram impl1 R m) s1
            (objProgram impl2 R m) s2
      /\ forall R' (p1: Program impl1.(ObjInstruction) R') p2,
          PDRel p1 s1 p2 s2 ->
          PPRel (objProgram impl1 R m) (objProgram impl2 R m)
                p1 p2.

  (*
       ((s1 = init1 /\ s2 = init2)
       \/ (exists (R': Type), exists (p1: Program impl1.(ObjInstruction) R') p2,
             PDRel p1 s1 p2 s2)) ->
  *)

  Hypothesis retRel:
    forall R (r: R) s1 p2 s2,
      PDRel (Return r) s1 p2 s2 ->
      DRel s1 s2 ->
      p2 = Return r.

  Hypothesis stepRel:
    forall R' (m: impl1.(ObjInstruction) R') R (f: R' -> Program impl1.(ObjInstruction) R) s1 p2 s2 s1' r,
      DRel s1 s2 ->
      PDRel (Bind m f) s1 p2 s2 ->
      impl1.(objBase).(EvalInstruction) m s1 r s1' ->
      exists p2' s2',
        ProgramSteps impl2.(objBase) _ p2 s2 p2' s2'
        /\ DRel s1' s2'
        /\ PDRel (f r) s1' p2' s2'
        /\ (forall R (p1t: Program impl1.(ObjInstruction) R) p2t,
            PDRel p1t s1 p2t s2 ->
            PPRel (Bind m f) p2 p1t p2t (* s1 s2 *) ->
            PDRel p1t s1' p2t s2'
            /\ PPRel (f r) p2' p1t p2t (* s1' s2' *)).


  Inductive PPsRel : forall R,
      Program impl1.(ObjInstruction) R ->
      Program impl2.(ObjInstruction) R ->
      ProgsState impl1 ->
      ProgsState impl2 ->
      (* DataState impl1 ->
      DataState impl2 -> *)
      Prop :=

  | NilPPsRel:
      forall R p1 p2 (* s1 s2 *),
        PPsRel R p1 p2 nil nil (* s1 s2 *)
  | ConsPPsRel:
      forall R p1 ps1 p2 ps2 R' p1' p2' (* s1 s2 *),
        PPRel p1 p2 p1' p2' (* s1 s2 *) ->
        PPsRel R p1 p2 ps1 ps2 (* s1 s2 *) ->
        PPsRel R p1 p2 (existT _ R' p1' :: ps1) (existT _ R' p2' :: ps2) (* s1 s2 *).
  (* Arguments PPsRel {R} _ _ _ _ _ _. *)
  Arguments PPsRel {R} _ _ _ _.

  Lemma invertPPsRel
        (R: Type)
        (p1: Program impl1.(ObjInstruction) R)
        (p2: Program impl2.(ObjInstruction) R)
        (ps1: ProgsState impl1)
        (ps2: ProgsState impl2)
        (* (ds1: DataState impl1)
        (ds2: DataState impl2) *):
    @PPsRel R p1 p2 ps1 ps2 (* ds1 ds2 *) ->
    (ps1 = nil /\ ps2 = nil) \/
    (exists R p1' ps1' p2' ps2',
        existT _ R p1' :: ps1' = ps1 /\ existT _ R p2' :: ps2' = ps2 /\
        PPRel p1 p2 p1' p2' (* ds1 ds2 *) /\
        PPsRel p1 p2 ps1' ps2' (* ds1 ds2 *)).

  Proof.
    clear.
    introv H1.
    destruct H1.
    - auto.
    - right.
      repeat esplit; assumption.
  Qed.

  Lemma PPsRelInd:
    forall (P: forall R,
          Program impl1.(ObjInstruction) R ->
          Program impl2.(ObjInstruction) R ->
          ProgsState impl1 ->
          ProgsState impl2 ->
          (* DataState impl1 ->
          DataState impl2 -> *)
          Prop),

      (forall R p1 p2 (* s1 s2 *),
          P R p1 p2 nil nil (* s1 s2 *)) ->
      (forall R p1 ps1 p2 ps2 R' p1' p2' (* s1 s2 *),
          PPRel p1 p2 p1' p2' (* s1 s2 *) ->
          P R p1 p2 ps1 ps2 (* s1 s2 *) ->
          P R p1 p2 (existT _ R' p1' :: ps1) (existT _ R' p2' :: ps2) (* s1 s2 *)) ->
      forall R p1 p2 ps1 ps2 (* s1 s2 *),
        PPsRel p1 p2 ps1 ps2 (* s1 s2 *) ->
        P R p1 p2 ps1 ps2 (* s1 s2 *).

  Proof.
    clear invRel retRel stepRel noopRel initRel.
    introv H1 H2 H3.
    induction H3; eauto.
  Qed.

  Inductive PsDRel :
    ProgsState impl1 -> DataState impl1 ->
    ProgsState impl2 -> DataState impl2 -> Prop :=
  | NilPsDRel:
      forall s1 s2,
        DRel s1 s2 ->
        PsDRel nil s1 nil s2
  | ConsPsDRel:
      forall R p1 s1 ps1 p2 s2 ps2,
        (* DRel s1 s2 -> *)
        PDRel p1 s1 p2 s2 ->
        PPsRel p1 p2 ps1 ps2 (* s1 s2 *) ->
        PsDRel ps1 s1 ps2 s2 ->
        PsDRel (existT _ R p1 :: ps1) s1 (existT _ R p2 :: ps2) s2.

  Lemma PsDRelImplDRel:
    forall (ps1 : ProgsState impl1) s1
      ps2 s2,

      PsDRel ps1 s1 ps2 s2 ->
      DRel s1 s2.

  Proof.
    introv H1.
    clear -H1.
    induction H1; assumption.
  Qed.

  Lemma PsDRelInd:
    forall (P : ProgsState impl1 -> DataState impl1 ->
                ProgsState impl2 -> DataState impl2 -> Prop),

      (forall s1 s2,
          DRel s1 s2 ->
          P nil s1 nil s2) ->
      (forall R p1 s1 ps1 p2 s2 ps2,
          PDRel p1 s1 p2 s2 ->
          DRel s1 s2 ->
          PPsRel p1 p2 ps1 ps2 (* s1 s2 *) ->
          P ps1 s1 ps2 s2 ->
          P (existT _ R p1 :: ps1) s1 (existT _ R p2 :: ps2) s2) ->
      forall p s p' s',
        PsDRel p s p' s' -> P p s p' s'.

  Proof.
    introv H1 H2 H3.
    induction H3.
    - auto.
    - eapply H2; try eassumption.
      eapply PsDRelImplDRel. eassumption.
  Qed.


  Lemma invertPsDRel (ps1 : ProgsState impl1) ds1 (ps2 : ProgsState impl2) ds2 :
    PsDRel ps1 ds1 ps2 ds2 ->
    (ps1 = nil
     /\ ps2 = nil
     /\ DRel ds1 ds2) \/
    (exists R p1 p2 ps1' ps2',
        existT _ R p1 :: ps1' = ps1 /\ existT _ R p2 :: ps2' = ps2 /\
        PDRel p1 ds1 p2 ds2
        /\ DRel ds1 ds2
        /\ PPsRel p1 p2 ps1' ps2' (* ds1 ds2 *)
        /\ PsDRel ps1' ds1 ps2' ds2).

  Proof.
    clear.
    introv H1.
    destruct H1; simpl in *; intuition.
    right.
    repeat esplit; auto.
    induction H1 using PsDRelInd; assumption.
  Qed.



  Lemma PPsRelPres:
    forall (R R': Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R')
      (p2' : Program impl2.(ObjInstruction) R')
      (ps1 : ProgsState impl1) (ps2 : ProgsState impl2),

      PPsRel p1 p2 ps1 ps2 ->

      (forall (Rt : Type)
         (p1t : Program impl1.(ObjInstruction) Rt)
         (p2t : Program impl2.(ObjInstruction) Rt),
          PPRel p1 p2 p1t p2t ->
          PPRel p1' p2' p1t p2t) ->
  (*
      (forall (Rt : Type) Rs
         (p1t : Program (Instruction (objBase impl1)) Rt) (r1: ProgsState impl1 Rs)
         (p2t : Program (Instruction (objBase impl2)) Rt) r2,
          ps1 = (p1t, r1) ->
          ps2 = (p1t, r2) ->
          PPRel p1 p2 p1t p2t  ->
          PPRel p1' p2' p1t p2t) ->
  *)
      PPsRel p1' p2' ps1 ps2.

  Proof.
    introv H1 H2.
    clear -H1 H2.
    induction H1 using PPsRelInd; intros.
    * constructor.
    * constructor; eauto.
  Qed.

  Lemma PPsRelPres':
    forall R
      (p1 : Program impl1.(ObjInstruction) R)
      (p1' : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (p2' : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1))
      (ds2 : State (objBase impl2)),

      PPsRel p1 p2 ps1 ps2 ->
      PsDRel ps1 ds1 ps2 ds2 ->
      (forall (R : Type)
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          PDRel p1t ds1 p2t ds2 ->
          PPRel p1 p2 p1t p2t ->
          PPRel p1' p2' p1t p2t) ->
      PPsRel p1' p2' ps1 ps2.

  Proof.
    introv H1 H2 H3.
    clear -H1 H2 H3.
    induction H1 using PPsRelInd.
    - constructor.
    - apply invertPsDRel in H2.
      destruct H2 as [(? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ? & ? & ?)].
      + discriminate.
      + inverts H0.
        inverts H1.
        constructor; auto.
  Qed.

  (*
  Lemma DUpdatePPsRelPres:
    forall (R: Type) (Rs: list Type)
      (p1 : Program impl1.(ObjInstruction) R)
      (p2 : Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1 Rs) (ps2 : ProgsState impl2 Rs)
      (ds1 ds1' : State (objBase impl1)) (ds2 ds2' : State (objBase impl2)),

      PPsRel p1 p2 ps1 ps2 ds1 ds2 ->
      (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1)
         (p1' : Program impl2.(ObjInstruction) R1)
         (p2 : Program impl1.(ObjInstruction) R2)
         (p2' : Program impl2.(ObjInstruction) R2),
          PPRel p1 p1' p2 p2' ds1 ds2 ->
          PPRel p1 p1' p2 p2' ds1' ds2') ->
      PPsRel p1 p2 ps1 ps2 ds1' ds2'.

  Proof.
    introv H1 H2.
    induction H1 using PPsRelInd.
    * constructor.
    * constructor; eauto.
  Qed.
  *)

  Lemma PsDRelPres:
    forall
      (ps1: ProgsState impl1)
      (ps2: ProgsState impl2)
      (ds1 ds1': State (objBase impl1))
      (ds2 ds2': State (objBase impl2)),

      PsDRel ps1 ds1 ps2 ds2 ->
      (forall R (p1 : Program impl1.(ObjInstruction) R) p2,
          PDRel p1 ds1 p2 ds2 ->
          PDRel p1 ds1' p2 ds2') ->
      (* (forall (R1 R2 : Type)
         (p1 : Program impl1.(ObjInstruction) R1)
         (p1' : Program impl2.(ObjInstruction) R1)
         (p2 : Program impl1.(ObjInstruction) R2)
         (p2' : Program impl2.(ObjInstruction) R2),
          PPRel p1 p1' p2 p2' ds1 ds2 ->
          PPRel p1 p1' p2 p2' ds1' ds2') -> *)
      DRel ds1' ds2' ->
      PsDRel ps1 ds1' ps2 ds2'.

  Proof.
    clear noopRel invRel retRel stepRel.
    introv H1 H2 H3 (* H4 *).
    induction H1.
    - constructor.
      assumption.
    - constructor; eauto.
      (* eapply DUpdatePPsRelPres; eauto. *)
  Qed.


  Lemma PsDRelPres':
    forall R
      (p1: Program impl1.(ObjInstruction) R)
      (p2: Program impl2.(ObjInstruction) R)
      (ps1 : ProgsState impl1)
      (ps2 : ProgsState impl2)
      (ds1 : State (objBase impl1))
      (ds2 : State (objBase impl2))
      ds1' ds2',

      PPsRel p1 p2 ps1 ps2 ->
      PsDRel ps1 ds1 ps2 ds2 ->
      DRel ds1' ds2' ->
      (forall (R : Type)
         (p1t : Program impl1.(ObjInstruction) R)
         (p2t : Program impl2.(ObjInstruction) R),
          PDRel p1t ds1 p2t ds2 ->
          PPRel p1 p2 p1t p2t ->
          PDRel p1t ds1' p2t ds2') ->
      PsDRel ps1 ds1' ps2 ds2'.

  Proof.
    introv H1 H2 H3 H4.
    clear -H1 H2 H3 H4.
    induction H2 as [ s1 s2 H2 | R' Rs p1t s1 ps1 p2t s2 ps2 H21 H22 ] using PsDRelInd.
    - constructor. assumption.
    - apply invertPPsRel in H1.
      destruct H1 as [(? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ?)].
      + discriminate.
      + inverts H.
        inverts H0.
        constructor; auto.
  Qed.



  Definition Rel (s1 : ThreadsState impl1) (s2 : ThreadsState impl2) : Prop :=
    PsDRel (fst s1) (snd s1)
           (fst s2) (snd s2).


  Lemma ThreadAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            PPsRel p1 p2 (fst s1) (fst s2) (* (snd s1) (snd s2) *) ->
            PDRel p1 (snd s1') p2 (snd s2')
            /\ PPsRel p1 p2 (fst s1') (fst s2') (* (snd s1') (snd s2') *)).

        (* /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2
             R' (p3: Program impl1.(ObjInstruction) R') p4,
              PDRel p1 (snd s1) p2 (snd s2) ->
              PDRel p3 (snd s1) p4 (snd s2) ->
              PPsRel p1 p2 (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPsRel p3 p4 (fst s1) (fst s2) (snd s1) (snd s2) ->
              PPRel p3 p4 p1 p2 (snd s1) (snd s2) ->
              PPRel p3 p4 p1 p2 (snd s1') (snd s2')).
        *)

  Proof.
    introv H1 H2.
    revert PPRelSym stepRel noopRel.
    clear -H1 H2.
    introv PPRelSym stepRel noopRel.
    unfold Rel in *.

    gen s2.
    induction H2 as [ p1 ps1 ds1 ps1' ds1'
                    | Ret ps1 ds1 R m f r ds1'
                    | Ret ps1 ds1 p1 ];
      introv H11;
      destruct s2 as (pps2 & ds2); destruct pps2 as [| p2 ps2]; simpl in *.

    - apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R & p1' & p2' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; discriminate.
    - (* TACons *)
      apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R1 & p1' & p2' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      forwards (s2' & H3 & H4 & H5): IHThreadAdvance (ps2, ds2); simpl; clear IHThreadAdvance.
      + assumption.
      + destruct s2' as (ps2' & ds2'). simpl in *.
        exists (existT _ _ p2' :: ps2', ds2'). simpl in *. splits.
        * apply advanceCons. eassumption.
        * { econstructor.
            - apply H5; assumption.
            - apply H5; assumption.
            - assumption.
          }
        * { intros R p3 p4 H7 H8.
            apply invertPPsRel in H8.
            destruct H8 as [(? & ?) | (R' & p1'' & ps1'' & p2'' & ps2'' & Hps1'' & Hps2'' & H81 & H82)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            split.
            - apply H5; assumption.
            - constructor.
              + eauto.
              + apply H5; assumption.
          }
        (* * intros R p3 p4 R' p5 p6 H7 H8 H9 H10 H11.
          apply invertPPsRel in H9. destruct H9 as (H91 & H92). simpl in *.
          apply invertPPsRel in H10. destruct H10 as (H101 & H102). simpl in *.
          eauto.
        *)

    - apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R' & p1' & p2' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; discriminate.
    - (* TABind *)
      apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      forwards (p2' & ds2' & H2 & H3 & H4 & H5): stepRel; clear stepRel.
      + eassumption.
      + eassumption.
      + eassumption.
      + exists (existT _ Ret p2' :: ps2, ds2').
        simpl. splits.
        * eapply ProgramStepsImplThreadsAdvance in H2. exact H2.
        * { econstructor.
            - assumption.
            - eapply PPsRelPres'.
              eassumption.
              eassumption.
              intros. eapply H5; eassumption.
            - eapply PsDRelPres'.
              + eassumption.
              + eassumption.
              + eassumption.
              + intros R' p3 p4 H6. eapply H5. assumption.
          }
        * { intros R' p1t p2t H6 H7.
            apply invertPPsRel in H7.
            destruct H7 as [(? & ?) | (R'' & p1 & p2 & ps1' & ps2' & Hps1' & Hps2' & H71 & H72)]; [discriminate |].
            inverts Hps1'.
            inverts Hps2'.
            split.
            - apply H5.
              + assumption.
              + apply PPRelSym. assumption.

            - constructor.
              + apply PPRelSym. apply H5; eauto.
              + assumption.
          }

    - apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; discriminate.
    - (* TANoop *)
      apply invertPsDRel in H11.
      destruct H11 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112 & H113 & H114)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      apply noopRel in H111.
      destruct H111 as [(H11 & H12) | [(p2' & H11 & H12 & H13) | (p2' & H11 & H12 & H13)]].

      + exists (existT _ Ret p2'' :: ps2, ds2). simpl in *.
        splits.
        * constructor.
        * { econstructor.
            - assumption.
            - eapply PPsRelPres'; eauto.
            - eapply PsDRelPres'; eauto.
          }
        * { intros R' p1t p2t H2 H3.
            apply invertPPsRel in H3.
            destruct H3 as [(? & ?) | (R'' & p1' & p2' & ps1' & ps2' & Hps1' & Hps2' & H31 & H32)]; [discriminate |].
            inverts Hps1'.
            inverts Hps2'.
            split.
            - assumption.
            - constructor.
              + apply PPRelSym. eauto.
              + assumption.
          }

      + subst p2''.
        exists (existT _ Ret p2' :: ps2, ds2). simpl in *.
        splits.
        * (* assert (A := @TANoop Method impl2). simpl in A. apply A. constructor. *)
          apply ThreadAdvanceImplThreadsAdvance.
          econstructor.
        * { econstructor.
            - assumption.
            - eapply PPsRelPres'; eauto.
            - eapply PsDRelPres'; eauto.
          }
        * { intros R' p1t p2t H2 H3.
            apply invertPPsRel in H3.
            destruct H3 as [(? & ?) | (R'' & p1'' & p2'' & ps1' & ps2' & Hps1' & Hps2' & H31 & H32)]; [discriminate |].
            inverts Hps1'.
            inverts Hps2'.
            split.
            - assumption.
            - constructor.
              + apply PPRelSym. eauto.
              + assumption.
          }

      + exists (existT _ Ret p2' :: ps2, ds2). simpl in *.
        splits.
        * apply ProgramStepsImplThreadsAdvance. eassumption.
        * { econstructor.
            - assumption.
            - eapply PPsRelPres'; eauto.
            - eapply PsDRelPres'; eauto.
          }
        * { intros R' p1t p2t H2 H3.
            apply invertPPsRel in H3.
            destruct H3 as [(? & ?) | (R'' & p1''' & p2''' & ps1' & ps2' & Hps1' & Hps2' & H31 & H32)]; [discriminate |].
            inverts Hps1'.
            inverts Hps2'.
            split.
            - assumption.
            - constructor.
              + apply PPRelSym. eauto.
              + assumption.
          }
  Qed.


  Lemma ThreadsAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadsAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'
        (* /\ forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            PDRel p1 (snd s1') p2 (snd s2'). *)
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
              PDRel p1 (snd s1) p2 (snd s2) ->
              PPsRel p1 p2 (fst s1) (fst s2) (* (snd s1) (snd s2) *) ->
              PDRel p1 (snd s1') p2 (snd s2')
              /\ PPsRel p1 p2 (fst s1') (fst s2') (* (snd s1') (snd s2') *)).

  Proof.
    introv H1 H2.
    clear -H1 H2 PPRelSym noopRel stepRel.
    unfold ThreadsAdvance in H2.
    gen s2.
    induction H2 as [s1 | s1 s1m s1']; introv H1.
    - exists s2. splits. constructor. assumption.
      eauto.
    - forwards (s2m & H3 & H4 & H5): ThreadAdvanceRel.
      + eassumption.
      + eassumption.
      + forwards (s2' & H6 & H7 & H8): IHclos_refl_trans_1n.
        * eassumption.
        * { exists s2'. splits.
            - unfold ThreadsAdvance. eapply ThreadsAdvanceTrans; eassumption.
            - eassumption.
            - introv H9 H10. eapply H8; eapply H5; eassumption.
              (* intros. eauto. *)
          }
  Qed.


  Lemma ThreadsInteractionRel:
    forall Ret s1 s2 s1' int a,
      Rel s1 s2 ->
      ThreadsInteraction impl1 Ret int s1 a s1' ->
      exists s2',
        ThreadsInteraction impl2 Ret int s2 a s2'
        /\ Rel s1' s2'
        (* /\ forall R (p1: Program impl1.(ObjInstruction) R) p2,
            PDRel p1 (snd s1) p2 (snd s2) ->
            PDRel p1 (snd s1') p2 (snd s2'). *)
        /\ (forall R (p1: Program impl1.(ObjInstruction) R) p2,
              PDRel p1 (snd s1) p2 (snd s2) ->
              PPsRel p1 p2 (fst s1) (fst s2) (* (snd s1) (snd s2) *) ->
              PDRel p1 (snd s1') p2 (snd s2')
              /\ PPsRel p1 p2 (fst s1') (fst s2') (* (snd s1') (snd s2') *)).

  Proof.
    introv H1 H2.
    revert invRel retRel PPRelSym.
    clear -H1 H2.
    introv invRel retRel PPRelSym.
    destruct H2.
    revert s2 H1.
    induction H as [ Ret m ps1
                   | Ret r ps1
                   | Ret int p1 ps1 r ps1'];
      intros (ps2 & ds2) H1;
      unfold Rel in *;
      simpl in *.

    - (* TISpawn *)
      clear retRel.
      exists (existT _ Ret (objProgram impl2 Ret m) :: ps2, ds2). simpl.
      splits.
      + constructor. constructor.
      + econstructor.
        * apply invRel. apply invertPsDRel in H1. destruct H1; jauto.
        * { induction H1 using PsDRelInd.
            - constructor.
            - constructor.
              + eapply invRel; eassumption.
              + assumption.
          }
        * assumption.
      + intros R' p1t p2t H2 H3.
        splits.
        * assumption.
        * { constructor. apply PPRelSym. eapply invRel.
            - apply invertPsDRel in H1; destruct H1; jauto.
            - assumption.
            - assumption.
          }


    - (* TIJoin *)
      clear invRel.
      apply invertPsDRel in H1.
      destruct H1 as [(? & ?) | (R & p1' & p2' & ps1' & ps2' & Hps1' & <- & H11 & H12 & H13 & H14)]; [discriminate |].
      inverts Hps1'.
      simpl in *.
      forwards A: retRel. eassumption. assumption. subst.
      exists (ps2', ds2).
      simpl. splits.
      + constructor. constructor.
      + assumption.
      + intros R' p1t p2t H2 H3.
        splits.
        * assumption.
        * apply invertPPsRel in H3; simpl in H3.
          destruct H3 as [(? & ?) | (R & p1' & ps1'' & p2' & ps2'' & Hps1'' & Hps2'' & H31 & H32)]; [discriminate |].
          inverts Hps1''.
          inverts Hps2''.
          assumption.


    - (* TICons *)
      clear invRel retRel.
      apply invertPsDRel in H1.
      destruct H1 as [(? & ?) | (R & p1' & p2' & ps1'' & ps2'' & Hps1'' & Hps2'' & H11 & H12 & H13 & H14)]; [discriminate |].
      destruct ps2 as [| p2 ps2]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      forwards (s2' & H2 & H3 & H4): IHThreadsInteraction' (ps2, ds2).
      + simpl. assumption.
      + destruct s2' as [ps2' ds2']. simpl in *.
        exists (existT _ _ p2' :: ps2', ds2'). splits.
        * (* econstructor. *)
          eapply (invertThreadsInteraction impl2) in H2.
          destruct H2 as (H21 & H22).
          simpl in *. subst.
          econstructor. econstructor. assumption.
        * simpl in *.
          { econstructor.
            + apply H4; assumption.
            + apply H4; assumption.
            + assumption.
          }
        * { simpl in *.
            intros R' p1t p2t H5 H6.
            apply invertPPsRel in H6.
            destruct H6 as [(? & ?) | (R1 & p1'' & ps1'' & p2'' & ps2'' & Hps1'' & Hps2'' & H61 & H62)]; [discriminate |].
            inverts Hps1''.
            inverts Hps2''.
            splits.
            + apply H4; assumption.
            + constructor.
              * assumption.
              * apply H4; assumption.
          }

  Qed.

  Theorem simulatesCoInd':
    forall (s1 : ThreadsState impl1) (s2 : ThreadsState impl2),
      Rel s1 s2 ->
      Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2) s1 s2.
  Proof.
    clear initRel.
    cofix self.
    introv H1.
    constructor.
    introv H2.
    unfold ThreadsBehavior in H2.
    destruct H2 as [ts1m H2 H3].
    forwards (ts2m & H4 & H5 & _): ThreadsAdvanceRel.
    - eassumption.
    - eassumption.
    - forwards (ts2' & H6 & H7 & _): ThreadsInteractionRel.
      + eassumption.
      + eassumption.
      + econstructor.
        * unfold ThreadsBehavior.
          econstructor. exact H4. eassumption.
        * eapply self.
          eassumption.
  Qed.

  Theorem simulatesCoInd:
    ConcSimulates
      impl1 impl2.
  Proof.
    eapply simulatesCoInd'.
    unfold Rel. simpl. constructor.
    assumption.
  Qed.

End SimulatesCoInd.
End SimulatesCoInd.


Module SimulatesCoInd2.
Section SimulatesCoInd2.

  (* This co-ind can be reproved usign the previous co-ind. *)
  Context
    {Method : Type -> Type}
    (impl1 impl2 : Object Method).

  Definition ProgsState {M} (impl : Object M) :=
    list (sigT (Program impl.(ObjInstruction))).

  Definition DataState {M} (impl : Object M) :=
    impl.(objBase).(State).


  Variable (PRel: forall R, Program impl1.(ObjInstruction) R -> Program impl2.(ObjInstruction) R -> Prop).
  Variable (DRel: impl1.(objBase).(State) -> impl2.(objBase).(State) -> Prop).

  Hypothesis initRel: DRel impl1.(objBase).(initialState) impl2.(objBase).(initialState).

  Hypothesis PDRelNoop:
    forall Ret p1 p2,
      PRel Ret (Noop p1) p2 ->
      PRel Ret p1 p2 \/
      (exists p2', p2 = Noop p2' /\ PRel Ret p1 p2').

  Hypothesis invRel:
    forall Ret m,
      PRel Ret (objProgram impl1 Ret m) (objProgram impl2 Ret m).

  Hypothesis retRel:
    forall Ret r p2,
      PRel Ret (Return r) p2 ->
      p2 = Return r.

  Hypothesis stepRel:
    forall Ret R (m: impl1.(ObjInstruction) R) f s1 p2 s2 s1' r,
      PRel Ret (Bind m f) p2 ->
      DRel s1 s2 ->
      impl1.(objBase).(EvalInstruction) m s1 r s1' ->
      exists p2' s2',
        ProgramSteps impl2.(objBase) _ p2 s2 p2' s2'
        /\ PRel Ret (f r) p2'
        /\ DRel s1' s2'.
  (*
    Definition PRel' R p1 p2 :=
      PRel R p1 p2
      /\ exists m s1 s2, ProgramSteps (impl1.(objBase)) R (impl1.(objProgram) R m) s1 p1 s2.
   *)

  Inductive PsRel : ProgsState impl1 -> ProgsState impl2 -> Prop :=
  | NilRel:
      PsRel nil nil
  | ConsRel:
      forall R p1 ps1 p2 ps2,
        (* PRel' R p1 p2 -> *)
        PRel R p1 p2 ->
        PsRel ps1 ps2 ->
        PsRel (existT _ R p1 :: ps1) (existT _ R p2 :: ps2).


  Lemma invertPsRel (ps1 : ProgsState impl1) (ps2 : ProgsState impl2) :
    PsRel ps1 ps2 ->
    (ps1 = nil
     /\ ps2 = nil) \/
    (exists R p1 p2 ps1' ps2',
        existT _ R p1 :: ps1' = ps1 /\ existT _ R p2 :: ps2' = ps2 /\
        PRel R p1 p2
        /\ PsRel ps1' ps2').

  Proof.
    clear.
    introv H1.
    destruct H1;
      [left | right];
      jauto.
  Qed.


  Definition Rel (s1 : ThreadsState impl1) (s2 : ThreadsState impl2) : Prop :=
    PsRel (fst s1) (fst s2)
    /\ DRel (snd s1) (snd s2).

  Lemma ThreadAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'.

  Proof.
    introv H1 H2.
    unfold Rel in *.
    destruct H1 as [H11 H12].
    gen s2.
    induction H2 as [ p1 ps1 ds1 ps1' ds1'
                    | Ret ps1 ds1 R m f r ds1'
                    | Ret ps1 ds1 p1 ];
      introv H11 H12;
      destruct s2 as (pps2 & ds2); destruct pps2 as [| p2 ps2]; simpl in *.

    - apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.

    - (* TACons *)
      clear stepRel.
      apply invertPsRel in H11. simpl in *.
      destruct H11 as [(? & ?) | (R & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      forwards (s2m & H3 & H4 & H5): IHThreadAdvance (ps2, ds2); simpl.
      + assumption.
      + assumption.
      + destruct s2m as (ps2m & ds2m).
        exists (existT _ _ p2'' :: ps2m, ds2m). simpl in *.
        splits.
        * apply advanceCons. eassumption.
        * econstructor; eassumption.
        * eassumption.

    - apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.

    - (* TABind *)
      apply invertPsRel in H11. simpl in *.
      destruct H11 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      forwards (p2' & ds2' & H2 & H3 & H4): stepRel.
      + eassumption.
      + eassumption.
      + eassumption.
      + exists (existT _ Ret p2' :: ps2, ds2').
        simpl. splits.
        * (* eapply ProgramStepImplThreadsAdvance. *)
          eapply ProgramStepsImplThreadsAdvance in H2. exact H2.

        * econstructor; eassumption.
        * eassumption.

    - apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (? & ? & ? & ? & ? & ? & ? & ? & ?)]; discriminate.

    - (* TANoop *)
      clear stepRel.
      apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (R' & p1'' & p2'' & ps1'' & ps2'' & Hps1'' & Hps2'' & H111 & H112)]; [discriminate |].
      inverts Hps1''.
      inverts Hps2''.
      apply PDRelNoop in H111. destruct H111 as [H111 | (p2' & H1111 & H1112)].
      + exists (existT _ Ret p2'' :: ps2, ds2).
        simpl in *.
        splits.
        * constructor.
        * econstructor; assumption.
        * assumption.
      + subst p2''.
        exists (existT _ Ret p2' :: ps2, ds2).
        simpl in *.
        splits.
        * (* assert (A := @TANoop Method impl2). simpl in A. apply A. constructor. *)
          apply ThreadAdvanceImplThreadsAdvance.
          econstructor.
        * econstructor; assumption.
        * assumption.
  Qed.

  Lemma ThreadsAdvanceRel:
    forall s1 s2 s1',
      Rel s1 s2 ->
      ThreadsAdvance impl1 s1 s1' ->
      exists s2',
        ThreadsAdvance impl2 s2 s2'
        /\ Rel s1' s2'.

  Proof.
    introv H1 H2.
    unfold ThreadsAdvance in H2.
    generalize dependent s2.
    induction H2 as [s1 | s1 s1m s1']; introv H1.
    - exists s2. splits. constructor. assumption.
    - forwards (s2m & H3 & H4): ThreadAdvanceRel.
      + eassumption.
      + eassumption.
      + forwards (s2' & H5 & H6): IHclos_refl_trans_1n.
        * eassumption.
        * { exists s2'. splits.
            - unfold ThreadsAdvance. eapply ThreadsAdvanceTrans; eassumption.
            - eassumption.
          }
  Qed.


  Lemma ThreadsInteractionRel:
    forall Ret s1 s2 s1' int a,
      Rel s1 s2 ->
      ThreadsInteraction impl1 Ret int s1 a s1' ->
      exists s2',
        ThreadsInteraction impl2 Ret int s2 a s2'
        /\ Rel s1' s2'.

  Proof.
    introv H1 H2.
    destruct H2.
    revert s2 H1.
    induction H as [ Ret m ps1
                   | Ret r ps1
                   | Ret int p1 ps1 r ps1' ];
      intros (pps2 & ds2) H1; simpl in *.

    - (* TISpawn *)
      exists (existT _ Ret (objProgram impl2 Ret m) :: pps2, ds2).
      splits.
      + constructor. constructor.
      + unfold Rel in *. destruct H1 as (H11 & H12). simpl in *.
        splits.
        * econstructor. apply invRel. assumption.
        * assumption.

    - (* TIJoin *)
      unfold Rel in *.
      simpl in *.
      destruct H1 as [H11 H12].
      apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (R & p1 & p2 & ps1' & ps2' & Hps1' & <- & H111 & H112)]; [discriminate |].
      inverts Hps1'.
      simpl in *.
      apply retRel in H111.
      exists (ps2', ds2).
      simpl. splits.
      + constructor. subst. econstructor.
      + assumption.
      + assumption.

    - (* TICons *)
      unfold Rel in H1.
      simpl in H1.
      destruct H1 as [H11 H12].
      apply invertPsRel in H11.
      destruct H11 as [(? & ?) | (R & p1' & p2' & ps1'' & ps2'' & Hps1'' & <- & H111 & H112)]; [discriminate |].
      inverts Hps1''.
      forwards (s2' & H2 & H3): IHThreadsInteraction' (ps2'', ds2).
      + unfold Rel. splits; simpl; assumption.

      + forwards (s2m' & H4 & H5): IHThreadsInteraction' (ps2'', ds2).
        * unfold Rel. simpl; splits; assumption.
        * { destruct s2m' as [ps2m' ds2m'].
            exists (existT _ _ p2' :: ps2m', ds2m'). splits.
            - (* econstructor. *)
              eapply (invertThreadsInteraction impl2) in H4.
              destruct H4 as (H41 & H42).
              simpl in *. subst.
              econstructor. econstructor. assumption.
            - unfold Rel in *. simpl in *. destruct H5 as (H51 & H52).
              splits. econstructor; assumption. assumption.
          }
  Qed.

  Theorem simulatesCoInd':
    forall s1 s2,
      Rel s1 s2 ->
      Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2) s1 s2.
  Proof.
    cofix self.
    introv H1.
    constructor.
    introv H2.
    unfold ThreadsBehavior in H2.
    destruct H2 as [ts1m H2 H3].
    forwards (ts2m & H4 & H5): ThreadsAdvanceRel.
    - eassumption.
    - eassumption.
    - forwards (ts2' & H6 & H7): ThreadsInteractionRel.
      + eassumption.
      + eassumption.
      + econstructor.
        * unfold ThreadsBehavior.
          econstructor. exact H4. eassumption.
        * eapply self.
          eassumption.
  Qed.

  Theorem simulatesCoInd:
    ConcSimulates impl1 impl2.
  Proof.
    eapply simulatesCoInd'.
    unfold Rel. simpl. split. constructor. apply initRel.
  Qed.

End SimulatesCoInd2.
End SimulatesCoInd2.


Module SimulatesSerImplCoInd.
  Section SimulatesSerImplCoInd.

  Context {M: Type -> Type}.
  Context (cSpec : Spec M).
  Context (aSpec : Spec M).


  Definition PRel R (p1 p2: Program M R): Prop :=
    p1 = p2
    /\ ((exists m, p1 = (x <- m; Return x)) \/ (exists r, p1 = Return r)).


  Variable DRel: cSpec.(State) -> aSpec.(State) -> Prop.

  Hypothesis initRel: DRel cSpec.(initialState) aSpec.(initialState).

  Hypothesis stepRel:
    forall R (m: M R) s1 s1' r s2,
      DRel s1 s2 ->
      cSpec.(EvalInstruction) m s1 r s1' ->
      exists s2',
        aSpec.(EvalInstruction) m s2 r s2'
        /\ DRel s1' s2'.

  (*
  Hypothesis stepRel:
    forall R R' (m: M R') f p2 s1 s1' r s2,
      PRel R (x <- m; f x) p2 ->
      DRel s1 s2 ->
      CEval R' m s1 r s1' ->
      exists p2' s2',
        ProgramSteps aSpec R p2 s2 p2' s2'
        /\ PRel R (f r) p2'
        /\ DRel s1' s2'.
  *)

  Theorem simulatesSerImplCoInd:
    ConcSimulates
      (serialObject cSpec) (serialObject aSpec).
  Proof.
    eapply (SimulatesCoInd2.simulatesCoInd (serialObject cSpec)
                           (serialObject aSpec)
                           PRel DRel).
    - apply initRel.
    - introv H1. exfalso. unfold PRel in H1. destruct H1 as [_ [[m H1] | [r H1]]]; inversion H1.
    - introv. simpl. unfold PRel. eauto.
    - introv H1. unfold PRel in H1. destruct H1 as (H1 & _). eauto.
    - introv H1 H2 H3. simpl in *.
      unfold PRel in H1. destruct H1 as (H1 & H1'). subst.
      forwards (s2' & A1 & A2): stepRel H2 H3.
      exists (f r). exists s2'. splits.
      + apply ProgramStepImplProgramSteps. constructor. simpl. assumption.
      + unfold PRel. split. reflexivity. destruct H1' as [[m' H11'] | [r' H12']].
        * right. inverts H11'. exists r. reflexivity.
        * inverts H12'.
      + assumption.
  Qed.

  End SimulatesSerImplCoInd.

  Lemma SeqRefImplConRef:
    forall {Instruction: Type -> Type}
      (obj: Spec Instruction) (obj': Spec Instruction),
      Simulates obj obj' ->
      ConcSimulates (serialObject obj) (serialObject obj').
  Proof.
    introv H1.

    eapply simulatesSerImplCoInd.
    - exact H1.
    - clear H1.
      introv H2 H3.
      inverts H2 as H2.
      forwards A: H2.
      simpl in *.
      eassumption.
      clear H2.
      destruct A as (s2', A1, A2). simpl in *.
      eexists. splits; eassumption.
  Qed.

  Lemma SeqRefImplConRefIO:
    forall {Instruction: Type -> Type}
      (impl: Object Instruction) (obj: Spec Instruction),
      Simulates (coarsen impl) obj ->
      ConcSimulates (serialObject (coarsen impl)) (serialObject obj).
  Proof.
    introv H1.
    eapply SeqRefImplConRef. assumption.
  Qed.

End SimulatesSerImplCoInd.
