Set Asymmetric Patterns.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Pair
  PairSimulates
  Threads
  LinCoInd
  FlattenSimulates
  SimulatesFacts
  Linearizability.

Section SimulatesflattenProgram.

  Context {Instruction : Type -> Type}.
  Variable (spec: Spec Instruction).
  Variable (M: Type -> Type).
  Variable (impl1 impl2: forall R, M R -> Program Instruction R).

  Definition PRel R (p1 p2: Program Instruction R): Prop :=
    programNEq p1 p2.
    (* /\ forall r, p1 = Return r -> p2 = Return r. *)

  Definition DRel (s1: spec.(State)) (s2: spec.(State)): Prop :=
    s1 = s2.

  Lemma initRel: DRel spec.(initialState) spec.(initialState).
  Proof. reflexivity. Qed.

  Lemma noopRel:
    forall Ret p1 p2,
      PRel Ret (Noop p1) p2 ->
      PRel Ret p1 p2 \/
      (exists p2', p2 = Noop p2' /\ PRel Ret p1 p2').

  Proof.
    introv H1.
    unfold PRel in *.
    inverts H1.
    - right. eexists. split. reflexivity. assumption.
    - left. assumption.
  Qed.

  Lemma invRel:
    implNEq impl1 impl2 ->
    forall Ret m,
      PRel Ret (impl1 Ret m) (impl2 Ret m).

  Proof.
    introv H1.
    intros.
    unfold PRel.
    apply H1.
  Qed.

  Lemma retRel:
    forall Ret r p2,
      PRel Ret (Return r) p2 ->
      p2 = Return r.
  Proof.
    introv H1.
    unfold PRel in H1.
    inverts H1.
    reflexivity.
  Qed.


  Lemma stepRel:
    forall Ret R (m: Instruction R) f s1 p2 s2 s1' r,
      PRel Ret (Bind m f) p2 ->
      DRel s1 s2 ->
      spec.(EvalInstruction) m s1 r s1' ->
      exists p2' s2',
        ProgramSteps spec _ p2 s2 p2' s2'
        /\ PRel Ret (f r) p2'
        /\ DRel s1' s2'.
  Proof.
    introv H1 H2 H3.
    unfold PRel in H1.
    unfold DRel in H2. subst s2.
    inverts H1.
    exists (f2 r) s1'.
    splits.
    + apply ProgramStepImplProgramSteps. constructor. assumption.
    + unfold PRel. auto.
    + unfold DRel. reflexivity.
  Qed.


  Lemma ImplNEqSimulates:
    implNEq impl1 impl2 ->
    ConcSimulates
      {| objBase := spec; objProgram := impl1 |}
      {| objBase := spec; objProgram := impl2 |}.

  Proof.
    introv H1.
    eapply (SimulatesCoInd2.simulatesCoInd {| objBase := spec; objProgram := impl1 |}
                                           {| objBase := spec; objProgram := impl2 |}
                                           PRel DRel initRel noopRel (invRel H1) retRel stepRel).
  Qed.

End SimulatesflattenProgram.

Lemma simulatesSerial {Instruction} spec M (impl: forall R, M R -> Program Instruction R) impl' :
    ConcSimulates
      {| objBase := spec;
         objProgram := impl |}
      impl' ->
    ConcSimulates
      (flattenObject (serialObject spec) _ impl)
      impl'.
Proof.
  introv H1.
  eapply simulatesTrans; [ | exact H1 ].
  clear H1. unfold serialObject. unfold flattenObject. simpl.
  eapply ImplNEqSimulates.
  unfold implNEq.
  intros.
  apply FlattenBindRetNEq.
Qed.

Lemma LinSubst:
  forall {Instruction Method}
    (baseObj: Spec Instruction) baseImpl
    (impl': Object Method) impl,

    Linearizable baseObj baseImpl ->
    ConcSimulates
      {| objBase := baseObj; objProgram := impl |} impl' ->
    ConcSimulates
      (flattenObject baseImpl _ impl) impl'.

Proof.
  introv H1 H2.
  unfold Linearizable in *.
  eapply flattenSimulates2.
  eassumption.
  eapply simulatesSerial. assumption.
Qed.

Lemma FlattenSerialRightNEq:
  forall {IL} (baseObjL : Spec IL) (MR : Type -> Type)
    (baseImplR : Object MR) (R' : Type)
    (p : Program (PairMethod IL MR) R'),
    programNEq
      (flattenProgram
         (objProgram
            (pairObject (serialObject baseObjL) baseImplR))
         p)
      (flattenProgramRight
         (objProgram baseImplR)
         p).

Proof.
  cofix CoInd.
  intros. destruct p.
  - destruct p.
    + match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end.
      constructor.
      match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
          simpl
      end.
      econstructor. intros.
      match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
          simpl
      end.
      econstructor.
      apply (@CoInd IL baseObjL).

    + match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
          rewriteFrobProgram p';
          simpl
      end.
      constructor.
      generalize (objProgram baseImplR A m).
      cofix CoInd'. intros.
      destruct p.
      * match goal with
        | [ |- programNEq ?p ?p' ] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end.
        constructor. intros.
        eapply CoInd'.

      * match goal with
        | [ |- programNEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end. constructor.
        apply (@CoInd IL baseObjL).

      * match goal with
        | [ |- programNEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end. constructor.
        eapply CoInd'.

  - setoid_rewrite frobProgramId. simpl.
    constructor.

  - match goal with
    | [ |- programNEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p';
      simpl
    end.
    constructor. apply (@CoInd IL baseObjL).
Qed.


Lemma LinSubstLeft:
  forall {IL} (baseObjL: Spec IL) (baseImplL: Object IL)
    {MR} (baseImplR: Object MR)
    {I} (impl': Object I)
    (impl: forall R, I R -> Program (PairMethod IL MR) R),

    Linearizable baseObjL baseImplL ->
    ConcSimulates
      {| objBase := pairSpec baseObjL baseImplR.(objBase);
         objProgram := flattenImplProgRight baseImplR.(objProgram) impl |}
      impl' ->
    ConcSimulates
      (flattenObject (pairObject baseImplL baseImplR) _ impl)
      impl'.

Proof.
  introv H1 H2.
  unfold Linearizable in *.
  eapply flattenSimulates2.
  - eapply pairSimulatesL. exact H1.
  - clear H1.
    eapply simulatesTrans; [ |exact H2 ].
    eapply ImplNEqSimulates.
    unfold implNEq. intros.
    unfold flattenImplProgRight.
    eapply FlattenSerialRightNEq.
Qed.

Lemma FlattenSerialLeftNEq:
  forall (ML: Type -> Type) (baseImplL : Object ML)
    {IR} (baseObjR : Spec IR)
    (R' : Type)(p : Program (PairMethod ML IR) R'),
    programNEq
      (flattenProgram
         (objProgram
            (pairObject  baseImplL (serialObject baseObjR))) p)
      (flattenProgramLeft (objProgram baseImplL) p).

Proof.
  cofix CoInd.
  intros. destruct p.
  - destruct p.

    + match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end.
      constructor.
      generalize (objProgram baseImplL A m).
      cofix CoInd'.
      intros. destruct p.
      * match goal with
        | [ |- programNEq ?p ?p' ] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end.
        constructor. intros.
        eapply CoInd'.

      * match goal with
        | [ |- programNEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end. constructor.
        apply (@CoInd ML baseImplL IR baseObjR).

      * match goal with
        | [ |- programNEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p';
            simpl
        end. constructor.
        eapply CoInd'.

    + match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end.
      econstructor.
      match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end.
      constructor.
      intros.
      match goal with
      | [ |- programNEq ?p ?p' ] =>
        rewriteFrobProgram p;
        rewriteFrobProgram p';
        simpl
      end. constructor.
      apply (@CoInd ML baseImplL IR baseObjR).

  - setoid_rewrite frobProgramId. simpl.
    constructor.

  - match goal with
    | [ |- programNEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p';
      simpl
    end.
    constructor. apply (@CoInd ML baseImplL IR baseObjR).
Qed.


Lemma LinSubstRight:
  forall {ML} (baseImplL: Object ML)
    {IR} (baseObjR: Spec IR)
    (baseImplR: Object IR)
    impl'
    (impl: forall R, IR R -> Program (PairMethod ML IR) R),

    Linearizable baseObjR baseImplR ->
    ConcSimulates
      {| objBase := pairSpec baseImplL.(objBase) baseObjR;
         objProgram := flattenImplProgLeft baseImplL.(objProgram) impl |}
      impl' ->
    ConcSimulates
      (flattenObject (pairObject baseImplL baseImplR) _ impl)
      impl'.

Proof.
  introv H1 H2.
  unfold Linearizable in *.
  eapply flattenSimulates2.
  - eapply pairSimulatesR. exact H1.
  - clear H1.
    eapply simulatesTrans; [ | exact H2 ].
    eapply ImplNEqSimulates.
    unfold implNEq. intros.
    eapply FlattenSerialLeftNEq.
Qed.

Ltac flatten thm :=
  match type of thm with
  | Linearizable ?o _ => eapply (@LinSubst _ _ o); [ apply thm | ]
  end.
