(** Expediting "these are kinda the same" kinds of reasoning. *)

Set Implicit Arguments.
Set Contextual Implicit.
Set Maximal Implicit Insertion.

From Coq Require Import List.

From C4.Util Require LibTactics.
From C4.Util Require Import Util.

Ltac injpair := LibTactics.injpair.

From C4 Require Import
  Program
  Object
  Threads
  ThreadsFacts
  Linearizability
  RSimulates
  Eutt.

Lemma ProgramSteps_eutt_Return {M} (sp : Spec M) R p2 s (r : R) s'
  : ProgramSteps sp R p2 s (Return r) s' ->
    forall p2', eutt ieq p2 p2' ->
    ProgramSteps sp R p2' s (Return r) s'.
Proof.
  unfold ProgramSteps at 1; unfold ProgramSteps'.
  pose (ps2 := (p2, s)). change (p2, s) with ps2. change p2 with (fst ps2). change s with (snd ps2).
  clearbody ps2; clear p2 s.
  assert (WW := Operators_Properties.rt1n_ind_right _ (ProgramStep' sp R)).
  intros II; pattern ps2; revert II; eapply WW; clear WW; cbn.
  - intros ? EE; apply Symmetric_eutt_ieq in EE; eapply eutt_ProgramSteps, EE.
  - intros. destruct H.
    + apply eutt_Bind_inv in H2. destruct H2 as (_ & k2 & [] & EE & UT).
      induction UT.
      * eapply ProgramFacts.ProgramStepStepsImplSteps.
        { constructor. } { auto. }
      * subst. eapply ProgramFacts.ProgramStepStepsImplSteps.
        { constructor; eauto. } { auto. }
    + apply eutt_Noop_L_inv in H2. auto.
Qed.

Inductive OneProgramStep_ {M} (spec : Spec M) {R} (s2 : State spec) (p2 : Program M R)
    (s1 : State spec) : Program M R -> Prop :=
| OneProgramStep_Noop p1
  : OneProgramStep_ spec s2 p2 s1 p1 -> OneProgramStep_ spec s2 p2 s1 (Noop p1)
| OneProgramStep_Bind R0 (e : M R0) (k : R0 -> Program M R) (r0 : R0)
  : EvalInstruction spec e s1 r0 s2 ->
    p2 = k r0 ->
    OneProgramStep_ spec s2 p2 s1 (Bind e k)
.

Notation OneProgramStep spec s1 p1 s2 p2 := (OneProgramStep_ spec s2 p2 s1 p1).

Module Export SMS.
Section Facts.

Context {M : Type -> Type}.
Context (o1 o2 : Object M).

Notation OI := ObjInstruction.
Notation OP := objProgram.

Record SimulatesManySteps (init_s1 init_s2 : State _) : Type :=
  { RR : forall R, Program (OI o1) R -> Program (OI o2) R -> Prop
  ; RS : State (objBase o1) -> State (objBase o2) -> Prop
  ; RS_init : RS init_s1 init_s2
  ; RR_Spawn : forall R (m : M R), RR (OP o1 _ m) (OP o2 _ m)
  ; RR_Step : forall s1 s2, RS s1 s2 ->
      forall R (p1 : Program _ R) (p2 : Program _ R), RR p1 p2 ->
      forall s1' p1', OneProgramStep (objBase o1) s1 p1 s1' p1' ->
      exists p1'' s2' p2', eutt ieq p1' p1'' /\ ProgramSteps _ _ p2 s2 p2' s2' /\
       RS s1' s2' /\ RR p1'' p2'
  ; RR_Join : forall R (p1 : _ R) (p2 : _ R) r, RR p1 p2 -> eutt ieq p1 (Return r) ->
     forall s, ProgramSteps (objBase o2) _ p2 s (Return r) s
  }.

Context (SMS : SimulatesManySteps o1.(objBase).(initialState) o2.(objBase).(initialState)).

Notation T1_ := {x : Type & Program (OI o1) x}.
Notation T2_ := {x : Type & Program (OI o2) x}.
Notation T1 := (list T1_).
Notation T2 := (list T2_).

Definition Steppable {R} p p' := forall s, ProgramSteps (objBase o2) R p s p' s.

Inductive exRR (rrr : forall X, Program (OI o2) X -> Program (OI o2) X -> Prop) : T1_ -> T2_ -> Prop :=
| ExRR X p1 p2 p1' p2'
  : RR SMS p1' p2' -> eutt ieq p1 p1' -> eutt ieq p2 p2' ->
    (forall r, p1 <> Return r) ->
    exRR rrr (existT _ X p1) (existT _ X p2)
| ExRR_Return X r p2
  : rrr _ p2 (Return r) ->
    exRR rrr (existT _ X (Return r)) (existT _ X p2)
.

Definition all_RR : T1 -> T2 -> Prop := Forall2 (exRR (@Steppable)).
Definition all_RR' : T1 -> T2 -> Prop := Forall2 (exRR (fun _ => eq)).

Notation TB := ThreadsBehavior.

Lemma Eval_OneProgramStep (o : Object M) R R1 (instr : _ R) (a : R) (p1 : _ R1) f st st'
  : EvalInstruction (objBase o) instr st a st' ->
    eutt ieq (x <- instr; f x) p1 ->
    exists p1',
      OneProgramStep (objBase o) st p1 st' p1' /\ eutt ieq (f a) p1'.
Proof.
  intros EVAL EQ. apply eutt_Bind_inv in EQ. destruct EQ as (m2 & k2 & -> & Ek & UNTAUS).
  induction UNTAUS as [ * ? IH | ].
  - destruct IH as (q & ? & ?); exists q; split; auto. constructor; auto.
  - exists (k2 a). split; auto. subst p; econstructor; eauto.
Qed.

Lemma OneProgramStep_ProgramSteps (o : Object M) R (p p' : _ R) st st'
  : OneProgramStep (objBase o) st p st' p' -> ProgramSteps (objBase o) R p st p' st'.
Proof.
  induction 1; subst.
  - eapply ProgramFacts.ProgramStepStepsImplSteps; [ constructor | eauto ].
  - eapply ProgramFacts.ProgramStepImplProgramSteps; constructor; auto.
Qed.

Lemma ProgramSteps_eutt {Ret} p p' p_ s s_
  : eutt ieq p p' ->
    ProgramSteps (objBase o2) Ret p s p_ s_ ->
    exists p_', eutt ieq p_ p_' /\ ProgramSteps (objBase o2) Ret p' s p_' s_.
Proof.
  intros EQ STEPS; revert EQ. red in STEPS. remember (p,s) as ps; remember (p_,s_) as ps_.
  revert p s p_ s_ p' Heqps Heqps_. induction STEPS as [ | * ? ? IH ]; intros; subst.
  - inversion Heqps_; subst; clear Heqps_.
    exists p'; split; auto; apply ProgramFacts.ProgramStepsRefl.
  - red in H; cbn in H. destruct y; cbn in H.
    destruct H.
    + eapply Eval_OneProgramStep in H; eauto. destruct H as (? & HH & H).
      apply OneProgramStep_ProgramSteps in HH.
      specialize (IH _ _ _ _ _ eq_refl eq_refl H).
      destruct IH as (p'' & ? & ?).
      exists p''; split; auto.
      eapply ProgramFacts.ProgramStepsStepsImplSteps; eauto.
    + eapply IH; eauto. apply eutt_Noop_L_inv; assumption.
Qed.

Lemma returnOrNot {MM R} (p : Program MM R)
  : (exists r, p = Return r) \/ (forall r, p <> Return r).
Proof.
  destruct p.
  - right; discriminate.
  - left; eauto.
  - right; discriminate.
Qed.

Lemma stepsReturn_ {R} p1 p1' p2 p2'
  : RR SMS p1 p2 -> eutt ieq p1 p1' -> eutt ieq p2 p2' ->
    exists p_,
      ((exists r, p1' = Return r /\ p_ = Return r) \/ ((forall r, p1' <> Return r) /\ p_ = p2'))
      /\ forall s, ProgramSteps (objBase o2) R p2' s p_ s.
Proof.
  destruct (returnOrNot (p := p1')) as [[r ->] | K]; intros.
  - exists (Return r). split.
    + left. eauto.
    + intros s; eassert (JJ := RR_Join SMS H H0 (s0 := s)).
      eauto using ProgramSteps_eutt_Return.
  - exists p2'. split.
    + right; auto.
    + intros; apply ProgramFacts.ProgramStepsRefl.
Qed.

Ltac forward H :=
  lazymatch type of H with
  | (?P -> _) => let H0 := fresh H in assert (H0 : P); [ | specialize (H H0); clear H0]
  end.

Lemma SMS_ThreadAdvance ts1 ts1'
  : ThreadAdvance o1 ts1 ts1' ->
    forall ts2 t2 s2,
    ts2 = (t2, s2) ->
    all_RR' (fst ts1) t2 ->
    RS SMS (snd ts1) s2 ->
    exists ts2',
      ThreadsAdvance o2 ts2 ts2' /\ all_RR' (fst ts1') (fst ts2') /\ RS SMS (snd ts1') (snd ts2').
Proof.
  induction 1 as [ * ? IH | | ]; intros * -> H_RR H_RS; cbn in *.
  - inversion H_RR; subst.
    specialize (IH _ _ _ eq_refl H4 H_RS).
    destruct IH as ([t2' s2'] & ? & ? & ?).
    exists (y :: t2', s2').
    split; [ | split ]; cbn.
    + apply advanceCons. auto.
    + constructor; auto.
    + auto.
  - inversion H_RR; subst.
    inversion H2; injpair.
    eapply Eval_OneProgramStep in H; [ | eassumption ]. destruct H as (? & ? & ?).
    eapply (RR_Step SMS) in H; try eassumption.
    destruct H as (? & ? & ? & ? & ? & ? & ?).
    eapply ProgramSteps_eutt in H1; [ | apply Symmetric_eutt_ieq; eassumption ].
    destruct H1 as (p_' & ? & ?).
    eassert (HH := stepsReturn_ (p1' := f a)).
    forward HH; [ eassumption |].
    forward HH; [ apply Symmetric_eutt_ieq; eapply Transitive_eutt_ieq; eassumption | ].
    forward HH; [ eassumption | ].
    destruct HH as (p__ & HH & ?).
    exists (existT _ _ p__ :: l', x1).
    split; [ | split ].
    + apply ProgramStepsImplThreadsAdvance.
      eapply ProgramFacts.ProgramStepsStepsImplSteps; eauto.
    + constructor; auto.
      destruct HH as [(? & -> & ?) | [] ]; [constructor 2; auto | econstructor 1].
      * apply H9.
      * eapply Transitive_eutt_ieq; eassumption.
      * subst p__. apply Symmetric_eutt_ieq; assumption.
      * assumption.
    + auto.
  - inversion H_RR; subst. inversion H1; injpair.
    apply eutt_Noop_L_inv in H4.
    eassert (HH := stepsReturn_ H2 (Symmetric_eutt_ieq _ _ H4) (Symmetric_eutt_ieq _ _ H5)).
    destruct HH as (p_ & ? & ?).
    exists (existT _ _ p_ :: l', s2). split; [ | split]; auto.
    + apply ProgramStepsImplThreadsAdvance; auto.
    + constructor; auto.
      destruct H as [(? & -> & ->) | [? ->]];
        [constructor 2 | econstructor 1 with (1 := H2)]; eauto.
Qed.

Lemma SMS_ThreadsAdvance ts1 ts1'
  : ThreadsAdvance o1 ts1 ts1' ->
    forall ts2 t2 s2,
    ts2 = (t2, s2) ->
    all_RR' (fst ts1) t2 ->
    RS SMS (snd ts1) s2 ->
    exists ts2',
      ThreadsAdvance o2 ts2 ts2' /\ all_RR' (fst ts1') (fst ts2') /\ RS SMS (snd ts1') (snd ts2').
Proof.
  induction 1 as [ | * ADV H_etc IH ].
  - intros; exists (t2, s2); subst; cbn; split; [ constructor | auto ].
  - intros _ * -> H_RR H_RS.
    eapply SMS_ThreadAdvance in ADV; eauto. destruct ADV as ([] & ? & ? & ?).
    cbn in *.
    edestruct IH as (ts2'' & ADV'' & H_RR'' & H_RS''); eauto.
    exists ts2''; split; [ | auto ].
    eapply ThreadsAdvanceTrans; eassumption.
Qed.

Lemma SMS_ThreadsAdvance' t1 s1
  : forall t2 s2,
    all_RR t1 t2 ->
    RS SMS s1 s2 ->
    exists t2',
      ThreadsAdvance o2 (t2, s2) (t2', s2) /\ all_RR' t1 t2'.
Proof.
  intros t2 s2 H_RR H_SMS.
  induction H_RR as [ | * ? ? IH ].
  - exists nil. split; [ apply Reflexive_ThreadsAdvance | constructor ].
  - destruct IH as (t2' & ? & ?).
    destruct H.
    + eassert (HH := stepsReturn_).
      forward HH; [ eassumption | ].
      forward HH; [ apply Symmetric_eutt_ieq; eassumption | ].
      forward HH; [ apply Symmetric_eutt_ieq; eassumption | ].
      destruct HH as (p_ & HH & ?).
      exists (existT _ _ p_ :: t2').
      destruct HH as [(? & -> & ->)| [? ->]].
      { split.
        * eapply ThreadsAdvanceTrans; [ eapply advanceCons; eassumption | eapply ProgramStepsImplThreadsAdvance; eauto ].
        * constructor; auto. constructor 2; eauto.
      }
      { split.
        * apply advanceCons. auto.
        * constructor; [ econstructor | ]; eauto.
      }
    + exists (existT _ X (Return r) :: t2'). split.
      * eapply ThreadsAdvanceTrans; [ eapply advanceCons; eassumption | eapply ProgramStepsImplThreadsAdvance; eauto ].
      * constructor; auto. constructor 2; eauto.
Qed.

Lemma RR_RR' p y
  : exRR (fun X : Type => eq) p y ->
    exRR (@Steppable) p y.
Proof.
  intros [ | ? ? ? -> ]; [econstructor 1| constructor 2]; eauto.
  intros ?; apply ProgramFacts.ProgramStepsRefl.
Qed.

Lemma all_RR_RR' t1 t2 : all_RR' t1 t2 -> all_RR t1 t2.
Proof.
  apply Forall2_impl_. auto using RR_RR'.
Qed.

Lemma SMS_ThreadsInteraction ts1 ts1' t2 s2 R (i : Interaction M R) (r : R)
  : all_RR' (fst ts1) t2 ->
    RS SMS (snd ts1) s2 ->
    ThreadsInteraction o1 R i ts1 r ts1' ->
    exists t2',
      ThreadsInteraction o2 R i (t2, s2) r (t2', s2) /\ all_RR (fst ts1') t2' /\ RS SMS (snd ts1') s2.
Proof.
  intros H_RR H_RS INT. destruct INT as [* INT]; cbn in *.
  revert t2 s2 H_RR H_RS; induction INT as [ | | * INT IH ]; cbn; intros.
  - exists (existT _ _ (OP o2 Ret m) :: t2).
    split; [ | split ].
    + constructor. constructor.
    + eassert (HH := stepsReturn_ (p1 := OP o1 _ m) (p2 := OP o2 _ m)).
      forward HH; [ eapply RR_Spawn | ].
      forward HH; [ apply Reflexive_eutt_ieq | ].
      forward HH; [ apply Reflexive_eutt_ieq | ].
      constructor; [ | revert H_RR; apply all_RR_RR'].
      destruct HH as (p_ & [(? & -> & ->) | [? ?]] & ?).
      { constructor 2; auto. }
      { econstructor; try eapply Reflexive_eutt_ieq; auto.
        { apply RR_Spawn. }
      }
    + auto.
  - inversion H_RR; subst. exists l'.
    inversion H1; injpair; [ congruence | ].
    split; [ | split ].
    + constructor. constructor.
    + revert H3; apply all_RR_RR'.
    + auto.
  - inversion H_RR; subst.
    specialize (IH _ _ H3 H_RS).
    destruct IH as (t2' & ? & ? & ?).
    exists (y :: t2').
    split; [ | split ].
    + inversion H; subst. constructor. constructor. auto.
    + constructor; auto. revert H1. apply RR_RR'.
    + auto.
Qed.

Lemma SMS_ConcS_ : forall t1 t2 s1 s2,
    all_RR t1 t2 -> RS SMS s1 s2 -> Simulates.Simulates_ (TB o1) (TB o2) (t1, s1) (t2, s2).
Proof.
  cofix SELF. intros * H_RR H_RS. constructor.
  intros * EVAL. destruct EVAL as [* ADV INT].
  eapply SMS_ThreadsAdvance' in H_RR; cbn; eauto. destruct H_RR as (t2_ & ADV_ & H_RR_).
  eapply SMS_ThreadsAdvance in ADV; cbn; eauto. destruct ADV as ([t2' s2'] & ADV & H_RR' & H_RS').
  eapply SMS_ThreadsInteraction in INT; eauto. destruct INT as (t2'' & INT & H_RR'' & H_RS'').
  destruct s1'.
  exists (t2'', s2'); [exists (t2', s2') | ]; eauto.
  eapply ThreadsAdvanceTrans; eassumption.
Qed.

Theorem SMS_ConcS : ConcSimulates o1 o2.
Proof.
  apply SMS_ConcS_.
  - constructor.
  - apply RS_init.
Qed.

End Facts.
End SMS.
