Set Asymmetric Patterns.
From Coq Require Import Eqdep List.

From C4.Core Require Import
  Program
  ProgramOps
  Threads
  ThreadsFacts
  Linearizability
  FlattenSimulates.

Set Implicit Arguments.

Definition progHead I R (p : Program I R) : option (sigT I) :=
  match p with
  | Bind _ m _ => Some (existT _ _ m)
  | _ => None
  end.

Definition progTail I R (p : Program I R) : option {T : Type & T -> Program I R} :=
  match p with
  | Bind _ _ f => Some (existT _ _ f)
  | _ => None
  end.

Lemma Some_inj : forall A (x y : A), Some x = Some y -> x = y.
Proof.
  congruence.
Qed.

Ltac inv' :=
  intuition subst; simpl;
  repeat match goal with
         | [ H : ?X = ?X |- _ ] => clear H
         | [ H : Some _ = Some _ |- _ ] => apply Some_inj in H; subst
         | [ H : Return _ = Return _ |- _ ] => inversion_clear H; subst
         | [ H : existT _ ?X _ = existT _ ?Y _ |- _ ] =>
           (unify X Y; apply inj_pair2 in H || fail 1)
             || (inversion H; clear H); subst
         | [ H : Bind _ _ = Bind _ _ |- _ ] =>
           generalize (f_equal (@progHead _ _) H); apply (f_equal (@progTail _ _)) in H;
           intro; simpl in *
         | _ => discriminate
         end.

Ltac inv H := inversion H; clear H; inv'.

Hint Extern 1 (ProgramSteps _ _ _ _ _ _) => hnf.

Ltac progStep := hnf; cbv beta;
  match goal with
  | [ |- Relation_Operators.clos_refl_trans_1n _ _ (Bind _ _, _) _ ] =>
    eapply Relation_Operators.rt1n_trans; [ red; simpl; instantiate (1 := (_, _)); simpl | ]
  end.

Hint Extern 1 (Relation_Operators.clos_refl_trans_1n _ _ _ _) => progStep.
Hint Extern 1 (Relation_Operators.clos_refl_trans_1n _ _ _ _) => apply Relation_Operators.rt1n_refl.
Hint Extern 1 (EvalInstruction _ _ _ _ _) => progress hnf.
Hint Extern 1 (ProgramStep _ _ (Bind _ _) _ _ _) => eapply EBind.
