From C4.Util Require Import
  LibTactics Util.
From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts
  Threads
  Object
  MapBehavior.

Section MapObjectBehavior.

  Context {Method Method' : Type -> Type}
          (impl : Object Method)
          (f : forall Ret, Method' Ret -> Method Ret).

  Lemma mapThreadAdvance st st' :
    ThreadAdvance impl st st' <->
    ThreadAdvance (mapObject f impl) st st'.
  Proof.
    split.
    - induction 1.
      + constructor 1 with (Method:=Method').
        assumption.
      + constructor 2 with (Method:=Method').
        assumption.
      + constructor 3 with (Method:=Method').
    - induction 1.
      + constructor 1 with (Method:=Method).
        assumption.
      + constructor 2 with (Method:=Method).
        assumption.
      + constructor 3 with (Method:=Method).
  Qed.

  Lemma mapThreadsAdvance st st' :
    ThreadsAdvance impl st st' <->
    ThreadsAdvance (mapObject f impl) st st'.
  Proof.
    split.
    - induction 1; econstructor.
      + edestruct mapThreadAdvance; eauto.
      + assumption.
    - induction 1; econstructor.
      + edestruct mapThreadAdvance; eauto.
      + assumption.
  Qed.

  Lemma mapThreadsInteraction' Ret int ps a ps' :
    ThreadsInteraction' impl Ret (mapInteraction f int) ps a ps' <->
    ThreadsInteraction' (mapObject f impl) Ret int ps a ps'.
  Proof.
    split.
    - intro ti. revert ps ps' ti; induction int; intros; inversions ti; injpair.
      + constructor 1 with (impl := mapObject f impl).
      + constructor 2 with (impl := mapObject f impl).
      + constructor 3 with (impl := mapObject f impl). auto.
    - intro ti. revert ps ps' ti; induction int; intros; inversions ti; injpair; cbn.
      all: constructor; auto.
  Qed.

  Lemma mapThreadsInteraction Ret int st a st' :
    ThreadsInteraction impl Ret (mapInteraction f int) st a st' <->
    ThreadsInteraction (mapObject f impl) Ret int st a st'.
  Proof.
    split; intros [].
    - split with (Method:=Method').
      apply mapThreadsInteraction'.
      assumption.
    - split with (Method:=Method).
      apply mapThreadsInteraction'.
      assumption.
  Qed.

  Lemma mapThreadsBehavior Ret (int : Interaction Method' Ret) st a st' :
    (ThreadsBehavior impl).(EvalInstruction) (mapInteraction f int) st a st' <->
    (ThreadsBehavior (mapObject f impl)).(EvalInstruction) int st a st'.
  Proof.
    split.
    - destruct 1.
      esplit.
      + apply mapThreadsAdvance.
        eassumption.
      + apply mapThreadsInteraction.
        assumption.
    - destruct 1.
      esplit.
      + apply mapThreadsAdvance.
        eassumption.
      + apply mapThreadsInteraction.
        assumption.
  Qed.

  Theorem mapBehaviorObject (st : ThreadsState impl) :
    Simulates_
      (MapBehavior f (ThreadsBehavior impl))
      (ThreadsBehavior (mapObject f impl))
      st
      st.
  Proof.
    eapply relationSimulates.
    - instantiate (1 := eq).
      intros.
      subst.
      eexists.
      + edestruct mapThreadsBehavior as [H _].
        eapply H.
        simpl in *.
        eassumption.
      + reflexivity.
    - reflexivity.
  Qed.

  Theorem mapObjectBehavior (st : ThreadsState impl) :
    Simulates_
      (ThreadsBehavior (mapObject f impl))
      (MapBehavior f (ThreadsBehavior impl))
      st
      st.
  Proof.
    eapply relationSimulates.
    - instantiate (1 := eq).
      intros.
      subst.
      eexists.
      + eapply mapThreadsBehavior; eassumption.
      + reflexivity.
    - reflexivity.
  Qed.

End MapObjectBehavior.


Section MapSimulates.

  Context {Method Method' : Type -> Type}
          (impl1 impl2 : Object Method)
          (f : forall Ret, Method' Ret -> Method Ret).

  Theorem mapSimulates :
    ConcSimulates impl1 impl2 ->
    ConcSimulates (mapObject f impl1) (mapObject f impl2).
  Proof.
    intro.
    eapply simulatesTrans, simulatesTrans.
    - apply mapObjectBehavior.
    - apply mapBehaviorSimulates.
      apply H.
    - apply mapBehaviorObject.
  Qed.

End MapSimulates.
