Require Import
        Coq.Lists.List
        Coq.Relations.Relations.

From C4.Core Require Import
  Program
  Object
  Simulates
  SimulatesFacts
  Threads.

Section MapState.

  Context
    {Method : Type -> Type}
    (impl1 : Object Method)
    (B2 : Spec (Interaction Method))
    (MapState : ThreadsState impl1 -> B2.(State) -> Prop)
    (advance :
       forall
         ts1 ts1' ts2
         (ms : MapState ts1 ts2)
         (ta1 : ThreadAdvance impl1 ts1 ts1'),
       exists2 ts2',
         Simulates_ B2 B2 ts2' ts2 &
         MapState ts1' ts2')
    (interact :
       forall
         Ret ts1 a ts1' ts2
         (ms : MapState ts1 ts2)
         (int : Interaction Method Ret)
         (ti : ThreadsInteraction impl1 Ret int ts1 a ts1'),
       exists2 ts2',
         B2.(EvalInstruction) int ts2 a ts2' &
         MapState ts1' ts2').

  Lemma mapStateAdvance
        ts1 ts1' ts2
        (ms : MapState ts1 ts2)
        (tas1 : ThreadsAdvance impl1 ts1 ts1') :
    exists2 ts2',
      Simulates_ B2 B2 ts2' ts2 &
      MapState ts1' ts2'.
  Proof.
    revert ts2 ms.
    induction tas1 as [| ts1 ts1' ts1'' tas1]; intros.
    - exists ts2.
      + apply simulatesRefl.
      + assumption.
    - destruct advance with ts1 ts1' ts2 as [ts2'].
      + assumption.
      + assumption.
      + destruct IHtas1 with ts2' as [ts2'' tas2' ms''].
        * assumption.
        * { exists ts2''.
            - eapply simulatesTrans_; eassumption.
            - assumption.
          }
  Qed.

  Lemma mapStateBehavior
        A int ts1 (a : A) ts1'' ts2
        (ms : MapState ts1 ts2)
        (behavior : (ThreadsBehavior impl1).(EvalInstruction) int ts1 a ts1'') :
    exists2 ts2'',
      B2.(EvalInstruction) int ts2 a ts2'' &
      exists2 ts2''',
        Simulates_ B2 B2 ts2''' ts2'' &
        MapState ts1'' ts2'''.
  Proof.
    induction behavior as [ts1' ta ti].
    edestruct mapStateAdvance with ts1 ts1' ts2 as [ts2' [S2] ms']; try assumption.
    edestruct interact; try eassumption.
    edestruct S2; eauto.
  Qed.

  CoFixpoint mapStateSimulates' ts1 ts2 ts2' :
    Simulates_ B2 B2 ts2' ts2 ->
    MapState ts1 ts2' ->
    Simulates_ (ThreadsBehavior impl1) B2
              ts1 ts2.
  Proof.
    destruct 1 as [S].
    split.
    intros.
    edestruct mapStateBehavior as [? ? [? S' ?]].
    - eassumption.
    - eassumption.
    - edestruct S.
      + eassumption.
      + eexists.
        * eassumption.
        * { eapply mapStateSimulates'.
            - eapply simulatesTrans_.
              + exact S'.
              + assumption.
            - assumption.
          }
  Qed.

  CoFixpoint mapStateSimulates ts1 ts2
             (ms : MapState ts1 ts2) :
    Simulates_ (ThreadsBehavior impl1) B2
              ts1 ts2.
  Proof.
    eapply mapStateSimulates'.
    - apply simulatesRefl.
    - assumption.
  Qed.

End MapState.
