From Coq Require Import List Relations.

From C4.Util Require Import LibTactics.

From C4.Core Require Import
  Tactics
  Program
  ProgramOps
  Object
  Simulates
  Threads
  ThreadsFacts
  MapState.

Section MethodSpecBehavior.

  Context
    {Instruction Instruction' Method : Type -> Type}
    (obj : Spec Instruction)
    (impl : forall Ret, Method Ret -> Program Instruction' Ret)
    (f : forall Ret, Instruction' Ret -> Instruction Ret).

  Definition mapProgramImpl := {| objBase := obj; objProgram Ret m := mapProgram f (impl Ret m) |}.
  Definition mapSpecImpl := {| objBase := mapSpec obj f; objProgram := impl |}.

  Definition mapPrograms :
    list (sigT (Program Instruction')) ->
    list (sigT (Program Instruction)) :=
    map (fun p => existT _ (projT1 p) (mapProgram f (projT2 p))).

  Lemma methodThreadAdvance ts2 ts1' :
    ThreadAdvance mapProgramImpl (mapPrograms (fst ts2), snd ts2) ts1' ->
    exists2 ts2',
      ThreadAdvance mapSpecImpl ts2 ts2' &
      ts1' = (mapPrograms (fst ts2'), snd ts2').
  Proof.
    intro ta.
    remember (mapPrograms (fst ts2), snd ts2) as ts1.
    revert ts2 Heqts1.
    induction ta; intros [[| [Ret' p2] ps2] st2] Heqts1; inversions Heqts1; injpair.
    - edestruct IHta with (ts2:=(ps2, st2)).
      + reflexivity.
      + inversion H0.
        subst.
        destruct x.
        eexists.
        * constructor 1 with (impl:=mapSpecImpl).
          eassumption.
        * reflexivity.
    - frobinvr p2 H2.
      eexists.
      + constructor 2 with (impl:=mapSpecImpl).
        simpl in *.
        eassumption.
      + reflexivity.
    - frobinvr p2 H1.
      eexists.
      + constructor 3 with (impl:=mapSpecImpl).
      + reflexivity.
  Qed.

  Lemma methodThreadsInteraction' Ret int ps2 a ps1' :
    ThreadsInteraction' mapProgramImpl Ret int (mapPrograms ps2) a ps1' ->
    exists2 ps2',
      ThreadsInteraction' mapSpecImpl Ret int ps2 a ps2' &
      ps1' = mapPrograms ps2'.
  Proof.
    intro ti'.
    remember (mapPrograms ps2) as ps1.
    revert ps2 Heqps1.
    induction ti'; intros ps2 Heqps1; subst.
    - eexists.
      + constructor 1.
      + reflexivity.
    - destruct ps2 as [| [Ret' p2] ps2']; [discriminate |].
      inversions Heqps1; injpair.
      frobinvr p2 H1.
      eexists.
      + constructor 2.
      + reflexivity.
    - destruct ps2 as [| [Ret'' p2] ps2']; [discriminate |].
      inversions Heqps1.
      edestruct IHti'.
      + reflexivity.
      + subst.
        eexists.
        * constructor 3.
          eassumption.
        * reflexivity.
  Qed.

  Theorem mapMethodSimulates
          (ps : list (sigT (Program Instruction')))
          (st : State obj) :
    Simulates_
      (ThreadsBehavior mapProgramImpl)
      (ThreadsBehavior mapSpecImpl)
      (mapPrograms ps, st)
      (ps, st).
  Proof.
    apply mapStateSimulates with (MapState:=fun ts1 ts2 => ts1 = (mapPrograms (fst ts2), snd ts2)).
    - intros.
      subst.
      edestruct methodThreadAdvance.
      + eassumption.
      + eexists.
        * apply simulateAdvance, clos_rt1n_step.
          eassumption.
        * assumption.
    - intros.
      subst.
      destruct ts2.
      inversions ti.
      edestruct methodThreadsInteraction'.
      + eassumption.
      + subst.
        repeat econstructor; eassumption.
    - reflexivity.
  Qed.

End MethodSpecBehavior.
