From C4.Core Require Import
  Program
  Pair
  Prod
  ProgramOps.

Definition coarsen {Method} (impl : Object Method) : Spec Method :=
  mapSpec (programSpec impl.(objBase)) impl.(objProgram).

Definition serialObject {Method} (obj : Spec Method) : Object Method :=
  {|
    objBase := obj;
    objProgram Ret m := Bind m Return
  |}.

Definition coarsenImpl {Method} (impl: Object Method): Object Method :=
  serialObject (coarsen impl).


Definition unitObject : Object UnitInstruction :=
  {|
    objBase := unitSpec;
    objProgram Ret m := Empty_set_rect (fun _ => _) m
  |}.

Definition mapImplProg
           {Instruction Method Method'}
           (f : forall Ret, Method' Ret -> Method Ret)
           (impl : forall Ret, Method Ret -> Program Instruction Ret)
           Ret (m : Method' Ret) :
  Program Instruction Ret :=
  impl Ret (f Ret m).

Definition mapObject
           {Method Method'}
           (f : forall Ret, Method' Ret -> Method Ret)
           (impl : Object Method) :
  Object Method' :=
  {|
    objBase := impl.(objBase);
    objProgram := mapImplProg f impl.(objProgram)
  |}.

Definition flattenImplProg
           {Method}{M}
           (base: forall Ret, Method Ret -> Program M Ret)
           {Method'}
           (impl : forall Ret, Method' Ret -> Program Method Ret):
  forall Ret, Method' Ret -> Program M Ret :=
  fun Ret m =>
    flattenProgram base (impl Ret m).

Definition flattenObject
           {Method}
           (base : Object Method)
           Method'
           (impl : forall Ret, Method' Ret -> Program Method Ret) :
  Object Method' :=
  {|
    objBase := base.(objBase);
    objProgram := flattenImplProg base.(objProgram) impl
  |}.   
  

Definition prodInjectProgram {T Instruction} (t : T) {Ret} :
  Program (Instruction t) Ret ->
  Program (ProdMethod Instruction) Ret :=
  mapProgram (fun A => existT (fun t' => Instruction t' A) t).

Definition prodObject
           {T Method} (impl : forall t : T, Object (Method t)) :
  Object (ProdMethod Method) :=
  {|
    objBase := prodSpec (fun t => (impl t).(objBase));
    objProgram Ret m :=
      let 'existT _ t m' := m in
      prodInjectProgram t (objProgram (impl t) Ret m')
  |}.

Definition pairObject
           {MethodL} (implL : Object MethodL)
           {MethodR} (implR : Object MethodR) :
  Object (PairMethod MethodL MethodR) :=
  {|
    objBase := pairSpec implL.(objBase) implR.(objBase);
    objProgram Ret m :=
      match m with
        | inl mL => mapProgram (fun _ => inl) (implL.(objProgram) Ret mL)
        | inr mR => mapProgram (fun _ => inr) (implR.(objProgram) Ret mR)
      end
  |}.

Definition flattenImplLeft
           {ML} (baseL: Object ML)
           {IR} (baseR: Spec IR)
           {M'} (impl: forall R, M' R -> Program (PairMethod ML IR) R) :
  Object M' :=
  {|
    objBase := pairSpec baseL.(objBase) baseR;
    objProgram := flattenImplProgLeft baseL.(objProgram) impl
  |}.

Definition flattenImplRight
           {IL} (baseL: Spec IL)
           {MR} (baseR: Object MR)
           {M'} (impl: forall R, M' R -> Program (PairMethod IL MR) R) :
  Object M' :=
  {|
    objBase := pairSpec baseL baseR.(objBase);
    objProgram := flattenImplProgRight baseR.(objProgram) impl
  |}.

Definition flattenImplLeft'
           {ML} (baseL: Object ML)
           {IR} (baseR: Spec IR)
           {M'} (impl: forall R, M' R -> Program (PairMethod ML IR) R) :
  Object M' :=
  {|
    objBase := pairSpec baseL.(objBase) baseR;
    objProgram := flattenImplProgLeft' baseL.(objProgram) impl
  |}.

Definition flattenImplRight'
           {IL} (baseL: Spec IL)
           {MR} (baseR: Object MR)
           {M'} (impl: forall R, M' R -> Program (PairMethod IL MR) R) :
  Object M' :=
  {|
    objBase := pairSpec baseL baseR.(objBase);
    objProgram := flattenImplProgRight' baseR.(objProgram) impl
  |}.


Definition OpStartImplProg 
           {M I} (impl: forall R, M R -> Program I R): Prop :=
  forall R m, OpStartProgram (impl R m).

Definition OpStartImpl
           {M} (impl: Object M): Prop :=
  OpStartImplProg (impl.(objProgram)).
