From C4.Util Require Import
  LibTactics
  MTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Object
  Simulates
  SimulatesFacts
  Pair.

Section PairObjSimulates.
  
  Context
    {ML MR : Type -> Type}
    (objL1 : Spec ML)
    (objR1 : Spec MR)
    (objL2 : Spec ML)
    (objR2 : Spec MR).

  Lemma pairObjSimulates stL1 stR1 stL2 stR2 :
    Simulates_
      objL1 objL2
      stL1 stL2 ->
    Simulates_
      objR1 objR2
      stR1 stR2 ->    
    Simulates_
      (pairSpec objL1 objR1)
      (pairSpec objL2 objR2)
      (stL1, stR1)
      (stL2, stR2).
  Proof.
    intros * H1 H2.
    eapplyRelationSimulates
      (fun (s1 : State objL1 * State objR1) (s2 : State objL2 * State objR2) =>
         let (stL1, stR1) := s1 in
         let (stL2, stR2) := s2 in
         Simulates_
           objL1 objL2
           stL1 stL2
         /\ Simulates_
             objR1 objR2
             stR1 stR2).
    intros * N1 N2.
    destruct s1 as (sL1 & sR1).
    destruct s1' as (sL1' & sR1').
    destruct s2 as (sL2 & sR2).
    destruct N1 as (N11 & N12).
    simpl in *.
    destruct int.
    - simpl in N2. destruct N2 as (N21 & N22). subst sR1'.
      inverts N11 as N11. forwardsIn N11. eassumption.
      destruct N11 as (sL2', N111, N112).
      exists (sL2', sR2). split.
      + simpl. assumption.
      + reflexivity.
      + split; assumption.
    - simpl in N2. destruct N2 as (N21 & N22). subst sL1'.
      inverts N12 as N12. forwardsIn N12. eassumption.
      destruct N12 as (sR2', N121, N122).
      exists (sL2, sR2'). split.
      + simpl. assumption.
      + reflexivity.
      + split; assumption.
    - split; assumption.
  Qed.

End PairObjSimulates.


Theorem methodPairSpec ML MR ML' MR'
        (objL : Spec ML) (objR : Spec MR)
        (fL : forall A, ML' A -> ML A)
        (fR : forall A, MR' A -> MR A)
        (st : objL.(State) * objR.(State)) :
  Simulates_
    (mapSpec
       (pairSpec objL objR)
       (fun A m =>
          match m with
          | inl m' => inl (fL A m')
          | inr m' => inr (fR A m')
          end))
    (pairSpec
       (mapSpec objL fL)
       (mapSpec objR fR))
    st
    st.
Proof.
  eapplyRelationSimulates (@eq (State objL * State objR)).
  - intros.
    subst.
    eexists.
    + destruct int; simpl in *; eassumption.
    + reflexivity.
  - reflexivity.
Qed.
