From C4.Core Require Import
  Program
  Object
  SimulatesFacts
  Threads
  Pair
  Prod
  Simulates
  ProdBehavior
  ProdSimulates
  MapBehavior
  MapSimulates
  MethodSimulates
  BaseSimulates
  ProgramEqSimulates
  ProgramOpsFacts.

Section PairSpec'.

  Context {MethodL : Type -> Type}
          (BL : Spec (Interaction MethodL))
          {MethodR : Type -> Type}
          (BR : Spec (Interaction MethodR)).

  Definition BoolProdMethod : bool -> Type -> Type :=
    bool_rect _ MethodL MethodR.

  Definition boolProdMethod Ret :
    PairMethod MethodL MethodR Ret ->
    ProdMethod BoolProdMethod Ret :=
    sum_rect _ (existT _ true) (existT _ false).

  Definition prodBoolMethod Ret :
    ProdMethod BoolProdMethod Ret ->
    PairMethod MethodL MethodR Ret :=
    sigT_rect _ (bool_rect _ inl inr).

  Definition BoolB : forall b, Spec (Interaction (BoolProdMethod b)) :=
    bool_rect _ BL BR.

  Definition PairState := ProdState bool BoolProdMethod BoolB.

  Definition PairSpec' : Spec (Interaction (PairMethod MethodL MethodR)) :=
    MapBehavior boolProdMethod (ProdBehavior BoolB).

End PairSpec'.

Arguments PairState : clear implicits.


Section PairSpec'Simulates.

  Context {MethodL MethodR : Type -> Type}
          (BL1 : Spec (Interaction MethodL))
          (BR1 : Spec (Interaction MethodR))
          (BL2 : Spec (Interaction MethodL))
          (BR2 : Spec (Interaction MethodR)).

  Theorem pairBehaviorSimulates :
    Simulates BL1 BL2 ->
    Simulates BR1 BR2 ->
    Simulates (PairSpec' BL1 BR1) (PairSpec' BL2 BR2).
  Proof.
    intros.
    apply mapBehaviorSimulates.
    apply prodBehaviorSimulates.
    - decide equality.
    - intro t.
      destruct t; assumption.
  Qed.

End PairSpec'Simulates.

Section PairProdObject.

  Context (MethodL MethodR : Type -> Type)
          (implL : Object MethodL)
          (implR : Object MethodR).

  Definition boolImpl : forall b, Object (@BoolProdMethod MethodL MethodR b) :=
    bool_rect _ implL implR.

  Theorem prodPairObject :
    ConcSimulates
      (mapObject boolProdMethod (prodObject boolImpl))
      (pairObject implL implR).
  Proof.
    eapply simulatesTrans, simulatesTrans.
    2: exact
         (mapMethodSimulates
            (prodSpec (fun t => objBase (boolImpl t)))
            (pairObject implL implR).(objProgram)
            (fun Ret => sum_rect _ (existT _ true) (existT _ false))
            nil (initialState _)).
    - apply programEqSimulates.
      intros.
      destruct m; symmetry; apply mapMapProgram.
    - simpl.
      apply baseSimulates.
      apply relationSimulates with (fun st : forall b, State (objBase (boolImpl b)) => eq (st true, st false)).
      + intros.
        simpl in *.
        subst.
        destruct int; destruct H0.
        * eexists (_, _).
          -- split.
             ++ eassumption.
             ++ reflexivity.
          -- f_equal.
             symmetry.
             apply H0.
             discriminate.
        * eexists (_, _).
          -- split.
             ++ eassumption.
             ++ reflexivity.
          -- f_equal.
             symmetry.
             apply H0.
             discriminate.
      + reflexivity.
  Qed.

  Theorem pairProdObject :
    ConcSimulates
      (pairObject implL implR)
      (mapObject boolProdMethod (prodObject boolImpl)).
  Proof.
    unfold pairObject.
    eapply simulatesTrans, simulatesTrans.
    2: exact
         (mapMethodSimulates
            (pairSpec (objBase implL) (objBase implR))
            (mapObject boolProdMethod (prodObject boolImpl)).(objProgram)
            (fun Ret => sigT_rect _ (bool_rect _ inl inr))
            nil (initialState _)).
    - apply programEqSimulates.
      intros.
      destruct m; symmetry; apply mapMapProgram.
    - simpl.
      apply baseSimulates.
      apply relationSimulates with (fun st' st => (st true, st false) = st').
      + intros.
        destruct s1'.
        simpl in *.
        subst.
        destruct int as [[]]; destruct H0; simpl in *; subst.
        * eexists (bool_rect _ _ _).
          -- split.
             ++ eassumption.
             ++ intros.
                destruct t'; tauto.
          -- reflexivity.
        * eexists (bool_rect _ _ _).
          -- split.
             ++ eassumption.
             ++ intros.
                destruct t'; tauto.
          -- reflexivity.
      + reflexivity.
  Qed.

End PairProdObject.

Section PairSimulates.

  Context
    {MethodL MethodR : Type -> Type}
    (implL1 : Object MethodL)
    (implR1 : Object MethodR)
    (implL2 : Object MethodL)
    (implR2 : Object MethodR).

  Theorem pairSimulates :
    ConcSimulates implL1 implL2 ->
    ConcSimulates implR1 implR2 ->
    ConcSimulates (pairObject implL1 implR1) (pairObject implL2 implR2).
  Proof.
    intros SL SR.
    eapply simulatesTrans, simulatesTrans.
    - apply pairProdObject.
    - apply mapSimulates, prodSimulates.
      + decide equality.
      + intro t.
        instantiate (1 := bool_rect _ _ _).
        destruct t; eassumption.
    - apply prodPairObject.
  Qed.

End PairSimulates.

Theorem pairSimulatesL:
  forall {ML MR}
    (implL1 implL2: Object ML)
    (implR1: Object MR),
    ConcSimulates implL1 implL2 ->
    ConcSimulates (pairObject implL1 implR1) (pairObject implL2 implR1).
Proof.
  intros. eapply pairSimulates. eassumption. eapply simulatesRefl.
Qed.

Theorem pairSimulatesR:
  forall {ML MR}
    (implL: Object ML)
    (implR1 implR2: Object MR),
    
    ConcSimulates implR1 implR2 ->
    ConcSimulates (pairObject implL implR1) (pairObject implL implR2).
Proof.
  intros. eapply pairSimulates. eapply simulatesRefl. eassumption.
Qed.
