From Coq Require Import List Eqdep FunctionalExtensionality Relations.

From C4.Util Require Import LibTactics.

From C4.Core Require Import
  Tactics
  Program
  ProgramOps
  Simulates
  SimulatesFacts
  Threads
  ThreadsFacts
  MapState
  Object
  Prod
  ProdBehavior.

Import ListNotations.


Definition splitSigT {T} {A : T -> Type} {t : T} (dec : forall t' : T, {t = t'} + {t <> t'}) :
  list (sigT A) -> list (A t) :=
  flat_map
    (fun e =>
       let 'existT _ t' a := e in
       match dec t' with
       | left E => [eq_rect t' A a t (eq_sym E)]
       | right _ => nil
       end).

Definition prodInject {T Instruction}
           (e : {t : T & sigT (Program (Instruction t))}) :
  sigT (Program (ProdMethod Instruction)) :=
  let 'existT _ t (existT _ Ret p) := e in
  existT _ Ret (prodInjectProgram t p).


Section ProdObjectBehavior.

  Context
    T
    (Tdec : forall t t' : T, {t = t'} + {t <> t'})
    (Method : T -> Type -> Type)
    (impl : forall T, Object (Method T)).

  Definition PBIRel
             (st1 : ProdState T Method (fun t => ThreadsBehavior (impl t)))
             (st2 : ThreadsState (prodObject impl)) :
    Prop :=
    exists es,
      prodTs _ st1 = map (@projT1 _ _) es /\
      map prodInject es = fst st2 /\
      forall t, prodState _ st1 t = (splitSigT (Tdec t) es, snd st2 t).

  Lemma prodThreadAdvance
        s s'
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))}) :
    ThreadAdvance (prodObject impl) (map prodInject es, s) s' ->
    exists es' t,
      ThreadAdvance (impl t) (splitSigT (Tdec t) es, s t) (splitSigT (Tdec t) es', snd s' t) /\
      map (@projT1 _ _) es = map (@projT1 _ _) es' /\
      fst s' = map prodInject es' /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es') /\
      (forall t', t <> t' -> s t' = snd s' t').
  Proof.
    intro ta.
    remember (map prodInject es, s) as s1.
    revert es Heqs1.
    induction ta; intros.
    - destruct es; [discriminate |].
      destruct s0 as [? [? ?]].
      inversions Heqs1.
      edestruct IHta as (? & ? & ? & ? & ? & ? & ?).
      + reflexivity.
      + eexists (existT _ x _ :: _), x2.
        simpl in *.
        split; [| split; [| split]].
        * destruct Tdec.
          -- destruct e.
             constructor 1.
             eassumption.
          -- eassumption.
        * f_equal.
          assumption.
        * f_equal.
          assumption.
        * split.
          -- intros.
             destruct Tdec.
             ++ destruct e.
                simpl.
                f_equal.
                intuition auto.
             ++ eauto.
          -- eauto.
    - destruct es; [discriminate |].
      destruct s0 as [? [? ?]].
      inversions Heqs1; injpair.
      frobinvr p H2.
      inversions H.
      eexists (existT _ x _ :: _), x.
      simpl in *.
      split; [| split; [| split; [| split]]].
      + destruct Tdec; [| congruence].
        rewrite <- eq_rect_eq.
        rewrite <- eq_rect_eq.
        simpl.
        constructor 2.
        eassumption.
      + reflexivity.
      + reflexivity.
      + intros.
        destruct Tdec; [congruence |].
        reflexivity.
      + assumption.
    - destruct es; [discriminate |].
      destruct s0 as [? [? ?]].
      inversions Heqs1; injpair.
      frobinvr p0 H1.
      eexists (existT _ x _ :: _), x.
      simpl in *.
      split; [| split; [| split; [| split]]].
      + destruct Tdec; [| congruence].
        rewrite <- eq_rect_eq.
        rewrite <- eq_rect_eq.
        simpl.
        constructor 3.
      + reflexivity.
      + reflexivity.
      + intros.
        destruct Tdec; [congruence |].
        reflexivity.
      + reflexivity.
  Qed.

  Lemma prodThreadAdvance_r
        (st2 : forall t, (impl t).(objBase).(State))
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))})
        (t : T)
        (st1 : (impl t).(objBase).(State))
        (s : (ThreadsBehavior (impl t)).(State)) :
    ThreadAdvance (impl t) (splitSigT (Tdec t) es, st1) s ->
    exists es',
      ThreadAdvance
        (prodObject impl)
        (map prodInject es, overrideDep (Tdec t) st1 st2)
        (map prodInject es', overrideDep (Tdec t) (snd s) st2) /\
      map (@projT1 _ _) es = map (@projT1 _ _) es' /\
      fst s = splitSigT (Tdec t) es' /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es').
  Proof.
    intro ta.
    remember (splitSigT (Tdec t) es, st1) as x.
    revert st2 es st1 Heqx.
    induction ta; intros; inversions Heqx.
    - induction es; inversions H0.
      simpl in *.
      destruct a.
      destruct Tdec.
      + subst.
        inversions H1.
        subst.
        edestruct IHta as (? & ? & ? & ? & ?).
        * reflexivity.
        * subst.
          eexists (existT (fun _ => _) _ _ :: _).
          simpl.
          split; [| split; [| split]].
          -- constructor.
             eassumption.
          -- f_equal.
             assumption.
          -- destruct Tdec.
             ++ rewrite <- eq_rect_eq.
                reflexivity.
             ++ congruence.
          -- intros.
             destruct Tdec.
             ++ congruence.
             ++ eauto.
      + destruct IHes as (? & ? & ? & ? & ?).
        * assumption.
        * destruct s.
          eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
          simpl.
          split; [| split; [| split]].
          -- constructor.
             eassumption.
          -- f_equal.
             assumption.
          -- destruct Tdec.
             ++ congruence.
             ++ assumption.
          -- intros.
             destruct Tdec.
             ++ f_equal; intuition auto.
             ++ eauto.
    - induction es; simpl in *; inversions H1; injpair.
      destruct a0.
      destruct Tdec.
      + subst.
        inversions H2.
        eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
        simpl.
        split; [| split; [| split]].
        * rewrite (frobProgramId (prodInjectProgram _ _)).
          apply (TABind (prodObject impl) _ _ _ _ _ (fun _ => prodInjectProgram _ _)).
          split.
          -- rewrite <- overrideDepEq, <- overrideDepEq.
             eassumption.
          -- intros.
             rewrite <- overrideDepNEq, <- overrideDepNEq; trivial.
        * reflexivity.
        * destruct Tdec.
          -- rewrite <- eq_rect_eq.
             reflexivity.
          -- congruence.
        * intros.
          destruct Tdec.
          -- congruence.
          -- reflexivity.
      + destruct IHes as (? & ? & ? & ? & ?).
        * assumption.
        * destruct s.
          eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
          simpl.
          split; [| split; [| split]].
          -- constructor.
             eassumption.
          -- f_equal.
             assumption.
          -- destruct Tdec.
             ++ congruence.
             ++ assumption.
          -- intros.
             destruct Tdec.
             ++ f_equal; intuition auto.
             ++ eauto.
    - induction es; simpl in *; inversions H0.
      destruct a.
      destruct Tdec.
      + simpl in *.
        subst.
        inversions H1.
        eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
        simpl.
        split; [| split; [| split]].
        * rewrite (frobProgramId (prodInjectProgram _ _)).
          constructor.
        * reflexivity.
        * destruct Tdec.
          -- rewrite <- eq_rect_eq.
             reflexivity.
          -- congruence.
        * intros.
          destruct Tdec.
          -- congruence.
          -- reflexivity.
      + simpl in *.
        destruct IHes as (? & ? & ? & ? & ?).
        * assumption.
        * destruct s.
          eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
          simpl.
          split; [| split; [| split]].
          -- constructor.
             eassumption.
          -- f_equal.
             assumption.
          -- destruct Tdec.
             ++ congruence.
             ++ assumption.
          -- intros.
             destruct Tdec.
             ++ f_equal; intuition auto.
             ++ eauto.
  Qed.

  Lemma prodThreadsAdvance_r
        (st2 : forall t, (impl t).(objBase).(State))
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))})
        (t : T)
        (st1 : (impl t).(objBase).(State))
        (s : (ThreadsBehavior (impl t)).(State)) :
    ThreadsAdvance (impl t) (splitSigT (Tdec t) es, st1) s ->
    exists es',
      ThreadsAdvance
        (prodObject impl)
        (map prodInject es, overrideDep (Tdec t) st1 st2)
        (map prodInject es', overrideDep (Tdec t) (snd s) st2) /\
      map (@projT1 _ _) es = map (@projT1 _ _) es' /\
      fst s = splitSigT (Tdec t) es' /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es').
  Proof.
    intro ta.
    remember (splitSigT (Tdec t) es, st1) as x.
    revert st1 st2 es Heqx.
    induction ta; intros st1 st2 es Heqx; subst.
    - repeat econstructor.
    - edestruct prodThreadAdvance_r as (es' & Hes'1 & Hes'2 & Hes'3 & Hes'4).
      + exact H.
      + destruct y.
        simpl in *.
        subst.
        edestruct IHta as (es'' & Hes''1 & Hes''2 & Hes''3 & Hes''4).
        * reflexivity.
        * exists es''.
          repeat split.
          -- econstructor 2; eassumption.
          -- rewrite Hes'2.
             assumption.
          -- assumption.
          -- intros t' N.
             rewrite Hes'4, Hes''4; trivial.
  Qed.

  Lemma prodThreadsInteraction'
        Ret pint a ps'
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))}) :
    ThreadsInteraction' (prodObject impl) Ret pint (map prodInject es) a ps' ->
    exists es' t int,
      ps' = map prodInject es' /\
      ProdInteraction t (map (@projT1 _ _) es) (map (@projT1 _ _) es') int pint /\
      ThreadsInteraction' (impl t) Ret int (splitSigT (Tdec t) es) a (splitSigT (Tdec t) es') /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es').
  Proof.
    intro ti.
    remember (map prodInject es) as ps.
    revert es Heqps.
    induction ti; intros; subst.
    - destruct m.
      eexists (existT (fun _ => _) _ (existT _ _ _) :: _), _, _.
      split; [| split; [| split]].
      + reflexivity.
      + constructor.
      + simpl.
        destruct Tdec; [| congruence].
        rewrite <- eq_rect_eq.
        constructor.
      + intros.
        simpl.
        destruct Tdec; [congruence |].
        reflexivity.
    - destruct es; [discriminate |].
      destruct s as [? [? ?]].
      inversions Heqps; injpair.
      frobinvr p H1.
      eexists _, _, _.
      split; [| split; [| split]].
      + reflexivity.
      + constructor.
      + simpl.
        destruct Tdec; [| congruence].
        rewrite <- eq_rect_eq.
        constructor.
      + simpl.
        intros.
        destruct Tdec; [congruence |].
        reflexivity.
    - destruct es; inversions Heqps.
      destruct s as [? [? ?]].
      edestruct IHti as (? & ? & ? & ? & ? & ? & ?).
      + reflexivity.
      + subst.
        eexists (existT (fun _ => _) _ (existT _ _ _) :: _), x2.
        destruct (Tdec x2 x).
        * subst.
          repeat esplit.
          -- constructor.
             eassumption.
          -- simpl.
             destruct Tdec.
             rewrite <- eq_rect_eq.
             constructor.
             ++ assumption.
             ++ congruence.
          -- intros.
             simpl.
             destruct Tdec; [congruence |].
             eauto.
        * repeat esplit.
          -- constructor; eassumption.
          -- simpl.
             destruct Tdec; [congruence |].
             assumption.
          -- intros.
             simpl.
             f_equal.
             intuition auto.
  Qed.

  Lemma prodThreadsInteraction
        Ret pint s a s'
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))}) :
    ThreadsInteraction (prodObject impl) Ret pint (map prodInject es, s) a s' ->
    exists es' t int,
      fst s' = map prodInject es' /\
      ProdInteraction t (map (@projT1 _ _) es) (map (@projT1 _ _) es') int pint /\
      ThreadsInteraction (impl t) Ret int (splitSigT (Tdec t) es, s t) a (splitSigT (Tdec t) es', snd s' t) /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es') /\
      (forall t', t <> t' -> s t' = snd s' t').
  Proof.
    intro ti; inversions ti.
    edestruct prodThreadsInteraction' as (? & ? & ? & ? & ? & ? & ?).
    - eassumption.
    - subst.
      repeat esplit; eassumption.
  Qed.

  Lemma prodThreadsInteraction_r A
        (st2 : forall t, (impl t).(objBase).(State))
        (es : list {t : T & sigT (Program (impl t).(ObjInstruction))})
        (t : T)
        (s : (ThreadsBehavior (impl t)).(State))
        (ts1' : list T)
        (int : Interaction (Method t) A)
        (pint : Interaction (ProdMethod Method) A)
        (a : A)
        (st1 : (impl t).(objBase).(State)) :
    ProdInteraction t (map (@projT1 _ _) es) ts1' int pint ->
    ThreadsInteraction (impl t) A int (splitSigT (Tdec t) es, st1) a s ->
    exists es',
      ThreadsInteraction
        (prodObject impl) A pint
        (map prodInject es, overrideDep (Tdec t) st1 st2) a
        (map prodInject es', overrideDep (Tdec t) (snd s) st2) /\
      ts1' = map (@projT1 _ _) es' /\
      fst s = splitSigT (Tdec t) es' /\
      (forall t', t <> t' -> splitSigT (Tdec t') es = splitSigT (Tdec t') es').
  Proof.
    intro pi.
    remember (map (@projT1 _ _) es) as ts.
    revert st2 es s a st1 Heqts.
    induction pi; intros; subst.
    - inversions H.
      inversions H4; injpair.
      eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
      split; [| split; [| split]].
      + split.
        apply (TISpawn (prodObject impl) _ (existT _ _ _)).
      + reflexivity.
      + simpl.
        destruct Tdec.
        * rewrite <- eq_rect_eq.
          reflexivity.
        * congruence.
      + intros.
        simpl.
        destruct Tdec.
        * congruence.
        * reflexivity.
    - inversions H.
      destruct es; inversions Heqts.
      simpl in *.
      destruct s.
      simpl in *.
      destruct Tdec; [| congruence].
      rewrite <- eq_rect_eq in H4.
      inversions H4. injpair.
      eexists.
      split; [| split; [| split]].
      + split.
        rewrite (frobProgramId (prodInjectProgram _ _)).
        constructor.
      + reflexivity.
      + reflexivity.
      + intros.
        destruct Tdec.
        * congruence.
        * reflexivity.
    - inversions H. destruct es; inversions Heqts.
      destruct s.
      simpl in *.
      destruct Tdec; [| congruence].
      inversions H4; injpair.
      edestruct (IHpi st2) as (? & ? & ? & ? & ?).
      + reflexivity.
      + split.
        eassumption.
      + simpl in *.
        subst.
        inversions H.
        eexists (existT (fun _ => _) _ _ :: _).
        simpl.
        split; [| split; [| split]].
        * split.
          constructor.
          eassumption.
        * reflexivity.
        * destruct Tdec.
          -- rewrite <- !eq_rect_eq.
             reflexivity.
          -- congruence.
        * intros.
          destruct Tdec.
          -- congruence.
          -- eauto.
      Unshelve. assumption.
    - destruct es; [discriminate |].
      destruct s0.
      simpl in *.
      destruct Tdec; [congruence |].
      inversions Heqts.
      edestruct IHpi as (? & ? & ? & ? & ?).
      + reflexivity.
      + eassumption.
      + destruct s.
        destruct s0.
        simpl in *.
        subst.
        inversions H1.
        eexists (existT (fun _ => _) _ (existT _ _ _) :: _).
        simpl.
        split; [| split; [| split]].
        * rewrite H8.
          split.
          constructor.
          eassumption.
        * reflexivity.
        * destruct Tdec.
          -- congruence.
          -- reflexivity.
        * intros.
          destruct Tdec.
          -- f_equal.
             intuition auto.
          -- eauto.
  Qed.

  Theorem prodBehaviorObject (st : forall t, State (objBase (impl t))) :
    Simulates_
      (ProdBehavior (fun t => ThreadsBehavior (impl t)))
      (ThreadsBehavior (prodObject impl))
      (Build_ProdState (fun t => ThreadsBehavior (impl t)) [] (fun t => (nil, st t)))
      (nil, st).
  Proof.
    apply relationSimulates with PBIRel.
    - simpl.
      intros A pint [ts1 s1] a [ts1' s1'] [ps2 st2] (es & Hes1 & Hes2 & Hes3) (t & int & I & [[ts1'' s1''] ta ti] & E').
      apply functional_extensionality_dep in Hes3.
      simpl in *.
      subst.
      edestruct prodThreadsAdvance_r as (es' & Hes'1 & Hes'2 & Hes'3 & Hes'4).
      + exact ta.
      + simpl in *.
        subst.
        rewrite Hes'2 in I.
        edestruct prodThreadsInteraction_r as (es'' & Hes''1 & Hes''2 & Hes''3 & Hes''4).
        * eassumption.
        * eassumption.
        * simpl in *.
          subst.
          eexists.
          -- eexists.
             ++ rewrite <- overrideDepNop in Hes'1.
                eassumption.
             ++ eassumption.
          -- eexists.
             repeat split.
             intro t'.
             unfold overrideDep.
             simpl.
             destruct Tdec.
             ++ subst.
                destruct s1'.
                simpl in *.
                subst.
                reflexivity.
             ++ rewrite <- Hes''4, <- Hes'4, E'; trivial.
    - eexists nil; repeat constructor.
  Qed.

  Theorem prodObjectBehavior_ st :
    Simulates_
      (ThreadsBehavior (prodObject impl))
      (ProdBehavior (fun t => ThreadsBehavior (impl t)))
      (nil, st)
      (Build_ProdState (fun t => ThreadsBehavior (impl t)) [] (fun t => (nil, st t))).
  Proof.
    eapply mapStateSimulates with (fun s1 s2 => PBIRel s2 s1).
    - intros [? ?] ? [? ?] (? & ? & ? & ?) ?.
      simpl in *.
      subst.
      apply prodThreadAdvance in ta1.
      destruct ta1 as (? & ? & ? & ? & ? & ? & ?).
      subst.
      eexists {| prodState t := _ |}.
      + eapply prodBehaviorSimulates'_.
        * assumption.
        * intro.
          rewrite H1.
          destruct (Tdec x1 t).
          -- subst.
             apply simulateAdvance, clos_rt1n_step.
             eassumption.
          -- rewrite H3, H4.
             ++ apply simulatesRefl.
             ++ assumption.
             ++ assumption.
      + eexists.
        eauto.
    - clear st.
      intros ? [? ?] ? [? ?] [? ?] (? & ? & ? & ?) ? ?.
      simpl in *.
      subst.
      apply prodThreadsInteraction in ti.
      destruct ti as (? & ? & ? & ? & ? & ? & ? & ?).
      subst.
      eexists {| prodState t := _ |}.
      + eexists _, _.
        simpl in *.
        split; [| split].
        * eassumption.
        * eexists.
          -- econstructor 1.
          -- rewrite H1.
             eassumption.
        * intros.
          rewrite H1.
          f_equal; intuition auto.
      + repeat esplit.
        congruence.
    - eexists nil; repeat constructor.
  Qed.

  Theorem prodObjectBehavior :
    Simulates
      (ThreadsBehavior (prodObject impl))
      (ProdBehavior (fun t => ThreadsBehavior (impl t))).
  Proof.
    apply prodObjectBehavior_.
  Qed.

End ProdObjectBehavior.


Section ProdSimulates.

  Context
    T
    (Tdec : forall t t' : T, {t = t'} + {t <> t'})
    (Method : T -> Type -> Type)
    (impl1 impl2 : forall T, Object (Method T)).

  Theorem prodSimulates :
    (forall t, ConcSimulates (impl1 t) (impl2 t)) ->
    ConcSimulates (prodObject impl1) (prodObject impl2).
  Proof.
    intro S.
    eapply simulatesTrans, simulatesTrans.
    - apply prodObjectBehavior.
      assumption.
    - apply prodBehaviorSimulates.
      + assumption.
      + intro.
        apply S.
    - apply prodBehaviorObject.
      assumption.
  Qed.

End ProdSimulates.
