Require Import
        Coq.Lists.List
        Coq.Program.Equality
        Coq.Relations.Relations.

From C4.Core Require Import
  MapState
  Program
  ProgramOps
  Threads
  ThreadsFacts
  Object.

Section ProgramEqSimulates.

  Context
    {I M}
    (obj : Spec I)
    (impl1 impl2 : forall R, M R -> Program I R)
    (ImplEq : forall R m, programEq (impl1 R m) (impl2 R m)).

  Inductive ProgramsEq :
    list (sigT (Program I)) -> list (sigT (Program I)) -> Prop :=
  | programsEqNil : ProgramsEq nil nil
  | programsEqCons R p1 p2 ps1 ps2 :
      programEq p1 p2 ->
      ProgramsEq ps1 ps2 ->
      ProgramsEq (existT _ R p1 :: ps1) (existT _ R p2 :: ps2).

  Definition ThreadStatesEq
             (ts1 : list (sigT (Program I)) * obj.(State))
             (ts2 : list (sigT (Program I)) * obj.(State)) :
    Prop :=
    ProgramsEq (fst ts1) (fst ts2) /\ snd ts1 = snd ts2.

  Lemma programEqThreadAdvance ts1 ts2 ts1' :
    ThreadStatesEq ts1 ts2 ->
    ThreadAdvance {| objBase:=obj; objProgram:=impl1 |} ts1 ts1' ->
    exists2 ts2',
      ThreadAdvance {| objBase:=obj; objProgram:=impl2 |} ts2 ts2' &
      ThreadStatesEq ts1' ts2'.
  Proof.
    intros E ta.
    revert ts2 E.
    induction ta; intros [ps2 st2] [E E']; simpl in *; destruct E'; dependent destruction E.
    - edestruct IHta with (ps2, st).
      + split; tauto.
      + destruct x.
        eexists.
        * constructor 1.
          eassumption.
        * split; [constructor |]; firstorder.
    - dependent destruction H0.
      eexists.
      + constructor 2.
        eassumption.
      + split; [constructor |]; firstorder.
    - dependent destruction H.
      repeat econstructor; eassumption.
  Qed.

  Lemma programEqThreadsInteraction' Ret int ps1 ps2 a ps1' :
    ProgramsEq ps1 ps2 ->
    ThreadsInteraction' {| objBase:=obj; objProgram:=impl1 |} Ret int ps1 a ps1' ->
    exists2 ps2',
      ThreadsInteraction' {| objBase:=obj; objProgram:=impl2 |} Ret int ps2 a ps2' &
      ProgramsEq ps1' ps2'.
  Proof.
    intros E ti'.
    revert ps2 E.
    induction ti'; intros ps2 E.
    - repeat econstructor; auto.
    - dependent destruction E.
      dependent destruction H.
      repeat econstructor; assumption.
    - dependent destruction E.
      edestruct IHti'; repeat econstructor; eassumption.
  Qed.

  Theorem programEqSimulates :
    ConcSimulates
      {| objBase:=obj; objProgram:=impl1 |}
      {| objBase:=obj; objProgram:=impl2 |}.
  Proof.
    apply mapStateSimulates with (MapState:=ThreadStatesEq).
    - intros.
      edestruct programEqThreadAdvance.
      + eassumption.
      + eassumption.
      + eexists.
        * apply simulateAdvance, clos_rt1n_step.
          eassumption.
        * assumption.
    - intros.
      edestruct ms, ts2, ti, programEqThreadsInteraction'.
      + eassumption.
      + eassumption.
      + repeat econstructor; eassumption.
    - split.
      + constructor.
      + reflexivity.
  Qed.

End ProgramEqSimulates.
