Set Asymmetric Patterns.
Require Import
        Coq.Relations.Relations
        Coq.Logic.Eqdep.
Import EqdepTheory.

From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts.

Section Program.

  Lemma invertBindType:
    forall M R
      R1 (m: M R1) (f: R1 -> Program M R)
      R2 (m': M R2) (f': R2 -> Program M R),
      Bind m f = Bind m' f' ->
      R1 = R2
      /\ forall e: (R1 = R2),
          match e in _ = R2 return M R2 -> (R2 -> Program M R) -> Prop with
          | eq_refl =>
            fun m' f' =>
              m = m'
              /\ f = f'
          end m' f'.

  Proof.
    introv H1.
    inverts H1.
    split. reflexivity.
    intros.
    forwards A: UIP_refl R2 e.
    subst. split; reflexivity.
  Qed.

  Lemma invertBind:
    forall M R (m: M R) R' (f: R -> Program M R') (m': M R) f',
      (x <- m; f x) = (x <- m'; f' x) ->
      m = m'
      /\ f = f'.

  Proof.
    introv H1.
    eapply invertBindType in H1.
    destruct H1 as (H11 & H12).
    specialize (H12 eq_refl). simpl in H12. destruct H12 as (H121 & H122).
    splits.
    - assumption.
    - assumption.
  Qed.


  Lemma invertBind':
    forall M R (m: M R) R' (f: Program M R') (m': M R) f',
      (x <- m; f) = (x <- m'; f' x) ->
      m = m'.
  (* /\ f = f'. *)

  Proof.
    introv H1.
    inverts H1.
    reflexivity.
  Qed.
  
End Program.

Section ProgramStep.

  Context Instruction (obj : Spec Instruction).

  Lemma invertProgramStep R
        (p: Program Instruction R) (s: obj.(State))
        (p': Program Instruction R) (s': obj.(State)):
    ProgramStep obj R p s p' s' ->
    match p with
    | Bind _ i f =>
      exists r,
      obj.(EvalInstruction) i s r s'
      /\ p' = f r
    | Return r =>
      False
    | Noop p =>
      p' = p
      /\ s = s'
    end.

  Proof.
    introv H1.
    destruct H1; jauto.
  Qed.


  Lemma ProgramStepsRefl:
    forall R p s,
      ProgramSteps obj R p s p s.
  Proof.
    intros.
    unfold ProgramSteps.
    unfold ProgramSteps'.
    econstructor.
  Qed.

  Lemma ProgramStepImplProgramSteps:
    forall R p1 s1 p2 s2,
      ProgramStep obj R p1 s1 p2 s2 ->
      ProgramSteps obj R p1 s1 p2 s2.
  Proof.
    introv H1.
    unfold ProgramSteps.
    unfold ProgramSteps'.
    eapply clos_rt1n_step.
    unfold ProgramStep'.
    simpl.
    assumption.      
  Qed.

  Lemma ProgramStepStepsImplSteps:
    forall R p1 s1 p2 s2 p3 s3,
      ProgramStep obj R p1 s1 p2 s2 ->
      ProgramSteps obj R p2 s2 p3 s3 ->
      ProgramSteps obj R p1 s1 p3 s3.
  Proof.
    introv H1 H2.
    unfold ProgramSteps in *.
    unfold ProgramSteps' in *.
    unfold ProgramStep' in *.

    econstructor.
    - simpl. instantiate (1 := (p2, s2)). simpl. eassumption.
    - eassumption.    
  Qed.

  Lemma ProgramStepsStepImplSteps:
    forall R p1 s1 p2 s2 p3 s3,
      ProgramSteps obj R p1 s1 p2 s2 ->
      ProgramStep obj R p2 s2 p3 s3 ->
      ProgramSteps obj R p1 s1 p3 s3.
  Proof.
    introv H1 H2.
    unfold ProgramSteps in *.
    unfold ProgramSteps' in *.
    unfold ProgramStep' in *.
    rewrite <- clos_rt_rt1n_iff in *.
    rewrite clos_rt_rtn1_iff in *.
    
    econstructor.
    - simpl. instantiate (1 := (p2, s2)). simpl. eassumption.
    - eassumption.
  Qed.

  Lemma ProgramStepsStepsImplSteps:
    forall R p1 s1 p2 s2 p3 s3,
      ProgramSteps obj R p1 s1 p2 s2 ->
      ProgramSteps obj R p2 s2 p3 s3 ->
      ProgramSteps obj R p1 s1 p3 s3.
  Proof.
    introv H1 H2.
    unfold ProgramSteps in *.
    unfold ProgramSteps' in *.
    rewrite <- clos_rt_rt1n_iff in *.
    eapply rt_trans; eassumption.
  Qed.

  
  Lemma ProgramStepsInv:
    forall R p s p' s',
      ProgramSteps obj R p s p' s' ->
      (p = p' /\ s = s')
      \/ match p with
        | Bind A m f =>
          exists s'' r,
          EvalInstruction obj m s r s''
          /\ ProgramSteps obj R (f r) s'' p' s'
        | Noop p =>
          ProgramSteps obj R p s p' s'
        | Return r =>
          p' = Return r
          /\ s = s'
        end.
  
  Proof.
    introv H1.
    unfold ProgramSteps in H1.
    unfold ProgramSteps' in H1.
    unfold ProgramStep' in H1.
    inverts H1.
    - left. auto.
    - right. destruct y. simpl in *. 
      destruct p.
      + inverts H. eauto.
      + inverts H.
      + inverts H. eauto.        
  Qed.        


  Lemma EvalProgramInv:
    forall R p s r s',
      EvalProgram obj R p s r s' ->
      match p with
      | Bind A m f =>
        exists s'' r',
        EvalInstruction obj m s r' s''
        /\ EvalProgram obj R (f r') s'' r s'
      | Noop p =>
        EvalProgram obj R p s r s'
      | Return r' =>
        r' = r
        /\ s = s'
      end.
  
  Proof.
    introv H1.
    unfold EvalProgram in H1.    
    apply ProgramStepsInv in H1.
    destruct H1 as [(H1 & H2) | H1].
    - subst. auto.
    - destruct p.
      + auto.
      + destruct H1 as (H1 & H2). inverts H1. auto.
      + auto.    
  Qed.        

  
  Lemma ProgramStepsInd:
    forall (P: forall R, Program Instruction R -> obj.(State) -> 
               Program Instruction R -> obj.(State) -> Prop),
      
      (forall R p s,
          P R p s p s) ->
      (forall R R' m (f: R -> Program Instruction R') s r s' p'' s'',
          obj.(EvalInstruction) m s r s' ->
          P R' (f r) s' p'' s'' ->
          P R' (Bind m f) s p'' s'') ->
      (forall R p s p' s',
          P R p s p' s' ->
          P R (Noop p) s p' s') ->      
      forall R p s p' s',
        ProgramSteps obj R p s p' s' -> P R p s p' s'.

  Proof.
    introv H1 H2 H3 H4.
    unfold ProgramSteps in *.
    unfold ProgramSteps' in *.
    unfold ProgramStep' in *.
    
    remember (fun s1 s2 : Program Instruction R * State obj =>
                ProgramStep obj R (fst s1) (snd s1) (fst s2) (snd s2)) as Rel.
    remember (p, s) as st1.
    remember (p', s') as st2.
    
    gen p s p' s'.
    
    induction H4; intros; subst.
    - inverts Heqst2. eauto.
    - destruct y. simpl in *.
      inverts H.
      + eapply H2. eassumption.
        eapply IHclos_refl_trans_1n. reflexivity. reflexivity.
      + eapply H3.
        eapply IHclos_refl_trans_1n. reflexivity. reflexivity.    
  Qed.

  
  Lemma EvalProgramInd:
    forall (P: forall R, Program Instruction R -> obj.(State) -> 
               R -> obj.(State) -> Prop),
      
      (forall R r s,
          P R (Return r) s r s) ->
      (forall R R' m (f: R -> Program Instruction R') s r s' p'' s'',
          obj.(EvalInstruction) m s r s' ->
          P R' (f r) s' p'' s'' ->
          P R' (Bind m f) s p'' s'') ->
      (forall R p s p' s',
          P R p s p' s' ->
          P R (Noop p) s p' s') ->      
      forall R p s r s',
        EvalProgram obj R p s r s' -> P R p s r s'.

  Proof.
    intros.    
    unfold EvalProgram in *.
    remember (Return r) as p'.

    (*
    gen r. generalize R p s p' s'.
    gen R p s p' s'.
    eapply ProgramStepsInd.    
    *)

    (* Set Printing Implicit. *)
    gen r.
    eapply ProgramStepsInd with (P := (fun R p s p' s' => forall r : R, p' = @Return Instruction R r -> P R p s r s')); intros; subst; eauto.
    
  Qed.
  
End ProgramStep.

Lemma programStepSimulates Method (obj1 obj2 : Spec Method)
      A (p p' : Program Method A) (s1 s1' : State obj1) (s2 : State obj2) :
  Simulates_ obj1 obj2 s1 s2 ->
  ProgramStep obj1 A p s1 p' s1' ->
  exists2 s2',
    ProgramStep obj2 A p s2 p' s2' &
    Simulates_ obj1 obj2 s1' s2'.
Proof.
  intros S P.
  destruct P.
  - destruct S as [S'].
    edestruct S'.
    + eassumption.
    + eexists.
      * constructor.
        eassumption.
      * assumption.
  - eexists.
    + constructor.
    + eassumption.
Qed.

Lemma programStepsSimulates Method (obj1 obj2 : Spec Method)
      A (p p' : Program Method A) (s1 s1' : State obj1) (s2 : State obj2) :
  Simulates_ obj1 obj2 s1 s2 ->
  ProgramSteps obj1 A p s1 p' s1' ->
  exists2 s2' : State obj2,
                ProgramSteps obj2 A p s2 p' s2' &
                Simulates_ obj1 obj2 s1' s2'.
Proof.
  unfold ProgramSteps.
  remember (p, s1) as q.
  remember (p', s1') as q'.
  intros S P.
  gen p s1 s2 Heqq S.
  induction P; intros p s1 s2 Heqq S; subst.
  - inversion Heqq; subst.
    eexists.
    + constructor.
    + assumption.
  - edestruct programStepSimulates.
    + eassumption.
    + eassumption.
    + destruct y.
      edestruct IHP.
      * reflexivity.
      * reflexivity.
      * eassumption.
      * eexists.
        -- eapply ProgramStepStepsImplSteps.
           ++ eassumption.
           ++ eassumption.
        -- eassumption.
Qed.

Lemma mapSpecSimulates_
        Method Method' (obj1 obj2 : Spec Method)
        (f : forall Ret, Method' Ret -> Method Ret) st1 st2 :
  Simulates_ obj1 obj2 st1 st2 ->
  Simulates_ (mapSpec obj1 f) (mapSpec obj2 f) st1 st2.
Proof.
  apply relationSimulates.
  simpl.
  intros.
  destruct H.
  eauto.
Qed.

Theorem mapSpecSimulates
        Method Method' (obj1 obj2 : Spec Method)
        (f : forall Ret, Method' Ret -> Method Ret) :
  Simulates obj1 obj2 -> Simulates (mapSpec obj1 f) (mapSpec obj2 f).
Proof.
  apply mapSpecSimulates_.
Qed.
