Set Asymmetric Patterns.

From C4.Core Require Import
  Program
  Pair.

CoFixpoint bindProgram {Instruction A B} (p : Program Instruction A) (f : A -> Program Instruction B) : Program Instruction B :=
  match p with
    | Bind _ instr f' => a <- instr; bindProgram (f' a) f
    | Return a => f a
    | Noop p' => Noop (bindProgram p' f)
  end.
Notation "x <-- f ; m" := (bindProgram f (fun x => m)) (at level 80, right associativity).

CoFixpoint mapProgram
           {Instruction Instruction'}
           (f : forall A, Instruction A -> Instruction' A)
           {Ret}
           (p : Program Instruction Ret) :
  Program Instruction' Ret :=
  match p with
    | Bind A instr f' => a <- f A instr; mapProgram f (f' a)
    | Return a => Return a
    | Noop p' => Noop (mapProgram f p')
  end.

CoFixpoint mapProgramDep
           {Instruction Instruction'}
           {B : forall A, Instruction A -> Type}
           (f : forall A (instr : Instruction A),
               Instruction' (B A instr) *
               (B A instr -> A))
           {Ret}
           (p : Program Instruction Ret) :
  Program Instruction' Ret :=
  match p with
  | Bind A instr' f' =>
    let (instr, g) := f A instr' in
    b <- instr;
      mapProgramDep f (f' (g b))
  | Return a => Return a
  | Noop p' => Noop (mapProgramDep f p')
  end.

(*
Definition mapProgram
           {Instruction Instruction'}
           (f : forall A, Instruction A -> Instruction' A)
           {Ret} :
  Program Instruction Ret -> Program Instruction' Ret :=
  mapProgramDep (fun A instr => (f A instr, fun b => b)).
*)

CoFixpoint foldProgram
           {Instruction Instruction' Accumulator}
           (f : forall A,
               Instruction' A ->
               Accumulator ->
               Instruction (A * Accumulator)%type)
           {Ret}
           (p : Program Instruction' Ret)
           (acc : Accumulator) :
  Program Instruction (Ret * Accumulator)%type :=
  match p with
  | Bind A instr f' =>
    b <- f A instr acc;
      foldProgram f (f' (fst b)) (snd b)
  | Return a => Return (a, acc)
  | Noop p' => Noop (foldProgram f p' acc)
  end.

CoFixpoint optionProgram {Instruction Instruction'}
           (f : forall A,
               Instruction' A ->
               Instruction (option A))
           {Ret}
           (p : Program Instruction' Ret) :
  Program Instruction (option Ret) :=
  match p with
  | Bind A instr f' =>
    oa <- f A instr;
      match oa with
      | Some a => optionProgram f (f' a)
      | None => Return None
      end
  | Return a => Return (Some a)
  | Noop p' => Noop (optionProgram f p')
  end.

(*
Definition flattenProgram_Step
           (flattenProgram:
              forall {Method Instruction}
                (impl : forall A, Method A -> Program Instruction A)
                {Ret}
                (p : Program Method Ret),
                Program Instruction Ret)
           (bindFlattenProgram:
              forall {M M'}
                (impl : forall A, M' A -> Program M A)
                {R R'} (f: R -> Program M' R') (p: Program M R), Program M R')
           {Method Instruction}
           (impl : forall A, Method A -> Program Instruction A)
           {Ret}
           (p : Program Method Ret) :
  Program Instruction Ret :=
  match p with
    | Bind A m f =>
      Noop (bindFlattenProgram impl f (impl A m))
    | Return a => Return a
    | Noop p' => Noop (flattenProgram impl p')
  end.

Definition bindFlattenProgram
           {M M'}
           (impl : forall A, M' A -> Program M A)
           {R R'} (f: R -> Program M' R') (p: Program M R) :=
  match p with
  | Bind _ m' f' => r <- m'; bindFlattenProgram impl f (f' r)
  | Return a => Noop (flattenProgram impl (f a))
  | Noop p'' => Noop (bindFlattenProgram impl f p'')
  end.


CoFixpoint flattenProgram
           {Method Instruction}
           (impl : forall A, Method A -> Program Instruction A)
           {Ret}
           (p : Program Method Ret) :
  Program Instruction Ret :=
  
  @flattenProgram_Step (@flattenProgram) (@bindFlattenProgram)
                       Method Instruction impl Ret p.
*)
           
CoFixpoint flattenProgram
           {Method Instruction}
           (impl : forall A, Method A -> Program Instruction A)
           {Ret}
           (p : Program Method Ret) :
  Program Instruction Ret :=
  match p with
    | Bind A m f =>
      (* bindProgram (impl A m) (fun a => Noop (flattenProgram impl (f a))) *)
      (*
      (cofix bindFlattenProgram p' :=
         match p' with
           | Bind _ instr f' => a <- instr; bindFlattenProgram (f' a)
           | Return a => Noop (flattenProgram impl (f a))
           | Noop p'' => Noop (bindFlattenProgram p'')
         end)
        (impl A m)
      *)
      Noop (bindFlattenProgram impl f (impl A m))
    | Return a => Return a
    | Noop p' => Noop (flattenProgram impl p')
  end

with bindFlattenProgram
           {M M'} (impl : forall A, M' A -> Program M A)
           {R R'} (f: R -> Program M' R') (p: Program M R) :=
  match p with
  | Bind _ m' f' => r <- m'; bindFlattenProgram impl f (f' r)
  | Return a => Noop (flattenProgram impl (f a))
  | Noop p'' => Noop (bindFlattenProgram impl f p'')
  end.


CoInductive correlation {R M M'}:
    (forall R', M R' -> M' R' -> Prop) ->
    Program M R -> Program M' R -> Prop :=
| BindCoRel R' (rel: forall R', M R' -> M' R' -> Prop)
            (m: M R') (f: R' -> Program M R)
            (m': M' R') (f': R' -> Program M' R):
    rel R' m m' ->
    (forall (r: R'), correlation rel (f r) (f' r)) ->
    correlation rel (Bind m f) (Bind m' f')
| ReturnCoRel rel r:
    correlation rel (Return r) (Return r)
| NoopCoRe rel p1 p2:
    correlation rel p1 p2 ->
    correlation rel (Noop p1) (Noop p2).    
           
           
Definition frobProgram {Instruction Ret} (p : Program Instruction Ret) :
  Program Instruction Ret :=
  match p with
    | Bind _ instr f => Bind instr f
    | Return a => Return a
    | Noop p' => Noop p'
  end.

Theorem frobProgramId {Instruction Ret} (p : Program Instruction Ret) :
  p = frobProgram p.
Proof.
  destruct p; reflexivity.
Defined.

Ltac progFrobSimpl :=
  setoid_rewrite frobProgramId;
  simpl.

Ltac progFrobSimplAt i :=
  match i with
  | 0 => setoid_rewrite frobProgramId at 0
  | 1 => setoid_rewrite frobProgramId at 1
  | 2 => setoid_rewrite frobProgramId at 2
  | 3 => setoid_rewrite frobProgramId at 3
  | 4 => setoid_rewrite frobProgramId at 4
  | 5 => setoid_rewrite frobProgramId at 5
  | 6 => setoid_rewrite frobProgramId at 6
  | 7 => setoid_rewrite frobProgramId at 7
  | 8 => setoid_rewrite frobProgramId at 8
  | 9 => setoid_rewrite frobProgramId at 9
  | _ => fail
  end;
  simpl.

Ltac progFrobSimplIn H :=
  setoid_rewrite frobProgramId in H;
  simpl in H.

Ltac progFrobSimplInAt H i :=
  match i with
  | 0 => setoid_rewrite frobProgramId at 0 in H
  | 1 => setoid_rewrite frobProgramId at 1 in H
  | 2 => setoid_rewrite frobProgramId at 2 in H
  | 3 => setoid_rewrite frobProgramId at 3 in H
  | 4 => setoid_rewrite frobProgramId at 4 in H
  | 5 => setoid_rewrite frobProgramId at 5 in H
  | 6 => setoid_rewrite frobProgramId at 6 in H
  | 7 => setoid_rewrite frobProgramId at 7 in H
  | 8 => setoid_rewrite frobProgramId at 8 in H
  | 9 => setoid_rewrite frobProgramId at 9 in H
  | _ => fail
  end;
  simpl in H.

CoInductive programEq {I} {R}: Program I R -> Program I R -> Prop :=
| BindEq R' (i: I R') (f1: R' -> Program I R) (f2: R' -> Program I R):
    (forall v, programEq (f1 v) (f2 v)) ->
    programEq (Bind i f1) (Bind i f2)
| ReturnEq r:
    programEq (Return r) (Return r)
| NoopEq p1 p2:
    programEq p1 p2 ->
    programEq (Noop p1) (Noop p2).


Definition serialImplProg M:
  forall R, M R -> Program M R :=
  fun R m => Bind m Return.

Definition pairImplProg
           {M1 I1 M2 I2}
           (implL : forall A, M1 A -> Program I1 A)
           (implR : forall A, M2 A -> Program I2 A):
  forall A, (PairMethod M1 M2) A -> Program (PairMethod I1 I2) A :=
  fun Ret m =>
    match m with
    | inl mL => mapProgram (fun _ => inl) (implL Ret mL)
    | inr mR => mapProgram (fun _ => inr) (implR Ret mR)
    end.  


Definition flattenProgramLeft
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {R} (p: Program (PairMethod M1 M2) R):
  Program (PairMethod I1 M2) R :=

  flattenProgram
    (pairImplProg
       impl
       (serialImplProg M2))
    p.


Definition bindFlattenProgramLeft
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {R R'} (p: Program (PairMethod I1 M2) R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod I1 M2) R' :=

  bindFlattenProgram
    (pairImplProg
       impl
       (serialImplProg M2))
    f p.


Definition flattenProgramRight
           {M1 M2 I2} (impl : forall R, M2 R -> Program I2 R)
           {R} (p : Program (PairMethod M1 M2) R):
  Program (PairMethod M1 I2) R :=

  flattenProgram
    (pairImplProg
       (serialImplProg M1)
       impl)
    p.
  

Definition bindFlattenProgramRight
           {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
           {R R'} (p: Program (PairMethod M1 I2) R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod M1 I2) R' :=

  bindFlattenProgram
    (pairImplProg
       (serialImplProg M1)       
       impl)
    f p.
  

Definition flattenImplProgLeft
           {ML IL} (base: forall R, ML R -> Program IL R)
           {M' MR} (impl : forall R, M' R -> Program (PairMethod ML MR) R):
  forall R, M' R -> Program (PairMethod IL MR) R :=
  fun R m =>
    flattenProgramLeft base (impl _ m).

Definition flattenImplProgRight
           {MR IR} (base: forall R, MR R -> Program IR R)
           {M' ML} (impl: forall R, M' R -> Program (PairMethod ML MR) R):
  forall R, M' R -> Program (PairMethod ML IR) R :=
  fun R m =>
    flattenProgramRight base (impl _ m).



CoFixpoint flattenProgramLeft'
           {M1 M2 I1} (impl : forall R, M1 R -> Program I1 R)
           {R} (p : Program (PairMethod M1 M2) R):
  Program (PairMethod I1 M2) R :=
  match p with
    | Bind R m f =>
      match m with
      | inl lm =>
        Noop (bindFlattenProgramLeft' impl (impl R lm) f)
      | inr rm =>
        x <- inr rm;
        flattenProgramLeft' impl (f x)
      end 
    | Return a => Return a
    | Noop p' => Noop (flattenProgramLeft' impl p')
  end

with bindFlattenProgramLeft'
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {R R'} (p: Program I1 R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod I1 M2) R' :=
  match p with
  | Bind _ m' f' => r <- inl m'; bindFlattenProgramLeft' impl (f' r) f
  | Return r => Noop (flattenProgramLeft' impl (f r))
  | Noop p'' => Noop (bindFlattenProgramLeft' impl p'' f)
  end.


CoFixpoint flattenProgramRight'
           {M1 M2 I2} (impl : forall R, M2 R -> Program I2 R)
           {R} (p : Program (PairMethod M1 M2) R):
  Program (PairMethod M1 I2) R :=
  match p with
    | Bind R m f =>
      match m with
      | inl lm =>
        x <- inl lm;
        flattenProgramRight' impl (f x)
      | inr rm =>
        Noop (bindFlattenProgramRight' impl (impl R rm) f)
      end 
    | Return a => Return a
    | Noop p' => Noop (flattenProgramRight' impl p')
  end

with bindFlattenProgramRight'
           {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
           {R R'} (p: Program I2 R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod M1 I2) R' :=
  match p with
  | Bind _ m' f' => r <- inr m'; bindFlattenProgramRight' impl (f' r) f
  | Return r => Noop (flattenProgramRight' impl (f r))
  | Noop p'' => Noop (bindFlattenProgramRight' impl p'' f)
  end.
  

Definition flattenImplProgLeft'
           {ML IL} (base: forall R, ML R -> Program IL R)
           {M' MR} (impl : forall R, M' R -> Program (PairMethod ML MR) R):
  forall R, M' R -> Program (PairMethod IL MR) R :=
  fun R m =>
    flattenProgramLeft' base (impl _ m).

Definition flattenImplProgRight'
           {MR IR} (base: forall R, MR R -> Program IR R)
           {M' ML} (impl: forall R, M' R -> Program (PairMethod ML MR) R):
  forall R, M' R -> Program (PairMethod ML IR) R :=
  fun R m =>
    flattenProgramRight' base (impl _ m).





(*  
CoFixpoint flattenProgramLeft
           {M1 M2 I1}
           (impl : forall R, M1 R -> Program I1 R)
           {R} (p : Program (PairMethod M1 M2) R):
  Program (PairMethod I1 M2) R :=
  match p with
    | Bind R m f =>
      (* bindProgram (impl A m) (fun a => Noop (flattenProgram impl (f a))) *)
      match m with
      | inl lm =>
        (cofix flattenProgramLeft' p' :=
         match p' with
           | Bind _ m f' => r <- inl m; flattenProgramLeft' (f' r)
           | Return a => Noop (flattenProgramLeft impl (f a))
           | Noop p'' => Noop (flattenProgramLeft' p'')
         end) (impl R lm)
      | inr rm =>
        x <- inr rm;
        flattenProgramLeft impl (f x)
      end 
    | Return a => Return a
    | Noop p' => Noop (flattenProgramLeft impl p')
  end.


CoFixpoint flattenProgramLeft'
           {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
           {R R'} (p: Program I1 R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod I1 M2) R' :=
  match p with
  | Bind _ m' f' => r <- inl m'; flattenProgramLeft' impl (f' r) f
  | Return r => Noop (flattenProgramLeft impl (f r))
  | Noop p'' => Noop (flattenProgramLeft' impl p'' f)
  end.

Lemma flattenProgramLeftEq:
  forall {M1 M2 I1} (impl: forall R, M1 R -> Program I1 R)
    {R R'} (p: Program I1 R) (f: R -> Program (PairMethod M1 M2) R'),

    programEq
      (flattenProgramLeft' impl p f)
      ((cofix flattenProgramLeft' p' :=
          match p' with
           | Bind _ m f' => r <- inl m; flattenProgramLeft' (f' r)
           | Return a => Noop (flattenProgramLeft impl (f a))
           | Noop p'' => Noop (flattenProgramLeft' p'')
          end) p).

Proof.
  intros.
  
  gen p.
  cofix Ind.
  intros.
  unfold flattenProgramLeft'.

  match goal with
  | [ |- programEq ?x ?y] =>
    rewriteFrobProgram x;
    rewriteFrobProgram y      
  end.

  destruct p; simpl.  
  - constructor. intros. eapply Ind.
  - reflexivity.
  - constructor. eapply Ind.
Qed.
 *)


(*
CoFixpoint flattenProgramRight
           {M1 M2 I2}
           (impl : forall R, M2 R -> Program I2 R)
           {R} (p : Program (PairMethod M1 M2) R):
  Program (PairMethod M1 I2) R :=
  match p with
    | Bind R m f =>
      (* bindProgram (impl A m) (fun a => Noop (flattenProgram impl (f a))) *)
      match m with
      | inl lm =>
        x <- inl lm;
        flattenProgramRight impl (f x)
      | inr rm =>
        (* flattenProgramRight' impl (impl R rm) f *)        
        (cofix flattenProgramRight' p' :=
         match p' with
           | Bind _ m f' => a <- inr m; flattenProgramRight' (f' a)
           | Return a => Noop (flattenProgramRight impl (f a))
           | Noop p'' => Noop (flattenProgramRight' p'')
         end) (impl R rm)
      end 
    | Return a => Return a
    | Noop p' => Noop (flattenProgramRight impl p')
  end.

CoFixpoint flattenProgramRight'
           {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
           {R R'} (p: Program I2 R) (f: R -> Program (PairMethod M1 M2) R'):
  Program (PairMethod M1 I2) R' :=
  match p with
  | Bind _ m' f' => r <- inr m'; flattenProgramRight' impl (f' r) f
  | Return r => Noop (flattenProgramRight impl (f r))
  | Noop p'' => Noop (flattenProgramRight' impl p'' f)
  end.


Lemma flattenProgramRightEq:
  forall {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
    {R R'} (p: Program I2 R) (f: R -> Program (PairMethod M1 M2) R'),
    programEq
      (flattenProgramRight' impl p f)
      ((cofix flattenProgramRight' p' :=
          match p' with
          | Bind _ instr f' => a <- inr instr; flattenProgramRight' (f' a)
          | Return a => Noop (flattenProgramRight impl (f a))
          | Noop p'' => Noop (flattenProgramRight' p'')
          end) p).

Proof.
  intros.
  
  gen p.
  cofix Ind.
  intros.
  unfold flattenProgramRight'.

  match goal with
  | [ |- programEq ?x ?y] =>
    rewriteFrobProgram x;
    rewriteFrobProgram y      
  end.

  destruct p; simpl.  
  - constructor. intros. eapply Ind.
  - reflexivity.
  - constructor. eapply Ind.
Qed.
*)

(*
Inductive OpStartProgram {I R}: Program I R -> Prop :=
| OpStartBind R' (i: I R') (f: R' -> Program I R): OpStartProgram (r <- i; f r)
| OpStartReturn r: OpStartProgram (Return r).
*)

CoInductive OpStartProgram {I R}: Program I R -> Prop :=
| OpStartBind R' (i: I R') (f: R' -> Program I R): OpStartProgram (r <- i; f r)
| OpStartReturn r: OpStartProgram (Return r).

Section ProgramNEq.

  Context {Instruction : Type -> Type}.

  
  CoInductive programNEq {I} {R}: Program I R -> Program I R -> Prop :=
  | BindNEq R' (i: I R') (f1: R' -> Program I R) (f2: R' -> Program I R):
      (forall v, programNEq (f1 v) (f2 v)) ->
      programNEq (Bind i f1) (Bind i f2)
  | ReturnNEq r:
      programNEq (Return r) (Return r)
  | NoopNEq1 p1 p2:
      programNEq p1 p2 ->
      programNEq (Noop p1) (Noop p2)
  | NoopNEq2 p1 p2:
      programNEq p1 p2 ->
      programNEq (Noop p1) p2.
  (*
  | NoopNEq' p1 p2:
      programNEq p1 p2 ->
      programNEq p1 (Noop p2).
  *)

  (*
  Lemma ProgramNEqNoopLeft:
    forall R (p1 p2: Program Instruction R),
      programNEq (Noop p1) p2 ->
      programNEq p1 p2.

  Proof.
    cofix CoInd.
    introv H1.
    inverts H1.
    - constructor. assumption.
    - assumption.
    - constructor. apply CoInd. assumption.
   Qed.

  Lemma ProgramNEqNoopRight:
    forall R (p1 p2: Program Instruction R),
      programNEq (Noop p1) p2 ->
      programNEq p1 p2.

  Proof.
    cofix CoInd.
    introv H1.
    inverts H1.
    - constructor. assumption.
    - assumption.
    - constructor. apply CoInd. assumption.
   Qed.
  
  Lemma ProgramNEqNoopSym:
    forall R (p1 p2: Program Instruction R),
      programNEq (Noop p1) p2 ->
      programNEq p1 (Noop p2).
  Proof.
    cofix CoInd.
    introv H1.
    inverts H1.
    - constructor. constructor. assumption.
    - constructor. assumption.
    - constructor. apply CoInd. assumption.
   Qed.
  *)
  
  Definition implNEq {M} (impl1 impl2: forall R, M R -> Program Instruction R) :=
    forall R' (m: M R'),
      programNEq (impl1 _ m) (impl2 _ m).

End ProgramNEq.  
