Set Asymmetric Patterns.
Require Import
        Coq.Lists.List
        Coq.Logic.FunctionalExtensionality
        Coq.Relations.Relations.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  Pair
  ProgramOps
  Simulates
  SimulatesFacts.

Require Import Setoid.
Require Import SetoidClass.
Require Import Morphisms.

Lemma mapProgramEqBind:
  forall I I' (f: forall R, I R -> I' R) R (p : Program I R) A (m : I' A) (g : A -> Program I' R),
    mapProgram f p = Bind m g ->
    exists m' g', p = Bind m' g' /\ m = f A m' /\ g = (fun a => mapProgram f (g' a)).
Proof.
  intros.
  destruct p; progFrobSimplIn H; inverts H.
  eexists _, _.
  splits; reflexivity.
Qed.

Lemma mapProgramEqRet:
  forall I I' (f: forall R, I R -> I' R) R (p : Program I R) r,
    mapProgram f p = Return r ->
    p = Return r.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  reflexivity.
Qed.

Lemma mapProgramEqNoop:
  forall I I' (f: forall R, I R -> I' R) R (p : Program I R) p',
    mapProgram f p = Noop p' ->
    exists p'', p = Noop p'' /\ p' = mapProgram f p''.

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  eauto.
Qed.

Lemma mapProgramInv I I' (f : forall R, I R -> I' R) R (p : Program I R) (p' : Program I' R) :
  mapProgram f p = p' ->
  match p' with
  | Return r =>
    p = Return r
  | Bind A m g =>
    exists m' g',
      p = Bind m' g' /\ m = f A m' /\ g = (fun a => mapProgram f (g' a))
  | Noop p' =>
    exists p'',
      p = Noop p'' /\ p' = mapProgram f p''
  end.
Proof.
  destruct p'.
  - apply mapProgramEqBind.
  - apply mapProgramEqRet.
  - apply mapProgramEqNoop.
Qed.


(*
Lemma bindFlattenProgramRightEqNoop:
  forall {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
    {R R'} (p: Program I2 R) (f: R -> Program (PairMethod M1 M2) R') p',
    
    bindFlattenProgramRight impl p f = Noop p' ->
    ((exists r, p = Return r /\ p' = flattenProgramRight impl (f r))
     \/ (exists p'', p = Noop p'' /\ p' = bindFlattenProgramRight impl p'' f)).

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  - left. eauto.
  - right. eauto.
Qed.
*)

Ltac rewriteFrobProgram x :=
  let r := fresh "r" in
  let H := fresh "H" in
  remember x as r eqn: H;
    revert H;    
    match goal with
    | [ |- r = ?x -> ?y ] =>
      refine (
          match eq_sym (frobProgramId x)
                in _ = X
                return r = X -> y
          with
          | eq_refl => _
          end)
    end;
    intro H; subst r.

Lemma eqProgramEq:
  forall I R p,
    @programEq I R p p.

Proof.
  intros I R.
  cofix self.
  intros.
  destruct p;
  constructor;
  auto.
Qed.

Lemma eqProgramEq':
  forall I R p1 p2,
    p1 = p2 ->
    @programEq I R p1 p2.
Proof.
  introv H1.
  rewrite H1.
  apply eqProgramEq.
Qed.

Lemma programEqSym:
  forall I R p1 p2,
    @programEq I R p1 p2 ->
    @programEq I R p2 p1.

Proof.
  intros I R.
  cofix self.
  introv H.
  destruct p1; destruct p2; inverts H.
  - constructor. intros. eauto.
  - constructor.
  - constructor. eauto.
Qed.

Lemma programEqTrans:
  forall I R p1 p2 p3,
    @programEq I R p1 p2 ->
    @programEq I R p2 p3 ->
    @programEq I R p1 p3.

Proof.
  intros I R.
  cofix self.
  introv H1 H2.
  destruct p1; destruct p2; inverts H1; destruct p3; inverts H2.
  - constructor. intros. eauto.
  - constructor.
  - constructor. eauto.
Qed.


Add Parametric Relation I R: (@Program I R) (@programEq I R)
    reflexivity proved by (eqProgramEq I R)
    symmetry proved by (programEqSym I R)
    transitivity proved by (programEqTrans I R)                        
      as programEq_rel.

Add Parametric Morphism I R A: (@Bind I R A)
    with signature eq ==> pointwise_relation _ eq ==> eq
      as bind_eq_eq.
Proof.
  unfold pointwise_relation.
  intros. f_equal. apply functional_extensionality. assumption.
Qed.

(*
Add Parametric Morphism I R A: (@Bind I R A)
    with signature eq ==> pointwise_relation _ programEq ==> eq
      as bind_programEq_eq.
Proof.
Qed.
 *)

Add Parametric Morphism I R A: (@Bind I R A)
    with signature eq ==> pointwise_relation _ programEq ==> programEq
      as bind_programEq_programEq.
Proof.
  unfold pointwise_relation.
  intros. constructor. assumption.
Qed.

(*
*)

Lemma mapProgramProgramEq:
  forall (I I' : Type -> Type) (f : forall A : Type, I A -> I' A) 
    (R : Type) (x y : Program I R),
    programEq x y -> programEq (mapProgram f x) (mapProgram f y).

Proof.
  cofix self.
  intros.
  match goal with
  | [ |- programEq ?p ?p'] =>
    rewriteFrobProgram p;
      rewriteFrobProgram p'
  end.
  destruct x; destruct y; simpl in *; inverts H.
  - constructor. eauto.
  - reflexivity.
  - constructor. eauto.
Qed.

Add Parametric Morphism I I' f R: (@mapProgram I I' f R)
    with signature programEq ==> programEq
      as mapProgram_programEq.
Proof.
  apply mapProgramProgramEq.
Qed.

CoFixpoint flattenProgramProgramEq
           Method Instruction
           (impl impl' : forall A, Method A -> Program Instruction A)
           Ret
           (p p' : Program Method Ret) :
  (forall A m, programEq (impl A m) (impl' A m)) ->
  programEq p p' ->
  programEq (flattenProgram impl p) (flattenProgram impl' p')
with bindFlattenProgramProgramEq
       Method Instruction
       (impl impl' : forall A, Method A -> Program Instruction A)
       Ret Ret'
       (f f' : Ret -> Program Method Ret')
       (p p' : Program Instruction Ret) :
  (forall A m, programEq (impl A m) (impl' A m)) ->
  (forall r, programEq (f r) (f' r)) ->
  programEq p p' ->
  programEq (bindFlattenProgram impl f p) (bindFlattenProgram impl' f' p').
Proof.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; inversion H0; injpair_; subst; constructor; auto.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; inversion H1; injpair_; subst; constructor; auto.
Qed.

Lemma ProgramStepsProgramEq:
  forall {Instruction} (o : Spec Instruction) (R : Type) (p1 p2 : Program Instruction R)
    (s : State o) (p1' : Program Instruction R) 
    (s' : State o),
    programEq p1 p2 ->
    ProgramSteps o R p1 s p1' s' ->
    exists p2',
      ProgramSteps o R p2 s p2' s'
      /\ programEq p1' p2'.

Proof.
  introv H1 H2.
  gen H1.
  induction H2 using @ProgramStepsInd; intros.
  + symmetry in H1.
    eexists. splits.
    eapply ProgramStepsRefl.
    symmetry. assumption.
  + inverts H1.
    forwards (p2' & A1 & A2): IHProgramSteps. eauto.
    eexists. splits.
    eapply ProgramStepStepsImplSteps. econstructor. eassumption. eassumption.
    assumption.
  + inverts H1.
    forwards (p2' & A1 & A2): IHProgramSteps. eauto.
    eexists. splits.
    eapply ProgramStepStepsImplSteps. econstructor. eassumption.
    assumption.
Qed.

Lemma EvalProgramReturn {M} (obj : Spec M) (R : Type) (s1 : State obj) (r : R) :
  EvalProgram obj R (Return r) s1 r s1.
Proof.
  red. apply ProgramStepsRefl.
Qed.

Lemma EvalProgramNoop {M} (obj : Spec M) (R : Type) (s1 s' : State obj) (p : Program M R) (r : R)
  : EvalProgram obj R p s1 r s' ->
    EvalProgram obj R (Noop p) s1 r s'.
Proof.
  intros H. eapply ProgramStepStepsImplSteps; [ constructor | eassumption ].
Qed.

Lemma EvalProgramBind {M} (obj : Spec M) (R S : Type) (e : M R) (f : R -> Program M S)
    (r : R) (s : S) (z z' z'' : State obj)
  : EvalInstruction obj e z r z' ->
    EvalProgram obj S (f r) z' s z'' ->
    EvalProgram obj S (Bind e f) z s z''.
Proof.
  intros H1 H2.
  eapply ProgramStepStepsImplSteps; [ | eassumption ].
  constructor. assumption.
Qed.


(*
Add Parametric Morphism o R : (@ProgramSteps o R)
    with signature programEq ==> eq ==> eq ==> eq ==> iff
      as ProgramSteps_eq.
Proof.
  introv H1.
  intros.
  split.
  apply ProgramStepsProgramEq.  
Qed.
*)

Lemma EvalProgramProgramEq:
  forall {Instruction} (o : Spec Instruction) (R : Type) (p1 p2 : Program Instruction R)
    (s : State o) (s' : State o) r,
    programEq p1 p2 ->
    EvalProgram o R p1 s r s' ->
    EvalProgram o R p2 s r s'.

Proof.
  intros.
  unfold EvalProgram in *.
  forwards (p2' & A1 & A2): @ProgramStepsProgramEq Instruction.
  eassumption.
  eassumption.
  inverts A2. assumption.
Qed.


Add Parametric Morphism Instruction o R : (@EvalProgram Instruction o R)
    with signature programEq ==> eq ==> eq ==> eq ==> iff
      as EvalProgram_eq.
Proof.
  intros.
  split.
  apply EvalProgramProgramEq; assumption.
  apply EvalProgramProgramEq. symmetry. assumption.
Qed.


Lemma bindProgramInv:
  forall M R R' (p': Program M R') (p: Program M R) (f: R -> Program M R'),
    programEq p' (bindProgram p f) ->
    match p' with
    | Bind R m' f' =>
      (exists pr, p = (x <- m'; pr x))
      \/ (exists r, p = Return r /\ programEq p' (f r))

    | Return r =>
      exists r',
      p = Return r'
      /\ programEq (f r') (Return r)
                      
    | Noop p'' =>
      (exists p''', 
          p = Noop p'''
          /\ programEq p'' (bindProgram p''' f))
      \/ (exists r',
            p = Return r'
            /\ programEq (f r') (Noop p''))
    end.

Proof.
  intros.
  setoid_rewrite frobProgramId in H at 2. simpl in H.
  destruct p'.
  - destruct p eqn: E.
    + left. inverts H. eexists. econstructor.
    + right. eexists. split. reflexivity. destruct (f r); assumption.
    + inverts H.
      
  - destruct p eqn: E. 
    + inverts H.      
    + eexists. split. reflexivity. destruct (f r0); symmetry; assumption.
    + inverts H.
      
  - destruct p eqn: E. 
    + inverts H.      
    + right. eexists. split. reflexivity. destruct (f r); symmetry; assumption.
    + inverts H. left. eexists. eauto.
Qed.

(*
Lemma flattenProgramNoop:
  forall M R (p: @Program M R) p',
    (flattenProgram (fun _ m => Bind m Return) p) = Noop p' ->
    exists p'', p = Noop p''
           /\ p' = flattenProgram (fun _ m => Bind m Return) p''.
Proof.
  introv H1.
  destruct p;
    rewrite frobProgramId in H1 at 1;
    simpl in H1.
  - inversion H1.
  - inversion H1.
  - inversion H1. eauto.      
Qed.

Lemma flattenProgramBind:
  forall M R (p: @Program M R) R' (m: M R') f,
    (flattenProgram (fun _ m => Bind m Return) p) = Bind m f ->
    exists f', p = Bind m f'
          /\ forall v, Noop (flattenProgram (fun _ m => Bind m Return) (f' v)) = (f v).

Proof.
  introv H1.
  destruct p;
    rewrite frobProgramId in H1 at 1;
    simpl in H1.
  - inverts H1.
    exists p. splits.
    reflexivity.
    intros. rewrite frobProgramId. simpl. reflexivity.
    
  - inversion H1.
  - inversion H1.
Qed.


Lemma flattenProgramRet:
  forall M R (p: Program M R) M' (impl: forall R, M R -> Program M' R) r,
    (flattenProgram impl p) = @Return M' R r ->
    p = Return r.

Proof.
  intros.
  destruct p;
    rewrite frobProgramId in H at 1;
    simpl in H.
  - destruct (impl A m).
    + inversion H.
    + inversion H.
    + inversion H.
  - inverts H. reflexivity.
  - inversion H.
Qed.
*)

(*
CoFixpoint bindFlattenProgram
           {M M'}
           (impl : forall A, M' A -> Program M A)
           {R R'} (f: R -> Program M' R') (p: Program M R) :=
  match p with
  | Bind _ m' f' => r <- m'; bindFlattenProgram impl f (f' r)
  | Return a => Noop (flattenProgram impl (f a))
  | Noop p'' => Noop (bindFlattenProgram impl f p'')
  end.



Lemma bindFlattenProgramEq:
  forall (M M' : Type -> Type) (R : Type)
    (impl : forall R : Type, M R -> Program M' R) 
    (R' : Type) (f : R' -> Program M R) p,
    programEq
      (bindFlattenProgram impl f p)
      ((cofix bindFlattenProgram (p' : Program M' R') : 
         Program M' R :=
         match p' with
         | Bind A1 instr f' => (a <- instr; bindFlattenProgram (f' a))
         | Return a => Noop (flattenProgram impl (f a))
         | Noop p'' => Noop (bindFlattenProgram p'')
         end) p).

Proof.
  intros.
  
  gen p.
  cofix Ind.
  intros.
  unfold bindFlattenProgram.

  match goal with
  | [ |- programEq ?x ?y] =>
    rewriteFrobProgram x;
    rewriteFrobProgram y      
  end.

  destruct p; simpl.  
  - constructor. intros. eapply Ind.
  - reflexivity.
  - constructor. eapply Ind.
Qed.
*)

Lemma bindFlattenProgramBindProgram:
  forall {M M'}
    (impl : forall A, M' A -> Program M A)
    {R R'} (f: R -> Program M' R') (p: Program M R),
    programEq (bindFlattenProgram impl f p)
              (bindProgram p (fun r : R => Noop (flattenProgram impl (f r)))).

Proof.
  cofix Ind.
  intros.
  unfold bindFlattenProgram.
  match goal with
  | [ |- programEq ?x ?y] =>
    rewriteFrobProgram x;
    rewriteFrobProgram y
  end.
  simpl.
  destruct p.
  - constructor. intros. apply Ind.
  - reflexivity.
  - constructor. apply Ind.
Qed.

(*
Lemma functional_extensionality_program {A M R} (f g : A -> Program M R):
  (forall x, programEq (f x) (g x)) -> f = g.
Proof.
Qed.
*)

Lemma flattenProgramInv:
  forall M M' R (pa: @Program M R) p
    (impl : forall R, M R -> Program M' R),
    p = flattenProgram impl pa ->
    match p with
    | Bind _ m f =>
      False

    | Return r =>
      pa = Return r
                  
    | Noop p' =>
      (exists p'',
          pa = Noop p''
          /\ p' = flattenProgram impl p'')
        
      \/ (exists R' ma fa,
          pa = Bind ma fa
          /\ p' = bindFlattenProgram impl fa (impl R' ma))
          
      \/ (exists R' ma fa r,
            pa = Bind ma fa
            /\ impl R' ma = Return r
            /\ p' = flattenProgram impl (fa r))
          
      \/ (exists {R'} (ma: M R') fa p'',
            pa = Bind ma fa
            /\ impl R' ma = Noop p''
            /\ programEq p' (bindFlattenProgram impl fa p''))
    end.

Proof.
  intros.
  subst.
  destruct pa eqn:P; simpl.
  - right. left.
    eexists. eexists. eexists. splits.
    reflexivity. reflexivity.
  - reflexivity.    
  - left. eauto.
Qed.

Lemma bindFlattenProgramRightEqNoop:
  forall {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
    {R R'} (p: Program (PairMethod M1 I2) R) (f: R -> Program (PairMethod M1 M2) R') p',
    
    bindFlattenProgramRight impl p f = Noop p' ->
    ((exists r, p = Return r /\ p' = flattenProgramRight impl (f r))
     \/ (exists p'', p = Noop p'' /\ p' = bindFlattenProgramRight impl p'' f)).

Proof.
  introv H1.
  destruct p; progFrobSimplIn H1; inverts H1.
  - left. eauto.
  - right. eauto.
Qed.

Lemma bindFlattenProgramEqRet:
  forall {M1 M2 I2} (impl: forall R, M2 R -> Program I2 R)
    {R R'} (p: Program (PairMethod M1 I2) R) (f: R -> Program (PairMethod M1 M2) R') (r: R'),
    
    bindFlattenProgramRight impl p f = Return r ->
    False.

Proof.
  intros.
  destruct p;
    rewrite frobProgramId in H at 1;
    simpl in H.
  - inverts H.
  - inverts H.
  - inverts H.
Qed.

Lemma EvalProgramBindFlattenRet:
  forall M M' (impl : forall R, M' R -> Program M R)
    R (p: Program M R)
    o s r s',
    EvalProgram o R p s r s' ->
    EvalProgram o R (bindFlattenProgram impl (fun r: R => Return r) p) s r s'.

Proof.
  introv H1.

  gen R p s r s'.
  eapply EvalProgramInd.
  - intros. progFrobSimpl.
    eapply ProgramStepStepsImplSteps.
    econstructor. progFrobSimpl.
    apply ProgramStepsRefl.
  - introv H1 H2. progFrobSimpl.
    eapply ProgramStepStepsImplSteps.
    econstructor. eassumption. eassumption.
  - introv H1. progFrobSimpl.
    eapply ProgramStepStepsImplSteps.
    econstructor. eassumption.
Qed.


Lemma mapProgramRefl:
  forall M R (p: Program M R),
    programEq
      (mapProgram (fun R m => m) p)
      p.

Proof.
  intros.
  gen p.
  cofix self.
  intros.
  destruct p;
    match goal with
    | [ |- programEq ?p _] =>
      rewriteFrobProgram p
    end;
    simpl.
  - constructor. eauto.
  - constructor.
  - constructor. eauto.    
Qed.

Lemma mapMapProgram:
  forall M M' M'' (f : forall R, M R -> M' R) (f' : forall R, M' R -> M'' R) R (p: Program M R),
    programEq
      (mapProgram f' (mapProgram f p))
      (mapProgram (fun R m => f' R (f R m)) p).

Proof.
  intros.
  gen p.
  cofix self.
  intros.
  destruct p.
  - match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end.
    simpl.
    constructor.
    eauto.
  - match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end.
    simpl.
    reflexivity.
  - match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end.
    simpl.
    constructor.
    eauto.
Qed.
(*
Lemma Lem1:  
  forall R A M M' M'' (p : A -> Program M'' R) (p0 : Program M A) (f: forall R, M R -> M' R) (impl: forall R, M'' R -> Program M R) p1 p2,
    (forall R (p: Program M'' R),
        programEq (mapProgram f (flattenProgram impl p))
                  (flattenProgram (fun R m => mapProgram f (impl R m)) p)) ->
    programEq (mapProgram f (bindFlattenProgram impl p p0))
              p1 ->
    programEq (bindFlattenProgram (fun R m => mapProgram f (impl R m)) p (mapProgram f p0))
              p2 ->
    programEq p1
              p2.

Proof.
  introv H1 H2 H3.
  gen p0 p1 p2 H2 H3.
  cofix Ind.
  intros.

  destruct p0;
  rewrite frobProgramId in H2 at 1; simpl in H2;
  inverts H2;
  rewrite frobProgramId in H3 at 1; simpl in H3;
  inverts H3.
  - constructor.
    intros. eauto.
    
  - constructor.
    etransitivity. symmetry. eassumption.
    etransitivity. eapply H1. eassumption.
    
  - constructor.
    eauto.
Defined.
*)

Lemma Lem:  
  forall R A M M' M'' (p : A -> Program M'' R) (p0 : Program M A) (f: forall R, M R -> M' R) (impl: forall R, M'' R -> Program M R),
    (forall R (p: Program M'' R),
        programEq (mapProgram f (flattenProgram impl p))
                  (flattenProgram (fun R m => mapProgram f (impl R m)) p)) ->
    programEq (mapProgram f (bindFlattenProgram impl p p0))
              (bindFlattenProgram (fun R m => mapProgram f (impl R m)) p (mapProgram f p0)).

Proof.
  introv H1.
  gen p0 p.
  cofix Ind.
  intros.
  destruct p0;
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end;
    simpl.
   
  - constructor.
    intros. eauto.
    
  - constructor.
    eauto.
    
  - constructor.
    eauto.  
Defined.

Lemma mapFlattenProgram:
  forall M M' M'' (f: forall R, M R -> M' R) (impl: forall R, M'' R -> Program M R) R (p: Program M'' R),
    programEq
      (mapProgram f (flattenProgram impl p))
      (flattenProgram (fun R (m: M'' R) => mapProgram f (impl R m)) p).

Proof.
  intros.

  gen R p.
  cofix Ind.
      
  destruct p.
  + match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end.
    simpl.

    destruct (impl A m).


    * {
        constructor. intros.
        
        match goal with
        | [ |- programEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p'
        end; simpl.
        constructor. intros.
        
        generalize (p0 v).
        cofix Ind'.
        intros.
        destruct p1.
        match goal with
        | [ |- programEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p'
        end; simpl.
        
        - constructor.
          intros. apply Ind'.
          
        - match goal with
          | [ |- programEq ?p ?p'] =>
            rewriteFrobProgram p;
              rewriteFrobProgram p'
          end; simpl.
          constructor.
          apply Ind.
          
        - match goal with
          | [ |- programEq ?p ?p'] =>
            rewriteFrobProgram p;
              rewriteFrobProgram p'
          end; simpl.
          constructor.
          apply Ind'. 
          
      }

    * constructor.
      match goal with
      | [ |- programEq ?p ?p'] =>
        rewriteFrobProgram p;
          rewriteFrobProgram p'
      end; simpl.
      constructor.
      apply Ind.      

    * constructor.

      match goal with
      | [ |- programEq ?p ?p'] =>
        rewriteFrobProgram p;
          rewriteFrobProgram p'
      end; simpl.
      constructor.
      
      {
      generalize p0.
      cofix Ind'.
      intros.
      destruct p1;
        match goal with
        | [ |- programEq ?p ?p'] =>
          rewriteFrobProgram p;
            rewriteFrobProgram p'
        end;
        simpl.
      
      - constructor.
        intros. eauto.
        
      - constructor.
        eauto.
        
      - constructor.
        eauto.  

      }
      
  + match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p;
      rewriteFrobProgram p'
    end.
    simpl.
    reflexivity.

  + match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p
      ; rewriteFrobProgram p'
    end.
    simpl.
    constructor.
    eauto.

Qed.

CoFixpoint flattenMapProgram
           M M' I (impl : forall R, M' R -> Program I R) (f : forall R, M R -> M' R) R (p : Program M R) :
  programEq
    (flattenProgram impl (mapProgram f p))
    (flattenProgram (fun R m => impl R (f R m)) p)
with bindFlattenMapProgram
       M M' I (impl : forall R, M' R -> Program I R) (f : forall R, M R -> M' R) R R' (g : R' -> Program M R) (p : Program I R') :
  programEq
    (bindFlattenProgram impl (fun a => mapProgram f (g a)) p)
    (bindFlattenProgram (fun A m => impl A (f A m)) g p).
Proof.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; constructor; auto.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; constructor; auto.
Qed.

CoFixpoint flattenFlattenProgram M M1 M2
  (implProg1 : forall R, M R -> Program M1 R)
  (implProg2 : forall R, M1 R -> Program M2 R)
  R (p : Program M R) :
  programEq
    (flattenProgram implProg2 (flattenProgram implProg1 p))
    (flattenProgram (fun R m => flattenProgram implProg2 (implProg1 R m)) p)
with
flattenBindFlattenProgram M M1 M2
  (implProg1 : forall R, M R -> Program M1 R)
  (implProg2 : forall R, M1 R -> Program M2 R)
  A R (f : A -> Program M R) (p : Program M1 A) :
  programEq
    (flattenProgram implProg2 (bindFlattenProgram implProg1 f p))
    (bindFlattenProgram
       (fun R m => flattenProgram implProg2 (implProg1 R m)) f
       (flattenProgram implProg2 p))
with
bindFlattenBindFlattenProgram M M1 M2
  (implProg1 : forall R, M R -> Program M1 R)
  (implProg2 : forall R, M1 R -> Program M2 R)
  A B R (f : B -> Program M R) (g : A -> Program M1 B) (p : Program M2 A) :
  programEq
    (bindFlattenProgram
       implProg2 (fun a => bindFlattenProgram implProg1 f (g a))
       p)
    (bindFlattenProgram
       (fun R m => flattenProgram implProg2 (implProg1 R m)) f
       (bindFlattenProgram implProg2 g p)).

Proof.
  - match goal with
    | [ |- programEq ?p ?p' ] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; constructor.
    + apply flattenBindFlattenProgram.
    + apply flattenFlattenProgram.
  - match goal with
    | [ |- programEq ?p ?p' ] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; constructor.
    + apply bindFlattenBindFlattenProgram.
    + apply flattenFlattenProgram.
    + apply flattenBindFlattenProgram.
  - match goal with
    | [ |- programEq ?p ?p' ] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; constructor.
    + intro.
      apply bindFlattenBindFlattenProgram.
    + apply flattenBindFlattenProgram.
    + apply bindFlattenBindFlattenProgram.
Qed.


CoFixpoint flattenProgramLeft'ProgramEq
           Method Method' Instruction
           (impl impl' : forall A, Method A -> Program Instruction A)
           Ret
           (p p' : Program (PairMethod Method Method') Ret) :
  (forall A m, programEq (impl A m) (impl' A m)) ->
  programEq p p' ->
  programEq (flattenProgramLeft' impl p) (flattenProgramLeft' impl' p')
with bindFlattenProgramLeft'ProgramEq
       Method Method' Instruction
       (impl impl' : forall A, Method A -> Program Instruction A)
       Ret Ret'
       (f f' : Ret -> Program (PairMethod Method Method') Ret')
       (p p' : Program Instruction Ret) :
  (forall A m, programEq (impl A m) (impl' A m)) ->
  (forall r, programEq (f r) (f' r)) ->
  programEq p p' ->
  programEq (bindFlattenProgramLeft' impl p f) (bindFlattenProgramLeft' impl' p' f').
Proof.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p as [? [|] ? | |]; inverts H0; constructor; auto.
  - intros.
    match goal with
    | [ |- programEq ?p ?p'] =>
      rewriteFrobProgram p; rewriteFrobProgram p'
    end.
    destruct p; inverts H1; constructor; auto.
Qed.

CoFixpoint mapFlattenProgramLeft'
           {M M' M'' M'''}
           (f : forall R : Type, M'' R -> M' R)
           (implProg : forall R : Type, M R -> Program M'' R)
           R (p : Program (PairMethod M M''') R) :
  programEq
    (mapProgram
       (fun R m =>
          match m with
          | inl m' => inl (f R m')
          | inr m' => inr m'
          end)
       (flattenProgramLeft' implProg p))
    (flattenProgramLeft' (fun R m => mapProgram f (implProg R m)) p)
with mapBindFlattenProgramLeft'
       {M M' M'' M'''}
       (f : forall R : Type, M'' R -> M' R)
       (implProg : forall R : Type, M R -> Program M'' R)
       R A (p : Program M'' A)
       (p' : A -> Program (PairMethod M M''') R) :
       programEq
         (mapProgram
            (fun R m =>
               match m with
               | inl m' => inl (f R m')
               | inr m' => inr m'
               end)
            (bindFlattenProgramLeft' implProg p p'))
         (bindFlattenProgramLeft'
            (fun R m => mapProgram f (implProg R m))
            (mapProgram f p) p').
Proof.
  - match goal with
    | [ |- programEq ?p1 ?p2 ] =>
      rewrite (frobProgramId p2), (frobProgramId p1)
    end.
    destruct p as [? [|] ? | |]; constructor; intros.
    + apply mapBindFlattenProgramLeft'.
    + apply mapFlattenProgramLeft'.
    + apply mapFlattenProgramLeft'.
  - match goal with
    | [ |- programEq ?p1 ?p2 ] =>
      rewrite (frobProgramId p2), (frobProgramId p1)
    end.
    destruct p; constructor; intros.
    + apply mapBindFlattenProgramLeft'.
    + apply mapFlattenProgramLeft'.
    + apply mapBindFlattenProgramLeft'.
Qed.



Section ProgramBigStep.

  Context {Instruction : Type -> Type}.
  Context (obj : Spec Instruction).
  Context (Method : Type -> Type).
  Context (impl: forall R, Method R -> Program (Instruction) R).
  
  Inductive ProgramBigStep :
    forall R, Program Method R -> obj.(State) ->
         Program Method R -> obj.(State) -> Prop :=
    
  | BigEBind {R m s} {r : R} {s' R' f}:
      EvalProgram obj R (impl R m) s r s' ->
      ProgramBigStep R' (Bind m f) s (f r) s'

  | BigENoop {R} p s:
      ProgramBigStep R (Noop p) s p s.

  Definition ProgramBigStep' R
             (s1: Program Method R * obj.(State))
             (s2: Program Method R * obj.(State)): Prop :=
    ProgramBigStep R (fst s1) (snd s1) (fst s2) (snd s2).
  
  
  Definition ProgramBigSteps' R := clos_refl_trans_1n
                                  (Program Method R * obj.(State))
                                  (ProgramBigStep' R).

  Definition ProgramBigSteps R
             (p1: Program Method R) (s1: obj.(State))
             (p2: Program Method R) (s2: obj.(State)): Prop :=
    ProgramBigSteps' R (p1, s1) (p2, s2).

  Definition BigEvalProgram R p s r s' :=
    ProgramBigSteps R p s (Return r) s'.

  Lemma ProgramBigStepsRefl:
    forall R p s,
      ProgramBigSteps R p s p s.
  Proof.
    intros.
    unfold ProgramBigSteps.
    unfold ProgramBigSteps'.
    econstructor.
  Qed.

  Lemma ProgramBigStepImplProgramBigSteps:
    forall R p1 s1 p2 s2,
      ProgramBigStep R p1 s1 p2 s2 ->
      ProgramBigSteps R p1 s1 p2 s2.
  Proof.
    introv H1.
    unfold ProgramBigSteps.
    unfold ProgramBigSteps'.
    eapply clos_rt1n_step.
    unfold ProgramStep'.
    simpl.
    assumption.      
  Qed.

  Lemma ProgramBigStepBigStepsImplBigSteps:
    forall R p1 s1 p2 s2 p3 s3,
      ProgramBigStep R p1 s1 p2 s2 ->
      ProgramBigSteps R p2 s2 p3 s3 ->
      ProgramBigSteps R p1 s1 p3 s3.
  Proof.
    introv H1 H2.
    unfold ProgramBigSteps in *.
    unfold ProgramBigSteps' in *.
    unfold ProgramBigStep' in *.

    econstructor.
    - simpl. instantiate (1 := (p2, s2)). simpl. eassumption.
    - eassumption.    
  Qed.
  
  Lemma ProgramBigStepInv:
    forall R p s p' s',
      ProgramBigStep R p s p' s' ->
      match p with
      | Bind R' m f =>
        exists r,
        EvalProgram obj R' (impl R' m) s r s'
        /\ p' = f r
      | Noop p =>
        s = s'
        /\ p' = p
      | Return r =>
        False
      end.

  Proof.
    intros.
    destruct H.
    - eexists. eauto.
    - eauto.
  Qed.

  
  Lemma ProgramBigStepsInv:
    forall R p s p' s',
      ProgramBigSteps R p s p' s' ->
      ((p' = p /\ s' = s)
       \/ match p with
         | Bind R' m f =>
           exists s'' r,
           EvalProgram obj R' (impl R' m) s r s''
           /\ ProgramBigSteps R (f r) s'' p' s'
         | Noop p =>
           ProgramBigSteps R p s p' s'
         | Return r =>
           p' = Return r
           /\ s = s'
         end).

  Proof.
    introv H1.
    unfold ProgramBigSteps in H1.
    unfold ProgramBigSteps' in H1.
    unfold ProgramBigStep' in H1.
    remember (p, s) as st1.
    remember (p', s') as st2.    
    gen p s p' s'.
    inverts H1; intros.
    - subst. inversion Heqst2. subst. left. auto.
    - right. destruct y. destruct st1. simpl in *. subst. inversion Heqst1. subst.
      eapply ProgramBigStepInv in H.
      destruct p.
      + destruct H as (r & A21 & A22). subst. eexists. eexists. split. eassumption. hnf. eassumption.
      + inversion H.
      + destruct H as (A21 & A22). subst. hnf. eassumption.
  Qed.        


  Lemma BigEvalProgramInv:
    forall R p s r s',
      BigEvalProgram R p s r s' ->
      match p with
      | Bind R' m f =>
        exists s'' r'',
        EvalProgram obj R' (impl R' m) s r'' s''
        /\ BigEvalProgram R (f r'') s'' r s'
      | Noop p =>
        BigEvalProgram R p s r s'
      | Return r' =>
        r' = r
        /\ s = s'
      end.
  
  Proof.
    introv H1.
    unfold BigEvalProgram in H1.
    (* remember (flattenProgram impl p) as p1. *)
    apply ProgramBigStepsInv in H1.
    destruct H1 as [(H1 & H2) | H1].
    - subst. auto.
    - destruct p.
      + auto.
      + destruct H1 as (H1 & H2). inverts H1. auto.
      + auto.    
  Qed.

  Lemma ProgramBigStepsInd:
    forall (P: forall R, Program Method R -> obj.(State) -> 
               Program Method R -> obj.(State) -> Prop),

      (forall R p s,
          P R p s p s) ->
      (forall R R' m (f: R -> Program Method R') s r s' r'' s'',
          EvalProgram obj R (impl R m) s r s' ->
          P R' (f r) s' r'' s'' ->
          P R' (Bind m f) s r'' s'') ->
      (forall R p s r s',
          P R p s r s' ->
          P R (Noop p) s r s') ->
      forall R p s p' s',
        ProgramBigSteps R p s p' s' -> P R p s p' s'.
  
  Proof.    
    introv H1 H2 H3 H4.
    (* hnf in H4 *)    
    unfold BigEvalProgram in *.
    unfold ProgramBigSteps in *.
    unfold ProgramBigSteps' in *.
    unfold ProgramBigStep' in H4.
    
    (* remember (fun s1 s2 : Program (Instruction obj) R * State obj =>
                ProgramBigStep obj obj' impl R (fst s1) (snd s1) (fst s2) (snd s2)) as Rel. *)
    remember (p, s) as st1.
    remember (p', s') as st2.
    
    gen p s p' s'.
    induction H4; intros; subst.

    - inverts Heqst2. auto.
    - destruct y. simpl in *.
      forwards A: ProgramBigStepInv H.

      destruct p.
      + destruct A as (r' & A1 & A2). subst. eauto.
        (* eapply H2. eassumption. eapply IHclos_refl_trans_1n. reflexivity. reflexivity. *)
      + inversion A.
      + destruct A as (A1 & A2). subst. eauto.
        (* eapply H3. eapply IHclos_refl_trans_1n. reflexivity. reflexivity. *)
  Qed.

  
  Lemma BigEvalProgramInd:
    forall (P: forall R, Program Method R -> obj.(State) -> 
               R -> obj.(State) -> Prop),

      (forall R r s,
          P R (Return r) s r s) ->
      (forall R R' m (f: R -> Program Method R') s r s' r'' s'',
          EvalProgram obj R (impl R m) s r s' ->
          P R' (f r) s' r'' s'' ->
          P R' (Bind m f) s r'' s'') ->
      (forall R p s r s',
          P R p s r s' ->
          P R (Noop p) s r s') ->
      forall R p s r s',
        BigEvalProgram R p s r s' -> P R p s r s'.
  
  Proof.
    intros.    
    unfold BigEvalProgram in *.
    remember (Return r) as p'.

    (* Set Printing Implicit. *)
    
    gen r.
    eapply ProgramBigStepsInd with
    (P := (fun R p s p' s' => forall r : R, p' = @Return Method R r -> P R p s r s'));
      intros; subst; eauto.    
  Qed.


(*
CoFixpoint flattenProgram
           {Method Instruction}
           (impl : forall A, Method A -> Program Instruction A)
           {Ret}
           (p : Program Method Ret) :
  Program Instruction Ret :=
  match p with
    | Bind A m f =>
      (* bindProgram (impl A m) (fun a => Noop (flattenProgram impl (f a))) *)

      (cofix bindFlattenProgram p' :=
         match p' with
           | Bind _ instr f' => a <- instr; bindFlattenProgram (f' a)
           | Return a => Noop (flattenProgram impl (f a))
           | Noop p'' => Noop (bindFlattenProgram p'')
         end)
        (impl A m)

    | Noop p' => Noop (flattenProgram impl p')
    | Return a => Return a
  end.
*)

  
  Definition Rel R pa si (p: Program Instruction R) s :=

    match p with
    | Bind _ m f =>
      exists R' pr ma fa,
      pa = Bind ma fa
      /\ programEq p (bindProgram pr (fun (r: R') => (Noop (flattenProgram impl (fa r)))))
      /\ ProgramSteps obj R' (impl R' ma) si pr s

    | Return r =>
      pa = Return r
      /\ s = si
      
    | Noop p' =>
      (exists R' ma fa,
          pa = Bind ma fa
          /\ programEq p'
                      (bindProgram
                               (impl R' ma)
                               (fun (r: R') => (Noop (flattenProgram impl (fa r)))))
          /\ s = si)
        
      \/ (exists pa',
          pa = Noop pa'
          /\ programEq p' (flattenProgram impl pa')
          /\ s = si)

      \/ (exists R' pr ma fa,
            pa = Bind ma fa
            /\ programEq p (bindProgram pr (fun (r: R') => (Noop (flattenProgram impl (fa r)))))
            /\ ProgramSteps obj R' (impl R' ma) si pr s)                  
    end.

  

  Lemma RelEqProgramEq:
    forall (R : Type) (pa : Program Method R) 
      (si : State obj) (p1 p2 : Program Instruction R) 
      (s : State obj), programEq p1 p2 -> Rel R pa si p1 s -> Rel R pa si p2 s.
  
  Proof.
    introv H1 H2.
    destruct H1.
    - unfold Rel in *.
      destruct H2 as (R'' & pr & ma & fa & H21 & H22 & H23).
      eexists. eexists. eexists. eexists. splits.
      eassumption.
      etransitivity. Focus 2. eassumption. constructor. intros. symmetry. eauto.
      eassumption.
    - assumption.
    - unfold Rel in *.
      destruct H2 as [(R' & ma & fa & H21 & H22 & H23) | [(pa' & H21 & H22 & H23) | (R'' & pr & ma & fa & H21 & H22 & H23)]].
      + left.
        eexists. eexists. eexists. splits.
        eassumption.
        etransitivity. Focus 2. eassumption. symmetry. assumption.
        eassumption.
      + right. left. eexists. splits. eassumption. etransitivity. symmetry. eassumption. eassumption. assumption.
      + right. right.
        eexists. eexists. eexists. eexists. splits.
        eassumption.
        etransitivity. Focus 2. eassumption. constructor. intros. symmetry. eauto.
        eassumption.
  Qed.
        
  Add Parametric Morphism R: (Rel R)
      with signature eq ==> eq ==> programEq ==> eq ==> iff
        as Rel_eq_programEq_eq.
  Proof.
    intros pa si p1 p2 H1 s.
    split.
    apply RelEqProgramEq. assumption.
    apply RelEqProgramEq. symmetry. assumption.
  Qed.

    
  Lemma RelFlattenProgram:
    forall R pa s,
      Rel R pa s (flattenProgram impl pa) s.
      
  Proof.
    intros.
    unfold Rel.
    destruct (flattenProgram impl pa) eqn:E; symmetry in E; apply flattenProgramInv in E.
    - contradiction.
      
    - auto.
      
    - destruct E as [(p'' & E1 & E2) | [(R' & ma & fa & E1 & E2) | [(R' & ma & fa & r & E1 & E2 & E3) | (R' & ma & fa & p'' & E1 & E2 & E3)]]].
      
      + right. left.
        eexists. splits.
        eassumption.
        subst. reflexivity.
        reflexivity.

      + left.
        subst.      
        eexists. eexists. eexists. splits.      
        reflexivity.
        apply bindFlattenProgramBindProgram.
        reflexivity.
        
      + right. right. subst.
        eexists. eexists. eexists. eexists. splits.
        reflexivity.
        Focus 2. eapply ProgramStepsRefl.      
        rewrite E2. setoid_rewrite frobProgramId at 4. simpl.
        reflexivity.
      
      + right. right. subst.
        eexists. eexists. eexists. eexists. splits.      
        reflexivity.
        Focus 2. eapply ProgramStepsRefl.
        rewrite E2. setoid_rewrite frobProgramId at 3. simpl.
        constructor. rewrite E3. apply bindFlattenProgramBindProgram.      
  Qed.


  Lemma RelBindProgram:
    forall R s R'
      (ma : Method R') (fa : R' -> Program Method R),
      Rel R
          (x <- ma; fa x)
          s
          (bindProgram (impl R' ma)
                       (fun r : R' => Noop (flattenProgram impl (fa r))))
          s.
  
  Proof.
    intros.
    destruct ((bindProgram (impl R' ma) (fun r : R' => Noop (flattenProgram impl (fa r))))) eqn: E;
      unfold Rel.
    * eexists. eexists. eexists. eexists. splits. reflexivity. rewrite E. reflexivity. constructor.
    * exfalso. apply eqProgramEq' in E. symmetry in E. apply bindProgramInv in E. destruct E as (r' & E1 & E2). inverts E2.
    * right. right.
      subst. eexists. eexists. eexists. eexists. splits.
      reflexivity.
      symmetry. apply eqProgramEq' in E. eassumption.
      constructor.
  Qed.
  
  Lemma ProgramStepsProgramBigSteps:
    forall R pa si p s p' s',
      Rel R pa si p s ->
      ProgramSteps obj R p s p' s' ->
      exists pa' si',
        Rel R pa' si' p' s'
        /\ ProgramBigSteps R pa si pa' si'.
  
  Proof.
    introv H1 H2.
    gen H1.

    gen pa si.
    
    applys @ProgramStepsInd Instruction
           (fun R p s p' s' =>
              forall (pa : Program Method R) (si : State obj),
                Rel R pa si p s ->
                exists pa' si', Rel R pa' si' p' s' /\ ProgramBigSteps R pa si pa' si')
           H2; clear H2; intros; subst.
    
    - eexists. eexists. split. eassumption. unfold Rel in H. apply ProgramBigStepsRefl.

    - eapply H0.
      unfold Rel in H1. destruct H1 as (R'' & pr & ma & fa & H1 & H2 & H3).
      
      (* Need to show that Rel is preserved after m steps. *)
      unfold Rel.
      (* Is pr only m or more. *)
      forwards A: bindProgramInv H2.
      
      destruct A as [(pr' & A1) | (ret & A1 & A2)].
      (* destruct (f r). *)
      + rewrite H1 in *.
        rewrite A1 in *.
        setoid_rewrite frobProgramId in H2 at 2. simpl in H2. inverts H2.
        specialize (H5 r).
        
        destruct (pr' r) eqn:E.
        * (* The execution of the object of ma is not yet finished. *)
          (* There is a method call *)
          destruct (f r).

          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.
          eexists. exists (x <- i; p0 x). eexists. eexists. splits.
          reflexivity.
          setoid_rewrite frobProgramId at 2. simpl. constructor. assumption.
          eapply ProgramStepsStepImplSteps. eassumption. forwards B: @EBind Instruction pr' H. rewrite E in B. eassumption.

          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.

          setoid_rewrite frobProgramId in H5 at 3; simpl in H5. inverts H5.

          
        * destruct (f r).

          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.

          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.

          setoid_rewrite frobProgramId in H5 at 3; simpl in H5; inverts H5.
          
          (* The execution of the object of ma is finished. *)
          right. right.
          eexists. exists (pr' r). eexists. eexists. splits.          
          reflexivity.
          rewrite E. rewrite frobProgramId. simpl. constructor. eassumption.
          eapply ProgramStepsStepImplSteps. eassumption. constructor. assumption.
          
        * destruct (f r).

          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.
          setoid_rewrite frobProgramId in H5 at 2; simpl in H5; inverts H5.
          setoid_rewrite frobProgramId in H5 at 3; simpl in H5; inverts H5.
          
          (* The execution of the object of ma is not yet finished. *)
          (* There is a noop *)
          right. right.
          eexists. exists (pr' r). eexists. eexists. splits.
          reflexivity.
          rewrite E. rewrite frobProgramId. simpl. constructor. eassumption.
          eapply ProgramStepsStepImplSteps. eassumption. constructor. assumption.

      + inverts A2.

    - unfold Rel in H0.
      destruct H0 as [(R' & ma & fa & H1 & H2 & H3) | [(pa' & H1 & H2 & H3) | (R' & pr & ma & fa & H1 & H2 & H3)]].


      + forwards (pai & sm & A1 & A2): H pa si. subst. rewrite H2. apply RelBindProgram.
        subst. eexists. eexists. splits; eassumption.
       
      + subst.
        forwards (pai & sm & A1 & A2): H pa' si. rewrite H2. apply RelFlattenProgram.
        eexists. eexists. splits.
        eassumption.
        eapply ProgramBigStepBigStepsImplBigSteps. constructor. eassumption.       
        
     + forwards A: bindProgramInv H2.
       destruct A as [(pi & A1 & A2) | (r' & A1 & A2)].
       * eapply H. subst.         
         unfold Rel.
         setoid_rewrite frobProgramId in H2 at 3. simpl in H2. inverts H2. {
         destruct p0.

         - destruct pi; setoid_rewrite frobProgramId in H4 at 2; simpl in H4; inverts H4.
           eexists. eexists. eexists. eexists. splits.
           reflexivity.
           eassumption.
           eapply ProgramStepsStepImplSteps. eassumption. constructor.

         - destruct pi; setoid_rewrite frobProgramId in H4 at 2; simpl in H4; inverts H4.

         - destruct pi; setoid_rewrite frobProgramId in H4 at 3; simpl in H4; inverts H4.
           right. right. eexists. eexists. eexists. eexists. splits.
           reflexivity.
           eassumption.
           eapply ProgramStepsStepImplSteps. eassumption. constructor.

           right. right. eexists. eexists. eexists. eexists. splits.
           reflexivity.
           eassumption.
           eapply ProgramStepsStepImplSteps. eassumption. constructor.
         }
         
       * inverts A2.

         subst.
         forwards (pai & sm & A1 & A2): H (fa r') s0.
         rewrite <- H5. apply RelFlattenProgram.
         
         eexists. eexists. splits.
         eassumption.
         eapply ProgramBigStepBigStepsImplBigSteps. constructor. eassumption. eassumption.
       
  Qed.  

  (* Small step evaluation implies big step evaluation. *)
  Lemma EvalProgramFlattenBigEvalProgram:
    forall R p s r s',
      EvalProgram obj R (flattenProgram impl p) s r s' ->
      BigEvalProgram R p s r s'.

  Proof.
    intros.
    unfold EvalProgram in *.
    forwards (pa' & si' & A1 & A2): ProgramStepsProgramBigSteps p s H.
    - apply RelFlattenProgram.
        
    - unfold BigEvalProgram.
      unfold Rel in A1. destruct A1. subst. assumption.
    
  Qed.


  Lemma EvalProgramFlattenInd:
    forall (P: forall R, Program Method R -> obj.(State) -> 
               R -> obj.(State) -> Prop),

      (forall R r s,
          P R (Return r) s r s) ->
      (forall R R' m (f: R -> Program Method R') s r s' r'' s'',
          EvalProgram obj R (impl R m) s r s' ->
          P R' (f r) s' r'' s'' ->
          P R' (Bind m f) s r'' s'') ->
      (forall R p s r s',
          P R p s r s' ->
          P R (Noop p) s r s') ->
      forall R p s r s',
        EvalProgram obj R (flattenProgram impl p) s r s' -> P R p s r s'.
  
  Proof.
    intros.
    apply EvalProgramFlattenBigEvalProgram in H2.
    eapply BigEvalProgramInd; eassumption.    
  Qed.

End ProgramBigStep.




Section EvalProgramFlattenRef.

  Context {AInstruction} {aObj: Spec AInstruction}.
  Context {CInstruction} {cObj: Spec CInstruction}.

  Variable DRel: cObj.(State) -> aObj.(State) -> Prop.

  Variable cInit: cObj.(State).
  Variable aInit: aObj.(State).


  Hypothesis initRel: DRel cInit aInit.

  Variable impl: forall R, AInstruction R -> Program CInstruction R.

  Hypothesis stepRel:
    forall R (m: AInstruction R) s1 s1' r s2,
      DRel s1 s2 ->
      EvalProgram cObj R (impl R m) s1 r s1' ->
      exists s2',
        aObj.(EvalInstruction) m s2 r s2'
        /\ DRel s1' s2'.

  Lemma EvalProgramFlattenRef:
    forall R (p: Program AInstruction R)
      (cst': cObj.(State)) (r : R),
      EvalProgram cObj R
                  (flattenProgram impl p)
                  cInit r cst' ->
      exists (ast': aObj.(State)),
        EvalProgram aObj R 
                    p
                    aInit r ast'
        /\ DRel cst' ast'.

  Proof.
    intros.
    eapply (EvalProgramFlattenInd cObj AInstruction impl
                                  (fun R p cInit r cst' =>
                                     forall aInit0 : State aObj,
                                       DRel cInit aInit0 ->
                                       exists ast', EvalProgram aObj R p aInit0 r ast' /\ DRel cst' ast')) in H.
    - eassumption.
    - intros. eexists. split. constructor. assumption.      
    - intros.
      forwards (s2' & A1 & A2): stepRel.
      eassumption. eassumption.
      forwards (ast' & B1 & B2): H1.
      eassumption.
      eexists. split.
      + eapply ProgramStepStepsImplSteps.
        * constructor. eassumption.
        * eassumption.
      + eassumption.
    - introv H1 H2.
      forwards (ast' & B1 & B2): H1.
      eassumption.      
      eexists. split. eapply ProgramStepStepsImplSteps.
      + constructor.
      + eassumption.
      + assumption.
    - assumption.

  Qed.
    
  
End EvalProgramFlattenRef.

Lemma FlattenBindRetNEq:
  forall {M} (R' : Type) (p : Program M R'),
    programNEq
      (flattenProgram (fun R (m : M R) => x <- m; Return x) p)
      p.

Proof.
  cofix CoInd.
  intros. destruct p.
  - match goal with
    | [ |- programNEq ?p _] =>
      rewriteFrobProgram p; simpl
    end.
    constructor.
    match goal with
    | [ |- programNEq ?p _] =>
      rewriteFrobProgram p; simpl
    end.
    constructor.
    intros.
    match goal with
    | [ |- programNEq ?p _] =>
      rewriteFrobProgram p; simpl
    end.
    constructor. apply CoInd.
    
  - setoid_rewrite frobProgramId at 1. simpl.
    constructor.

  - match goal with
    | [ |- programNEq ?p _] =>
      rewriteFrobProgram p; simpl
    end.
    constructor. auto.
Qed.


Lemma programNEqProgramSteps:
  forall R I (obj: Spec I) (p1 p1' p2: Program I R) s1 s1',
    programNEq p1 p2 ->
    ProgramSteps obj _ p1 s1 p1' s1' ->
    exists p2',
      ProgramSteps obj _ p2 s1 p2' s1'
      /\ programNEq p1' p2'.

Proof.
  introv H1 H2.
  induction H2 using ProgramStepsInd.
  - eexists. split. eapply ProgramStepsRefl. assumption.
  - inverts H1. forwards (p2' & H1 & H2): IHProgramSteps. eapply H5.
    eexists. splits.
    + eapply ProgramStepStepsImplSteps. constructor. eassumption. eassumption.
    + assumption.
  - { inverts H1.
      - forwards (p2' & H1 & H2): IHProgramSteps. eassumption.
        eexists. splits.
        + eapply ProgramStepStepsImplSteps. constructor. eassumption.
        + assumption.
      - auto.
    }
Qed.

Lemma evalMapProgram Method Method'
      (obj : Spec Method)
      (f : forall A, Method' A -> Method A)
      Ret
      (p : Program Method' Ret)
      (ret : Ret)
      (st st' : obj.(State)) :
  EvalProgram obj Ret (mapProgram f p) st ret st' ->
  EvalProgram (mapSpec obj f) Ret p st ret st'.
Proof.
  remember (mapProgram f p) as p2.
  intro H.
  gen p Heqp2.
  induction H using EvalProgramInd; intros q Heqp2; symmetry in Heqp2; apply mapProgramInv in Heqp2.
  - subst.
    constructor.
  - destruct Heqp2 as (? & ? & ? & ? & ?).
    subst.
    eapply ProgramStepStepsImplSteps.
    + constructor.
      simpl.
      eassumption.
    + apply IHEvalProgram.
      reflexivity.
  - destruct Heqp2 as (? & ? & ?).
    subst.
    eapply ProgramStepStepsImplSteps.
    + constructor.
    + apply IHEvalProgram.
      reflexivity.
Qed.
