From Coq Require Import
  Setoid Relations Morphisms.

From C4.Util Require Import
  Util LibTactics.

From C4.Core Require Import
  Program ProgramFacts
  ProgramOps ProgramOpsFacts
  Superposition 
  Object
  Pair PairFacts
  Simulates SimulatesFacts
  RSimulates
  Eutt
  ImplSimulates.

Lemma programSpecSimulates_ Method (obj1 obj2 : Spec Method) st1 st2 :
  Simulates_ obj1 obj2 st1 st2 ->
  Simulates_ (programSpec obj1) (programSpec obj2) st1 st2.
Proof.
  intro S.
  eapplyRelationSimulates (Simulates_ obj1 obj2).
  - intros * H1 H2.
    simpl.
    eapply programStepsSimulates.
    + eassumption.
    + assumption.
  - assumption.
Qed.

Lemma programSpecSimulates Method (obj1 obj2 : Spec Method) :
  Simulates obj1 obj2 -> Simulates (programSpec obj1) (programSpec obj2).
Proof.
  apply programSpecSimulates_.
Qed.

Lemma flattenProgramSimulates Method Method' (obj : Spec Method) (implProg : forall Ret, Method' Ret -> Program Method Ret) st :
  Simulates_
    (mapSpec
       (programSpec obj)
       (@flattenProgram _ _ implProg))
    (programSpec
       (mapSpec (programSpec obj) implProg))
    st st.
Proof.
  cbn in st.
  eapplyRelationSimulates (@eq (State obj)).
  intros * H1 H2.
  - subst.
    simpl in *.
    eexists.
    + edestruct (@EvalProgramFlattenRef Method' (mapSpec (programSpec obj) implProg) Method obj eq) as [? [? ?]].
      * reflexivity.
      * intros.
        subst.
        eexists.
        split.
        -- simpl.
          eassumption.
        -- reflexivity.
      * eassumption.
      * subst.
        eassumption.
    + reflexivity.
  - reflexivity.
Qed.

Theorem programSpecSteps Method Method' (obj : Spec Method)
        (implProg : forall Ret : Type, Method' Ret -> Program Method Ret)
        A (p p' : Program Method' A) (st st' : State obj) :
  ProgramSteps (mapSpec (programSpec obj) implProg) A p st p' st' ->
  ProgramSteps obj A (flattenProgram implProg p) st (flattenProgram implProg p') st'.
Proof.
  induction 1 using ProgramStepsInd.
  - apply ProgramStepsRefl.
  - progFrobSimplAt 1.
    eapply ProgramStepStepsImplSteps.
    + econstructor.
    + induction H using EvalProgramInd.
      * progFrobSimplAt 1.
        eapply ProgramStepStepsImplSteps.
        -- econstructor.
        -- assumption.
      * progFrobSimplAt 1.
        eapply ProgramStepStepsImplSteps.
        -- econstructor.
           eassumption.
        -- auto.
      * progFrobSimplAt 1.
        eapply ProgramStepStepsImplSteps.
        -- econstructor.
        -- auto.
  - progFrobSimplAt 1.
    eapply ProgramStepStepsImplSteps.
    + econstructor.
    + assumption.
Qed.

Lemma flattenProgramSimulates_r Method Method' (obj : Spec Method) (implProg : forall Ret, Method' Ret -> Program Method Ret) st :
  Simulates_
    (programSpec
       (mapSpec (programSpec obj) implProg))
    (mapSpec
       (programSpec obj)
       (@flattenProgram _ _ implProg))
    st st.
Proof.
  eapplyRelationSimulates (@eq (State obj)).
  intros * H1 H2.
  - subst.
    apply programSpecSteps in H2.
    progFrobSimplInAt H2 3.
    simpl.
    eauto.
  - reflexivity.
Qed.

Theorem programPairSpec ML MR
        (objL : Spec ML) (objR : Spec MR)
        (st : objL.(State) * objR.(State)) :
  Simulates_
    (mapSpec
       (programSpec
          (pairSpec objL objR))
       (fun R p =>
          match p with
          | inl p' => mapProgram (fun _ => inl) p'
          | inr p' => mapProgram (fun _ => inr) p'
          end))
    (pairSpec (programSpec objL) (programSpec objR))
    st st.
Proof.
  eapplyRelationSimulates eq.
  - simpl.
    intros.
    subst.
    destruct st as [stL stR].
    destruct int.
    + simpl.
      remember (mapProgram _ _) as p1.
      revert p Heqp1.
      induction H0 using EvalProgramInd;
        intros q Heqp1; symmetry in Heqp1; apply mapProgramInv in Heqp1.
      * subst.
        eexists.
        -- split.
           ++ apply ProgramStepsRefl.
           ++ reflexivity.
        -- reflexivity.
      * destruct Heqp1 as (? & ? & ? & ? & ?).
        subst.
        destruct H.
        edestruct IHEvalProgram as [? [? ?] ?].
        -- reflexivity.
        -- eexists.
           ++ split.
              ** eapply ProgramStepStepsImplSteps.
                 --- constructor.
                     eassumption.
                 --- eassumption.
              ** congruence.
           ++ assumption.
      * destruct Heqp1 as (? & ? & ?).
        subst.
        edestruct IHEvalProgram as [? [? ?] ?].
        -- reflexivity.
        -- eexists.
           ++ split.
              ** eapply ProgramStepStepsImplSteps.
                 --- constructor.
                 --- eassumption.
              ** assumption.
           ++ assumption.
    + simpl.
      remember (mapProgram _ _) as p1.
      revert p Heqp1.
      induction H0 using EvalProgramInd;
        intros q Heqp1; symmetry in Heqp1; apply mapProgramInv in Heqp1.
      * subst.
        eexists.
        -- split.
           ++ apply ProgramStepsRefl.
           ++ reflexivity.
        -- reflexivity.
      * destruct Heqp1 as (? & ? & ? & ? & ?).
        subst.
        destruct H.
        edestruct IHEvalProgram as [? [? ?] ?].
        -- reflexivity.
        -- eexists.
           ++ split.
              ** eapply ProgramStepStepsImplSteps.
                 --- constructor.
                     eassumption.
                 --- eassumption.
              ** congruence.
           ++ assumption.
      * destruct Heqp1 as (? & ? & ?).
        subst.
        edestruct IHEvalProgram as [? [? ?] ?].
        -- reflexivity.
        -- eexists.
           ++ split.
              ** eapply ProgramStepStepsImplSteps.
                 --- constructor.
                 --- eassumption.
              ** assumption.
           ++ assumption.
  - reflexivity.
Qed.

Lemma EvalInstruction_superposition_Noop {M} (o : Spec M)
      {R} (p : Program M R) (r : R)
      (ss2 ss2' : State o -> Prop)
  : EvalInstruction (superposition (programSpec o)) p ss2 r ss2' ->
    EvalInstruction (superposition (programSpec o)) (Noop p) ss2 r ss2'.
Proof.
  cbn.
  intros []; split; auto; intros.
  destruct (H0 _ H1) as (? & ? & ?).
  eexists; split; eauto.
  apply EvalProgram_Noop.
  assumption.
Qed.

Lemma RSimulates_eutt_EvalProgram {M1 M2} (RR : IRel M1 M2)
      (o1 : Spec M1) (o2 : Spec M2)
      s1 s1' R p1 (r : R)
  : EvalProgram o1 R p1 s1 r s1' ->
    forall p2,
    eutt RR p1 p2 ->
    forall ss2,
    (exists s2, ss2 s2) ->
    RSimulates_ RR o1 (superposition o2) s1 ss2 ->
    exists ss2',
    RSimulates_ RR o1 (superposition o2) s1' ss2' /\
    EvalInstruction (superposition (programSpec o2)) p2 ss2 r ss2'.
Proof.
  induction 1 using EvalProgramInd;
    intros p2 Heutt.
  - apply eutt_Ret_inv in Heutt. induction Heutt.
    + intros ss2 Hss2 Hss2'. specialize (IHHeutt ss2 Hss2 Hss2').
      destruct IHHeutt as (ss2' & ? & ?).
      exists ss2'; split; [ eauto | ].
      apply EvalInstruction_superposition_Noop.
      assumption.
    + destruct H. intros.
      exists ss2. split; auto. cbn. split; auto. intros.
      exists st'. split; auto. apply ProgramStepsRefl.
  - apply eutt_Bind_inv in Heutt.
    destruct Heutt as (m2 & k2 & Rm2 & EUTTk2 & Heutt).
    induction Heutt.
    + intros ss2 Hss2 Hss2'. specialize (IHHeutt ss2 Hss2 Hss2').
      destruct IHHeutt as (ss2' & ? & ?).
      exists ss2'; split; [ eauto | ].
      apply EvalInstruction_superposition_Noop.
      assumption.
    + destruct H0. intros.
      destruct H1.
      eapply H1 in Rm2; [ | eassumption ].
      destruct Rm2 as [ ? ? ? ].
      apply (IHEvalProgram (k2 r)) in H3;
        [ | auto | eauto using EvalInstruction_superposition_nonempty ].
      destruct H3 as (ss2' & []).
      exists ss2'.
      split; auto.
      split; [ eauto using EvalInstruction_superposition_nonempty |].
      intros.
      destruct H4.
      apply H6 in H5.
      destruct H5 as (? & ? & ?).
      destruct H2.
      apply H8 in H5.
      destruct H5 as (? & ? & ?).
      exists x1; split; auto.
      cbn. red. eapply ProgramStepStepsImplSteps.
      * constructor. eauto.
      * auto.
  - apply IHEvalProgram.
    apply eutt_Noop_L_inv. assumption.
Qed.

(************)

Lemma Proper_programSpec {M} (o1 o2 : Spec M) s1 s2
  : HRSimulates_ ieq o1 o2 s1 s2 ->
    HRSimulates_ (eutt ieq) (programSpec o1) (programSpec o2) s1 s2.
Proof.
  unfold HRSimulates.
  enough (H : forall s1 ss2,
    (exists s2, ss2 s2) ->
    RSimulates_ ieq o1 (superposition o2) s1 ss2 ->
    RSimulates_ (eutt ieq) (programSpec o1) (superposition (programSpec o2)) s1 ss2).
  { apply H; auto. exists s2. constructor. }
  clear.
  cofix self. intros. constructor. intros. cbn in H2.
  eapply RSimulates_eutt_EvalProgram in H1; eauto.
  destruct H1 as (? & ? & ?); eauto.
  eexists; eauto.
  eapply self; eauto.
  eapply EvalInstruction_superposition_nonempty in H3; eauto.
Qed.
