(** Heterogeneous simulation, where the specs may not have the same
    interface. *)

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  Object
  Superposition
  Abortable.

Section Simulates.

  Context {I1 I2 : Type -> Type}.
  Context (RR : IRel I1 I2)
          (spec1 : Spec I1) (spec2 : Spec I2).

  CoInductive RSimulates_ (s1 : spec1.(State)) (s2 : spec2.(State)) : Prop :=
  | RSimStep :
      (forall R i1 i2 (r : R) s1',
         RR _ i1 i2 ->
         spec1.(EvalInstruction) i1 s1 r s1' ->
         exists2 s2',
           spec2.(EvalInstruction) i2 s2 r s2' &
           RSimulates_ s1' s2') ->
      RSimulates_ s1 s2.

  Definition RSimulates : Prop := RSimulates_ spec1.(initialState) spec2.(initialState).

End Simulates.

Definition HRSimulates_ {I1 I2} (RR : IRel I1 I2) spec1 spec2 s1 s2 :=
  RSimulates_ RR spec1 (superposition spec2) s1 (eq s2).

Definition HRSimulates {I1 I2} (RR : IRel I1 I2) spec1 spec2 :=
  RSimulates RR spec1 (superposition spec2).

Inductive eq_Abortable {M} (RR : IRel M M) : forall R (p1 p2 : Abortable M R), Prop :=
| lift_eq_Abortable {R} u1 u2
  : RR _ u1 u2 -> eq_Abortable RR (option R) (abortable u1) (abortable u2).
