From C4.Util Require Import
  Util
  LibTactics.

From C4.Core Require Import
  Program
  Object
  Simulates
  Superposition
  Abortable
  RSimulates.

Lemma RSimulates_Simulates {I} (o1 o2 : Spec I)
  : forall s1 s2, RSimulates_ ieq o1 o2 s1 s2 -> Simulates_ o1 o2 s1 s2.
Proof.
  cofix self.
  intros; constructor; intros.
  destruct H. edestruct H; eauto.
Qed.

Lemma Simulates_RSimulates {I} (o1 o2 : Spec I)
  : forall s1 s2, Simulates_ o1 o2 s1 s2 -> RSimulates_ ieq o1 o2 s1 s2.
Proof.
  cofix self.
  intros; constructor; intros.
  destruct H0.
  destruct H. edestruct H; eauto.
Qed.

Lemma RSimulates_ieq {M} (o : Spec M) : forall s, RSimulates_ ieq o o s s.
Proof.
  cofix self.
  constructor; intros.
  destruct H.
  eauto.
Qed.

Lemma HRSimulates_ieq {M} (o : Spec M) : HRSimulates ieq o o.
Proof.
  enough (forall s (ss : State o -> Prop),
             ss s ->
             RSimulates_ ieq o (superposition o) s ss).
  { apply H. constructor. }
  cofix self. constructor; intros.
  destruct H0.
  exists (fun s_ => exists s0_, ss s0_ /\ EvalInstruction o i1 s0_ r s_).
  - constructor; eauto.
  - eauto.
Qed.

Lemma RSimulates_HRSimulates {M1 M2} (RR : IRel M1 M2) (o1 : Spec M1) (o2 : Spec M2)
  : RSimulates RR o1 o2 -> HRSimulates RR o1 o2.
Proof.
  enough (forall s1 s2 (ss2 : _ -> Prop),
             ss2 s2 ->
             RSimulates_ RR o1 o2 s1 s2 ->
             RSimulates_ RR o1 (superposition o2) s1 ss2).
  { apply H. constructor. }
  cofix self. constructor; intros.
  destruct H0. edestruct H0; eauto.
  exists (fun s22 => x = s22).
  - constructor. eauto. intros ? []. eauto.
  - eauto.
Qed.

Lemma Proper_mapSpec {M N}
      (R : IRel N N)
      (o1 o2 : Spec N)
      (f1 f2 : forall T, M T -> N T)
  : (ieq ==> R)%isig f1 f2 ->
    HRSimulates R o1 o2 ->
    HSimulates (mapSpec o1 f1) (mapSpec o2 f2).
Proof.
  intros Hf.
  unfold HRSimulates, HSimulates.
  enough (forall s1 ss2,
             RSimulates_ R o1 (superposition o2) s1 ss2 ->
             Simulates_ (mapSpec o1 f1) (superposition (mapSpec o2 f2)) s1 ss2).
  { apply H. }
  cofix self; intros.
  constructor; intros.
  destruct H. edestruct H.
  - eapply Hf. constructor.
  - eapply H0.
  - eexists; [ eapply H1 | eauto ].
Qed.

Lemma Proper_abortableSpec {M} (R : IRel M M) (o1 o2 : Spec M)
  : HRSimulates R o1 o2 ->
    HRSimulates (eq_Abortable R) (abortableSpec o1) (abortableSpec o2).
Proof.
  unfold HRSimulates.
  enough (forall s1 ss2,
             (exists s2, ss2 s2) ->
             RSimulates_ R o1 (superposition o2) s1 ss2 ->
             RSimulates_ (eq_Abortable R) (abortableSpec o1) (superposition (abortableSpec o2)) s1 ss2).
  { apply H. eexists _. constructor. }
  cofix self; intros.
  constructor; intros.
  destruct H1.
  inversion H2.
  - injpair_. subst. clear H4 H3.
    destruct H0. edestruct H0; eauto. cbn in H3. destruct H3.
    eexists.
    + split; [ apply H3 | ].
      intros. edestruct H5 as (? & ? & ?). eauto.
      eexists; split; eauto.
      constructor. auto.
    + eauto.
  - injpair_. subst. clear H5 H4.
    exists ss2.
    + constructor; eauto.
      intros. eexists; split; eauto. constructor.
    + auto.
Qed.      

Lemma Proper_mapAbortable {M N} (r : IRel N N) (f1 f2 : forall R, M R -> N R)
  : (ieq ==> r)%isig f1 f2 ->
    (ieq ==> eq_Abortable r)%isig (mapAbortable f1) (mapAbortable f2).
Proof.
  intros Hf ? [] ? [].
  constructor.
  apply Hf. constructor.
Qed.

Lemma mapSpecRSimulates {M N} (f : forall R, M R -> N R)
      (o : Spec N) s
  : RSimulates_ (irel_fun f) o (mapSpec o f) s s.
Proof.
  revert s. cofix self. constructor.
  intros. cbn. destruct H. eauto.
Qed.

Lemma mapSpecHRSimulates {M N} (f : forall R, M R -> N R)
      (o : Spec N)
  : HRSimulates (irel_fun f) o (mapSpec o f).
Proof.
  apply RSimulates_HRSimulates, mapSpecRSimulates.
Qed.
