From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  ProgramFacts
  ProgramOps
  ProgramOpsFacts
  Simulates
  SimulatesFacts
  FlattenSimulates
  Program
  Object.


Module SeqRefCoInd.
Section SeqRefCoInd.

  Context
    {M: Type -> Type}
    (cObj aObj: Spec M).

  Variable DRel: cObj.(State) -> aObj.(State) -> Prop.
  Hypothesis initRel: DRel (initialState cObj) (initialState aObj).

  Hypothesis stepRel:
    forall R (m: M R) s1 s1' r s2,
      DRel s1 s2 ->
      cObj.(EvalInstruction) m s1 r s1' ->
      exists s2',
        aObj.(EvalInstruction) m s2 r s2'
        /\ DRel s1' s2'.
  
  Theorem simulatesSerImplCoInd:
    Simulates
      cObj aObj.
  Proof.
    apply relationSimulates with (Rel:=DRel).
    - intros.
      edestruct stepRel as [? [? ?]].
      + eassumption.
      + eassumption.
      + eexists; eassumption.
    - assumption.
  Qed.
      
End SeqRefCoInd.
End SeqRefCoInd.

Module SeqRefCoInd2.
Section SeqRefCoInd2.

  Import SeqRefCoInd.
  
  Context
    {M: Type -> Type}
    (obj: Spec M)
    (impl: Object M).

  Variable DRel: impl.(objBase).(State) -> obj.(State) -> Prop.
  Hypothesis initRel: DRel impl.(objBase).(initialState) (initialState obj).

  Hypothesis stepRel:
    forall R (m: M R) s1 s1' r s2,
      DRel s1 s2 ->
      EvalProgram _ _ (impl.(objProgram) _ m) s1 r s1' ->
      exists s2',
        obj.(EvalInstruction) m s2 r s2'
        /\ DRel s1' s2'.
  
  Theorem simulatesSerImplCoInd:
    Simulates (coarsen impl) obj.
  Proof.
    eapply simulatesSerImplCoInd; eassumption.
  Qed.
      
End SeqRefCoInd2.
End SeqRefCoInd2.

Module SeqRefCoInd3.
Section SeqRefCoInd3.

  Import SeqRefCoInd.
  
  Context
    {M: Type -> Type}
    (impl: Object M)
    (obj: Spec M).

  Variable DRel: obj.(State) -> impl.(objBase).(State) -> Prop.
  Hypothesis initRel: DRel (initialState obj) impl.(objBase).(initialState).

  Hypothesis stepRel:
    forall R (m: M R) s1 s1' r s2,
      DRel s1 s2 ->
      obj.(EvalInstruction) m s1 r s1' ->
      exists s2',
        EvalProgram _ _ (impl.(objProgram) _ m) s2 r s2'
        /\ DRel s1' s2'.
  
  Theorem simulatesSerImplCoInd:
    Simulates
      obj (coarsen impl).
  Proof.
    eapply simulatesSerImplCoInd; eassumption.
  Qed.
      
End SeqRefCoInd3.
End SeqRefCoInd3.

Module SeqRefCoInd4.
Section SeqRefCoInd4.

  Import SeqRefCoInd.
  
  Context
    {M: Type -> Type}
    (impl impl':Object M).

  Variable DRel: impl.(objBase).(State) -> impl'.(objBase).(State) -> Prop.
  Hypothesis initRel: DRel impl.(objBase).(initialState) impl'.(objBase).(initialState).

  Hypothesis stepRel:
    forall R (m: M R) s1 s1' r s2,
      DRel s1 s2 ->
      EvalProgram _ _ (impl.(objProgram) _ m) s1 r s1' ->
      exists s2',
        EvalProgram _ _ (impl'.(objProgram) _ m) s2 r s2'
        /\ DRel s1' s2'.
  
  Theorem simulatesSerImplCoInd:
    Simulates
      (coarsen impl) (coarsen impl').
  Proof.
    eapply simulatesSerImplCoInd; eassumption.
  Qed.
      
End SeqRefCoInd4.
End SeqRefCoInd4.

Lemma serialSeqRef:
  forall I (obj: Spec I),
    Simulates
      obj (coarsen (serialObject obj)).
Proof.
  intros.
  eapply SeqRefCoInd3.simulatesSerImplCoInd.
  - exact eq_refl.
  - introv H1. subst s2. eexists. simpl. splits. unfold EvalProgram.
    eapply ProgramStepStepsImplSteps. econstructor. eassumption.
    eapply ProgramStepsRefl. reflexivity.
Qed.

Lemma serialSeqRef':
  forall I (impl: Object I) (obj: Spec I),
    Simulates (coarsen impl) obj ->
    Simulates (coarsen impl) (coarsen (serialObject obj)).
Proof.
  introv H1.
  eapply simulatesTrans.
  eassumption.
  eapply serialSeqRef.
Qed.

Lemma implNEqSeqRef:
  forall {Instruction : Type -> Type}
    (obj: Spec Instruction)
    (M: Type -> Type)
    (impl1 impl2: forall R, M R -> Program Instruction R),
    implNEq impl1 impl2 ->
    Simulates
      (coarsen {| objBase := obj; objProgram := impl1 |})
      (coarsen {| objBase := obj; objProgram := impl2 |}).
Proof.
  introv H1.
  eapply SeqRefCoInd4.simulatesSerImplCoInd.
  - exact eq_refl.
  - simpl in *. introv H2 H3. subst s2.
    eexists.
    unfold implNEq in H1. specialize (H1 R m).
    splits.
    forwards (p & A1 & A2): programNEqProgramSteps H1 H3. inverts A2.
    eassumption.
    reflexivity.
Qed.


Lemma flattenSerialSeqRef:
  forall {M'} obj M (impl: forall R, M R -> Program M' R),
    Simulates
      (coarsen (flattenObject (serialObject obj) _ impl))
      (coarsen {| objBase := obj; objProgram := impl |}).

Proof.
  intros.
  unfold serialObject. unfold flattenObject. simpl in *.
  eapply implNEqSeqRef.
  unfold implNEq.
  intros.
  apply FlattenBindRetNEq.
Qed.

  
Lemma seqRefSubst:
  forall I (impl1: Object I) (base2: Spec I) M (implProg: forall R, M R -> Program I R),
    Simulates (coarsen impl1) base2 ->
    Simulates (coarsen (flattenObject impl1 M implProg)) (coarsen (object base2 implProg)).

Proof.
  introv H1.
  eapply simulatesTrans.
  - eapply flattenSeqRef.
    eapply serialSeqRef'. eassumption.  
  - eapply flattenSerialSeqRef.
Qed.
