Require Import
        Coq.Program.Equality.

From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts.


Section Sig.

  Context {Method : Type -> Type}
          {T : Type}
          (obj : T -> Spec Method).

  Definition SigState := {t : T & (obj t).(State)}.

  Inductive EvalSigInstruction :
    forall Ret, Method Ret -> SigState -> Ret -> SigState -> Prop :=
  | evalSigInstruction t Ret m st ret st' :
      (obj t).(EvalInstruction) m st ret st' ->
      EvalSigInstruction Ret m (existT _ t st) ret (existT _ t st').

  Definition sigSpec {t : T} (initial : (obj t).(State)) : Spec Method :=
    Build_Spec Method SigState EvalSigInstruction (existT _ t initial).

  Theorem sigSimulates t (initial s : (obj t).(State)) :
    Simulates_ (sigSpec initial) (obj t) (existT _ t s) s.
  Proof.
    eapplyRelationSimulates (fun st1 st2 => st1 = existT _ t st2).
    - intros A m st1 a st1' st2 H1 H2.
      destruct H2.
      dependent destruction H1.
      eexists.
      + eassumption.
      + reflexivity.
    - reflexivity.
  Qed.

  Theorem sigSimulates_r t (initial s : (obj t).(State)) :
    Simulates_ (obj t) (sigSpec initial) s (existT _ t s).
  Proof.
    eapplyRelationSimulates (fun st1 st2 => existT _ t st1 = st2).
    - intros A m st1 a st1' st2 H1 H2.
      subst.
      eexists.
      + split.
        eassumption.
      + reflexivity.
    - reflexivity.
  Qed.

End Sig.
