From Coq Require Import JMeq.
From C4.Util Require Import
  LibTactics
  MTactics.
From C4.Core Require Import
  Program
  Simulates.

Definition JMeq_eq {a} (x y : a) : JMeq x y -> x = y.
Proof.
  inversion 1; injpair_. assumption.
Qed.

Section RelationSimulates.

  Context {Interaction : Type -> Type}
          (spec1 : Spec Interaction)
          (spec2 : Spec Interaction)
          (Rel : spec1.(State) -> spec2.(State) -> Prop)
          (step : forall A int s1 (a : A) s1' s2,
              Rel s1 s2 ->
              spec1.(EvalInstruction) int s1 a s1' ->
              exists2 s2',
                spec2.(EvalInstruction) int s2 a s2' &
                Rel s1' s2').

  CoFixpoint relationSimulates s1 s2 (r : Rel s1 s2) :
    Simulates_ spec1 spec2 s1 s2.
  Proof.
    split.
    intros.
    edestruct step.
    - eassumption.
    - eassumption.
    - eexists; eauto.
  Qed.

End RelationSimulates.

Tactic Notation "eapplyRelationSimulates" uconstr(Rel) :=
  match goal with
    [ |- Simulates_ ?obj1 ?obj2 ?st1 ?st2 ] =>
    eapply (relationSimulates obj1 obj2 Rel)
  end.

CoFixpoint simulatesRefl {Method} (spec: Spec Method) (ts: spec.(State)) :
  Simulates_ spec spec ts ts.
Proof.
  split.
  intros.
  eexists.
  - eassumption.
  - apply simulatesRefl.
Qed.

CoFixpoint simulatesTrans_ {Method}
           (spec1 spec2 spec3 : Spec Method)
           (ts1: spec1.(State))
           (ts2: spec2.(State))
           (ts3: spec3.(State)):
  Simulates_ spec1 spec2 ts1 ts2 ->
  Simulates_ spec2 spec3 ts2 ts3 ->
  Simulates_ spec1 spec3 ts1 ts3.
Proof.
  intros [S] [S'].
  split.
  intros.
  edestruct S.
  - eassumption.
  - edestruct S'.
    + eassumption.
    + eexists.
      * eassumption.
      * eapply simulatesTrans_; eassumption.
Qed.

Definition simulatesTrans {Method} (spec1 spec2 spec3 : Spec Method) :
  Simulates spec1 spec2 -> Simulates spec2 spec3 -> Simulates spec1 spec3.
Proof.
  apply simulatesTrans_.
Qed.

Theorem simulatesJMeq
        {Method} (spec spec' : Spec Method)
        (ts : spec.(State))
        (ts' : spec'.(State)) :
  spec = spec' ->
  JMeq ts ts' ->
  Simulates_ spec spec' ts ts'.
Proof.
  intros H; revert ts'; destruct H; intros ts' H. apply JMeq_eq in H. subst.
  apply simulatesRefl.
Qed.

Definition FSimulates
           {M} (o1 o2: Spec M)
           (rel : o1.(State) -> o2.(State) -> Prop)
           (s1i : o1.(State)) (s2i : o2.(State)) : Prop :=

  rel s1i s2i
  /\ (forall R s1 s2 int (r : R) s1',
        rel s1 s2 ->
        o1.(EvalInstruction) int s1 r s1' ->
        exists s2', o2.(EvalInstruction) int s2 r s2'
               /\ rel s1' s2').

Lemma ForSimulatesSimulates:
  forall {M} (o1 o2: Spec M) (s1 : o1.(State)) (s2 : o2.(State)) rel,
    FSimulates o1 o2 rel s1 s2 ->
    Simulates_ o1 o2 s1 s2.
Proof.
  introv H1.
  unfold FSimulates in *. destruct H1 as [H1 H2].
  eapply (relationSimulates _ _ rel).
  - intros. forwardsIn H2; try eassumption. destruct H2 as (s2' & H21 & H22).
    eexists; eassumption.
  - assumption.
Qed.

Lemma SimulatesForSimulates:
  forall {M} (o1 o2: Spec M) (s1 : o1.(State)) (s2 : o2.(State)),
    Simulates_ o1 o2 s1 s2 ->
    FSimulates o1 o2 (@Simulates_ _ _ _) s1 s2.
Proof.
  intros.
  unfold FSimulates. splits.
  - assumption.
  - introv H1 H2. inverts H1 as H1. eapply H1 in H2.
    inverts H2 as H2. eauto.
Qed.
