From C4.Core Require Import
  Program
  Simulates
  SimulatesFacts.

Definition superposition {Method} (obj : Spec Method) :
  Spec Method :=
  {|
    State := obj.(State) -> Prop;
    EvalInstruction A m sst a sst' :=
      ex sst' /\
      (forall st',
          sst' st' -> exists st,
            sst st /\ obj.(EvalInstruction) m st a st');
    initialState := eq obj.(initialState)
  |}.

Definition HSimulates_ {M: Type -> Type} (obj1 : Spec M) (obj2 : Spec M) s1 s2 : Prop :=
    Simulates_ obj1 (superposition obj2) s1 s2.

Definition HSimulates {M: Type -> Type} (obj1 : Spec M) (obj2 : Spec M) : Prop :=
    Simulates obj1 (superposition obj2).

Theorem superpositionSingleSimulates {Method} (obj : Spec Method) st (sst : obj.(State) -> Prop) :
  sst st ->
  Simulates_ obj (superposition obj) st sst.
Proof.
  intro.
  apply relationSimulates with (fun st sst => sst st); repeat esplit; eauto.
Qed.

Theorem superpositionSubsetSimulates {Method} (obj : Spec Method) (sst1 sst2 : obj.(State) -> Prop) :
  (forall st, sst1 st -> sst2 st) ->
  Simulates_ (superposition obj) (superposition obj) sst1 sst2.
Proof.
  eapply
    (relationSimulates
       (superposition obj) (superposition obj)
       (fun sst1 sst2 => forall st, sst1 st -> sst2 st)).
  clear sst1 sst2.
  intros A int sst1 a sst1' sst2 H1 [H2 H3].
  exists sst1'.
  - split.
    + eassumption.
    + intros.
      edestruct H3 as [? [? ?]]; eauto.
  - auto.
Qed.

Theorem superpositionSimulates {Method}
        (obj1 obj2 : Spec Method)
        (sst1 : obj1.(State) -> Prop)
        (sst2 : obj2.(State) -> Prop) :
  (forall st1, sst1 st1 -> Simulates_ obj1 (superposition obj2) st1 sst2) ->
  Simulates_ (superposition obj1) (superposition obj2) sst1 sst2.
Proof.
  intros H2.
  eapplyRelationSimulates
    (fun sst1 sst2 =>
       forall st1, sst1 st1 -> Simulates_ obj1 (superposition obj2) st1 sst2).
  - clear sst1 sst2 H2.
    intros A int sst1 a sst1' sst2 H2 [H3 H4].
    exists (fun st2' => exists st2, sst2 st2 /\ (obj2).(EvalInstruction) int st2 a st2').
    + split.
      * edestruct H3.
        edestruct H4 as [? [? ?]]; [eassumption |].
        edestruct H2 as [H21]; [eassumption |].
        edestruct H21 as [? [[? ?] ?] ?]; eauto.
      * intros.
        assumption.
    + intros st1' Hst1'.
      edestruct H4 as [? [? ?]]; [eassumption |].
      edestruct H2 as [H21]; [eassumption |].
      edestruct H21 as [? [? ?] ?]; [eassumption |].
      eapply simulatesTrans_.
      * eassumption.
      * apply superpositionSubsetSimulates.
        eauto.
  - assumption.
Qed.

Theorem relationBackwardSimulates
        {Method : Type -> Type}
        (obj1 obj2 : Spec Method)
        (Rel : obj1.(State) -> obj2.(State) -> Prop)
        (step :
           forall A int s1 (a : A) s1',
             ex (Rel s1) ->
             obj1.(EvalInstruction) int s1 a s1' ->
             ex (Rel s1') /\
             (forall s2',
                 Rel s1' s2' ->
                 exists s2, Rel s1 s2 /\ EvalInstruction obj2 int s2 a s2'))
        (initS1 : obj1.(State)) :
  ex (Rel initS1) ->
  Simulates_ obj1 (superposition obj2) initS1 (Rel initS1).
Proof.
  intros initR.
  apply relationSimulates with
  (fun s1 ss2 => ex (Rel s1) /\ forall s2, ss2 s2 <-> Rel s1 s2).
  - intros A int s1 a s1' ss2 [I Hss2] E.
    edestruct step as [I' S].
    + exact I.
    + eassumption.
    + exists (Rel s1'); firstorder.
  - tauto.
Qed.
