Require Import
        Coq.Lists.List
        Coq.Relations.Relations.

From C4.Util Require Import Util.
From C4.Core Require Import
  Simulates
  Program
  Object.

Section Threads.

  Context
    {Method : Type -> Type}
    (impl : Object Method).

  Inductive Interaction : Type -> Type :=
  | Spawn {Ret} (m : Method Ret) : Interaction unit
  | Join {Ret} : Interaction Ret
  | Up {Ret} : Interaction Ret -> Interaction Ret.
  Arguments Interaction : clear implicits.

  Definition ThreadsState : Type :=
    (list (sigT (Program impl.(ObjInstruction))) * impl.(objBase).(State))%type.

  Inductive ThreadAdvance :
    ThreadsState -> ThreadsState -> Prop :=
  | TACons p ps st ps' st' :
      ThreadAdvance (ps, st) (ps', st') ->
      ThreadAdvance (p :: ps, st) (p :: ps', st')
  | TABind Ret ps st A instr f (a : A) st' :
      impl.(objBase).(EvalInstruction) instr st a st' ->
      ThreadAdvance (existT _ Ret (Bind instr f) :: ps, st) (existT _ Ret (f a) :: ps, st')
  | TANoop Ret ps st p :
      ThreadAdvance (existT _ Ret (Noop p) :: ps, st) (existT _ Ret p :: ps, st).

  Definition ThreadsAdvance := clos_refl_trans_1n ThreadsState ThreadAdvance.

  Inductive ThreadsInteraction' :
    forall Ret,
      Interaction Ret ->
      list (sigT (Program impl.(ObjInstruction))) ->
      Ret ->
      list (sigT (Program impl.(ObjInstruction))) ->
      Prop :=
  | TISpawn Ret m ps :
      ThreadsInteraction' unit (Spawn m) ps tt (existT _ Ret (impl.(objProgram) Ret m) :: ps)
  | TIJoin Ret ret ps :
      ThreadsInteraction' Ret Join (existT _ Ret (Return ret) :: ps) ret ps
  | TICons Ret' int p ps ret ps' :
      ThreadsInteraction' Ret' int ps ret ps' ->
      ThreadsInteraction' Ret' (Up int) (p :: ps) ret (p :: ps').

  Inductive ThreadsInteraction Ret (int : Interaction Ret)
    : ThreadsState ->
      Ret ->
      ThreadsState ->
      Prop :=
  | TI ps a ps' st :
      ThreadsInteraction' Ret int ps a ps' ->
      ThreadsInteraction Ret int (ps, st) a (ps', st).

  Definition ThreadsBehavior : Spec Interaction :=
    {|
      State := ThreadsState;
      EvalInstruction Ret int ts ret ts'' :=
        exists2 ts' : ThreadsState,
          ThreadsAdvance ts ts' &
          ThreadsInteraction Ret int ts' ret ts'';
      initialState := (nil, impl.(objBase).(initialState))
    |}.

End Threads.

Arguments Interaction : clear implicits.

Inductive eq_Interaction {M N} (RR : IRel M N)
  : IRel (Interaction M) (Interaction N) :=
| eq_Interaction_Spawn R m n
  : RR R m n -> eq_Interaction RR unit (Spawn m) (Spawn n)
| eq_Interaction_Join R : eq_Interaction RR R Join Join
| eq_Interaction_Up R m n
  : eq_Interaction RR R m n -> eq_Interaction RR R (Up m) (Up n)
.

Definition ConcSimulates {M} (impl impl': Object M) :=
  Simulates (ThreadsBehavior impl) (ThreadsBehavior impl').
