Require Import
        Coq.Logic.Eqdep
        Coq.Lists.List
        Coq.omega.Omega
        Coq.Relations.Relations
        Coq.Program.Equality.
From C4.Util Require Import
  LibTactics.
From C4.Core Require Import
  Program
  ProgramFacts
  Object
  Simulates
  SimulatesFacts
  Threads.

Lemma clos_rt1n_trans A R : transitive A (clos_refl_trans_1n A R).
Proof.
  induction 1; intros.
  - assumption.
  - right with y; tauto.
Qed.

Lemma concRefRefl:
  forall M (impl: Object M),
    ConcSimulates impl impl.
Proof.
  intros.
  apply simulatesRefl.
Qed.

CoFixpoint concRefTrans {M} (impl1: Object M) (impl2: Object M) (impl3: Object M) :
  ConcSimulates impl1 impl2 -> ConcSimulates impl2 impl3 -> ConcSimulates impl1 impl3.
Proof.
  intros.
  eapply simulatesTrans; eassumption.
Qed.

Lemma Reflexive_ThreadsAdvance {M} {i : Object M}
  : forall s, ThreadsAdvance i s s.
Proof.
  constructor.
Qed.

Global Hint Resolve Reflexive_ThreadsAdvance : core.

Section Steps.

  Context
    {Method : Type -> Type}
    (impl : Object Method).

  Theorem advanceCons p ps st ps' st' :
    ThreadsAdvance impl (ps, st) (ps', st') ->
    ThreadsAdvance impl (p :: ps, st) (p :: ps', st').
  Proof.
    intro ta.
    dependent induction ta.
    - constructor 1.
    - destruct y.
      econstructor 2.
      + apply (TACons impl).
        eassumption.
      + eapply IHta; auto.
  Qed.

  Theorem advanceInteract' A int a ps ps' ps'' st st' :
    ThreadsInteraction' impl A int ps a ps' ->
    ThreadAdvance impl (ps, st) (ps'', st') ->
    exists2 ps''',
      ThreadsInteraction' impl A int ps'' a ps''' &
      ThreadAdvance impl (ps', st) (ps''', st').
  Proof.
    intro ti.
    revert ps'' st st'.
    induction ti; intros ps'' st st'.
    - repeat econstructor; assumption.
    - intro ta.
      dependent destruction ta.
      repeat econstructor; assumption.
    - intro ta.
      dependent destruction ta.
      + edestruct IHti; repeat econstructor; eassumption.
      + repeat econstructor; eassumption.
      + repeat econstructor; assumption.
  Qed.

  Theorem advanceInteract A int a ts ts' ts'' :
    ThreadsInteraction impl A int ts a ts' ->
    ThreadAdvance impl ts ts'' ->
    exists2 ts''',
      ThreadsInteraction impl A int ts'' a ts''' &
      ThreadAdvance impl ts' ts'''.
  Proof.
    destruct 1, ts''.
    intro.
    edestruct advanceInteract'.
    - eassumption.
    - eassumption.
    - eexists.
      + split.
        eassumption.
      + assumption.
  Qed.

  Theorem interactDet' {A int ps a ps' ps''} :
    ThreadsInteraction' impl A int ps a ps' ->
    ThreadsInteraction' impl A int ps a ps'' ->
    ps' = ps''.
  Proof.
    intros ti.
    revert ps''.
    induction ti; intros ps'' ti'; dependent destruction ti'.
    - reflexivity.
    - reflexivity.
    - f_equal.
      apply IHti.
      assumption.
  Qed.

  Theorem interactDet {A int ts a ts' ts''} :
    ThreadsInteraction impl A int ts a ts' ->
    ThreadsInteraction impl A int ts a ts'' ->
    ts' = ts''.
  Proof.
    intros ti ti'.
    dependent destruction ti.
    dependent destruction ti'.
    f_equal.
    eapply interactDet'; eassumption.
  Qed.

  Theorem advancesInteract A int a ts ts' ts'' :
    ThreadsInteraction impl A int ts a ts' ->
    ThreadsAdvance impl ts ts'' ->
    exists2 ts''',
      ThreadsInteraction impl A int ts'' a ts''' &
      ThreadsAdvance impl ts' ts'''.
  Proof.
    intros ti ta.
    revert ts' ti.
    dependent induction ta.
    - eexists.
      + eassumption.
      + constructor.
    - intros ts' ti.
      edestruct advanceInteract.
      + exact ti.
      + eassumption.
      + edestruct IHta.
        * eassumption.
        * { eexists.
            - eassumption.
            - econstructor 2; eassumption.
          }
  Qed.

  Lemma advanceLength ts ts' :
    ThreadAdvance impl ts ts' ->
    length (fst ts) = length (fst ts').
  Proof.
    induction 1; simpl; auto.
  Qed.

  Lemma advancesLength ts ts' :
    ThreadsAdvance impl ts ts' ->
    length (fst ts) = length (fst ts').
  Proof.
    induction 1.
    - reflexivity.
    - etransitivity.
      + apply advanceLength.
        eassumption.
      + assumption.
  Qed.

End Steps.


Lemma invertThreadsInteraction {M} (impl: Object M) Ret int ts a ts':
  ThreadsInteraction impl Ret int ts a ts' ->
  ThreadsInteraction' impl Ret int (fst ts) a (fst ts')
  /\ snd ts = snd ts'.

Proof.
  introv H1.
  destruct H1.
  jauto.
Qed.

Section Simulates.

  Context
    {Method : Type -> Type}
    (impl : Object Method).

  Theorem simulateAdvance ts ts' :
    ThreadsAdvance impl ts ts' ->
    Simulates_ (ThreadsBehavior impl) (ThreadsBehavior impl) ts' ts.
  Proof.
    intros ta.
    split.
    intros A int a ts'' [ts''' ta' ti'].
    exists ts''.
    - esplit.
      * eapply clos_rt1n_trans; eassumption.
      * assumption.
    - apply simulatesRefl.
  Qed.

End Simulates.


Section Simulates2.

  Context
    {Method : Type -> Type}
    (impl1 impl2 : Object Method).

  Theorem simulateInteraction {A} int ts1 ts2 a ts1' ts2' :
    ThreadsInteraction impl1 A int ts1 a ts1' ->
    ThreadsInteraction impl2 A int ts2 a ts2' ->
    Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2)
              ts1 ts2 ->
    Simulates_ (ThreadsBehavior impl1) (ThreadsBehavior impl2)
              ts1' ts2'.
  Proof.
    intros ti1 ti2 [S].
    edestruct S as [ts2'' [ts2''' ta2 ti2'] S'].
    - esplit.
      * apply rt1n_refl.
      * eassumption.
    - edestruct (advancesInteract impl2) as [ts2'''' ti2'' S''].
      + exact ti2.
      + eassumption.
      + eapply simulatesTrans_.
        * eassumption.
        * eapply simulateAdvance.
          replace ts2'' with ts2'''' by (eapply interactDet; eassumption).
          assumption.
  Qed.


  (* To be moved *)
  Lemma clos_refl_trans_1n_trans:
    forall A R x y z,
      clos_refl_trans_1n A R x y ->
      clos_refl_trans_1n A R y z ->
      clos_refl_trans_1n A R x z.
  Proof.
    intros.
    apply clos_rt_rt1n_iff.
    eapply rt_trans; apply clos_rt_rt1n_iff; eassumption.
  Qed.


  Lemma ThreadAdvanceImplThreadsAdvance:
    forall {M} (impl : Object M) ts1 ts2,
      ThreadAdvance impl ts1 ts2 ->
      ThreadsAdvance impl ts1 ts2.

  Proof.
    intros. eapply clos_rt1n_step. assumption.
  Qed.

  Lemma ThreadsAdvanceTrans:
    forall {M} (impl : Object M) ts1 ts2 ts3,
      ThreadsAdvance impl ts1 ts2 ->
      ThreadsAdvance impl ts2 ts3 ->
      ThreadsAdvance impl ts1 ts3.
  Proof.
    intros. eapply clos_refl_trans_1n_trans; eassumption.
  Qed.

  Lemma ProgramStepImplThreadAdvance:
    forall {M} (impl: Object M) Ret p ps ds p' ds',
      ProgramStep (objBase impl) Ret p ds p' ds' ->
      ThreadAdvance impl (existT _ Ret p :: ps, ds) (existT _ Ret p' :: ps, ds').
  Proof.
    intros.
    destruct p; apply invertProgramStep in H.
    - destruct H as (r & H1 & H2). subst. econstructor. eassumption.
    - inversion H.
    - destruct H as (H1 & H2). subst. econstructor.
  Qed.

  Lemma ProgramStepsImplThreadsAdvance:
    forall {M} (impl: Object M) Ret p ps ds p' ds',
      ProgramSteps (objBase impl) Ret p ds p' ds' ->
      ThreadsAdvance impl (existT _ Ret p :: ps, ds) (existT _ Ret p' :: ps, ds').

  Proof.
    intros.
    unfold ProgramSteps in H.
    remember (p, ds) as s1 eqn: S1.
    remember (p', ds') as s2 eqn: S2.
    gen p ds p' ds'.

    induction H; intros; subst.

    - inversion S2. subst.
      econstructor.

    - destruct y as (pm & dsm).
      unfold ProgramStep' in H. simpl in H.
      eapply ProgramStepImplThreadAdvance in H.
      econstructor.
      eassumption.
      eapply IHclos_refl_trans_1n; f_equal.
  Qed.

  (* End to be moved *)

End Simulates2.

