(** Main theorems:

- [ConcSimulates_ConcTraceIncl]
- [ConcTrace_flatten]
- [ConcTrace_flatten_inv]
- [ConcTrace_flattenObject]
- [ConcTrace_flattenObject_inv]
- [ConcTrace_pairProgram]
- [ConcTrace_pairProgram_inv]
- [ConcTrace_pairObject]
- [ConcTrace_pairObject_inv]
- [ConcTraceSpec_pairSpec]
- [ConcTraceSpec_pairSpec_inv]

*)

From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  Trace.

From C4.Core Require Export
  Trace_Basics
  Trace_ConcSimulates
  Trace_SerialObject
  Trace_MapObject
  Trace_Flatten
  Trace_Pair
  Trace_Linearizable.

Set Implicit Arguments.
Set Contextual Implicit.
Unset Strict Implicit.

(* Lemma 4.1 *)
Lemma ConcTraceIncl_serialObject {M} (spec spec' : Spec M)
  : Simulates spec spec' -> ConcTraceIncl (serialObject spec) (serialObject spec').
Proof.
  intros H.
  apply ConcSimulates_ConcTraceIncl.
  apply simulatesHSimulates.
  apply baseSimulates.
  apply H.
Qed.

Lemma lin_sim_lin {M} (o : Object M) (spec spec' : Spec M)
  : ConcTraceIncl o (serialObject spec) -> Simulates spec spec' -> ConcTraceIncl o (serialObject spec').
Proof.
  intros H1 H2.
  eapply Transitive_ConcTraceIncl.
  - apply H1.
  - apply ConcTraceIncl_serialObject. apply H2.
Qed.
