From Paco Require Import paco.

From Coq Require Import
  List
  Arith
  Setoid
  Morphisms
  FunctionalExtensionality.

From C4.Util Require LibTactics.

Ltac injpair := LibTactics.injpair.

From C4.Util Require Import Util.
From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Threads
  Superposition
  Simulates
  BaseSimulates
  HConcSimulates
  RSimulates
  RSimulatesFacts
  Eutt
  ImplSimulates
  Trace
  Trace_Basics
  Trace_ConcSimulates
  Trace_Flatten.

Definition once {M N} (f : M ~> N) : M ~> Program N :=
  fun _ m => Bind (f _ m) Return.

Lemma HImplSimulates_mapObject_flattenObject {M N} (f : M ~> N) o
  : HImplSimulates (mapObject f o) (flattenObject o _ (once f)).
Proof.
  apply (mk_HImplSimulates (mapObject f o) (fI o M (once f)) (ieq (M := ObjInstruction o))).
  - cbn. apply HRSimulates_ieq.
  - red. intros ? e _ <-. cbn. unfold flattenImplProg, once, mapImplProg.
    apply Symmetric_eutt_ieq, flattenTrigger.
Qed.

Lemma HImplSimulates_flattenObject_mapObject {M N} (f : M ~> N) o
  : HImplSimulates (flattenObject o _ (once f)) (mapObject f o).
Proof.
  apply (mk_HImplSimulates (fI o M (once f)) (mapObject f o) (ieq (M := ObjInstruction o))).
  - cbn. apply HRSimulates_ieq.
  - red. intros ? e _ <-. cbn. unfold flattenImplProg, once, mapImplProg.
    apply flattenTrigger.
Qed.

Lemma ConcTraceIncl_mapObject {M N} (f : M ~> N) o1 o2
  : ConcTraceIncl o1 o2 -> ConcTraceIncl (mapObject f o1) (mapObject f o2).
Proof.
  intros H.
  eapply Transitive_ConcTraceIncl;
   [ apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq, HImplSimulates_mapObject_flattenObject | ].
  eapply Transitive_ConcTraceIncl;
   [ | apply ConcSimulates_ConcTraceIncl, HConcSimulatesEq, HImplSimulates_flattenObject_mapObject].
  apply ConcTraceIncl_flatten.
  assumption.
Qed.
