Require Import
        Coq.Lists.List
        Coq.Relations.Relations.

From C4.Core Require Import
  Program
  Object
  Simulates
  SimulatesFacts
  Threads
  Superposition
  MapState.


Section MapStateBackward.

  Context
    {Method : Type -> Type}
    (impl : Object Method)
    (obj : Spec (Interaction Method))
    (MapState : ThreadsState impl -> obj.(State) -> Prop).

  Definition MapState' ts1 sst2 :=
    ex (MapState ts1) /\ forall st2, sst2 st2 <-> MapState ts1 st2.

  Context
    (advance :
       forall ts1 ts1',
         ex (MapState ts1) ->
         ThreadAdvance impl ts1 ts1' ->
         ex (MapState ts1') /\
         (forall st2',
             MapState ts1' st2' ->
             Simulates_ obj (superposition obj) st2' (MapState ts1)))
    (interact :
       forall Ret int ts1 ret ts1',
         ex (MapState ts1) ->
         ThreadsInteraction impl Ret int ts1 ret ts1' ->
         ex (MapState ts1') /\
         (forall st2',
             MapState ts1' st2' ->
             exists st2,
               MapState ts1 st2 /\
               obj.(EvalInstruction) int st2 ret st2')).

  CoFixpoint mapStateBackwardSimulates ts1 :
    ex (MapState ts1) ->
    Simulates_ (ThreadsBehavior impl) (superposition obj)
              ts1 (MapState ts1).
  Proof.
    simpl in interact.
    intro.
    eapply mapStateSimulates with MapState'.
    - clear ts1 H.
      intros.
      destruct ms.
      exists (MapState ts1').
      + eapply simulatesTrans_.
        * apply superpositionSimulates.
          intros.
          edestruct advance; eauto.
        * apply superpositionSimulates.
          intros.
          apply superpositionSingleSimulates, H0.
          assumption.
      + split.
        * edestruct advance; eassumption.
        * reflexivity.
    - clear ts1 H.
      intros.
      destruct ms.
      edestruct interact.
      + eassumption.
      + eassumption.
      + eexists.
        * split.
          -- eassumption.
          -- intros.
             edestruct H2 as [? [? ?]].
             ++ eassumption.
             ++ eexists.
                split.
                --- apply H0.
                    eassumption.
                --- assumption.
        * split.
          -- assumption.
          -- reflexivity.
    - split.
      + assumption.
      + reflexivity.
  Qed.

End MapStateBackward.
