Set Asymmetric Patterns.
From Coq Require Import
  Arith Bool List Permutation.

From C4.Util Require Import LibTactics.

Import ListNotations.


Section Predefs.

  Parameter MaxTId: nat.

  Definition TId := nat.
  Definition eq_tid_dec := eq_nat_dec.

  Definition beq_tid := beq_nat.
  Definition below_in (n: nat): list nat -> Prop := 
    fun bnats => forall n', n' < n <-> In n' bnats.
  Definition no_dups: list nat -> Prop := 
    fun bnats => NoDup bnats.

  Fixpoint bnats (n: nat): @sig2 (list nat) (below_in n) no_dups.
    refine(
      match n return (sig2 (below_in n) no_dups) with
        | 0 =>
            exist2 (below_in 0) no_dups nil _ _
        | S n' => 
            match (bnats n') with
              | exist2 pbnats H H' => 
                exist2 (below_in (S n')) no_dups 
                       (n'::pbnats)
                       _ _
            end
      end).

    unfold below_in. intros. split; intros. subst. inversion H. subst. inversion H.
    unfold no_dups. constructor.

    unfold below_in. intros. split; intros; subst.

      apply lt_n_Sm_le in H0. apply le_lt_or_eq in H0. destruct H0; subst. 
      simpl. right. unfold below_in in H. apply H. assumption.
      simpl. left. reflexivity.
      
      simpl in H0. destruct H0; subst.
      (*  (_ < S _). *) apply lt_n_Sn.
      unfold below_in in H. apply H in H0. apply lt_S. assumption.

    unfold no_dups. constructor. unfold not. intros. unfold below_in in H. apply H in H0. 
    clear bnats pbnats H H'. induction n'. inversion H0. apply IHn'. apply lt_S_n. assumption.
    unfold no_dups in H'. assumption.
  Defined.


  Definition tids: list TId.        
    refine(
        match (bnats MaxTId) with
          | exist2 l H1 H2 => l
        end).
  Defined.


  Lemma below_max_in_tids: forall (n: TId), n < MaxTId <-> In n tids.
    Proof.
      intros.
      unfold tids.
      destruct (bnats MaxTId) as [tids B N].
      unfold below_in in B.
      apply B.
    Qed.

  Lemma nodup_tids:
    NoDup tids.
    Proof.
      intros.
      unfold tids.
      destruct (bnats MaxTId) as [tids B N].
      assumption.
    Qed. 

  Definition init_tid := MaxTId.

  Lemma init_not_in_tids:
    not (In init_tid tids).

    Proof.
      intro.
      unfold init_tid in H.
      assert (A := below_max_in_tids MaxTId).
      apply A in H.
      apply lt_irrefl in H.
      assumption.
    Qed.
  
  Definition Key := nat.
  (* Definition Val := nat. *)
  Definition key_eq_dec := eq_nat_dec.

  
  Definition override {V: Type} (m: nat -> V) (k: nat) (v: V): nat -> V :=
    fun (k': nat) =>
      if (eq_nat_dec k' k) then v else (m k').

    
  Lemma eq_nat_dec_eq:
    forall (A: Type) n (s1 s2: A),
      (if eq_nat_dec n n then s1 else s2) =
      s1.

  Proof.
    intros.
    destruct (eq_nat_dec n n).
    reflexivity.

    exfalso. apply n0. reflexivity.
  Qed.

  Lemma eq_nat_dec_neq:
    forall (A: Type) n n' (s1 s2: A),
      (not (n = n'))
      -> (if eq_nat_dec n n' then s1 else s2) =
        s2.

  Proof.
    intros.

    destruct (eq_nat_dec n n').
    exfalso. apply H in e. assumption.

    reflexivity.    
  Qed.

  Lemma override_new_val:
    forall {V: Type} m k (v: V),
      (override m k v) k = v.

    Proof.
      intros.
      unfold override.
      rewrite eq_nat_dec_eq.
      reflexivity.
    Qed.

  Lemma override_new_val':
    forall {V: Type} m k1 k2 (v: V),
      k1 = k2 ->
      (override m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold override.
      rewrite eq_nat_dec_eq.
      reflexivity.
    Qed.

  Lemma override_new_val'':
    forall {V: Type} m k1 k2 (v: V),
      k2 = k1 ->
      (override m k1 v) k2 = v.

    Proof.
      intros.
      subst.
      unfold override.
      rewrite eq_nat_dec_eq.
      reflexivity.
    Qed.


  Lemma override_old_val:
    forall {V: Type} m k k' (v: V),
      (not (k = k'))
      -> (override m k v) k' = m k'.

    Proof.
      intros.
      unfold override.
      rewrite eq_nat_dec_neq.
      reflexivity.
      unfold not. intro. symmetry in H0. apply H in H0. assumption.
    Qed.

  Lemma override_old_val':
    forall {V: Type} m k k' (v: V),
      (not (k' = k))
      -> (override m k v) k' = m k'.

  Proof.
      intros.
      unfold override.
      rewrite eq_nat_dec_neq.
      reflexivity.
      assumption.
    Qed.

  
  Definition update_key {V: Set} (m: Key -> V) (k: Key) (v: V): Key -> V :=
    fun (k': Key) =>
      if (key_eq_dec k' k) then v else (m k).

  Definition update_tid {V: Set} (m: TId -> V) (n: TId) (v: V): TId -> V :=
    fun (n': TId) =>
      if (eq_tid_dec n' n) then v else (m n).

  Definition Clock := nat.

End Predefs.


Notation "a =_? b" := (eq_nat_dec a b) (at level 20).
Notation "a <>_? b" := (negb (a =_? b)) (at level 20).
Notation "a <_? b" := (lt_dec a b) (at level 20).
Notation "a <=_? b" := (le_dec a b) (at level 20).
Notation "a >_? b" := (gt_dec a b) (at level 20).
Notation "a >=_? b" := (ge_dec a b) (at level 20).

Notation "a <>? b" := (negb (beq_nat a b)) (at level 70).
Notation "a =? b" := (beq_nat a b) (at level 70).
Notation "a <=? b" := (leb a b) (at level 70).

(*
  Ltac simpl_override :=
    try rewrite override_new_val;
    try (
    (rewrite override_new_val'; [ | eassumption]) ||
    (rewrite override_new_val''; [ | eassumption]) ||
    (rewrite override_old_val; [ | eassumption]) ||
    (rewrite override_old_val'; [ | eassumption]));
    simpl.

  Ltac simpl_override_in H :=
    try rewrite override_new_val in H;
    try (
    (rewrite override_new_val' in H; [ | eassumption]) ||
    (rewrite override_new_val'' in H; [ | eassumption]) ||
    (rewrite override_old_val in H; [ | eassumption]) ||
    (rewrite override_old_val' in H; [ | eassumption]));
    simpl in H.
 *)


  Ltac simpl_override :=
    repeat (
       rewrite override_new_val ||
       (rewrite override_new_val'; [ | eassumption]) ||
       (rewrite override_new_val''; [ | eassumption]) ||
       (rewrite override_old_val; [ | eassumption]) ||
       (rewrite override_old_val'; [ | eassumption]));
    simpl.

  
  Ltac simpl_override_in H :=
    repeat (
       rewrite override_new_val in H ||
       (rewrite override_new_val' in H; [ | eassumption]) ||
       (rewrite override_new_val'' in H; [ | eassumption]) ||
       (rewrite override_old_val in H; [ | eassumption]) ||
       (rewrite override_old_val' in H; [ | eassumption]));
    simpl in H.
  
  (*
  Ltac simpl_override_in_star :=
    repeat (
       rewrite override_new_val in * ||
       (rewrite override_new_val' in *; [ | eassumption]) ||
       (rewrite override_new_val'' in *; [ | eassumption]) ||
       (rewrite override_old_val in *; [ | eassumption]) ||
       (rewrite override_old_val' in *; [ | eassumption]));
    simpl in *.
   *)

  (*
  Ltac simpl_override_in_star :=  
    simpl_override;
    repeat
      match goal with
      | [ H: context[override ?m ?k ?v ?k'] |- _ ] => simpl_override_in H
      end.
  *)  

  Ltac simpl_override_in_star :=  

    repeat
      match goal with
      | [ |- context[override ?m ?k ?v ?k] ] =>
        rewrite override_new_val
      | [ H': ?k = ?k' |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_new_val' m k k' v H')
      | [ H': ?k' = ?k |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_new_val'' m k' k v H')
      | [ H': not (?k = ?k') |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_old_val m k k' v H')
      | [ H': ?k = ?k' -> False |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_old_val m k k' v H')
      | [ H': not (?k' = ?k) |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_old_val' m k k' v H')
      | [ H': ?k' = ?k -> False |- context[override ?m ?k ?v ?k'] ] =>
        rewrite (override_old_val' m k k' v H')
      end;

    repeat
      match goal with
      | [ H: context[override ?m ?k ?v ?k] |- _ ] =>
        rewrite override_new_val in H
      | [ H: context[override ?m ?k ?v ?k'], H': ?k = ?k' |- _ ] =>
        rewrite (override_new_val' m k k' v H') in H
      | [ H: context[override ?m ?k ?v ?k'], H': ?k' = ?k |- _ ] =>
        rewrite (override_new_val'' m k' k v H') in H
      | [ H: context[override ?m ?k ?v ?k'], H': not (?k = ?k') |- _ ] =>
        rewrite (override_old_val m k k' v H') in H
      | [ H: context[override ?m ?k ?v ?k'], H': ?k = ?k' -> False |- _ ] =>
        rewrite (override_old_val m k k' v H') in H
      | [ H: context[override ?m ?k ?v ?k'], H': not (?k' = ?k) |- _ ] =>
        rewrite (override_old_val' m k k' v H') in H
      | [ H: context[override ?m ?k ?v ?k'], H': ?k' = ?k -> False |- _ ] =>
        rewrite (override_old_val' m k k' v H') in H
      end.
  
  
  Lemma override_eq_val:
    forall (A: Type) (m: nat -> A) k v k',
      m k' = v ->
      override m k v k' = v.

  Proof.
    intros.
    destruct (eq_nat_dec k k').
    simpl_override. reflexivity.
    simpl_override. assumption.    
  Qed.
  
  Lemma simpl_override_test:
    forall (V: Type) m k1 k2 k3 (v: V),
      ((not (k2 = k1))
       /\(not (k1 = k3)))
      -> ((override m k1 v) k1 = v
          /\ (override m k1 v) k2 = m k2
          /\ (override m k1 v) k3 = m k3).

  Proof.
    intros.
    destruct H as [H1 H2].
    split.
    
    simpl_override.
    reflexivity.

    split.
    simpl_override.
    reflexivity.

    simpl_override.
    reflexivity.

  Qed.

  
  Lemma simpl_override_test2:
    forall m1 m2 k1 k2 k3 (v1 v2: nat),
      (k1 = k3 -> False)
      -> override m1 k1 v1 (override m2 k1 k2 k3) = override m1 k1 v1 (m2 k3).

  Proof.
    intros.
    simpl_override_in_star.
    reflexivity.
  Qed.
 
  Lemma simpl_override_test3:
    forall m k (v: nat),
      (k = 0 -> False) ->
      override m 0 v k = m k.

  Proof.
    intros.    
    simpl_override_in_star.
    reflexivity.
  Qed.
  
Ltac sauto := 
  try solve [eassumption | auto | eauto | jauto].


Tactic Notation "open_conjs" :=
  repeat(match goal with
           | [ H : _ /\ _ |- _ ] => destruct H
         end).

Tactic Notation "split_all" :=
  repeat(match goal with
           | [ |- _ /\ _ ] => split
         end).


Ltac specex H :=
  repeat match type of H with
           | forall x : ?T, _ =>
             match type of T with
               | Prop =>
                 fail 1
               | _ =>
                 let x := fresh "x" in
                   evar (x : T);
                   let x' := eval unfold x in x in
                     clear x; specialize (H x')
             end
         end.

Ltac depremise H :=
  match type of H with
    | forall x : ?T, _ =>
      let x := fresh "x" in
      assert (x: T);
      [ clear H | specialize (H x); clear x ]                  
  end.

Ltac deprem H :=
  repeat (match type of H with
          | forall x : ?T, _ =>
            let x := fresh "x" in
            assert (x: T);
          [ clear H | specialize (H x); clear x ]                  
          end).

Ltac specex_deprem H :=
  specex H; depremise H.

Ltac subv x :=
  match goal with 
    | H: x = _ |- _ => rewrite H; simpl
    | H: _ = x |- _ => rewrite <- H; simpl
    | _ => try unfold x; simpl
  end.

Ltac subv_in x H' :=
  match goal with 
    | H: x = _ |- _ => rewrite H in H'; simpl in H'
    | H: _ = x |- _ => rewrite <- H in H'; simpl in H'
    | _ => try unfold x in H'; simpl in H'
  end.

  (*
  Lemma subv_test:
    let m := 0 in
    forall a b c d e, 
      c = 2 + 2
      -> 2 + 2 = d
      -> m + 1 = e
      -> a = 0
      -> 0 = b
      -> a + b = m + 2
      -> a + b = m.

    Proof.
      intros.
      subv b.
      subv a.
      subv m.
      subv_in a H4.
      subv_in b H4.
      subv_in m H4.     
    Qed.
  *)

(*
Tactic Notation "depremise" ident(H) :=
  match goal with
    | [ H: ?ps -> _ |- _ ] => 
      assert (PS: ps);
      [ clear H | match goal with
                    | PS: _ |- _ => specialize (H PS); clear PS
                  end]
  end.
*)

Tactic Notation "rewrite_clear" ident(H) ident(H') :=
  rewrite H in H'; clear H.
      
Tactic Notation "r_rewrite_clear" ident(H) ident(H') :=
  rewrite <- H in H'; clear H.


Tactic Notation "bool_to_prop" :=
  try match goal with
    | [ |- negb _ = true ] =>  try apply negb_true_iff
    | [ |- negb _ = false ] =>  try apply negb_false_iff
  end;
  (apply andb_prop ||
   apply orb_prop ||
   apply beq_nat_true_iff ||
   apply beq_nat_false_iff ||
   apply leb_iff).

Tactic Notation "bool_to_prop_in" ident(H) :=
  try match goal with
    | [ H: negb _ = true |- _ ] =>  try apply negb_true_iff in H
    | [ H: negb _ = false |- _ ] =>  try apply negb_false_iff in H
  end;
  (apply andb_prop in H ||
   apply orb_prop in H ||
   apply beq_nat_true_iff in H ||
   apply beq_nat_false_iff in H ||
   apply leb_iff in H).


Lemma in_single_iff:
  forall A (l l': A),
    In l [l'] <->
    l = l'.

Proof.
  intros; split; intros.
  simpl in H. destruct H. symmetry. assumption. contradiction.
  subst l. apply in_eq.
Qed.

Lemma lt_neq:
  forall n m,
    n < m -> n <> m.

Proof.
  intros. intro.
  subst n.
  assert (H' := H).
  apply lt_asym in H'.
  apply H' in H.
  assumption.
Qed.


Definition eq_option_dec A (u u': option A): (forall a a': A, {a = a'} + {a <> a'}) -> {u = u'} + {u <> u'}.
Proof.
  decide equality.
Qed.


Module Type StepStarParams.

  Parameter State: Type.
  Parameter Label: Type.
  Parameter step: State -> Label -> State -> Prop.

End StepStarParams.

Module StepStar (StepStarArgs: StepStarParams).

  Import StepStarArgs.

  Inductive step_star: State -> list Label -> State -> Prop :=
  | refl s: 
      step_star s nil s
  | steps s1 ls s2 l s3: 
      step_star s1 ls s2
      -> step s2 l s3
      -> step_star s1 (ls ++ [l]) s3.

  Definition history (init: State)(ls: list Label): Prop :=
    exists (s: State), step_star init ls s.

  Lemma step_star_end:
    forall s l h s'',
      step_star s (l :: h) s''
      <-> exists s',
           step s l s'
           /\ step_star s' h s''.

    Proof.
      intros.
      split;
      intros.      

      generalize dependent s''.
      induction h using rev_ind.

      intros.
      exists s''.
      split.
      inversion H.
      subst.
      rewrite <- app_nil_l in H0.
      apply app_inj_tail in H0.
      destruct H0; subst.
      inversion H1.
      subst.
      assumption.
      subst.
      apply app_eq_nil in H0.
      destruct H0; subst.     
      inversion H4.
      constructor.

      intros.
      rewrite app_comm_cons in H.
      inversion H.
      rewrite app_comm_cons in H0.
      apply app_inj_tail in H0.
      destruct H0.
      subst.
      apply IHh in H1.
      destruct H1 as [s' [N1 N2]].
      exists s'.
      split.
      assumption.
      econstructor; eassumption.

      (* inverse *)
      generalize dependent s''. 
      induction h using rev_ind.

      intros.
      destruct H as [s' [H1 H2]].
      inversion H2.
      subst.
      clear H2.
      assert (A := steps s nil s l s'').
      depremise A.
      constructor.
      depremise A.
      assumption.
      rewrite app_nil_l in A.
      assumption.

      subst.
      apply app_eq_nil in H.
      destruct H. inversion H3.

      intros.
      destruct H as [s' [H1 H2]].
      remember (h ++ [x]) as hh eqn: Hh.
      destruct H2.
      symmetry in Hh. apply app_eq_nil in Hh.
      destruct Hh.
      subst.
      inversion H0.

      apply app_inj_tail in Hh.
      destruct Hh.
      subst.
      specialize (IHh s2).
      depremise IHh.
      exists s1.
      split.
      assumption.
      assumption.
      rewrite app_comm_cons.
      econstructor; eassumption.      
    Qed.


  Lemma step_star_app:
    forall s h1 h2 s'',
      step_star s (h1 ++ h2) s''
      <-> exists s',
           step_star s h1 s'
           /\ step_star s' h2 s''.

    Proof.
      intros.
      split.

      intros.
      generalize dependent s.
      generalize dependent h2.
      induction h1.
      (* -- *)
      intros.
      rewrite app_nil_l in H.
      exists s.
      split. apply refl. assumption.
      (* -- *)
      intros.
      rewrite <- app_comm_cons in H.
      apply step_star_end in H.
      destruct H as [s' [H1 H2]].
      specialize (IHh1 h2 s').
      depremise IHh1. assumption.
      destruct IHh1 as [sm [IHh1 IHh2]].
      exists sm.
      split. 
        apply step_star_end. exists s'. split; assumption.
        assumption.

      intros.
      generalize dependent s.
      generalize dependent h2.
      induction h1.
      (* -- *)
      intros.
      destruct H as [s' [H1 H2]].
      rewrite app_nil_l.
      inversion H1. 
        subst. assumption.
        subst. destruct ls. rewrite app_nil_l in H. inversion H. inversion H.
      (* -- *)
      intros.
      destruct H as [s' [H1 H2]].
      apply step_star_end in H1.
      destruct H1 as [sm [H11 H12]].
      specialize (IHh1 h2 sm).
      depremise IHh1.
      exists s'. split; assumption.
      rewrite <- app_comm_cons.
      eapply step_star_end.
      exists sm. split; assumption.      

    Qed.


  Lemma step_star_one:
    forall s l s',
      step_star s [l] s'
      <-> step s l s'.

    Proof.
      intros; split; intros.

      inversion H.
      destruct ls.
      subst.
      (* app.*)
      rewrite app_nil_l in H0.
      inversion H0.
      subst.
      inversion H1.
      subst.
      assumption.
      destruct ls.
      rewrite app_nil_l in H2.
      inversion H2.
      inversion H2.
      inversion H0.
      destruct ls.
      rewrite app_nil_l in H7.
      inversion H7.
      inversion H7.

      assert (A : [l] = nil ++ [l]).
      apply app_nil_l.
      rewrite A in *; clear A.
      eapply steps.
      constructor.
      assumption.
    Qed.

  Lemma step_star_app_one:
    forall s h l s'',
      step_star s (h ++ [l]) s''
      <-> exists s',
           step_star s h s'
           /\ step s' l s''.

    Proof.
      intros.
      assert (L := step_star_app).
      specialize (L s h [l] s'').
      split; intros.
      destruct L as [L _].
      depremise L.
      assumption.
      destruct L as [s' [L1 L2]].
      assert (L := step_star_one).
      specialize (L s' l s'').
      destruct L as [L _].
      depremise L.
      assumption.
      exists s'.
      split; assumption.

      destruct L as [_ L].      
      depremise L.
      destruct H as [s' [H1 H2]].
      exists s'.      
      split.
      assumption.
      assert (L := step_star_one).
      specialize (L s' l s'').
      destruct L as [_ L].
      depremise L. assumption.
      assumption.
      assumption.
    Qed.


  Hint Constructors step_star.

End StepStar.



