Require Import
        Coq.Lists.List
        Coq.Relations.Relations
        Coq.Sets.Ensembles.

From C4.Util Require Import
  Util
  LibTactics
  MTactics.
From C4.Core Require Import
  Program
  Object
  Simulates.

Section TraceSub.
  
  Record Label {M} :=
    {
      LabelR: Type;
      labelCall: M LabelR;
      labelRet: LabelR;
    }.
  Arguments Label : clear implicits.
  Definition label := Build_Label.

  Definition labeledStep
             {M} (o: Spec M)
             (s1: o.(State)) (l: Label M) (s2: o.(State)):
    Prop :=
    @EvalInstruction _ o (LabelR l) (labelCall l) s1 (labelRet l) s2.

  Definition Trace M := list (Label M).
    

  Inductive labeledStepStar
            {M} (o: Spec M):
    o.(State) -> Trace M -> o.(State) -> Prop :=
  | labeledStepStarRefl s:
      labeledStepStar o s nil s
  | labeledStepStarTrans (l: Label M) s1 s2 t s3:
      labeledStep o s1 l s2 ->
      labeledStepStar o s2 t s3 ->
      labeledStepStar o s1 (l :: t) s3.

  Definition traceSubset
             {M} (o1 o2: Spec M) (s1: o1.(State)) (s2: o2.(State)):
    Prop :=
    forall s1' (t: Trace M),
      labeledStepStar o1 s1 t s1' ->
      exists s2', labeledStepStar o2 s2 t s2'.


  Lemma SimulatesTraceSubset:
    forall M (o1 o2: Spec M) (s1: o1.(State)) (s2: o2.(State)),
      Simulates_ o1 o2 s1 s2 ->
      traceSubset o1 o2 s1 s2.

  Proof.
    introv H1 H2.
    gen s2.
    induction H2 as [ s | l s1 s1' t s1'' H21 H22 H23].
    - eexists. constructor.
    - introv H3. inverts H3 as H3.
      unfold labeledStep in *.
      forwardsIn H3; try eassumption. destruct H3 as (s2', H31, H32).
      forwardsIn H23; try eassumption. destruct H23 as (s2'' & H23).
      eexists. econstructor; eassumption.
  Qed.

End TraceSub.
