From C4.Util Require Import
  FunOverride.
From C4.Core Require Import
  Program
  Object.

Inductive ArrayMethod M: Type -> Type :=
| arrayCall {R}: nat -> M R -> ArrayMethod M R
| getSize: ArrayMethod M nat.
Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

Section AbsArray.

  Context {I: Type -> Type}.
  Context (spec: Spec I).
  (* Variable initObjState: spec.(State). *)
  
  Record ArrayState := {
    arrayMapState: nat -> spec.(State);                        
    arraySizeState: { n: nat | 0 < n }
  }.
  Definition arrayState := Build_ArrayState.
  Canonical Structure Default_ArrayState m s := Build_ArrayState m s.
  
  Inductive ArraySem:
    forall {R}, ArrayMethod I R -> ArrayState -> R -> ArrayState -> Prop :=
    
  | succArrayCall i R (m: I R) st r o':
      i < proj1_sig (arraySizeState st) ->
      spec.(EvalInstruction) m (arrayMapState st i) r o' ->
      ArraySem
        (arrayCall i m)
        st
        r
        (arrayState (noverride (arrayMapState st) i o') (arraySizeState st))

  | failArrayCall i R (m: I R) st v st':
      proj1_sig (arraySizeState st) <= i ->
      ArraySem
        (arrayCall i m) st
        v st'

  | getSizeCall st:
      ArraySem
        getSize st
        (proj1_sig (arraySizeState st)) st.

  (*
  | resizeCall s st:
      ArraySem
        (resize s) st
        tt (initArrayState initObjState s).
  *)
      
  Definition absArray len : Spec (ArrayMethod I) :=
    Build_Spec _
      ArrayState
      (@ArraySem)
      (arrayState (fun _ => spec.(initialState)) len).

End AbsArray.
Arguments arrayMapState {I} {spec} _ _.
Arguments arraySizeState {I} {spec} _.


Section ArrayImpl.
  
  (* Context {M: Type -> Type}. *)
  (* Context (impl: Object M). *)

  (* Definition arrayLift {M R} (p: Program M R) i: Program (ArrayMethod M) R := *)
  (*   mapProgram *)
  (*     (fun _ bm => arrayCall i bm) *)
  (*     p. *)


  
  (*
   * A parametric object of array.
   * Given an object, returns the object of an array of that object.
   *)
  Axiom arrayImpl:
    forall {M: Type -> Type},
      Object M ->
      Object (ArrayMethod M).
  (* Definition arrayImpl: Object (ArrayMethod M) := {| *)
  (*   ObjInstruction := *)
  (*     ArrayMethod (ObjInstruction impl); *)

  (*   objBase := *)
  (*     absArray (objBase impl); *)

  (*   objProgram := *)
  (*     fun R (m: ArrayMethod M R) => *)
  (*       match m with *)
  (*       | arrayCall i im => *)
  (*           arrayLift (objProgram impl _ im) i *)

  (*       | getSize => *)
  (*         s <- getSize; *)
  (*         Return s *)
  (*       end *)
  (* |}. *)


  (* 
   * Given an initial state for an object and a size,
   * returns an initial state for the array of object with that size.
   *)
  (* Definition initArrayObjState (st: impl.(objBase).(State))(s: {n: nat | 0 < n}): arrayImpl.(objBase).(State) := *)
  (*   initArrayState st s. *)

  Axiom initArrayObjState:
    forall {M: Type -> Type},
    forall {impl: Object M},
      impl.(objBase).(State) ->
      {n: nat | 0 < n} ->
      (arrayImpl impl).(objBase).(State).
  
  
End ArrayImpl.
(* Arguments initArrayObjState {M} {impl} _ _. *)


