(* Duplicated [CHashMap.v] to avoid a universe inconsistency error. *)

Require Import Arith.

From C4.Core Require Import
  Program
  Object
  Pair.
From C4.App Require Import
  Lock
  Maps.
From C4.Extraction Require Import
  ArrayExt
  SHashMapExt.

Import EMap.

Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

Module Type CHashMapParams.

  (* The object of lock. *)
  Parameter lockImpl: Object LockMethod.

End CHashMapParams.

Module CHashMapFun (E : HashableType) (CHashMapArgs : CHashMapParams).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.
  Definition hash := E.hash.
  Definition bucketMapImpl := E.bucketMapImpl.
  Import CHashMapArgs.

  Section CHashMapFun1.

  Context (Val : Type).

  Definition locks := arrayImpl lockImpl.
  Definition buckets := arrayImpl (bucketMapImpl Val).

  Definition base := pairObject locks buckets.

  Definition BaseState :=
    prod
      locks.(objBase).(State)
      buckets.(objBase).(State).

  Definition posMult (n1 n2: {n: nat | 0 < n}): {n: nat | 0 < n}.
    refine (exist _ ((proj1_sig n1)*(proj1_sig n2)) _).
      destruct n1. destruct n2. simpl. eapply Nat.mul_pos_pos; eassumption.
    Defined.

  Definition BaseMethod :=
    PairMethod
      (ArrayMethod LockMethod)
      (ArrayMethod (EMapMethod Key Val)).
  Definition callLocks {R} (m: ArrayMethod LockMethod R): BaseMethod R :=
    inl m.
  Definition callLocksArray (i: nat) {R} (m: LockMethod R): BaseMethod R :=
    inl (arrayCall i m).
  Definition callBuckets {R} (m: ArrayMethod (EMapMethod Key Val) R): BaseMethod R :=
    inr m.
  Definition callBucketsArray {R} (i: nat) (m: EMapMethod Key Val R): BaseMethod R :=
    inr (arrayCall i m).

  Definition impl {R} (m: EMapMethod Key Val R):
    Program BaseMethod R :=

    let k := keyArg m in
    ls <- callLocks getSize;
    let li := (hash k) mod ls in
    _ <- callLocksArray li doLock;
    bs <- callBuckets getSize;
    let bi := (hash k) mod bs in
    r <- callBucketsArray bi m;
    _ <- callLocksArray li doUnlock;
    Return r.

  Definition cHashMapImpl :=
    flattenObject
      base _
      (@impl).

  End CHashMapFun1.
End CHashMapFun.
