From C4.Util Require Import
  Util.
From C4.Core Require Import
  Program.
From C4.Extraction Require Import
  ExtractionIO.

From Coq Require Import
  Arith ZArith List.

From Coq.extraction Require Import
  ExtrHaskellBasic
  ExtrHaskellString
  ExtrHaskellNatInt
  ExtrHaskellZInt.

(* Axiom IO : Type -> Type. *)
(* Definition ProgramIO : Type -> Type := Program IO. *)
(* Definition bindIO : forall {b a}, IO a -> (a -> ProgramIO b) -> ProgramIO b := @Bind IO. *)
(* Definition returnIO : forall {a}, a -> ProgramIO a := @Return IO. *)
(* Definition noopIO : forall {a}, ProgramIO a -> ProgramIO a := @Noop IO. *)

(* Notation "x <-- f ; m" := (bindIO f (fun x => m)) (at level 80, right associativity). *)

(* Axiom bindProgramIO: forall {A B}, ProgramIO A -> (A -> ProgramIO B) -> ProgramIO B. *)
(* Extract Constant bindProgramIO => "(Prelude.>>=)". *)
(* Notation "x <== p ; m" := (bindProgramIO p (fun x => m)) (at level 85, right associativity). *)
(* (* CoFixpoint bindProgramIO {A B} (p : ProgramIO A) (f : A -> ProgramIO B) : ProgramIO B := *) *)
(* (*   match p with *) *)
(* (*     | Bind instr f' => a <-- instr; bindProgramIO (f' a) f *) *)
(* (*     | Return a => f a *) *)
(* (*     | Noop p' => Noop (bindProgramIO p' f) *) *)
(* (*   end. *) *)

(* Record ObjTrans {I} (o: Spec I) := { *)
(*   Rep: Type; *)
(*   newRep: o.(State) -> ProgramIO Rep; *)
(*   mTrans: Rep -> forall {R}, I R -> IO R *)
(* }. *)


(* -------------------------------------------------------------------------- *)
Axiom Int : Type.
Axiom IORef : Type -> Type.
Axiom Ticket : Type -> Type.
Axiom MutableArray : Type -> Type.

Axiom int_of_Z : Z -> Int.
Axiom Z_of_int : Int -> Z.
Axiom peekTicket : forall {a}, Ticket a -> a.
Axiom newIORef : forall {a}, a -> IO (IORef a).
Axiom readForCAS : forall {a}, IORef a -> IO (Ticket a).
Axiom casIORef : forall {a}, IORef a -> Ticket a -> a -> IO (bool * Ticket a).
Axiom newArray : forall {a}, Int -> a -> IO (MutableArray a).
Axiom readArrayElem : forall {a}, MutableArray a -> Int -> IO (Ticket a).
Axiom casArrayElem : forall {a}, MutableArray a -> Int -> Ticket a -> a -> IO (bool * Ticket a).
Axiom loadLoadBarrier : IO unit.
Axiom print_Z : Z -> IO unit.


Extraction Language Haskell.

Extract Inlined Constant Int => "Prelude.Int".
Extract Constant IO "a" => "Prelude.IO a".
(* Extract Inlined Constant IO "a" => "IO a". *)
(* Extract Constant IO "a" => "". *)
Extract Constant IORef "a" => "Runtime.IORef a".
Extract Constant Ticket "a" => "Runtime.Ticket a".
Extract Constant MutableArray "a" => "Runtime.MutableArray (Runtime.PrimState Prelude.IO) a".

Extract Inlined Constant int_of_Z => "Prelude.id".
Extract Inlined Constant Z_of_int => "Prelude.id".
Extract Inlined Constant peekTicket => "Runtime.peekTicket".
Extract Inlined Constant newIORef => "Runtime.newIORef".
Extract Inlined Constant readForCAS => "Runtime.readForCAS".
Extract Inlined Constant casIORef => "Runtime.casIORef".
Extract Inlined Constant newArray => "Runtime.newArray".
Extract Inlined Constant readArrayElem => "Runtime.readArrayElem".
Extract Inlined Constant casArrayElem => "Runtime.casArrayElem".
Extract Inlined Constant loadLoadBarrier => "Runtime.loadLoadBarrier".
Extract Inlined Constant print_Z => "Prelude.print".

Extract Constant Nat.modulo => "Prelude.mod".
Extract Constant Init.Nat.modulo => "Prelude.mod".

(* -------------------------------------------------------------------------- *)
(* ObjTrans moved to Program.v *)
(* Record ObjTrans {I} (o: Spec I) := { *)
(*   Rep: Type; *)
(*   newRep: o.(State) -> IO Rep; *)
(*   mTrans: Rep -> forall {R}, I R -> IO R *)
(* }. *)

(* MTrans moved to Program.v *)
(* Record MTrans M := { *)
(*   MState: Type; *)
(*   MRep: Type; *)
(*   mNewRep: MState -> IO MRep; *)
(*   mMTrans: MRep -> forall {R}, M R -> IO R *)
(* }. *)

Definition objMTrans {I} {o: Spec I} (ot: ObjTrans o): MTrans I := {|
  MState := o.(State);
  MRep := Rep o ot;
  mNewRep := newRep o ot;
  mMTrans := mTrans o ot;
|}.


(* -------------------------------------------------------------------------- *)
From C4 Require Import Register.


Axiom RegRep : Type -> Type.
(* Extract Constant RegRep "a" => "IORef a". *)
(* Extract Inlined Constant RegRep "t" => "IORef t". *)
Extract Constant RegRep "a" => "IORef a".

Definition Val := (atomicReg (Val := nat)).(State).
(* Definition Val := nat. *)

(* Axiom regNewRep: forall {T}, T -> IO (RegRep T). *)
(* Axiom regNewRep: forall {T}, Val T -> IO (RegRep (Val T)). *)
(* Extract Constant newRegRep "T" "t" => "newIORef t". *)
(* Extract Constant regNewRep => "newIORef". *)


Axiom natRegNewRep: Val -> ProgramIO (RegRep Val).
(* Extract Constant newRegRep "T" "t" => "newIORef t". *)
Extract Constant natRegNewRep => "newIORef".


(* Axiom cRead: forall {a}, RegRep a -> IO a. *)
(* Axiom cWrite: forall {a}, RegRep a -> a -> IO unit. *)
(* Axiom cCAS: forall {a}, RegRep a -> a -> a -> IO bool. *)

(* Axiom cReadN: RegRep nat -> IO nat. *)
(* Axiom cWriteN: RegRep nat -> nat -> IO unit. *)
(* Axiom cCASN: RegRep nat -> nat -> nat -> IO bool. *)

Axiom cReadN: RegRep Val -> IO Val.
Axiom cWriteN: RegRep Val -> Val -> IO unit.
Axiom cCASN: RegRep Val -> Val -> Val -> IO bool.



(* Require Import Coq.Strings.String. *)
(* Definition nl := String (Ascii.ascii_of_N 10) EmptyString. *)
(* Open Scope string. *)

Extract Constant cReadN => "readIORef".
Extract Constant cWriteN => "writeIORef".
Extract Constant cCASN =>
(* "ref" "v" "v'" => *)
"\ref v v' ->
       do {
          t <- readForCAS ref;
          let { tv = peekTicket t };
          if ((unsafeCoerce v) :: Prelude.Int)  Prelude.== ((unsafeCoerce tv) :: Prelude.Int) then do {
             ret <- casIORef ref t v';
             Prelude.return (Prelude.fst ret)
          } else
             Prelude.return Prelude.False
       }".


(* Definition natRegMTrans (rep: RegRep nat) {R'} (m: @RegisterMethod nat R'): IO R' := *)
Definition natRegMTrans (rep: RegRep Val) {R'} (m: @RegisterMethod Val R'): IO R' :=
  match m with
  | read => cReadN rep
  | write v => cWriteN rep v
  | cas v v' => cCASN rep v v'
  end.


Definition natRegObjTrans : ObjTrans (atomicReg (Val := nat)) :=
  {|
    Rep := RegRep nat;
    newRep := natRegNewRep;
    mTrans := natRegMTrans
  |}.

Definition natRegImplTrans: MTrans (@RegisterMethod nat) :=
  {|
    MState := nat;
    MRep := RegRep nat;
    mNewRep := natRegNewRep;
    mMTrans := natRegMTrans
  |}.


(* -------------------------------------------------------------------------- *)
From C4 Require Export Lock.

Axiom LockRep : Type.
Extract Constant LockRep => "Lock".

Definition B := atomicLock.(State).

Axiom newLockRep: B -> ProgramIO LockRep.
Extract Constant newLockRep =>
   "\ b ->
       if ((unsafeCoerce b) :: Prelude.Bool) then
          newAcquired
       else
          new".

Axiom cDoLock: LockRep -> IO unit.
Extract Constant cDoLock => "acquire".

Axiom cDoUnlock: LockRep -> IO unit.
Extract Constant cDoUnlock => "release".


(* Definition natRegMTrans (rep: RegRep nat) {R'} (m: @RegisterMethod nat R'): IO R' := *)
Definition lockMethTrans (rep: LockRep) {R'} (m: LockMethod R'): IO R' :=
  match m with
  | doLock => cDoLock rep
  | doUnlock => cDoUnlock rep
  end.


Definition lockObjTrans : ObjTrans atomicLock :=
  {|
    Rep := LockRep;
    newRep b := newLockRep b;
    mTrans := lockMethTrans
  |}.


Definition lockMTrans : MTrans LockMethod :=
  {|
    MState := bool;
    MRep := LockRep;
    mNewRep b := newLockRep b;
    mMTrans := lockMethTrans
  |}.


(* -------------------------------------------------------------------------- *)
From C4 Require Import Pair.

Definition pairObjTrans
           {I}{o: Spec I} (ot: ObjTrans o)
           {I'}{o': Spec I'} (ot': ObjTrans o'):
  ObjTrans (pairSpec o o') := {|
  Rep := prod (Rep o ot) (Rep o' ot');
  newRep :=
    fun (ss: (pairSpec o o').(State)) =>
      match ss with
        (s, s') =>
        r <== newRep o ot s;
        r' <== newRep o' ot' s';
        returnIO (r, r')
      end;
  mTrans :=
    fun ss _ i =>
      match i with
      | inl li =>
        mTrans o ot (fst ss) li
      | inr ri =>
        mTrans o' ot' (snd ss) ri
      end
|}.

Definition pairMTrans
           {M} (mt: MTrans M)
           {M'} (mt': MTrans M'):
  MTrans (PairMethod M M') :=
{|
  MState := prod (MState _ mt) (MState _ mt');
  MRep := prod (MRep _ mt) (MRep _ mt');
  mNewRep :=
    fun (ss: prod (MState _ mt) (MState _ mt')) =>
      match ss with
        (s, s') =>
        r <== mNewRep _ mt s;
        r' <== mNewRep _ mt' s';
        returnIO (r, r')
      end;
  mMTrans :=
    fun ss _ i =>
      match i with
      | inl li =>
        mMTrans _ mt (fst ss) li
      | inr ri =>
        mMTrans _ mt' (snd ss) ri
      end
|}.


(* -------------------------------------------------------------------------- *)
From C4 Require Import ArrayExt.
Axiom ArrayRep: Type -> Type.
Extract Constant ArrayRep "v" => "RArray v".

Axiom len : {n | n > 0}.

Axiom newArrayRep: forall {I}, forall (obj: Spec I), ObjTrans obj -> (absArray obj len).(State) -> ProgramIO (ArrayRep obj.(State)).
Extract Constant newArrayRep =>
"\ obj objTrans st ->
   case ((unsafeCoerce st) :: ArrayState Any) of {
      Build_ArrayState m s -> case objTrans of {
                                 Build_ObjTrans nr _ ->  do {
                                                            arr <- array_new (0, s Prelude.- 1);
                                                            Prelude.mapM_ 
                                                               (\ i ->
                                                                  do {
                                                                     r <- nr (m i);
                                                                     _ <- array_write arr i r;
                                                                     Prelude.return () })
                                                               [0..(s Prelude.- 1)];
                                                            Prelude.return ((unsafeCoerce (RArray arr s)) :: (ArrayRep a1)) }}}".

Axiom cArrayCall: forall {I}, forall (obj: Spec I), ObjTrans obj -> ArrayRep (obj.(State)) -> forall {R}, nat -> I R -> IO R.
Extract Constant cArrayCall =>
"\ obj objTrans rep i m -> do {
   o <- array_read (rarray rep) i;
   case objTrans of {
      Build_ObjTrans _ mt ->  do {
         r <- mt ((unsafeCoerce o) :: Any) () m;
         Prelude.return ((unsafeCoerce r) :: a2)
      }
   }
}".


Axiom cGetSize: forall {I}, forall (obj: Spec I), ArrayRep (obj.(State)) -> IO nat.
Extract Constant cGetSize =>
"\ obj rep ->
   Prelude.return (rarraysize rep)".

Definition arrayMTrans {I} (obj: Spec I) (objTrans: ObjTrans obj) (rep: ArrayRep (obj.(State))) {R} (m: ArrayMethod I R): IO R :=
  match m with
  | arrayCall i m => cArrayCall obj objTrans rep i m
  | getSize => cGetSize obj rep
  end.


Definition arrayTrans {I}(obj: Spec I)(objTrans: ObjTrans obj): ObjTrans (absArray obj len) :=
  {|
    Rep := ArrayRep obj.(State);
    newRep := newArrayRep obj objTrans;
    mTrans := arrayMTrans obj objTrans;
  |}.


(* --------------------------------------------------------- *)
(* arrayImpl Translation *)

(* Context {M: Type -> Type}. *)
(* Context (impl: Object M). *)
(* Definition arrayImpl: Object (ArrayMethod M). *)

(* Record MTrans M := { *)
(*   MState: Type; *)
(*   MRep: Type; *)
(*   mNewRep: MState -> ProgramIO MRep; *)
(*   mMTrans: MRep -> forall {R}, M R -> IO R *)
(* }. *)

Record PArrayState State := {
  arrayMapState: nat -> State;
  arraySizeState: { n: nat | 0 < n }
}.

Definition initPArrayState {State} (st: State) (s: {n: nat | 0 < n}): PArrayState State :=
  Build_PArrayState _ (fun n => st) s.


Axiom newArrayIRep:
  forall {M}, forall mTrans: MTrans M,
      PArrayState (MState _ mTrans) ->
      ProgramIO (ArrayRep (MRep _ mTrans)).

Extract Constant newArrayIRep =>
"\ mTrans st ->
   case ((unsafeCoerce st) :: PArrayState Any) of {
      Build_PArrayState m s -> 
        case mTrans of {
          Build_MTrans nr _ ->  do {
            arr <- array_new (0, s Prelude.- 1);
            Prelude.mapM_ 
              (\ i ->
                do {
                  r <- nr (m i);
                  _ <- array_write arr i r;
                  Prelude.return () })
              [0..(s Prelude.- 1)];
            Prelude.return ((unsafeCoerce (RArray arr s)) :: (ArrayRep a1)) }}}".


Axiom cIArrayCall:
  forall {M}, forall mTrans: MTrans M, 
    ArrayRep (MRep _ mTrans) ->
    forall {R}, nat -> M R ->
    IO R.

Extract Constant cIArrayCall =>
"\ mTrans rep i m -> do {
   o <- array_read (rarray rep) i;
   case mTrans of {
      Build_MTrans _ mt ->  do {
         r <- mt ((unsafeCoerce o) :: Any) () m;
         Prelude.return ((unsafeCoerce r) :: a2)
      }
   }
}".


Axiom cIGetSize:
  forall {M}, forall mTrans: MTrans M, 
    ArrayRep (MRep _ mTrans) ->
    IO nat.

Extract Constant cIGetSize =>
"\ obj rep ->
   Prelude.return (rarraysize rep)".


Definition arrayIMTrans
           {M} (mTrans: MTrans M)
           (rep: ArrayRep (MRep _ mTrans))
           {R} (m: (ArrayMethod M) R)
  : IO R :=

  match m with
  | arrayCall i m =>
    cIArrayCall mTrans rep i m
  | getSize =>
    cIGetSize mTrans rep
  end.

  
Definition arrayImplTrans
           {M} (mTrans: MTrans M)
  : MTrans (ArrayMethod M) :=
{|
   MState := PArrayState (MState _ mTrans);
   MRep := ArrayRep (MRep _ mTrans);
   mNewRep := newArrayIRep mTrans;
   mMTrans := arrayIMTrans mTrans;
|}.


(* -------------------------------------------------------------------------- *)
Axiom print: forall {T}, T -> IO unit.
Extract Inlined Constant print => "Prelude.print".
Axiom hpar: forall {R'}, list ((ProgramIO R') * (R' -> IO unit)) -> ProgramIO unit.
Extract Inlined Constant hpar => "Runtime.hpar".

(* -------------- *)
CoFixpoint pTrans
           {M} (it: MTrans M) (rep: MRep M it)
           {R'} (p: Program M R')
  : ProgramIO R' :=

  match p with
  | Bind m f => bindIO (mMTrans M it rep m) (fun r => pTrans it rep (f r))
  | Return r => returnIO r
  | Noop p' => noopIO (pTrans it rep p')
  end.


CoFixpoint oPTrans {I} (o: @Spec I) (ot: ObjTrans o) (rep: Rep o ot) {R'} (p: Program I R'): ProgramIO R' :=
  pTrans (objMTrans ot) rep p.

Definition oPsTrans {I} (o: @Spec I) (ot: ObjTrans o) (rep: Rep o ot) {R'} (ps: list (Program I R')): list (ProgramIO R') :=
  map (oPTrans o ot rep) ps.


Definition par
           {I} (o: @Spec I)
           (ot: ObjTrans o)
           (i: o.(State))
           {R'} (pps: list ((Program I R') * (R' -> IO unit)))
  : ProgramIO unit :=
  
    rep <== newRep o ot i;
    let ps := oPsTrans o ot rep (map fst pps) in
    hpar (zip ps (map snd pps)).


(* -------------- *)
From C4 Require Import ProgramOps Object.

CoFixpoint pITrans
           {I} (impl: Object I)
           (ot: ObjTrans impl.(objBase))
           (rep: Rep impl.(objBase) ot)
           {R'} (p: Program I R'):
  ProgramIO R' :=
  oPTrans
    impl.(objBase)
    ot
    rep
    (flattenProgram impl.(objProgram) p).


Definition psITrans
           {I} (impl: Object I)
           (ot: ObjTrans impl.(objBase))
           (rep: Rep impl.(objBase) ot)
           {R'} (ps: list (Program I R')):
  list (ProgramIO R') :=
  map (pITrans impl ot rep) ps.


Definition ipar
           {I} (impl: Object I)
           (ot: ObjTrans impl.(objBase))
           (i: impl.(objBase).(State))
           {R'} (pps: list ((Program I R') * (R' -> IO unit))):
  ProgramIO unit :=
  rep <== newRep impl.(objBase) ot i;
  let ps := psITrans impl ot rep (map fst pps) in
  hpar (zip ps (map snd pps)).


(* -------------- *)
Definition pIFTrans
           {M M'} (implProg : forall Ret, M' Ret -> Program M Ret)
           (it: MTrans M) (rep: MRep M it)
           {R'} (p: Program M' R')
  : ProgramIO R' :=
  pTrans
    it
    rep
    (flattenProgram implProg p).
  


Definition psIFTrans
           {M M'} (implProg : forall Ret, M' Ret -> Program M Ret)
           (it: MTrans M) (rep: MRep M it)
           {R'} (ps: list (Program M' R'))
  : list (ProgramIO R') :=
  map (pIFTrans implProg it rep) ps.


Definition iFlatPar
           {M M'} (implProg : forall Ret, M' Ret -> Program M Ret)
           (it: MTrans M)
           (s: MState _ it)           
           {R'} (pps: list ((Program M' R') * (R' -> IO unit))):
  ProgramIO unit :=

  rep <== mNewRep _ it s;
  let ps := (map fst pps) in
  let fs := (map snd pps) in
  let tps := psIFTrans implProg it rep ps in
  hpar (zip tps fs).
