From Coq Require Import
     List
     Equalities
     Arith.
From C4.Core Require Import
     Object
     Program.
From C4.App Require Import
     Lock
     Maps.
From C4.Extraction Require Import
     Extraction
     SHashMapExt
     CHashMapExt
     ExtractionIO.

Import ListNotations.
Import EMap.

Module SHashMapArgs <: HashableType.
  Definition t := nat.
  Definition eq_dec := eq_nat_dec.
  Definition hash := @id nat.

  Module EMapArgs <: MiniDecidableType.
    Definition t := nat.
    Definition eq_dec := eq_nat_dec.
  End EMapArgs.

  Module CEMap := EMapFun EMapArgs.
  Import CEMap.

  Definition bucketMapImpl (Val: Type): Object (EMapMethod t Val) :=
    serialObject (atomicEMap Val).
End SHashMapArgs.
Import SHashMapArgs.

Module SHashMap := SHashMapFun SHashMapArgs.
Import SHashMap.

Module CEMap := EMapFun SHashMapArgs.
Import CEMap.

Module CHashMapArgs <: CHashMapParams.

  Definition lockImpl: Object LockMethod :=
    serialObject atomicLock.
End CHashMapArgs.

Module CHashMap := CHashMapFun SHashMapArgs CHashMapArgs.
Import CHashMap.

Parameter Val : Type.

Parameter LockSpec : Type.
Parameter InScope : LockSpec -> Prop.

Parameter locks_getSize : LockSpec -> IO nat.
Parameter mmeq_locks_getSize
  : forall (o : LockSpec),
    InScope o ->
    mmeq (callLocks Val ArrayExt.getSize) (locks_getSize o).
Hint Resolve mmeq_locks_getSize : mmeq.

Parameter locks_lock : LockSpec -> nat -> IO unit.
Parameter mmeq_locks_lock
  : forall (o : LockSpec) (i : nat),
    InScope o ->
    mmeq (callLocksArray Val i doLock) (locks_lock o i).
Hint Resolve mmeq_locks_lock : mmeq.

Parameter locks_unlock : LockSpec -> nat -> IO unit.
Parameter mmeq_locks_unlock
  : forall (o : LockSpec) (i : nat),
    InScope o ->
    mmeq (callLocksArray Val i doUnlock) (locks_unlock o i).
Hint Resolve mmeq_locks_unlock : mmeq.

Parameter buckets_getSize : LockSpec -> IO nat.
Parameter mmeq_buckets_getSize
  : forall (o : LockSpec),
    InScope o ->
    mmeq (callBuckets Val ArrayExt.getSize) (buckets_getSize o).
Hint Resolve mmeq_buckets_getSize : mmeq.

Parameter buckets_call
  : forall {R}, LockSpec -> nat -> EMapMethod Key Val R -> IO R.
Parameter mmeq_buckets_call
  : forall R (o : LockSpec) (i : nat) (m : EMapMethod Key Val R),
    InScope o ->
    mmeq (callBucketsArray Val i m) (buckets_call o i m).
Hint Resolve mmeq_buckets_call : mmeq.

(** Extraction directives *)

Extract Constant Val => "Prelude.Int".
Extract Constant LockSpec => "LocksAndBuckets.LB Prelude.Int".

Extract Inlined Constant locks_getSize => "LocksAndBuckets.locksSize".
Extract Inlined Constant locks_lock => "LocksAndBuckets.lock".
Extract Inlined Constant locks_unlock => "LocksAndBuckets.unlock".
Extract Inlined Constant buckets_call => "LocksAndBuckets.bucketsCall".
Extract Inlined Constant buckets_getSize => "LocksAndBuckets.bucketsSize".

(** Object *)
Definition hashimpl {r} (o : LockSpec) (m : EMapMethod Key Val r)
           (SCOPE : InScope o)
  : { p : IO r | meq (impl _ m) p }.
Proof.
  eexists. unfold impl.
  automeq.
Qed.

(** Extraction *)

Extraction "extraction/ExtractionCHashMap.hs" hashimpl.
