From Coq Require Import Extraction.
From C4.Core Require Import Program ProgramOps.

Parameter IO : Type -> Type.
Extract Constant IO "a" => "Prelude.IO a".

Module IO.

Parameter ret : forall {a}, a -> IO a.
Extract Inlined Constant ret => "Prelude.pure".

Parameter bind : forall {a b}, IO a -> (a -> IO b) -> IO b.
Extract Inlined Constant bind => "(Prelude.>>=)".

End IO.

Parameter mmeq : forall {m a}, m a -> IO a -> Prop.

Parameter meq : forall {m a}, Program m a -> IO a -> Prop.
Parameter meq_bind
  : forall {m a b} (u : m a) (u' : IO a)
           (k : a -> Program m b) (k' : a -> IO b),
    mmeq u u' ->
    (forall x : a, meq (k x) (k' x)) ->
    meq (Bind u k) (IO.bind u' k').
Parameter meq_ret
  : forall {m a} (x : a), meq (@Return m _ x) (IO.ret x).

(* Dijkstra IO monad *)
Definition DIO {m a} (u : Program m a) : Type := { v : IO a | meq u v }.

(* Dijkstra method type *)
Definition DM {m a} (u : m a) : Type := { v : IO a | mmeq u v }.

Module DIO.

Definition ret : forall {m a} (x : a), DIO (m := m) (Return x).
Proof.
  econstructor; apply meq_ret.
Defined.

Definition bind {m a b} (u : m a) (k : a -> Program m b)
  : DM u ->
    (forall a, DIO (k a)) ->
    DIO (Bind u k).
Proof.
  intros v h; econstructor; eapply meq_bind.
  - apply (proj2_sig v).
  - intros x; apply (proj2_sig (h x)).
Defined.

End DIO.

Create HintDb mmeq discriminated.

Ltac automeq :=
  match goal with
  | [ |- meq (Bind _ _) _ ] =>
    eapply (meq_bind _ _ _ (fun _ => _)); [ eauto with mmeq | intros ?; try automeq ]
  | [ |- meq (Return _) _ ] =>
    eapply meq_ret
  end.

Ltac autodio :=
  match goal with
  | [ |- DIO (Return _) ] => apply DIO.ret
  | [ |- DIO (Bind _ ?k) ] =>
    let x :=
        match k with
        | (fun x => _) => fresh x
        | _ => fresh "x"
        end in
    apply DIO.bind; [ | intros x; try autodio ]
  | [ |- DIO (match ?w with _ => _ end) ] => destruct w
  end.

Definition ProgramIO : Type -> Type := Program IO.
Definition bindIO : forall {b a}, IO a -> (a -> ProgramIO b) -> ProgramIO b := @Bind IO.
Definition returnIO : forall {a}, a -> ProgramIO a := @Return IO.
Definition noopIO : forall {a}, ProgramIO a -> ProgramIO a := @Noop IO.

Notation "x <-- f ; m" := (bindIO f (fun x => m)) (at level 80, right associativity) : IO_scope.

Definition bindProgramIO : forall {A B} (p : ProgramIO A) (f : A -> ProgramIO B), ProgramIO B := @bindProgram IO.
Notation "x <== p ; m" := (bindProgramIO p (fun x => m)) (at level 85, right associativity).
(* CoFixpoint bindProgramIO {A B} (p : ProgramIO A) (f : A -> ProgramIO B) : ProgramIO B := *)
(*   match p with *)
(*     | Bind instr f' => a <-- instr; bindProgramIO (f' a) f *)
(*     | Return a => f a *)
(*     | Noop p' => Noop (bindProgramIO p' f) *)
(*   end. *)

Extract Constant ProgramIO "a" => "Prelude.IO a".
Extract Constant bindProgramIO => "(Prelude.>>=)".
Extract Constant bindIO => "(Prelude.>>=)".
Extract Constant returnIO => "Prelude.return".
Extract Constant noopIO => "Prelude.id".

Record ObjTrans {I} (o: Spec I) := {
  Rep: Type;
  newRep: o.(State) -> ProgramIO Rep;
  mTrans: Rep -> forall {R}, I R -> IO R
}.

Record MTrans M := {
  MState: Type;
  MRep: Type;
  mNewRep: MState -> ProgramIO MRep;
  mMTrans: MRep -> forall {R}, M R -> IO R
}.
