Require Import
        Coq.Sets.Ensembles
        Coq.Structures.Equalities.

From C4.Util Require Import
  FunOverride.
From C4.Core Require Import
  Program.
From C4.App Require Import
  Maps.
From C4.Extraction Require Import
  ExtractionIO.

Module BMapFunExt (E : MiniDecidableType) (F : Maps.BMap.BMapFunType E).
    Import Maps.BMap F.

    Axiom BMapRep : Type -> Type.
    Extract Constant BMapRep "val" => "(Runtime.BasicHashTable Key val)".

    Definition St Val := (atomicBMap Val).(State).

    (* Note that the initial value of the map must be empty. *)
    Axiom newBMapRep: forall Val, St Val -> ProgramIO (BMapRep Val).
    Extract Constant newBMapRep "val" =>
    "\ m ->
      Runtime.map_new".
      (* "(Runtime.map_new :: Runtime.BasicHashTable Key val)". *)

    Axiom cGet: forall Val, BMapRep Val -> Key -> IO (option Val).
    Extract Constant cGet => "Runtime.map_lookup".

    Axiom cPut: forall Val, BMapRep Val -> Key -> Val -> IO unit.
    Extract Constant cPut => "Runtime.map_insert".


    Definition bMapMTrans Val (rep: BMapRep Val) {R'} (m: BMapMethod Key Val R'): IO R' :=
      match m with
      | get k => cGet Val rep k
      | put k v => cPut Val rep k v
      end.

    Definition bMapObjTrans Val: ObjTrans (atomicBMap Val) :=
      {|
        Rep := BMapRep Val;
        newRep m := newBMapRep Val m;
        mTrans := bMapMTrans Val
      |}.

    Axiom newIBMapRep: forall Val, BMapState Val -> ProgramIO (BMapRep Val).
    Extract Constant newIBMapRep "val" =>
    "\ m ->
      Runtime.map_new".

    Definition bMapITrans Val: MTrans (BMapMethod Key Val) :=
      {|
        MState := BMapState Val;
        MRep := BMapRep Val;
        mNewRep m := newIBMapRep Val m;
        mMTrans := bMapMTrans Val
      |}.

End BMapFunExt.

Module EMapFunExt (E : MiniDecidableType) (F : Maps.EMap.EMapFunType E).
    Import Maps.EMap F.

    (* --------------------------------- *)
    Axiom NatEMapRep : Type.
    Extract Constant NatEMapRep => "(Runtime.BasicHashTable Key Prelude.Int)".

    Definition St := (atomicEMap nat).(State).

    (* Note that the initial value of the map must be empty. *)
    Axiom newNatEMapRep: St -> ProgramIO NatEMapRep.
    Extract Constant newNatEMapRep =>
    "\ m ->
      Runtime.map_new".
      (* "(Runtime.map_new :: Runtime.BasicHashTable Key val)". *)

    Axiom cGet: NatEMapRep -> Key -> IO (option nat).
    Extract Constant cGet => "Runtime.emap_get".

    Axiom cPut: NatEMapRep -> Key -> nat -> IO (option nat).
    Extract Constant cPut => "Runtime.emap_put".

    Axiom cPutIfAbsent: NatEMapRep -> Key -> nat -> IO (option nat).
    Extract Constant cPutIfAbsent => "Runtime.emap_putIfAbsent".
    
    Axiom cRemove: NatEMapRep -> Key -> IO (option nat).
    Extract Constant cRemove => "Runtime.emap_remove".
    
    Axiom cReplaceIt: NatEMapRep -> Key -> nat -> nat -> IO bool.
    Extract Constant cReplaceIt => "Runtime.emap_replace".


    Definition eMapMTrans (rep: NatEMapRep) {R'} (m: EMapMethod Key nat R'): IO R' :=
      match m with
      | get k => cGet rep k
      | put k v => cPut rep k v
      | remove k => cRemove rep k
      | putIfAbsent k v => cPutIfAbsent rep k v
      | replaceIt k v v' => cReplaceIt rep k v v'
      end.

    Definition bMapObjTrans: ObjTrans (atomicEMap nat) :=
      {|
        Rep := NatEMapRep;
        newRep m := newNatEMapRep m;
        mTrans := eMapMTrans
      |}.

    Axiom newIEMapRep: EMapState nat -> ProgramIO NatEMapRep.
    Extract Constant newIEMapRep =>
    "\ m ->
      Runtime.map_new".


    Definition bMapITrans: MTrans (EMapMethod Key nat) :=
      {|
        MState := EMapState nat;
        MRep := NatEMapRep;
        mNewRep m := newIEMapRep m;
        mMTrans := eMapMTrans
      |}.


    (* --------------------------------- *)
    (* Bacause of the comparison in repalce, Val should be of Eq class. *)
    
    (* Axiom EMapRep : Type -> Type. *)
    (* Extract Constant EMapRep "val" => "(Runtime.BasicHashTable Key val)". *)

    (* Definition St Val := (atomicEMap Val).(State). *)

    (* (* Note that the initial value of the map must be empty. *) *)
    (* Axiom newEMapRep: forall Val, St Val -> ProgramIO (EMapRep Val). *)
    (* Extract Constant newEMapRep "val" => *)
    (* "\ m -> *)
    (*   Runtime.map_new". *)
    (*   (* "(Runtime.map_new :: Runtime.BasicHashTable Key val)". *) *)

    (* Axiom cGet: forall Val, EMapRep Val -> Key -> IO (option Val). *)
    (* Extract Constant cGet => "emap_get". *)

    (* Axiom cPut: forall Val, EMapRep Val -> Key -> Val -> IO (option Val). *)
    (* Extract Constant cPut => "emap_put". *)

    (* Axiom cPutIfAbsent: forall Val, EMapRep Val -> Key -> Val -> IO (option Val). *)
    (* Extract Constant cPutIfAbsent => "emap_putIfAbsent". *)
    
    (* Axiom cRemove: forall Val, EMapRep Val -> Key -> IO (option Val). *)
    (* Extract Constant cRemove => "emap_remove". *)
    
    (* Axiom cReplaceIt: forall Val, EMapRep Val -> Key -> Val -> Val -> IO bool. *)
    (* Extract Constant cReplaceIt => "emap_replace".     *)


    (* Definition eMapMTrans Val (rep: EMapRep Val) {R'} (m: EMapMethod Key Val R'): IO R' := *)
    (*   match m with *)
    (*   | get k => cGet Val rep k *)
    (*   | put k v => cPut Val rep k v *)
    (*   | remove k => cRemove Val rep k *)
    (*   | putIfAbsent k v => cPutIfAbsent Val rep k v *)
    (*   | replaceIt k v v' => cReplaceIt Val rep k v v' *)
    (*   end. *)

    (* Definition bMapObjTrans Val: ObjTrans (atomicEMap Val) := *)
    (*   {| *)
    (*     Rep := EMapRep Val; *)
    (*     newRep m := newEMapRep Val m; *)
    (*     mTrans := eMapMTrans Val *)
    (*   |}. *)

    (* Axiom newIEMapRep: forall Val, EMapState Val -> ProgramIO (EMapRep Val). *)
    (* Extract Constant newIEMapRep "val" => *)
    (* "\ m -> *)
    (*   Runtime.map_new". *)


    (* Definition bMapITrans Val: MTrans (EMapMethod Key Val) := *)
    (*   {| *)
    (*     MState := EMapState Val; *)
    (*     MRep := EMapRep Val; *)
    (*     mNewRep m := newIEMapRep Val m; *)
    (*     mMTrans := eMapMTrans Val *)
    (*   |}. *)
End EMapFunExt.
