From Coq Require Import ZArith.
From C4.Core Require Import
  Program
  ProgramOps.
From C4.Extraction Require Import
  ExtractionIO
  Extraction.

CoInductive programIOEq {R}: Program IO R -> Program IO R -> Prop :=
| bindIOEq R' (i : IO R') (f1 : R' -> Program IO R) (f2 : R' -> ProgramIO R):
    (forall v, programIOEq (f1 v) (f2 v)) ->
    programIOEq (bindIO i f1) (bindIO i f2)
| returnIOEq r:
    programIOEq (returnIO r) (returnIO r)
| noopIOEq p1 p2:
    programIOEq p1 p2 ->
    programIOEq (noopIO p1) (noopIO p2).

CoFixpoint interpret {Ret} (p : Program IO Ret) : ProgramIO Ret :=
   match p with
   | Bind i f => bindIO i (fun a => interpret (f a))
   | Return ret => returnIO ret
   | Noop p' => noopIO (interpret p')
   end.

(* Programming in Coq with IO instructions and then transforming programs to haskell IO monad sequence. *)

CoFixpoint incrementLoop (ref : IORef Z) (old : Ticket Z) :
  Program IO (Ticket Z) :=
  ret <- casIORef ref old (peekTicket old + 1)%Z;
    (let '(success, new) := ret in
     if success : bool then Return new else incrementLoop ref new).

Definition increment (ref : IORef Z) : Program IO (Ticket Z) :=
  Bind (readForCAS ref) (incrementLoop ref).

(* Definition increment' ref := interpret (increment ref). *)

Theorem increment' : {increment' : _ -> ProgramIO _ | forall ref, programIOEq (increment ref) (increment' ref)}.
Proof.
  eexists.
  econstructor.
  instantiate (1:=cofix incrementLoop0 _ := _).
  cofix incrementLoop'.
  intro old.
  rewrite frobProgramId, (frobProgramId (incrementLoop _ _)).  instantiate (1:=bindIO _ _).
  econstructor.
  instantiate (1:=fun ret => _).
  intros ret.
  instantiate (1:=let (success, new) := ret in _).
  destruct ret as [success new].
  instantiate (1:=if success then _ else _).
  destruct success.
  * econstructor.
  * instantiate (1:=incrementLoop0 _).
    apply incrementLoop'.
Defined.


Extraction "extraction/ExtractionOpt.hs" increment'.
