(* Duplicated [SHashMap.v] to avoid a universe inconsistency error. *)

Require Import
        Coq.Structures.Equalities.

From C4.Core Require Import
  Program
  Object.
From C4.App Require Import
  Maps.
From C4.Extraction Require Import
  ArrayExt.

Import Nat.
Import EMap.

Arguments arrayCall {M} {R} _ _.
Arguments getSize {M}.

Module Type HashableType.

  Include MiniDecidableType.

  (* The hash function. *)
  Parameter hash : t -> nat.

  (* The sequential map object that is used in each bucket. *)
  Parameter bucketMapImpl : forall Val, Object (EMapMethod t Val).
End HashableType.


Module SHashMapFun (E : HashableType).

  Definition Key := E.t.
  Definition eq_key_dec := E.eq_dec.
  Definition hash := E.hash.
  Definition bucketMapImpl := E.bucketMapImpl.

  Section SHashMapFun1.

  Context (Val : Type).

  Definition buckets :=
    arrayImpl (bucketMapImpl Val).

  Definition BaseState :=
    buckets.(objBase).(State).

  Definition BaseMethod :=
    ArrayMethod (EMapMethod Key Val).
  Definition bucketsArrayCall {R} (i: nat) (m: EMapMethod Key Val R): BaseMethod R :=
    arrayCall i m.

  Definition impl {R} (m: EMapMethod Key Val R):
    Program BaseMethod R :=

    ls <- getSize;
    let k := keyArg m in
    let i := (hash k) mod ls in
    r <- bucketsArrayCall i m;
    Return r.

  Definition hashMapImpl :=
    flattenObject
      buckets _
      (@impl).

  End SHashMapFun1.
End SHashMapFun.
