# C4: Certified Composable Concurrency in Coq

A framework for Verified Transactional Objects in Coq.

- Formalization of concurrent objects, linearizability, strict
  serializability, and associated proof techniques.
- Verified linearizable concurrent hash map
- Verified strictly serializable TML
- Verified strictly serializable transaction-predicated map

Follow the instructions below to build, then open the file `Paper.v`, a summary
of top-level definitions and results from our paper.

## Getting Started

Step 0. Build the artifact.

### Dependencies

- paco (tested with 4.1.1)
- Coq (8.10; known NOT to build with 8.12 and above)

Install dependencies using Opam. More instructions to set up Coq and Opam
can be found [on Coq's official website](https://coq.inria.fr/opam-using.html).

```
# Install opam.
# Initialize opam:
opam init
eval $(opam env)
opam repo add coq-released https://coq.inria.fr/opam/released

# Install dependencies:
opam install coq.8.10.2 coq-paco.4.1.1
```

### Build

```
make -j
```

Warnings are to be expected.

Takes 4 minutes with 4 cores, not too much RAM (< 8GB).

## Step-by-step guide

Step 1. Continue by reading `Paper.v`.

There is no step 2.

## Axioms

- Excluded middle: `classic : forall P : Prop, P \/ ~ P`.
- Choice: `choice : forall (A B : Type) (R : A -> B -> Prop), (forall x, exists y, R x y) -> exists (f : A -> B), forall x, R x (f x)`.
- Functional extensionality (dependent): `functional_extensionality_dep : forall (A : Type) (B : A -> Type) (f g : forall x, B x), (forall x, f x = g x) -> f = g`.

These two other axioms, equivalent to UIP, are also among the axioms listed by
`Print Assumptions` on the main theorems. UIP is a consequence of excluded middle,
which we've already assumed.

- `JMeq.JMeq_eq : forall (A : Type) (x y : A), JMeq.JMeq x y -> x = y`
- `Eqdep.Eq_rect_eq.eq_rect_eq : forall (U : Type) (p : U) (Q : U -> Type) (x : Q p) (h : p = p), x = eq_rect p Q x p h`
