From Coq Require Import
  Ensembles.

From C4.Core Require Import
  Program
  ProgramOps
  Pair
  Simulates
  Superposition
  Sig.

From C4.CoreExtra Require Import
  BSimulatesFacts.

From C4.App Require Import
  Maps.

Import BMap.
Import EMap.
Import Map.


Inductive AllocatorMethod {Loc} : Type -> Type :=
| alloc : AllocatorMethod Loc.

Arguments AllocatorMethod : clear implicits.

Inductive EvalAllocatorMethod Loc (P : Loc -> Prop) :
  forall Ret, AllocatorMethod Loc Ret -> Ensemble Loc -> Ret -> Ensemble Loc -> Prop :=
| evalAlloc k keys :
    P k ->
    ~ keys k ->
    EvalAllocatorMethod Loc P Loc alloc keys k (Add _ keys k).

Definition atomicAllocator Loc P : Spec (AllocatorMethod Loc) :=
  Build_Spec (AllocatorMethod Loc) (Ensemble Loc) (EvalAllocatorMethod Loc P) (fun _ => False).

Section atomicAllocatorSplit.

Context {Loc : Type} {P : Loc -> Prop}.

Context (DecLoc : forall l l' : Loc, { l = l' } + { l <> l' }).

End atomicAllocatorSplit.

Definition Loc := nat.
Module LocTyp := PeanoNat.Nat.

Module LocBMap := BMapFun LocTyp.
