From Coq Require Import
  List Ensembles JMeq Arith SetoidDec.

From Coq Require Import FunctionalExtensionality.

From C4.Util Require Import
  Util.

From C4.Core Require Import
  Program
  ProgramOps
  Object
  Pair
  Simulates
  Superposition
  Sig
  Threads
  LinearizabilityFacts2
  Trace
  TraceFacts
  HConcSimulates.

From C4.App Require Import
  Maps.

From C4.AppExtra Require Import
  Counter.

From C4.TPred Require Import
  Allocator
  PairImplFunction.

Set Implicit Arguments.

Definition counterAlloc (R : Type) (m : AllocatorMethod nat R) : Program CounterMethod R :=
  match m with
  | alloc => Bind incAndFetch Return
  end.

Definition counterAllocObj : Object (AllocatorMethod nat) :=
  object
    atomicCounter
    counterAlloc.

Definition allocSpec {loc} : Spec (AllocatorMethod loc) := (atomicAllocator _ (Full_set _)).

Lemma counterAllocObj_simple
  : existT (fun o => State (objBase o)) counterAllocObj 0
  = existT _ (mapObject
      (fun _ m => match m in AllocatorMethod _ R return _ R with alloc => incAndFetch end)
      (serialObject atomicCounter)) 0.
Proof.
  cbv. lazymatch goal with
  | [ |- _ _ {| objProgram := ?p |} _ = _ _ {| objProgram := ?q |} _ ] =>
    replace p with q; [ reflexivity | ]
  end.
  apply functional_extensionality_dep; intros ?.
  apply functional_extensionality; intros []. reflexivity.
Qed.

Lemma mapObject_id {M} (o : Object M) (z : State (objBase o))
  : existT (fun o => State (objBase o)) o z
  = existT _ (mapObject (fun _ m => m) o) z.
Proof.
  destruct o. reflexivity.
Qed.

Inductive translate (M := CounterMethod) (N := AllocatorMethod nat)
  : {X & Program M X} -> {X & Program N X} -> Prop :=
| translate_Call : translate (existT _ _ (Bind incAndFetch Return)) (existT _ _ (Bind alloc Return))
| translate_Return (n : nat) : translate (existT _ _ (Return n)) (existT _ _ (Return n))
.

Definition rell (n : nat) (z : nat -> Prop) : Prop := (forall x, n <= x -> ~ z (S x)).

Lemma rell_S n z : rell n z -> rell (S n) (Add _ z (S n)).
Proof.
  intros H ? HS H0. inversion H0; subst.
  - apply H in H1; auto using Le.le_Sn_le.
  - inversion H1; subst. exact (Nat.nle_succ_diag_l _ HS).
Qed.

Theorem counterAllocObj_correct
  : ConcSimulates counterAllocObj (serialObject allocSpec).
Proof.
  unfold ConcSimulates.
  enough (GG :
    forall ts n ts' (z' : _ -> Prop),
      List.Forall2 translate ts ts' ->
      rell n z' ->
      Simulates_ (ThreadsBehavior counterAllocObj) (ThreadsBehavior (serialObject allocSpec))
        (ts, n) (ts', z')).
  { apply GG. { constructor. } { intros ? ? []. } }
  cofix SELF; intros * INV INVz.
  constructor; intros.
  cbn in *.
  destruct s1' as [ts2 n2].
  destruct H as [[ts1 n1] ADV INT].
  assert (ADV' : exists ts1' z1',
    List.Forall2 translate ts1 ts1' /\ rell n1 z1' /\
    ThreadsAdvance (serialObject allocSpec) (ts', z') (ts1', z1')).
  { remember (ts, n) as h eqn:E.
    remember (ts1, n1) as h1 eqn:E1.
    revert ts' z' ts n E ts1 n1 E1 INV INVz.
    clear - ADV. induction ADV as [ | ? [ts_ n_] ? ADV_ _ IH ].
    { intros; subst. inversion E1; subst; eauto. }
    intros. subst.
    assert (ADV_' : exists ts_' z_',
      List.Forall2 translate ts_ ts_' /\ rell n_ z_' /\
      ThreadAdvance (serialObject allocSpec) (ts', z') (ts_', z_')).
    { remember (ts, n) as h eqn:EE.
      remember (ts_, n_) as h_ eqn:EE_.
      revert ts' ts n EE ts_ n_ EE_ INV INVz.
      clear - ADV_. induction ADV_ as [ | | ]; intros.
      - inversion EE; inversion EE_; subst.
        inversion INV; subst.
        specialize (IHADV_ _ _ _ eq_refl _ _ eq_refl ltac:(eassumption) ltac:(eassumption)).
        destruct IHADV_ as (ts_' & z_' & INV_ & INVz_ & ADV_').
        eexists (_ :: ts_'), z_'.
        split; [ constructor; eauto | ]. split; [ auto | ].
        constructor; auto.
      - inversion EE; inversion EE_; subst.
        inversion INV; subst.
        inversion H2; injpair.
        inversion H; injpair.
        eexists (_ :: _), (Add _ z' (S n)).
        split; [ constructor; [ constructor | eauto ] | ].
        split; [ solve [auto using rell_S] | ].
        constructor. econstructor. constructor. apply (INVz _ (le_n _)).
      - inversion EE; inversion EE_; subst.
        inversion INV; subst.
        inversion H1.
    }
    destruct ADV_' as (ts_' & z_' & INV_ & INVz_ & ADV_').
    specialize (IH ts_' z_' _ _ eq_refl _ _ eq_refl INV_ INVz_).
    destruct IH as (ts1' & z1' & INV1 & INV1z & ADV').
    eexists _, _; split; [ eauto | split; [ eauto | ]].
    econstructor; eauto.
  }
  destruct ADV' as (ts1' & z1' & INV1 & INV1z & ADV').
  inversion INT as [? ? ? ? INT0]; subst.
  assert (INT' : exists ts2',
    List.Forall2 translate ts2 ts2' /\
    ThreadsInteraction' (serialObject allocSpec) R int ts1' r ts2').
  { clear ADV' INV1z.
    revert ts1' INV1.
    clear - INT0. induction INT0 as [ | | ]; intros.
    - destruct m; cbn.
      eexists (_ :: _).
      split; [ constructor; [ constructor | eauto ] | ].
      constructor.
    - inversion INV1; subst.
      inversion H1; injpair.
      eexists _.
      split; [ eauto | ].
      constructor.
    - inversion INV1; subst.
      specialize (IHINT0 _ ltac:(eassumption)).
      destruct IHINT0 as (ts2' & INV2' & INT').
      eexists (_ :: _).
      split; [ constructor; eauto | ].
      constructor; auto.
  }
  destruct INT' as (ts2' & INV2 & INT').
  econstructor.
  - econstructor; [ eassumption | econstructor; eassumption ].
  - apply SELF; auto.
Qed.

Set Contextual Implicit.

Inductive PureMapM (K V : Type) : Type -> Type :=
| PureLookup : K -> PureMapM K V V
.

Inductive MMapM (K V : Type) : Type -> Type :=
| Lookup : K -> MMapM K V (option V)
| PutIfAbsent : K -> V -> MMapM K V V
.

Inductive StepMMapM {K V} {EqDecK : EqDec K} (f0 : K -> option V) : forall R, MMapM K V R -> R -> (K -> option V) -> Prop :=
| StepMMapM_Lookup k
  : StepMMapM f0 (Lookup k) (f0 k) f0
| StepMMapM_PutIfAbsent k v
  : f0 k = None ->
    StepMMapM f0 (PutIfAbsent k v) v (FunOverride.override EqDecK f0 k (Some v))
| StepMMapM_PutIfAbsent_fail k v v0
  : f0 k = Some v0 ->
    StepMMapM f0 (PutIfAbsent k v) v0 f0
.

Definition mmapSpec {K V} {EqDecK : EqDec K} : Spec (MMapM K V) :=
  {| EvalInstruction := fun _ i s a s' => StepMMapM s i a s'
  ;  initialState := fun _ => None |}.

Definition emptyMap {K V DecK} : State (@mmapSpec K V DecK) := fun _ => None.

Definition lazyMap {K V} (R : Type) (m : PureMapM K V R)
  : Program (PairMethod (MMapM K V) (AllocatorMethod V)) R :=
  match m with
  | PureLookup k =>
    Bind (inl (Lookup k)) (fun ov =>
    match ov with
    | Some v => Return v
    | None =>
      Bind (inr alloc) (fun v0 =>
      Bind (inl (PutIfAbsent k v0)) (fun v =>
      Return v))
    end)
  end.

Definition lazyMapObj {K V EqDecK} : Object (PureMapM K V) :=
  object (pairSpec (mmapSpec (EqDecK := EqDecK)) allocSpec) lazyMap.

Definition lazyMapInit {K V DecK} : State (objBase (@lazyMapObj K V DecK)) := (emptyMap, fun _ => False).

Definition lookupSpec {K V} (f : K -> V) : forall R, PureMapM K V R -> R :=
  fun R m =>
    match m with
    | PureLookup k => f k
    end.

From Coq Require Import Sets.Image.

(** [lookupSpec f] is injective if [f] is injective. Assumes UIP. *)

Definition eqofPMM {K V} {R} (m : PureMapM K V R) : V = R :=
  match m with
  | PureLookup k => eq_refl
  end.

Definition KofPMM {K V} {R} (m : PureMapM K V R) : K :=
  match m with
  | PureLookup k => k
  end.

Lemma ffffff {K V} (f : K -> V) {R} (m : PureMapM K V R)
  : lookupSpec f m = eq_rect _ (fun A => A) (f (KofPMM m)) _ (eqofPMM m).
Proof. destruct m; reflexivity. Qed.

Lemma gggggg {K V R} (m : PureMapM K V R) : eq_rect _ _ (PureLookup (KofPMM m)) _ (eqofPMM m) = m.
Proof. destruct m. reflexivity. Qed.

Lemma injective_lookupSpec {K V} (f : K -> V)
  : injective K V f -> injective _ V (lookupSpec f (R := V)).
Proof.
  intros INJ x y H. rewrite 2 ffffff in H. rewrite 2 (UIP_refl _ _ (eqofPMM _)) in H; cbn in H.
  apply INJ in H.
  rewrite <- (gggggg (m := x)), <- (gggggg (m := y)).
  rewrite H. rewrite 2 (UIP_refl _ _ (eqofPMM _)). reflexivity.
Qed.

Section lazyMap_correct.

Context {K V : Type}.
Context {EqDecK : EqDec K}.
Context {EqDecV : EqDec V}.
Context (f0 : K -> V).
Context (INJ0 : injective K V f0).
Notation MMM := (MMapM K V +' AllocatorMethod V).

(* This construction is purely logical so it's fine if it's super not efficient. *)
Definition swap (f : K -> V) (k : K) (v : V) :=
  fun k' =>
    if eq_dec k k' then v
    else if eq_dec v (f k') then f k
    else f k'.

Lemma swap_eq f k v : swap f k v k = v.
Proof.
  unfold swap; destruct eq_dec; contradiction + auto.
Qed.

Lemma swap_neq f k v k' : k <> k' -> v <> f k' -> swap f k v k' = f k'.
Proof.
  unfold swap; destruct eq_dec; try contradiction.
  destruct eq_dec; try contradiction. auto.
Qed.

Lemma injective_swap (f : K -> V) (INJ : injective K V f) k v
  : injective K V (swap f k v).
Proof.
  intros x x'; unfold swap.
  destruct (eq_dec k), (eq_dec k); subst; auto.
  - destruct eq_dec.
    + intros ->. apply INJ in e; auto.
    + contradiction.
  - destruct eq_dec.
    + intros <-. apply INJ in e; auto.
    + congruence.
  - destruct eq_dec, eq_dec; subst; auto.
    + intros EE; apply INJ in EE; contradiction.
    + intros EE; apply INJ in EE; congruence.
Qed.

Definition oinjective {A B : Type} (f : A -> option B) : Prop :=
  forall x x' y, Some y = f x -> Some y = f x' -> x = x'.

Definition pia_v {X} (e : MMM X) : option V :=
  match e with
  | inl (PutIfAbsent _ v) => Some v
  | _ => None
  end.

Fixpoint lltrace (t : ConcTrace MMM) (f0 : K -> option V)
    (allocd : V -> Prop) (used : V -> Prop) (pending : nat -> option {R & MMM R}) : Prop :=
  match t with
  | nil => oinjective f0
  | ((n, e) :: t)%list =>
    match e with
    | CallEvent e =>
      let pre := match pia_v e with
                 | Some v => ~ used v
                 | None => True
                 end in
      let used' := match pia_v e with
                   | Some v => (fun x => v = x \/ used x)
                   | None => used
                   end in
      pre -> lltrace t f0 allocd used' (set n (Some (existT _ _ e)) pending)
    | (ReturnEvent r) as e0 =>
      match pending n with
      | None => False
      | Some (existT _ _ e) =>
        let pending := set n None pending in
        match e with
        | inl e =>
          match e with
          | Lookup k => (e0 = ReturnEvent (f0 k) \/ e0 = ReturnEvent (None (A := V)))
            /\ lltrace t f0 allocd used pending
          | PutIfAbsent k v =>
            exists v0, f0 k = Some v0 /\ e0 = ReturnEvent v0 /\ lltrace t f0 allocd used pending
          end
        | inr _e => exists v : V, e0 = ReturnEvent v /\ ~ allocd v /\ lltrace t f0 (fun x => v = x \/ allocd x) used pending
        end
      end
    end
  end.

Inductive IsConcTraceOfSpec'' (M : Type -> Type) (t : ConcTrace M) (s : Spec M) (z : State s) (ps : Pending M) (z' : State s) (ps' : Pending M)
 : Prop :=
| MkIsConcTraceOfSpec''
  : forall (t0 : list (nat * AtomicEvent M)),
    Steps PendingStep t0 ps ps' ->
    Steps (PendingEvalStep_ s) t0 z z' ->
    t = projAtomic t0 -> IsConcTraceOfSpec'' t s z ps z' ps'
.

Inductive IsConcTraceOfSpec' (M : Type -> Type) (t : ConcTrace M) (s : Spec M) (z : State s) (ps : Pending M) : Prop :=
| MkIsConcTraceOfSpec'
  : forall (t0 : list (nat * AtomicEvent M))
      (z' : State s) (ps' : Pending M),
    Steps PendingStep t0 ps ps' ->
    Steps (PendingEvalStep_ s) t0 z z' ->
    t = projAtomic t0 -> IsConcTraceOfSpec' t s z ps
.

Ltac set_dec n n0 :=
  destruct (eq_dec n n0) as [<- | ]; [ rewrite !set_eq | rewrite !set_neq by auto ].

Ltac blah :=
  lazymatch goal with
  | [ H : MkPendingEvent (ReturnEvent _) = ?ps, H' : MkPendingEvent (CallEvent _) = ?ps |- _ ] =>
    rewrite <- H in H'; discriminate
  end.

Lemma oinjective_set_None {T} {_ : EqDec T} (f : T -> option V) n
  : oinjective f -> oinjective (set n None f).
Proof.
  intros OINJ n1 n2; set_dec n n1; try discriminate.
  set_dec n n2; try discriminate. apply OINJ.
Qed.

Lemma lazyMap_correct_A_
  : forall
      (t0 : AtomicTrace MMM)
      (z z' : (K -> option V) * Ensemble V)
      (ps ps' : Pending (ObjInstruction lazyMapObj))
      (H : Steps PendingStep t0 ps ps')
      (H0 : Steps (PendingEvalStep_ (objBase lazyMapObj)) t0 z z')
    , (forall k v, Some v = fst z k -> Some v = fst z' k)
    /\ forall
      (OINJ : oinjective (fst z))
      (preallocd : nat -> option V) (OINJpreallocd : oinjective preallocd)
      (tmp : nat -> option V) (OINJtmp : oinjective tmp)
      (allocd used : _ -> Prop) pending
      (Hpreallocd : forall n v, Some v = preallocd n -> ~ allocd v /\ snd z v)
      (Hallocd : forall v, allocd v -> snd z v)
      (Hused : forall v k, ~used v -> Some v <> fst z k)
      (Htmp : forall n v, Some v = tmp n -> used v)
      (Htmpz : forall n v k, Some v = tmp n -> Some v <> fst z k)
      (Hpending : forall n,
        match pending n with
        | None => NoPendingEvent = ps n
        | Some (existT _ _ e) =>
          ( MkPendingEvent (CallEvent e) = ps n /\
            match pia_v e with
            | Some v => Some v = tmp n /\ (forall k', Some v <> fst z k')
            | _ => True
            end
          ) \/
          match e with
          | inl (Lookup k) => MkPendingEvent (ReturnEvent (fst z' k)) = ps n \/ MkPendingEvent (ReturnEvent (None (A := V))) = ps n
          | inl (PutIfAbsent k v) => exists v', Some v' = fst z' k /\ MkPendingEvent (ReturnEvent v') = ps n
          | inr _ => exists v, snd z v /\ Some v = preallocd n /\ MkPendingEvent (ReturnEvent v) = ps n
          end
        end)
    , lltrace (projAtomic t0) (fst z') allocd used pending.
Proof.
  induction t0 as [ | [n e] t0 IH ]; intros; cbn.
  { (* nil *) apply Steps_nil in H0; destruct H0; auto. }
  (* const *)
  apply Steps_cons_inv in H; destruct H as (ps0 & [STEPps CONSTANTps] & STEPSps).
  apply Steps_cons_inv in H0; destruct H0 as (z0 & STEPz & STEPSz).
  specialize (IH z0 z' ps0 ps' STEPSps STEPSz).
  destruct IH as [MONO IH].
  split.
  { destruct e as [ | ? [ [] | ] ]; cbn in STEPz.
    - subst; auto.
    - destruct (proj1 STEPz); auto.
      intros k1 v1. specialize (MONO k1 v1). unfold FunOverride.override in MONO.
      destruct EqDecK as [ <- | ] in MONO; [ congruence | auto ].
    - destruct (proj1 STEPz); auto.
      intros k1 v1. specialize (MONO k1 v1). unfold FunOverride.override in MONO.
      destruct EqDecK as [ <- | ] in MONO; [ congruence | auto ].
    - rewrite (proj2 STEPz); auto. }
  intros.
  destruct e as [e | R m r]; cbn in *.
  - (* MkAtomicEvent *)
    destruct STEPz.
    specialize (IH OINJ).
    destruct e as [ R e' | R r ] eqn:Ee.
    + (* CallEvent *)
      specialize (IH preallocd OINJpreallocd).
      intros ZZ.
      destruct (pia_v e') as [v | ] eqn:Epia.
      * specialize (IH (set n (Some v) tmp)).
        forward IH. {
          intros n1 n2 y; set_dec n n1; set_dec n n2; auto.
          - injection 1; intros <- HH. exfalso; eauto.
          - intros HH; injection 1; intros <-. exfalso; eauto.
          - apply OINJtmp. }
        specialize (IH allocd (fun x => v = x \/ used x) (set n (Some (existT _ _ e')) pending)).
        do 3 (forward IH; [ solve[auto] | ]).
        forward IH. {
          intros n1 v1; set_dec n n1; eauto.
          left; congruence. }
        forward IH. {
          intros n1 v1; set_dec n n1; eauto. intros ?; injection 1; intros <-. eauto. }
        forward IH.
        { intros n0; set_dec n n0.
          - left. subst e. split. { apply (proj2 STEPps). }
            rewrite Epia; eauto.
          - rewrite <- CONSTANTps by auto.
            specialize (Hpending n0); destruct (pending n0) as [ [? [ [] | ]] | ]; eauto.
        }
        auto.
      * specialize (IH tmp OINJtmp).
        specialize (IH allocd used (set n (Some (existT _ _ e')) pending)).
        do 5 (forward IH; [ solve[auto] | ]).
        forward IH.
        { intros n0; set_dec n n0.
          - left. subst e. split. { apply (proj2 STEPps). }
            rewrite Epia; auto.
          - rewrite <- CONSTANTps by auto; apply Hpending.
        }
        auto.
    + (* ReturnEvent *)
      destruct STEPps as [Hps Hps0].
      assert (Hpen := Hpending n).
      destruct (pending n) as [ [ ? [[] | []]] | ] eqn:Epn.
      * (* Lookup *) destruct Hpen as [[Hpen _] | Hpen]; [ blah | ].
        specialize (IH preallocd OINJpreallocd).
        specialize (IH tmp OINJtmp).
        specialize (IH allocd used (set n None pending)).
        do 5 (forward IH; [ solve[auto] | ]).
        forward IH.
        { intros n0; set_dec n n0; [ apply Hps0 | rewrite <- CONSTANTps by auto; apply Hpending ]. }
        split; auto.
        rewrite <- Hps in Hpen.
        destruct Hpen as [Hp | Hp]; [left | right]; inversion Hp; injpair; auto.
      * (* PutIfAbsent *) destruct Hpen as [[Hpen _] | Hpen]; [ blah | ].
        destruct Hpen as [v0 [Hv0 Hpen]].
        rewrite <- Hps in Hpen. inversion Hpen. subst R. injpair; clear Hpen.
        specialize (IH preallocd OINJpreallocd).
        specialize (IH tmp OINJtmp).
        specialize (IH allocd used (set n None pending)).
        do 5 (forward IH; [ solve[auto] | ]).
        forward IH.
        { intros n0; set_dec n n0.
          - apply Hps0.
          - rewrite <- CONSTANTps by auto. apply (Hpending n0).
        }
        eauto.
      * (* alloc *) destruct Hpen as [[Hpen _] | Hpen]; [ blah | ].
        destruct Hpen as (v & Hv & Hpreallocd' & Hpen).
        rewrite <- Hps in Hpen. inversion Hpen. subst R. injpair; clear Hpen.
        specialize (IH (set n None preallocd)).
        forward IH. { auto using @oinjective_set_None. }
        specialize (IH tmp OINJtmp).
        specialize (IH (fun x => r = x \/ allocd x) used (set n None pending)).
        forward IH. {
          intros n0 ?; set_dec n n0; try discriminate.
          intros Hvv. destruct (Hpreallocd _ _  Hvv) as [Hvv1 Hvv2]. split; auto.
          intros [ <- |].
          - apply n1; revert Hpreallocd' Hvv. apply OINJpreallocd.
          - contradiction. }
        forward IH. { intros ? [<- | ]; auto. }
        do 3 (forward IH; [solve[auto] | ]).
        forward IH. {
          intros n0; set_dec n n0; auto.
          rewrite <- CONSTANTps by auto.
          specialize (Hpending n0). destruct (pending n0) as [ [? []] | ]; eauto.
        }
        destruct (Hpreallocd _ _ Hpreallocd').
        eauto.
      * (* None *) rewrite <- Hps in Hpen; discriminate.
  - (* AtomicStepEvent *)
    assert (Hpen := Hpending n).
    destruct (pending n) as [ [? e] | ] eqn:Epn.
    2:{ exfalso. rewrite <- (proj1 STEPps) in Hpen; discriminate. }
    destruct Hpen as [[Hpen Hpen2] | Hpen].
    2:{ exfalso. destruct e as [[] | ].
        - rewrite <- (proj1 STEPps) in Hpen. destruct Hpen; discriminate.
        - destruct Hpen as (? & ? & Hpen); rewrite <- (proj1 STEPps) in Hpen; discriminate.
        - destruct Hpen as (? & ? & ? & Hpen); rewrite <- (proj1 STEPps) in Hpen; discriminate. }
    rewrite <- (proj1 STEPps) in Hpen; inversion Hpen; injpair; clear Hpen.
    destruct m.
    + destruct STEPz as [LLL RRR]; cbn in LLL, RRR.
      inversion LLL; subst R; injpair.
      * clear H1 H3.
        rewrite <- RRR, <- H5 in IH.
        specialize (IH OINJ preallocd OINJpreallocd tmp OINJtmp allocd used pending).
        do 5 (forward IH; [ solve[auto] | ]).
        forward IH.
        { intros n0; destruct (eq_dec n n0) as [ <- | ].
          - rewrite Epn, <- (proj2 STEPps), H5.
            right. destruct (fst z0 k) eqn:E; eauto. symmetry in E; apply MONO in E. rewrite <- E; eauto.
          - rewrite <- CONSTANTps by auto. apply Hpending. }
        auto.
      * clear H H3. rewrite <- RRR, <- H5 in IH.
        cbn in Hpen2.
        forward IH.
        { unfold FunOverride.override; intros k1 k2 y.
          destruct EqDecK, EqDecK.
          - congruence.
          - intros. exfalso. apply (proj2 Hpen2 k2). congruence.
          - intros. exfalso. apply (proj2 Hpen2 k1). congruence.
          - apply OINJ. }
        specialize (IH preallocd OINJpreallocd).
        specialize (IH (set n None tmp)).
        forward IH. { auto using @oinjective_set_None. }
        specialize (IH allocd used pending).
        forward IH. { auto. }
        forward IH. { auto. }
        forward IH.
        { intros v1 k1; unfold FunOverride.override. destruct EqDecK.
          - intros Hus HSome; injection HSome; intros <-.
            contradiction Hus. apply (Htmp _ _ (proj1 Hpen2)).
          - auto. }
        forward IH. {
          intros n1 ?; set_dec n n1; discriminate + eauto. }
        forward IH. {
          intros n1 ? ?; set_dec n n1; try discriminate.
          intros Etmp.
          unfold FunOverride.override; destruct EqDecK; [ | eauto].
          - injection 1; intros <-. apply n0. revert Etmp. generalize (proj1 Hpen2).
            apply OINJtmp. }
        forward IH.
        { intros n0; set_dec n n0.
          - rewrite Epn. right. exists r. split; [ | apply (proj2 STEPps)].
            apply MONO. rewrite <- H5. rewrite FunOverride.override_new_val. reflexivity.
          - rewrite <- CONSTANTps by auto.
            specialize (Hpending n0). destruct (pending n0) as [ [? [ [] | ]] | ]; cbn; eauto.
            destruct Hpending as [[? Hp] | Hp]; [left|right]; eauto.
            split; auto. destruct Hp; split; auto.
            intros k'.
            unfold FunOverride.override; destruct EqDecK as [-> | ]; [ | eauto ].
            injection 1; intros <-; apply n1. revert H0. generalize (proj1 Hpen2); apply OINJtmp. }
        auto.
      * clear H H3. cbn in Hpen2. rewrite <- H5, <- RRR in IH.
        forward IH. { auto. }
        specialize (IH preallocd OINJpreallocd).
        specialize (IH (set n None tmp)).
        forward IH. { auto using @oinjective_set_None. }
        specialize (IH allocd used pending).
        forward IH. { auto. }
        forward IH. { auto. }
        forward IH. { auto. }
        forward IH. {
          intros n1 ?; set_dec n n1; discriminate + eauto. }
        forward IH. {
          intros n1 ? ?; set_dec n n1; discriminate + eauto. }
        forward IH.
        { intros n0; set_dec n n0.
          - rewrite Epn. right. exists r. split; [ | apply (proj2 STEPps)].
            apply MONO. rewrite <- H5. auto.
          - rewrite <- CONSTANTps by auto. apply Hpending. }
        auto.
    + clear Hpen2. destruct STEPz as [W Ez]. cbn in W. inversion W. subst R; injpair.
      cbn in *. clear H H2 H4. rewrite <- H6, <- Ez in IH.
      forward IH. { auto. }
      specialize (IH (set n (Some r) preallocd)).
      forward IH. {
        intros n1 n2 y; set_dec n n1; set_dec n n2; auto.
        - injection 1; intros <- HH. apply Hpreallocd in HH. contradiction (proj2 HH).
        - intros HH; injection 1; intros <-. apply Hpreallocd in HH. contradiction (proj2 HH).
        - apply OINJpreallocd. }
      specialize (IH tmp OINJtmp).
      specialize (IH allocd used pending).
      forward IH. {
        intros n0 ?; set_dec n n0.
        - injection 1; intros <-. split; [ intros HH; contradiction H7; auto | ].
          constructor 2. constructor.
        - intros HH; apply Hpreallocd in HH. destruct HH; split; auto.
          constructor 1; assumption. }
      forward IH. { intros ? HH; apply Hallocd in HH; constructor 1; assumption. }
      do 3 (forward IH; [ solve[auto] | ]).
      forward IH.
      { intros n0. set_dec n n0.
        - rewrite Epn. right; eexists. split; [ constructor 2; constructor | eauto using (proj2 STEPps) ].
        - rewrite <- CONSTANTps by auto.
          specialize (Hpending n0). destruct (pending n0) as [ [? [ [] | ]] | ]; cbn; eauto.
          destruct Hpending as [[? Hp] | Hp]; [left|right]; eauto.
          destruct Hp as (? & ? & ?); eexists; split; [ | eauto ].
          constructor 1; assumption. }
      auto.
Qed.

Lemma lazyMap_correct_A
  : forall
      (t0 : AtomicTrace MMM)
      (z' : (K -> option V) * Ensemble V)
      (ps' : Pending (ObjInstruction lazyMapObj))
      (H : Steps PendingStep t0 noPending ps')
      (H0 : Steps (PendingEvalStep_ (objBase lazyMapObj)) t0 lazyMapInit z')
    , lltrace (projAtomic t0) (fst z') (fun _ => False) (fun _ => False) (fun _ => None).
Proof.
  intros; apply (proj2 (lazyMap_correct_A_ H H0)) with (preallocd := fun _ => None) (tmp := fun _ => None); discriminate + contradiction + auto.
Qed.

Inductive threadRel : Thread MMM -> PendingEvent MMM -> PendingEvent (PureMapM K V) -> Prop :=
| threadRel_MkThread R (p : Program _ R) ee
  : threadRel (MkThread p) NoPendingEvent (MkPendingEvent ee)
| threadRel_CallingThread_CallEvent R X (h : X -> Program _ R) (e : _ X) ee
  : threadRel (CallingThread h) (MkPendingEvent (CallEvent e)) (MkPendingEvent ee)
| threadRel_CallingThread_ReturnEvent R X (h : X -> Program _ R) (r : X) ee
  : threadRel (CallingThread h) (MkPendingEvent (ReturnEvent r)) (MkPendingEvent ee)
| threadRel_NoThread : threadRel NoThread NoPendingEvent NoPendingEvent
.

Inductive thread_status (f1 : K -> option V) (k0 : K) (th : Thread MMM) (pp : option {R & MMM R})
   : option V -> Type :=
| thread_status_LOOKUP_pre (etc : option V -> Program MMM V)
  : thread_status f1 k0 (CallingThread etc) (Some (existT _ _ (inl (Lookup k0)))) None ->
    th = MkThread (Bind (inl (Lookup k0)) etc) ->
    pp = None ->
    thread_status f1 k0 th pp None
| thread_status_LOOKUP_post (etc : option V -> Program MMM V)
  : (forall ov, (ov = f1 k0 \/ ov = None) -> thread_status f1 k0 (MkThread (etc ov)) None None) ->
    th = CallingThread etc ->
    pp = Some (existT _ _ (inl (Lookup k0))) ->
    thread_status f1 k0 th pp None
| thread_status_FRESH_pre (etc : V -> Program MMM V)
  : thread_status f1 k0 (CallingThread etc) (Some (existT _ _ (inr alloc))) None ->
    th = MkThread (Bind (inr alloc) etc) ->
    pp = None ->
    thread_status f1 k0 th pp None
| thread_status_FRESH_post (etc : V -> Program MMM V)
  : (forall v, thread_status f1 k0 (MkThread (etc v)) None (Some v)) ->
    th = CallingThread etc ->
    pp = Some (existT _ _ (inr alloc)) ->
    thread_status f1 k0 th pp None
| thread_status_PUT_pre v (etc : V -> Program MMM V)
  : thread_status f1 k0 (CallingThread etc) (Some (existT _ _ (inl (PutIfAbsent k0 v)))) None ->
    th = MkThread (Bind (inl (PutIfAbsent k0 v)) etc) ->
    pp = None ->
    thread_status f1 k0 th pp (Some v)
| thread_status_PUT_post v (etc : V -> Program MMM V)
  : (forall v', Some v' = f1 k0 -> thread_status f1 k0 (MkThread (etc v')) None None) ->
    th = CallingThread etc ->
    pp = Some (existT _ _ (inl (PutIfAbsent k0 v))) ->
    thread_status f1 k0 th pp None
| thread_status_RETURN v
  : Some v = f1 k0 ->
    th = MkThread (Return v) ->
    pp = None ->
    thread_status f1 k0 th pp None
.

Lemma thread_status_lazyMap f1 k0 : thread_status f1 k0 (MkThread (lazyMap (PureLookup k0))) None None.
Proof.
  econstructor; [ .. | reflexivity | reflexivity ].
  econstructor; [ .. | reflexivity | reflexivity ].
  intros ov Hov.
  assert (H2ov : { v | ov = Some v /\ Some v = f1 k0 } + { ov = None }).
  { destruct ov as [ v | ]; [ left; exists v | right; reflexivity ].
    split; [ auto | ].
    destruct Hov; [ auto | discriminate ].
  }
  destruct H2ov as [ [? [-> ?]] | -> ]; cbn.
  - econstructor 7; eauto.
  - repeat (econstructor; [ .. | reflexivity | reflexivity ]). (* backtracking FTW *)
    auto.
Qed.

Definition isNotCallingThread {M} (t : Thread M) : Prop :=
  match t with
  | CallingThread _ => False
  | MkThread _ | NoThread => True
  end.

Ltac my_solve :=
  match goal with
  | [ H : _ = ?t |- isNotCallingThread ?t ] => rewrite <- H; exact I
  | [ H : _ = ?t /\ _ |- isNotCallingThread ?t ] => rewrite <- (proj1 H); exact I
  | [ H : MkThread (Bind _ _) = ?t, H' : ?t = MkThread (Return _) |- _ ] =>
    exfalso; rewrite <- H in H'; discriminate
  end.

Section MkF.

Context (classicT : forall P : Prop, { P } + { ~ P }).

Fixpoint mkf (w : nat -> option K) (t : ConcTrace (PureMapM K V)) (kont : K -> V) : K -> V :=
  match t with
  | nil => kont
  | cons (n, e) t =>
    match e with
    | CallEvent (PureLookup k) => mkf (set n (Some k) w) t kont
    | @ReturnEvent _ R r =>
      match @classicT (R = V), w n with
      | left H, Some k => mkf (set n None w) t (swap kont k (eq_rect _ _ r _ H))
      | _, _ => f0
      end
    end
  end.

Lemma LEM_refl (X : Type) : @classicT (X = X) = left eq_refl.
Proof.
  destruct classicT; [ f_equal; apply UIP_refl | contradiction ].
Qed.

Lemma lazyMap_correct_B
    (t1 : list (BiThreadEvent (PureMapM K V) (ObjInstruction (lazyMapObj (K := K)))))
  : forall
      (t : ConcTrace (PureMapM K V))
      (H1 : t = projHigh t1)
      (ts finalThreads : Threads (ObjInstruction lazyMapObj))
      (H4 : ThreadsSteps (objProgram (lazyMapObj (EqDecK := @EqDecK))) t1 ts finalThreads)
      (kont : K -> V)
      (w : nat -> option K)
      (f := mkf w t kont)
      (qs : Pending _)
      (qsNothread : forall n, NoThread = ts n -> NoPendingEvent = qs n)
      (INJk : injective _ _ kont)
      (wNone : forall n, w n = None -> ts n = NoThread)
      (tmp : nat -> option V)
      ff allocd used pending
      (Hlow : lltrace (projLow t1) ff allocd used pending)
      (Hused_allocd : forall v, used v -> allocd v)
      (Hallocd : forall n v, Some v = tmp n -> allocd v)
      (Hused : forall n v, Some v = tmp n -> ~ used v)
      (Htmp : oinjective tmp)
      (Hpending : forall n, isNotCallingThread (ts n) -> None = pending n)
      (seen : K -> Prop)
      (Hseen : forall k, seen k -> exists v, Some v = ff k /\ (oinjective ff -> v = kont k))
      (wSome : forall n k, w n = Some k -> thread_status ff k (ts n) (pending n) (tmp n) * (qs n = MkPendingEvent (ReturnEvent (f k))) )
    , (injective (PureMapM K V V) V (lookupSpec f (R:=V)) /\
       oinjective ff /\
       (forall k, seen k -> Some (f k) = ff k)
      ) /\
      IsConcTraceOfSpec' t (functionSpec (lookupSpec f)) tt qs.
Proof.
  induction t1 as [ | [n e] t1 IH ]; intros; subst.
  { (* nil *) split.
    - split; [ apply injective_lookupSpec; auto | ].
      split; [ auto | ].
      intros k Hk; apply Hseen in Hk. destruct Hk as [v [Hff Hkont]].
      specialize (Hkont Hlow).
      rewrite Hkont in Hff. exact Hff.
    - econstructor 1 with (t0 := nil); reflexivity + apply Steps_nil; reflexivity.
  }
  specialize (IH _ eq_refl).
  apply Steps_cons_inv in H4; destruct H4 as (ts0 & STEPts & STEPSts).
  destruct e as [ _ [k0] | | | | ];
    destruct STEPts as [STEP CONSTANT]; cbn beta iota delta - [f] in STEP, CONSTANT, Hlow, f |- *.
  - (* HighCall PureLookup *)
    specialize (IH ts0 finalThreads STEPSts).
    specialize (IH kont (set n (Some k0) w)).
    (* Do the abstract steps immediately. (skip CallEvent) *)
    specialize (IH (set n (MkPendingEvent (ReturnEvent (f k0))) qs)).
    forward IH.
    { intros n0; set_dec n n0.
      - intros BAD; rewrite <- (proj2 STEP) in BAD; discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH INJk).
    forward IH.
    { intros n0; set_dec n n0.
      - discriminate.
      - rewrite <- CONSTANT by auto; apply wNone. }
    specialize (IH (set n None tmp)).
    specialize (IH _ _ _ _ Hlow Hused_allocd).
    forward IH. { intros n0 v; set_dec n n0; discriminate + eauto. }
    forward IH. { intros n0 v; set_dec n n0; discriminate + eauto. }
    forward IH. { intros n0 n1 y; set_dec n n0; discriminate + set_dec n n1; discriminate + apply Htmp. }
    forward IH.
    { intros n0; destruct (eq_dec n n0) as [ <- | ].
      - intros; apply Hpending; my_solve.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH seen Hseen).
    forward IH.
    { intros n0 k; set_dec n n0.
      - rewrite <- (proj2 STEP); intros H; inversion H; subst; clear H.
        rewrite <- Hpending by my_solve.
        split; [ apply thread_status_lazyMap | reflexivity ].
      - rewrite <- CONSTANT by auto. apply wSome. }
    destruct IH as [(IH1 & ? & ?) []]; split; [ auto | ].
    econstructor 1 with
      (t0 := ((n, MkAtomicEvent (CallEvent (PureLookup k0)))
             :: (n, AtomicStepEvent (PureLookup k0) (f k0)) :: t0)%list).
    + apply StepsMore with (a1 := set n (MkPendingEvent (CallEvent (PureLookup k0))) qs).
      { constructor; cbn.
        - split; [ apply qsNothread, STEP | rewrite set_eq; reflexivity ].
        - intros; rewrite set_neq by auto; auto. }
      eapply StepsMore; [ | eassumption ].
      { constructor; cbn.
        - split; [ rewrite set_eq; reflexivity | rewrite set_eq; reflexivity ].
        - intros; rewrite 2 set_neq by auto; auto. }
    + apply StepsMore with (a1 := tt); [ constructor | ].
      apply StepsMore with (a1 := tt); [ constructor | eassumption ].
    + cbn. f_equal; auto.
  - (* HighReturn *)
    specialize (IH ts0 finalThreads STEPSts).
    assert (Ew : exists k, w n = Some k).
    { destruct (w n) eqn:Eww; eauto. apply wNone in Eww.
      rewrite <- (proj1 STEP) in Eww; discriminate. }
    destruct Ew as [k0 Ew].
    destruct (wSome _ _ Ew) as [STATUS Eqs].
    assert (RV : exists (e : V = R) (v0 : V), Some v0 = ff k0 /\ r = eq_rect _ _ v0 _ e).
    { generalize (proj1 STEP). destruct STATUS;
        lazymatch goal with
        | [ H : ts n = _ |- _ ] => rewrite H
        end; try discriminate.
      intros W; inversion W; injpair; clear W.
      exists eq_refl; eauto. }
    destruct RV as [e [v0 [Ev0 EE]]].
    subst R r.
    revert f Hseen wSome Eqs; rewrite LEM_refl, Ew; intros; cbn in f.
    specialize (IH (swap kont k0 v0) (set n None w)).
    specialize (IH (set n NoPendingEvent qs)).
    forward IH.
    { intros n0; set_dec n n0.
      - auto.
      - rewrite <- CONSTANT by auto. auto. }
    forward IH; [ auto using injective_swap | ].
    forward IH.
    { intros n0; set_dec n n0.
      - rewrite <- (proj2 STEP). auto.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH tmp).
    specialize (IH _ _ _ _ Hlow).
    do 4 (forward IH; [ solve[auto] | ]).
    forward IH.
    { intros n0; destruct (eq_dec n n0) as [ <- | ].
      - intros; apply Hpending; my_solve.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH (fun k' => k' = k0 \/ seen k')).
    forward IH.
    { intros k' [-> | Hk]; eauto.
      - rewrite swap_eq. eauto.
      - destruct (eq_dec k0 k') as [<- | ].
        + rewrite swap_eq; eauto.
        + apply Hseen in Hk. destruct Hk as [v' [Eff Ekont]].
          exists v'; split; auto.
          intros OINJ.
          rewrite swap_neq; auto.
          specialize (Ekont OINJ).
          rewrite Ekont in Eff.
          intros ->.
          contradiction n0.
          revert Ev0 Eff.
          apply OINJ.
    }
    forward IH.
    { intros n0; set_dec n n0.
      - discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    destruct IH as [(IH1 & ? & Hf) []]; split; [ auto | ].
    econstructor 1 with
      (t0 := ((n, MkAtomicEvent (ReturnEvent (f k0))) :: t0)%list).
    + eapply StepsMore; [ | eassumption ].
      { constructor; cbn.
        - split; [ auto | rewrite set_eq; auto ].
        - intros; rewrite set_neq by auto; auto. }
    + eapply StepsMore; [ | eassumption ].
      { constructor. }
    + cbn. f_equal; [ f_equal | auto ].
      f_equal.
      specialize (Hf _ (or_introl eq_refl)).
      rewrite <- Hf in Ev0.
      inversion Ev0; auto.
  - (* LowCall *)
    destruct STEP as (R & k & Hts & Hts0).
    specialize (IH ts0 finalThreads STEPSts).
    specialize (IH kont w).
    specialize (IH qs).
    forward IH.
    { intros n1; destruct (eq_dec n n1) as [ <- | ].
      - rewrite <- Hts0; discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH INJk).
    forward IH.
    { intros n1; destruct (eq_dec n n1) as [ <- | ].
      - intros Hwn; apply wNone in Hwn; rewrite <- Hts in Hwn; discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    assert (Ew : exists k, w n = Some k).
    { destruct (w n) eqn:Ewn; [ eauto | exfalso ].
      apply wNone in Ewn. rewrite <- Hts in Ewn; discriminate.
    }
    destruct Ew as [k0 Ew].
    assert (WWW :
      match pia_v n0 with
      | None => match n0 with inl (PutIfAbsent _ _) => False | _ => True end
      | Some v => Some v = tmp n
      end).
    { assert (STAT := wSome _ _ Ew).
      destruct STAT as [STAT ?].
      rewrite <- Hts in STAT.
      inversion STAT; injpair; try discriminate.
      - inversion H0; injpair; cbn; auto.
      - inversion H0; injpair; cbn; auto.
      - inversion H0; injpair; cbn; eauto. }
    destruct (pia_v n0) eqn:Epia.
    + cbn in Hlow; forward Hlow; [ solve [ eauto ] | ].
      specialize (IH (set n None tmp)).
      specialize (IH ff _ _ _ Hlow).
      forward IH. { intros ? [ <- | ]; eauto. }
      forward IH. { intros n1 ?. set_dec n n1; discriminate + eauto. }
      forward IH. { intros n1 v1 ?. revert H; set_dec n n1; [ discriminate | ].
        intros HH [<- | ].
        - apply n2; revert WWW HH; eauto.
        - contradict H; eauto. }
      forward IH. {
        intros n1 n2 y. set_dec n n1; try discriminate.
        set_dec n n2; try discriminate.
        apply Htmp. }
      forward IH.
      { intros n1; set_dec n n1.
        - rewrite <- Hts0; contradiction.
        - rewrite <- CONSTANT by auto; auto. }
      specialize (IH seen Hseen).
      forward IH.
      { intros n1 k1 Hwn1; set_dec n n1.
        - rewrite <- Hts0.
          { specialize (wSome n k1 Hwn1).
            remember (pending n); remember (tmp n).
            destruct wSome as [[] ?]; try solve [congruence | my_solve ].
            all: rewrite <- Hts in e; inversion e; injpair; clear e; auto.
          }
        - rewrite <- CONSTANT by auto; auto. }
      destruct IH as [? []]; split; [ auto | ].
      econstructor 1 with (t0 := t0); eassumption.
    + cbn in Hlow; forward Hlow; [ solve [ eauto ] | ].
      specialize (IH tmp ff _ _ _ Hlow).
      do 4 (forward IH; [ solve [eauto] | ]).
      forward IH.
      { intros n1; set_dec n n1.
        - rewrite <- Hts0; contradiction.
        - rewrite <- CONSTANT by auto; auto. }
      specialize (IH seen Hseen).
      forward IH.
      { intros n1 k1 Hwn1; set_dec n n1.
        - rewrite <- Hts0.
          { specialize (wSome n k1 Hwn1).
            remember (pending n); remember (tmp n).
            destruct wSome as [[] ?]; try solve [contradiction | congruence | my_solve ].
            all: rewrite <- Hts in e; inversion e; injpair; clear e; auto; try contradiction.
          }
        - rewrite <- CONSTANT by auto; auto. }
      destruct IH as [? []]; split; [ auto | ].
      econstructor 1 with (t0 := t0); eassumption.
  - (* LowReturn *)
    destruct STEP as (R & k & Hts & Hts0).
    specialize (IH ts0 finalThreads STEPSts).
    specialize (IH kont w).
    specialize (IH qs).
    forward IH.
    { intros n0; destruct (eq_dec n n0) as [<- | ].
      - rewrite <- Hts0. discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    specialize (IH INJk).
    assert (Ew : exists k, w n = Some k).
    { destruct (w n) eqn:Ewn; [ eauto | exfalso ].
      apply wNone in Ewn. rewrite <- Hts in Ewn; discriminate.
    }
    destruct Ew as [k0 Ew].
    forward IH.
    { intros n0; destruct (eq_dec n n0) as [<- | ].
      - intros Hw; rewrite Ew in Hw; discriminate.
      - rewrite <- CONSTANT by auto; auto. }
    destruct (wSome _ _ Ew) as [STAT Hqs].
    remember (tmp n) as tmpn.
    destruct STAT;
      lazymatch goal with
      | [ e : pending n = _ |- _ ] => rewrite e in Hlow
      end; try contradiction.
    + (* Lookup *) destruct Hlow as [WELL Hlow]. specialize (IH tmp _ _ _ _ Hlow).
      do 4 (forward IH; [ solve[auto] | ]).
      forward IH.
      { intros n0. set_dec n n0; auto. rewrite <- CONSTANT by auto; auto. }
      specialize (IH seen Hseen).
      rewrite <- Hts in e; inversion e; clear e; injpair.
      forward IH.
      { intros n0 k1 Hw; set_dec n n0.
        - rewrite Ew in Hw; inversion Hw; clear Hw; subst k1.
          constructor.
          + rewrite <- Heqtmpn, <- Hts0. apply t.
            destruct WELL as [WELL|WELL]; inversion WELL; injpair; auto.
          + assumption.
        - rewrite <- CONSTANT by auto. auto. }
      destruct IH as [(IH1 & ? & Jf) []]; split; [ auto | ].
      econstructor 1 with (t0 := t0); eassumption.
    + (* alloc *) destruct Hlow as (v0 & HReturn & Hv0 & Hlow).
      specialize (IH (set n (Some v0) tmp) _ _ _ _ Hlow).
      forward IH. { eauto. }
      forward IH.
      { intros n0 v; set_dec n n0; eauto.
        injection 1; intros <-; auto. }
      forward IH. { intros n0 v; set_dec n n0; eauto. injection 1; intros <-; auto. }
      forward IH.
      { intros n0 n1 y; set_dec n n0; set_dec n n1.
        - reflexivity.
        - injection 1; intros <- contra; exfalso; eauto.
        - injection 2; intros <-; exfalso; eauto.
        - apply Htmp. }
      forward IH.
      { intros n0. set_dec n n0; auto. rewrite <- CONSTANT by auto; auto. }
      specialize (IH seen Hseen).
      rewrite <- Hts in e; inversion e; clear e; injpair.
      forward IH.
      { intros n0 k1 Hw; set_dec n n0.
        - rewrite Ew in Hw; inversion Hw; clear Hw; subst k1.
          constructor.
          + replace x with v0 in * by (inversion HReturn; injpair; auto); clear x.
            rewrite <- Hts0. apply t.
          + assumption.
        - rewrite <- CONSTANT by auto. auto. }
      destruct IH as [? []]; split; [ auto | ].
      econstructor 1 with (t0 := t0); eassumption.
    + (* PutIfAbsent *)
      destruct Hlow as (v0 & Hff & Hv0 & Hlow).
      specialize (IH tmp _ _ _ _ Hlow).
      do 4 (forward IH; [ solve [eauto] | ]).
      forward IH.
      { intros n0. set_dec n n0; auto. rewrite <- CONSTANT by auto; auto. }
      specialize (IH seen Hseen).
      rewrite <- Hts in e; inversion e; clear e; injpair.
      forward IH.
      { intros n0 k1 Hw; set_dec n n0.
        - rewrite Ew in Hw; inversion Hw; clear Hw; subst k1.
          constructor.
          * replace x with v0 in * by (inversion Hv0; injpair; auto); clear x.
            rewrite <- Hts0, <- Heqtmpn. apply t. auto.
          * assumption.
        - rewrite <- CONSTANT by auto. auto. }
      destruct IH as [? []]; split; [ auto | ].
      econstructor 1 with (t0 := t0); eassumption.
  - (* LowSilentStep *) exfalso.
    destruct STEP as (R & p & Hts & Hts0).
    destruct (w n) eqn:Ew.
    + apply wSome in Ew. destruct (fst Ew); try
        lazymatch goal with
        | [ H : ts n = _ |- _ ] => rewrite <- Hts in H; solve [inversion H]
        end.
    + apply wNone in Ew. rewrite <- Hts in Ew; discriminate.
Qed.

End MkF.

Theorem lazyMap_correct
  : forall t : ConcTrace (PureMapM K V),
      IsConcTraceOf t (lazyMapObj (EqDecK := @EqDecK)) ->
      exists f : K -> V,
        injective _ _ (lookupSpec f (R := V)) /\
        IsConcTraceOfSpec t (functionSpec (lookupSpec f)).
Proof.
  destruct classicT_inhabited as [classicT].
  intros t [? [[? ? ? Elow STEPSth] [? ? ? STEPSps STEPSz ->]]].
  assert (YY := lazyMap_correct_A STEPSps STEPSz).

  cbn in Elow. rewrite Elow in YY.
  eapply lazyMap_correct_B with (classicT := classicT) (w := fun _ => None) (tmp := fun _ => None)
    (qs := fun _ => NoPendingEvent) (seen := fun _ => False) in YY;
    try (eassumption + discriminate + contradiction + eauto).
  destruct YY as [INJ []].
  eexists; split; [ apply INJ | econstructor; eauto ].
Qed.

End lazyMap_correct.

Section Blah.

Import BMap.
Notation MMM K L V := (PairMethod (PureMapM K L) (BMapMethod L V)).

Definition predMap {Loc K V}
  : forall R, BMapMethod K V R -> Program (MMM K Loc V) R :=
  fun R m =>
    match m in BMapMethod _ _ R return Program (MMM K Loc V) R with
    | get k =>
      Bind (inl (PureLookup k)) (fun r : Loc =>
      Bind (inr (get r)) Return)
    | put k v =>
      Bind (inl (PureLookup k)) (fun r : Loc =>
      Bind (inr (put r v)) Return)
    end.

Definition K0 := nat.

Definition topSpec {V} : Spec (BMapMethod K0 V) := LocBMap.atomicBMap _.

Section pppp.

Ltac step H :=
  apply ProgramFacts.EvalProgramInv in H.

Theorem predMapCorrect_ {V}
  : forall lmap : forall Ret, PureMapM K0 Allocator.Loc Ret -> Ret,
      injective _ Loc (lmap Loc) ->
      forall s s',
      s' = (fun k => s (lmap _ (PureLookup k))) ->
      Simulates_
        (mapSpec
           (programSpec (LocBMap.atomicBMap V))
           (flattenImplProg (replaceFunction lmap) predMap))
        topSpec
        s
        s'.
Proof.
  intros lmap INJ; cofix SELF; intros s _ ->; constructor; intros R int r s' EVAL; cbn in *.
  exists (fun k => s' (lmap _ (PureLookup k))); [ | apply SELF; reflexivity ].
  step EVAL; destruct int; cbn in EVAL.
  - do 3 step EVAL; cbn in EVAL.
    destruct EVAL as [s'' [r'' [H EVAL]]]. inversion H; injpair; clear H.
    do 2 step EVAL; cbn in EVAL.
    destruct EVAL; subst.
    red. constructor 1 with (map := (fun k0 => s' (lmap _ (PureLookup k0)))).
  - do 3 (step EVAL; cbn in EVAL).
    destruct EVAL as [s'' [r'' [H EVAL]]]. inversion H; injpair; clear H.
    do 2 step EVAL; cbn in EVAL.
    destruct EVAL; subst.
    match goal with
    | [ |- _ ?W ] =>
      let t := type of W in
      evar (e : t);
      replace W with e; subst e; [ constructor | ]
    end.
    apply FunctionalExtensionality.functional_extensionality; intros x.
    unfold FunOverride.override.
    do 2 destruct LocBMap.eq_key_dec; auto.
    all: exfalso.
    + subst; contradiction.
    + apply INJ in e. congruence.
Qed.

(* ASSUMPTION: [predMap] is "correct" wrt. [topSpec] when
   it implements the sequential spec  *)
Theorem predMapCorrect {V}
  : forall lmap : forall Ret, PureMapM _ _ Ret -> Ret,
      injective _ Loc (lmap Loc) ->
      Simulates
        (mapSpec
           (programSpec (LocBMap.atomicBMap V))
           (flattenImplProg (replaceFunction lmap) predMap))
        topSpec.
Proof.
  intros lmap INJ; apply (predMapCorrect_ INJ). reflexivity.
Qed.

End pppp.

End Blah.

Print Assumptions predMapCorrect.
