From C4.Core Require Import
  Program
  Object
  Pair
  ProgramOps.

Definition shareRightPairProg
           {R I1 I2 I3}
           (p: Program (PairMethod (PairMethod I1 I3) (PairMethod I2 I3)) R):
  Program (PairMethod (PairMethod I1 I2) I3) R :=
  
  mapProgram
    (fun R' m' =>
       match m' with
       | inl mL =>
         match mL with
         | inl m1 => inl (inl m1)
         | inr m2 => inr m2
         end
       | inr mR =>
         match mR with
         | inl m1 => inl (inr m1)
         | inr m2 => inr m2
         end
       end)
    p.

Definition shareRightPairImplProg
           {M I1 I2 I3}
           (obj: forall R, M R -> Program (PairMethod (PairMethod I1 I3) (PairMethod I2 I3)) R):
  forall R, M R -> Program (PairMethod (PairMethod I1 I2) I3) R :=

  fun R m =>
    shareRightPairProg (obj R m).


Definition factorRightPairProg
           {I3 R}
           (p: Program (PairMethod I3 I3) R):
  Program (PairMethod UnitInstruction I3) R :=

  mapProgram
    (fun R' m' =>
       match m' with
       | inl mL => inr mL
       | inr mR => inr mR
       end)
      p.

Definition factorRightPairImplProg
           {M I3}
           (obj: forall R, M R -> Program (PairMethod I3 I3) R):
  forall R, M R -> Program (PairMethod UnitInstruction I3) R :=

  fun R m =>
    factorRightPairProg (obj R m).


Definition factorPairProg
           {I3 R}
           (p: Program (PairMethod I3 I3) R):
  Program I3 R :=
    mapProgram
      (fun R' m' =>
         match m' with
         | inl mL => mL
         | inr mR => mR
         end)
      p.

Definition factorPairImplProg
           {M I3}
           (obj: forall R, M R -> Program (PairMethod I3 I3) R):
  forall R, M R -> Program I3 R :=

  fun R m =>
    factorPairProg (obj R m).



Definition shareRightPairImpl
           {M I1 I2 I3}
           (o1 : Spec I1)
           (o2 : Spec I2)
           (o3 : Spec I3)
           (obj: forall R, M R -> Program (PairMethod (PairMethod I1 I3)
                                                 (PairMethod I2 I3)) R):
  Object M :=
  {|
    objBase := pairSpec (pairSpec o1 o2) o3;
    objProgram := shareRightPairImplProg obj
  |}.


Definition shareLeftPairImplProg
           {M I I1 I2}
           (obj: forall R, M R -> Program (PairMethod (PairMethod I I1) (PairMethod I I2)) R):
  forall R, M R -> Program (PairMethod I (PairMethod I1 I2)) R :=
  
  fun R m =>
    mapProgram
      (fun R' m' =>
         match m' with
         | inl mL =>
           match mL with
           | inl m1 => inl m1
           | inr m2 => inr (inl m2)
           end
         | inr mR =>
           match mR with
           | inl m1 => inl m1
           | inr m2 => inr (inr m2)
           end
         end)
      (obj R m).
  
Definition shareLeftPairImpl
           {M I I1 I2}
           (o : Spec I)
           (o1 : Spec I1)
           (o2 : Spec I2)
           (obj: forall R, M R -> Program (PairMethod (PairMethod I I1)
                                                 (PairMethod I I2)) R):
  Object M :=
  {|
    objBase := pairSpec o (pairSpec o1 o2);
    objProgram := shareLeftPairImplProg obj
  |}.


Definition shareRightFlattenPairImpl
           {M I1 I2 I3}
           (obj1: Object I1) (obj2: Object I2) (obj3: Object I3)
           (obj: forall R, M R -> Program (PairMethod (PairMethod I1 I3) (PairMethod I2 I3)) R):
  Object M :=

  flattenObject
    (pairObject (pairObject obj1 obj2) obj3) _
    (shareRightPairImplProg obj).

                                       
Definition shareLeftFlattenPairImpl
           {M I I1 I2}
           (obj1: Object I) (obj2: Object I1) (obj3: Object I2)
           (obj: forall R, M R -> Program (PairMethod (PairMethod I I1) (PairMethod I I2)) R):
  Object M :=

  flattenObject
    (pairObject obj1 (pairObject obj2 obj3)) _
    (shareLeftPairImplProg obj).


(*
Definition shareRightFactorImpl
           {M1 M2 I1 I2 I3}
           (obj1: Object I1) (obj2: Object I2) (obj3: Object I3)
           (objL: forall R, M1 R -> Program (PairMethod I1 I3) R)
           (objR: forall R, M2 R -> Program (PairMethod I2 I3) R):
  Object (PairMethod M1 M2) :=
  {|
    objBase :=
      pairSpec (pairSpec obj1.(objBase) obj2.(objBase))
                 obj3.(objBase);

    objProgram R m :=
      match m with
      | inl mL => mapProgram
                   (fun _ m =>
                      match m with
                      | inl m' => inl (inl m')
                      | inr m' => inr m'
                      end)
                   (flattenProgram (pairObject obj1 obj3).(objProgram) (objL R mL))
      | inr mR => mapProgram
                   (fun _ m =>
                      match m with
                      | inl m' => inl (inr m')
                      | inr m' => inr m'
                      end)
                   (flattenProgram (pairObject obj2 obj3).(objProgram) (objR R mR))
      end
  |}.
*)


Definition combineShareRightImplProg {Method1 LMethod1 Method2 LMethod2 RMethod}
           (objProg1 : forall Ret, Method1 Ret -> Program (PairMethod LMethod1 RMethod) Ret)
           (objProg2 : forall Ret, Method2 Ret -> Program (PairMethod LMethod2 RMethod) Ret)
           Ret (m : PairMethod Method1 Method2 Ret) :
  Program (PairMethod (PairMethod LMethod1 LMethod2) RMethod) Ret :=
  match m with
  | inl m' =>
    mapProgram
      (fun _ i =>
         match i with
         | inl i' => inl (inl i')
         | inr i' => inr i'
         end)
      (objProg1 _ m')
  | inr m' =>
    mapProgram
      (fun _ i =>
         match i with
         | inl i' => inl (inr i')
         | inr i' => inr i'
         end)
      (objProg2 _ m')
  end.
