From C4.Util Require Import
  LibTactics.

(* ---------------------------------------- *)
(* Ltacs *)

(*
    Ltac openAndInRec H H' :=
      let H1 := fresh H "_" in
      let H2 := fresh H "_" in
      match type of H' with
      | _ /\ _  =>
        destruct H' as (H1 & H2);
        try openAndInRec H H2
      end.
 *)

Ltac openAndIn H :=
  let rec openAndInRec H H' :=
      match type of H' with
      | _ /\ _  =>
        let H2 := fresh H "_" in
        destruct H' as (H' & H2);
      try openAndInRec H H2
      end in
  let H' := fresh H "_" in
  let H0 := fresh H "_" in
  let H1 := fresh H "_" in
  pose (H' := False);
pose (H0 := False);
pose (H1 := H);
openAndInRec H H1;
clear H H' H0.


Goal forall A B C,
    A /\ B /\ C -> C.
Proof.
  introv H1.
  openAndIn H1.
  assumption.
Qed.      


Ltac openAnds :=
  repeat (
      match goal with
      | [ H : _ /\ _ |- _ ] =>
        openAndIn H
      end).

Goal forall A B C D,
    A /\ B /\ C -> B /\ C /\ D -> C /\ D.
Proof.
  introv H1 H2.
  openAnds.
  split; assumption.
Qed.

(*
    Ltac deprem H :=
      match type of H with
      | forall x : ?T, _ =>
        let x := fresh "x" in
        assert (x: T);
          [ clear H
          | specialize (H x); clear x ]                  
      end.
 *)

Ltac forwardsInH H :=
  let H' := fresh in
  (* let args := forwards_build_app_arg H in
      lets H': H; *)
  forwards H': H;
    clear H;
    match goal with
    | [H': _ |- _] =>
      assert (H := H');
        clear H'
    | _ =>
      idtac
    end.

Ltac forwardsInH1 H A :=
  let H' := fresh in
  (* let args := forwards_build_app_arg H in
      lets H': H A; *)
  forwards H': H A;
    clear H;
    match goal with
    | [H': _ |- _] =>
      assert (H := H');
        clear H'
    | _ =>
      idtac
    end.

Ltac forwardsInH2 H A1 A2 :=
  let H' := fresh in
  forwards H': H A1 A2;
    clear H;
    match goal with
    | [H': _ |- _] =>
      assert (H := H');
        clear H'
    | _ =>
      idtac
    end.

Tactic Notation "forwardsIn" constr(E) :=
  forwardsInH E.
Tactic Notation "forwardsIn" constr(E0) constr(A1) :=
  forwardsInH1 E0 A1.
Tactic Notation "forwardsIn" constr(E0) constr(A1) constr(A2) :=
  forwardsInH2 E0 A1 A2.

Goal forall A B C,
    A ->
    (A -> B) ->
    (B -> C) ->
    C.
Proof.
  introv H1 H2 H3.
  forwardsIn H2. assumption.
  forwardsIn H3. assumption.
  
  Restart.

  introv H1 H2 H3.
  forwardsIn H2 H1.
  forwardsIn H3 H2.
  
  assumption.
Qed.


Ltac specex H :=
  repeat
    match type of H with
    | forall x : ?T, _ =>
      match type of T with
      | Prop =>
        fail 1
      | _ =>
        let x := fresh "x" in
        evar (x : T);
          (* to instantiate H with x *)
          let x' := eval unfold x in x in
              clear x; specialize (H x')
      end
    end.

Ltac specexDeprem H :=
  specex H;
  forwardsIn H.


Ltac subv x :=
  match goal with 
  | H: x = _ |- _ => rewrite H; simpl
  | H: _ = x |- _ => rewrite <- H; simpl
  | _ => try unfold x; simpl
  end.

Ltac subvIn x H' :=
  match goal with 
  | H: x = _ |- _ => rewrite H in H'; simpl in H'
  | H: _ = x |- _ => rewrite <- H in H'; simpl in H'
  | _ => try unfold x in H'; simpl in H'
  end.

Lemma subv_test:
  let m := 0 in
  forall a b c d e, 
    c = 2 + 2
    -> 2 + 2 = d
    -> m + 1 = e
    -> a = 0
    -> 0 = b
    -> a + b = m + 2
    -> a + b = m.

Proof.
  intros.
  subv b.
  subv a.
  subv m.
  subvIn a H4.
  subvIn b H4.
  subvIn m H4.
  reflexivity.
Qed.

Tactic Notation "applyClear" ident(H) :=
  apply H; clear H.

Tactic Notation "applyInClear" ident(H) ident(H') :=
  apply H in H'; clear H.

Tactic Notation "rewriteClear" ident(H) ident(H') :=
  rewrite H in H'; clear H.

Tactic Notation "rewriteRClear" ident(H) ident(H') :=
  rewrite <- H in H'; clear H.
